/*
 * Name: cifsmod.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-13
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: cifsmod.c,v 1.26 2002/06/03 21:58:17 cs Exp $
 */

#include "stdheaders.h"
#include "backend.h"
#include "module.h"
#include "hashtable.h"
#include "process.h"
#include "errcodes.h"
#include "util.h"
#include "user.h"
#include "server.h"
#include "share.h"
#include "file.h"
#include "cache.h"
#include "cifsconst.h"
#include "userdb.h"
#include "cfgparse.h"
#include "main.h"
#include "licman.h"
#include "ipc.h"
#include "mainmodule.h"
#include "guimanager.h"
#include "../../uiLib/maininterface.h"
#include "../../uiLib/cifsinterface.h"
#include "unicode.h"
#include "logging.h"
#include "lmb.h"
#include "cifsmod.h"
#include "frontend.h"
#include "notification.h"
#include "blockpool.h"
#include "raprpc.h"
#include "smb.h"
#include "ipcserver.h"
#include "pslayer.h"
#include "mainconst.h"

/* ------------------------------------------------------------------------- */

#define	CIFSFILE_MODULE_NAME	"cifsFile"

userdb_t	*cifsUserDb = NULL;

module_t	*cifsFileModule;

/* ------------------------------------------------------------------------- */

btint	cifsModGlobalCfgPath(struct cfgProperty **path, btint maxEntries)
{
cfgProperty_t	*p;
btint			i = 0;

	p = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.dbFile));
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	p = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.cfgFile));
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	return i;
}

static inline btint	cifsModHierarchyCfgPath(btu8 *section, btu8 *entity, struct cfgProperty *auxSection, struct cfgProperty **path, btint maxEntries)
{
cfgProperty_t	*p, *db, *cfg;
btint			i = 0;

	db = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.dbFile));
	cfg = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.cfgFile));
	db = cfgQueryWithName(section, 1, db);
	cfg = cfgQueryWithName(section, 1, cfg);
	p = cfgQueryWithName(entity, 1, db);
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	p = cfgQueryWithName(entity, 1, cfg);
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	if(auxSection != NULL && i < maxEntries)
		path[i++] = auxSection;
	p = cfgQueryWithName(GUIMAN_DEFAULT_ENTRY, 1, db);
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	p = cfgQueryWithName(GUIMAN_DEFAULT_ENTRY, 1, cfg);
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	return i;
}

btint	cifsModServerCfgPath(btu8 *srvName, struct cfgProperty *srvClassSection, struct cfgProperty **path, btint maxEntries)
{
	return cifsModHierarchyCfgPath("servers", srvName, srvClassSection, path, maxEntries);
}

btint	cifsModBrowserCfgPath(btu8 *browserName, struct cfgProperty **path, btint maxEntries)
{
	return cifsModHierarchyCfgPath("browsers", browserName, NULL, path, maxEntries);
}

/* ------------------------------------------------------------------------- */

static btint	cifsShareConfigPath(share_t *shr, struct cfgProperty **path, btint maxEntries)
{
	return cifsModServerCfgPath(srvName(shrServer(shr)), NULL, path, maxEntries);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static void	*cifsShareRootFile(share_t *share)
{
void	*file = shrRootFile(share);

	fileRetain(file);
	return file;
}

static btbool	cifsShareIsCaseSensitive(void *obj)
{
	return srvIsCaseSensitive(shrServer(obj));
}

static void	cifsUserAuthenticate(struct share *share, btint unixUid, btint unixGid, void *caller)
{
	userAuthenticate(shrServer(share), NULL, unixUid, unixGid, caller);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static void	cifsFileCreate(void *parent, btuni *name, btbool isDir, void *user, struct beFileSetAttr *attr, void *caller)
{
	fileCreateUnix(parent, name, isDir, 0, user, attr, caller);
}

static void	cifsFileSetOpenMode(void *obj, btint mode, btint modeMask, btbool autoclose, void *user, void *caller)
{
	fileSetOpenMode(obj, user == NULL ? -1 : userSmbUid(user), mode, modeMask, autoclose, 0, NULL, 0, 0, 1, 0, caller);
/* the return values don't match here: cifsFileSetOpenMode() should call back
 * with beFileRes_t whereas fileSetOpenMode calls back with filePrivRes_t.
 * Luckily this works by coincidence because only the error code is used and
 * this is at the same offset in both structures.
 */
}

static btbool	cifsFileOperationsRunning(void *obj)
{
cifsFile_t	*file = obj;

	return file->openLock || file->needOpen || (file->cache.any != NULL && cacheOperationsRunning(file->cache.any) > 0);
}

/* ------------------------------------------------------------------------- */
/* ------------------------ configuration handling ------------------------- */
/* ------------------------------------------------------------------------- */

static guiTypedesNumeric_t	numdesNbnsInitialTimeout = {
/*	div		mpy		inc		min		max			unit	*/
	1000,	1,		100,	100,	10000,		"seconds"
};

static guiTypedesNumeric_t	numdesNbnsMaxTimeout = {
/*	div		mpy		inc		min		max			unit	*/
	1000,	1,		100,	100,	1000000,	"seconds"
};

static guiTypedesNumeric_t	numdesPort = {
/*	div		mpy		inc		min		max			unit	*/
	1,		1,		1,		1,		65535,		""
};

static guiTypedesNumeric_t	numdesCifsTimeout = {
/*	div		mpy		inc		min		max			unit	*/
	1000,	1,		100,	100,	1000000,	"seconds"
};

static btu8	*exmapExtNames[] = {"Archive", "System", "Hidden", "Always On", "Always Off", NULL};
btu8	*cifsModExmapEnum[] = {"archive", "system", "hidden", "on", "off", NULL};
/* must match SRV_EXMAP_* in server.h */
static guiTypedesEnum_t	enumdesExecMap = {exmapExtNames, cifsModExmapEnum};

static btu8	*casemapExtNames[] = {"None", "Lowercase", "Uppercase", NULL};
btu8	*cifsModCasemapEnum[] = {"none", "lower", "upper", NULL};
/* must match SRV_CASE * in server.h */
static guiTypedesEnum_t	enumdesCaseMap = {casemapExtNames, cifsModCasemapEnum};

btu8	*cifsModSslVersEnum[] = {"ssl2or3", "ssl2", "ssl3", "tls1", NULL};

static btu8	*fakeSecurityExtNames[] = {"Server decides", "User Level", "Share Level", NULL};
btu8	*cifsFakeSecurityModeEnum[] = {"notFaked", "fakeUser", "fakeShare", NULL};
/* must match SRV_FAKE_SECURITY_* in server.h */
static guiTypedesEnum_t	enumdesFakeSecurity = {fakeSecurityExtNames, cifsFakeSecurityModeEnum};

#if defined USE_SSL
static btu8	*sslVersExtNames[] = {"SSL Version 2 or 3", "SSL Version 2", "SSL Version 3", "TSL Version 1", NULL};
/* must match SRV_SSL_* in server.h */
static guiTypedesEnum_t	enumdesSslVers = {sslVersExtNames, cifsModSslVersEnum};
#endif

/* ------------------------------------------------------------------------- */

#define	PERM_NORMAL	MAIN_GUI_PERM_AUTHORIZED
#define	PERM_EXPERT	(MAIN_GUI_PERM_AUTHORIZED | MAIN_GUI_PERM_EXPERT)

#define	DENY_NULL	0
#define	ALLOW_NULL	1
#define	FLG_0		0
#define	FLG_RC		MAIN_GUI_ENTRY_FLG_NEEDS_RECONNECT

static guiEntry_t	globalCfgEntries[] = {
	{	MAIN_GUI_ENTRY_STRING,	"Default Domain",
		NULL,					"cifsGlobalConfigDomain",
		PERM_NORMAL,			FLG_0, DENY_NULL,	"domain",
	},
	{	MAIN_GUI_ENTRY_STRING,	"WINS Server",
		NULL,					"cifsGlobalConfigWins",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"nbnsWinsIp",
	},
	{	MAIN_GUI_ENTRY_STRING,	"Scope-ID",
		NULL,					"cifsGlobalConfigScopeId",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"scopeID",
	},
	{	MAIN_GUI_ENTRY_NUMBER,	"Netbios NS Initial Timeout",
		&numdesNbnsInitialTimeout,	"cifsGlobalConfigNBInitTimeout",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"nbnsInitialTimeout",
	},
	{	MAIN_GUI_ENTRY_NUMBER,	"Netbios NS Maximum Timeout",
		&numdesNbnsMaxTimeout,	"cifsGlobalConfigNBMaxTimeout",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"nbnsTotalTimeout",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Enable Netbios Name Lookups",
		NULL,					"cifsGlobalConfigEnableNetbios",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"lookupTryNetbios",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Enable DNS Name Lookups",
		NULL,					"cifsGlobalConfigEnableDns",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"lookupTryDns",
	},
};

static guiEntry_t	serverCfgEntries[] = {
	{	MAIN_GUI_ENTRY_NUMBER,	"Connect on Port",
		&numdesPort,			"cifsServerConfigPort",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"tcpPort",
	},
	{	MAIN_GUI_ENTRY_STRING,	"Server\'s IP Address",
		NULL,					"cifsServerConfigIP",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"ipAddress",
	},
	{	MAIN_GUI_ENTRY_ENUM,	"Security Mode",
		&enumdesFakeSecurity,	"cifsServerConfigSecurityMode",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"fakeSecurityMode",
	},
	{	MAIN_GUI_ENTRY_NUMBER,	"CIFS Connect Timeout",
		&numdesCifsTimeout,		"cifsServerConfigConnectTimeout",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"connectTimeout",
	},
	{	MAIN_GUI_ENTRY_NUMBER,	"CIFS Request Timeout",
		&numdesCifsTimeout,		"cifsServerConfigRequestTimeout",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"requestTimeout",
	},
	{	MAIN_GUI_ENTRY_STRING,	"Remote Guest User Name",
		NULL,					"cifsServerConfigGuestName",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"guestRemoteUser",
	},
	{	MAIN_GUI_ENTRY_STRING,	"Guest User\'s Password",
		NULL,					"cifsServerConfigGuestPassword",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"guestPassword",
	},
	{	MAIN_GUI_ENTRY_ENUM,	"Take Execute Permission from",
		&enumdesExecMap,		"cifsServerConfigExecMap",
		PERM_NORMAL,			FLG_0, DENY_NULL,	"execMapping",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Invert Execute Permission",
		NULL,					"cifsServerConfigExecInvert",
		PERM_NORMAL,			FLG_0, DENY_NULL,	"execInvert",
	},
	{	MAIN_GUI_ENTRY_ENUM,	"Case Mapping",
		&enumdesCaseMap,		"cifsServerConfigCaseMapping",
		PERM_NORMAL,			FLG_0, DENY_NULL,	"caseMapping",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Enable faked Symbolic Links",
		NULL,					"cifsServerConfigEnableLinks",
		PERM_NORMAL,			FLG_RC, DENY_NULL,	"enableFakeLinks",
	},
#ifdef __APPLE__
	{	MAIN_GUI_ENTRY_BOOL,	"Disable Finder Trash",
		NULL,					"cifsServerPreventCreation",
		PERM_EXPERT,			FLG_0, DENY_NULL,	"preventCreationEnable",
	},
#endif
};

#if defined USE_SSL
static guiEntry_t	serverSslEntries[] = {
	{	MAIN_GUI_ENTRY_BOOL,	"Require SSL Encryption",
		NULL,					"cifsSslConfigRequireEncryption",
		PERM_EXPERT,			FLG_RC, DENY_NULL,	"sslRequireEncryption",
	},
	{	MAIN_GUI_ENTRY_ENUM,	"SSL Protocol",
		&enumdesSslVers,		"cifsSslConfigProtocol",
		PERM_EXPERT,			FLG_RC, DENY_NULL,	"sslVersion",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Bug Compatibility",
		NULL,					"cifsSslConfigCompatibility",
		PERM_EXPERT,			FLG_RC, DENY_NULL,	"sslCompatibility",
	},
	{	MAIN_GUI_ENTRY_FILE,	"Path to accepted Certificate(s)",
		NULL,					"cifsSslConfigCACert",
		PERM_EXPERT,			FLG_RC, DENY_NULL,	"sslCACertFile",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Require valid Server Certificate",
		NULL,					"cifsSslConfigRequireServerCert",
		PERM_EXPERT,			FLG_RC, DENY_NULL,	"sslRequireServerCert",
	},
	{	MAIN_GUI_ENTRY_FILE,	"PEM File with private Key",
		NULL,					"cifsSslConfigPrivateKey",
		PERM_EXPERT,			FLG_RC, DENY_NULL,	"sslKeyFile",
	},
	{	MAIN_GUI_ENTRY_FILE,	"PEM File with Certificate",
		&enumdesExecMap,		"cifsSslConfigCertificate",
		PERM_EXPERT,			FLG_RC, DENY_NULL,	"sslCertFile",
	},
};
#endif

static guiEntry_t	licenseEntries[] = {
	{	MAIN_GUI_ENTRY_STRING,	"Enter your License Key here",
		NULL,					"cifsLicenseConfigKeyInput",
		PERM_NORMAL,			FLG_0, DENY_NULL,	NULL,
	},
	{	MAIN_GUI_ENTRY_STRING,	"License Identifier",
		NULL,					"cifsLicenseConfigLicenseIdentifier",
		0,	/* permissions */	FLG_0, DENY_NULL,	NULL,
	},
	{	MAIN_GUI_ENTRY_STRING,	"License Type",
		NULL,					"cifsLicenseConfigLicenseType",
		0,	/* permissions */	FLG_0, DENY_NULL,	NULL,
	},
	{	MAIN_GUI_ENTRY_STRING,	"Max. active Clients",
		NULL,					"cifsLicenseConfigMaxClients",
		0,	/* permissions */	FLG_0, DENY_NULL,	NULL,
	},
	{	MAIN_GUI_ENTRY_STRING,	"Max. mounted Servers",
		NULL,					"cifsLicenseConfigMaxServers",
		0,	/* permissions */	FLG_0, DENY_NULL,	NULL,
	},
	{	MAIN_GUI_ENTRY_STRING,	"Expires",
		NULL,					"cifsLicenseConfigExpiry",
		0,	/* permissions */	FLG_0, DENY_NULL,	NULL,
	},
};

guiColumn_t	loginListColumns[] = {
	/*  name			align	 width	help */
	{	"Server",			0,		23,	"cifsFileLoginListColumnServer"		},
	{	"Local User",		0,		23,	"cifsFileLoginListColumnLocalUser"	},
	{	"Remote User",		0,		23,	"cifsFileLoginListColumnRemoteUser"	},
	{	"Domain",			0,		23,	"cifsFileLoginListColumnDomain"		},
	{	"State",			0,		8,	"cifsFileLoginListColumnState"		},
};

static guiButton_t	buttonEntries[] = {
	{	"Defaults",			MAIN_GUI_METHOD_GET,
		1,	/* parameter */	0,	/* flags */
		"cifsFileConfigButtonDefaults",
	},
	{	"Revert",			MAIN_GUI_METHOD_GET,
		0,	/* parameter */	MAIN_GUI_BUTTON_ACTIVE_IF_DIRTY,	/* flags */
		"cifsFileConfigButtonRevert",
	},
	{	"OK",				MAIN_GUI_METHOD_SET,
		0,	/* parameter */	MAIN_GUI_BUTTON_ACTIVE_IF_DIRTY,	/* flags */
		"cifsFileConfigButtonOk",
	},
};

static guiButton_t	loginListButtons[] = {
	{	"Login...",			MAIN_GUI_METHOD_LOGIN,
		0,	/* parameter */	0,	/* flags */
		"cifsFileLoginListButtonLogin",
	},
	{	"Logout",			MAIN_GUI_METHOD_LOGOUT,
		0,	/* parameter */	0,	/* flags */
		"cifsFileLoginListButtonLogout",
	},
	{	"Store",			MAIN_GUI_METHOD_STORELOGIN,
		0,	/* parameter */	0,	/* flags */
		"cifsFileLoginListButtonStore",
	},
	{	"Delete Stored",	MAIN_GUI_METHOD_DELSTORED,
		0,	/* parameter */	0,	/* flags */
		"cifsFileLoginListButtonDelStored",
	},
};

static guiObject_t	*cifsServerCfgDsource;
static guiObject_t	*cifsServerSslDsource = NULL;
static guiObject_t	*licenseDsource;

/* ------------------------------------------------------------------------- */

static void	marshalError(ipcMsg_t *reply, btint class, btint code)
{
btError	err;

	ipcAddInt(reply, class);
	ipcAddInt(reply, code);
	err.class = class;
	err.code = code;
	ipcAddAscii(reply, errDescription(&err));
}

/* ------------------------------------------------------------------------- */

typedef struct{
	PROC_VARSET;
	btint			uid;
	btint			index;
	ipcMsg_t		*reply;
	ipcMsg_t		*rq;
	btbool			listAll;
	hashtable_t		*servers;
	server_t		*server;
	share_t			*share;
	cifsUser_t		*user;
	ipcSrvClient_t	*client;
	btu8			*shrName;
	array_t			*clients;
}cifsCtx_t;

static blockpool_t	*cifsPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(cifsPool, block);
}

/* ------------------------------------------------------------------------- */

static btint	stringCompare(void *s1, void *s2){
	return utilStrCmp(s1, s2, 0);
}

static btu32	stringHashVal(void *s){
	return utilStrHash(s, 0);
}

static collectionObjectI_t	srvNameInterface = {
	stringCompare, stringHashVal, NULL, NULL
};

/* ------------------------------------------------------------------------- */

static void	getServerListProc(void *obj, void *arg)
{
cifsCtx_t		*this = obj;
cfgProperty_t	*p, *path[4];
btint			cnt, i, j, numEntries;
btu8			*domain;
lmbRes_t		*lrval = arg;
rapRes_t		*rrval = arg;
smbRequest_t	*smb = arg;
array_t			*all, *knownServers;
srvRes_t		*srval = arg;

	PROC_BEGIN;
	this->servers = htabNew(&srvNameInterface, 16, 75);
	/* First browse the servers. Do it before reading config and other arrays
	 * because we don't duplicate the strings and if browsing sleeps, the
	 * strings might be freed before we use them.
	 */
	 if(this->listAll){
	 	numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		domain = cfgStringValue(cfgQueryFromArray("domain", numEntries, path), "WORKGROUP");
		PROC_WAITFOR_X(1, lmbForDomain(domain, 0, this));
		if(lrval->err.class == 0){
			this->server = lrval->server;
			PROC_WAITFOR_X(2, srvConnect(this->server, this->uid, this));
			if(srval->err.class == 0 && shrConnectionStatus(srvIpcShare(this->server)) == PSL_CONNECTED){
				PROC_WAITFOR_X(3, srvWantRequest(this->server, this));
				numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
				domain = cfgStringValue(cfgQueryFromArray("domain", numEntries, path), "WORKGROUP");
				PROC_WAITFOR_X(4, rapEnumServers(smb, domain, 0, userSmbUid(srvNullSession(this->server)), shrTid(srvIpcShare(this->server)), this));
				if(rrval->err.class == 0 && (rrval->status == RAP_OK || rrval->status == RAP_ERR_MORE_DATA)){
					for(i=0;i<rrval->numEntries;i++){
						/* capitalize server names */
						for(j = 0; j < 15 && rrval->entries[i].name[j] != 0; j++){
							if(j == 0)
								rrval->entries[i].name[j] = asciiCharToupper(rrval->entries[i].name[j]);
							else
								rrval->entries[i].name[j] = asciiCharTolower(rrval->entries[i].name[j]);
						}
						rrval->entries[i].name[j] = 0;
		/* we need not copy the string because we won't sleep until marshalling */
						htabAddObject(this->servers, rrval->entries[i].name);
					}
				}
			}
			srvRelease(this->server);
		}
	 }
	for(j=0;j<2;j++){
		p = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(j ? gdata.dbFile : gdata.cfgFile));
		p = cfgQueryWithName("servers", 1, p);
		all = cfgQueryList(p);
		cnt = arrObjectCount(all);
		for(i=0;i<cnt;i++)
			htabAddObject(this->servers, arrObjectAtIndex(all, i));
		arrFree(all);
	}
	knownServers = NULL;
	if(this->listAll){
		/* get connected and remotely connected servers */
		knownServers = licAllServers();
		cnt = arrObjectCount(knownServers);
		for(i=0;i<cnt;i++)
			htabAddObject(this->servers, arrObjectAtIndex(knownServers, i));
		/* delay freeing of array because it also frees content strings */
		/* get browser servers */
		all = lmbAllServers();
		cnt = arrObjectCount(all);
		for(i=0;i<cnt;i++)
			htabAddObject(this->servers, srvName(arrObjectAtIndex(all, i)));
		arrFree(all);
	}
	all = htabAllObjects(this->servers);
	htabFree(this->servers);
	arrSort(all);
	PROC_RETURN_AND_CLEANUP(all,
		arrFree(all);
		if(knownServers != NULL)
			arrFree(knownServers);
	);
	PROC_END;
}

static void	getServerList(btbool listAll, btint uid, void *caller)
{
cifsCtx_t	*ctx;

	PROC_INIT(ctx, getServerListProc, caller, cifsPool, poolFreeBlock);
	ctx->listAll = listAll;
	ctx->uid = uid;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	listServersProc(void *obj, void *arg)
{
cifsCtx_t	*this = obj;
array_t		*rval = arg;
btint		cnt, i;

	PROC_BEGIN;
	PROC_WAITFOR(getServerList(this->listAll, this->uid, this));
	marshalError(this->reply, 0, 0);
	cnt = arrObjectCount(rval);
	ipcAddInt(this->reply, cnt);		/* number of entries */
	for(i=0;i<cnt;i++)
		ipcAddAscii(this->reply, arrObjectAtIndex(rval, i));
	PROC_END;
}

static void	listServers(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
cifsCtx_t	*ctx;

	PROC_INIT(ctx, listServersProc, caller, cifsPool, poolFreeBlock);
	ctx->uid = uid;
	ctx->reply = reply;
	ctx->listAll = ipcGetInt(rq);
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	addServer(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8			*name;
cfgProperty_t	*sect, *p;

	if(guimanAllowAccess(MAIN_GUI_PERM_AUTHORIZED, uid)){
		name = ipcGetAscii(rq);
		sect = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.dbFile));
		if(sect == NULL){
			sect = cfgBuildDictProperty();
			cfgSetProperty(CIFS_CONFIG_SECTION, cfgFilePlist(gdata.dbFile), sect);
		}
		p = cfgQueryWithName("servers", 1, sect);
		if(p == NULL){
			p = cfgBuildDictProperty();
			cfgSetProperty("servers", sect, p);
		}
		sect = p;
		p = cfgQueryWithName(name, 1, sect);
		if(p == NULL){
			p = cfgBuildDictProperty();
			cfgSetProperty(name, sect, p);
		}
		marshalError(reply, 0, 0);
		cfgUnparse(gdata.dbFile);
		mainModGuiUpdate(cifsServerCfgDsource);
		if(cifsServerSslDsource != NULL)
			mainModGuiUpdate(cifsServerSslDsource);
	}else{
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnotauth);
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	removeServer(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8			*name;
cfgProperty_t	*p;

	if(guimanAllowAccess(MAIN_GUI_PERM_AUTHORIZED, uid)){
		name = ipcGetAscii(rq);
		p = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.dbFile));
		p = cfgQueryWithName("servers", 1, p);
		if(p != NULL){
			cfgSetProperty(name, p, NULL);
		}
		marshalError(reply, 0, 0);
		cfgUnparse(gdata.dbFile);
		mainModGuiUpdate(cifsServerCfgDsource);
		if(cifsServerSslDsource != NULL)
			mainModGuiUpdate(cifsServerSslDsource);
	}else{
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnotauth);
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}


/* ------------------------------------------------------------------------- */

typedef struct{
	btu8	*serverName;
	btint	unixUid;
	btuni	*cifsName;
	btu8	*domain;
	btbool	isConnected;
	btbool	isStored;
}loginRecord_t;

static void	addLoginRecord(array_t *arr, btu8 *serverName, btint unixUid, btuni *cifsName, btu8 *domain, btbool isConnected, btbool isStored)
{
btint			cnt, i;
loginRecord_t	*rec;

	if(serverName == NULL || cifsName == NULL)
		return;	/* ignore invalid records */
	cnt = arrObjectCount(arr);
	for(i=0;i<cnt;i++){
		rec = arrObjectAtIndex(arr, i);
		if(utilStrCmp(rec->serverName, serverName, 0) == 0 && rec->unixUid == unixUid && uniStrCmp(rec->cifsName, cifsName, 0) == 0){
			rec->isStored |= isStored;
			rec->isConnected |= isConnected;
			if(rec->domain == NULL || isConnected){
				if(rec->domain != NULL)
					memFree(rec->domain);
				rec->domain = utilStrDup(domain);
			}
			return;
		}
	}
	rec = memCalloc(sizeof(loginRecord_t));
	rec->serverName = serverName;
	rec->unixUid = unixUid;
	rec->cifsName = cifsName;
	rec->domain = utilStrDup(domain);
	rec->isConnected = isConnected;
	rec->isStored = isStored;
	arrAddObject(arr, rec);
}

static btint	loginRecordCompare(void *obj, void *otherObj)
{
loginRecord_t	*this = obj, *other = otherObj;
btint			rval;

	rval = this->unixUid - other->unixUid;
	if(rval != 0)
		return rval;
	rval = utilStrCmp(this->serverName, other->serverName, 0);
	if(rval != 0)
		return rval;
	return uniStrCmp(this->cifsName, other->cifsName, 0);
}

static collectionObjectI_t	loginRecordInterface = {
	loginRecordCompare,	/* needed for sorting */
	NULL,
	NULL,
	memFree
};

static void	loginListGet(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
array_t			*records, *servers, *users;
server_t		*server;
cifsUser_t		*user;
btint			i, cnt, j, cnt1, k;
userRecord_t	*urec;
loginRecord_t	*rec;
btu8			buffer[512];
struct passwd	*pw;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "cifsmod: loginListGet(%d)\n", uid);
	records = arrNew(&loginRecordInterface, 8);
	/* users logged in at mounted servers and browser servers */
	for(k=0;k<2;k++){
		servers = k == 0 ? licOurServers() : lmbAllServers();
		cnt = arrObjectCount(servers);
		for(i=0;i<cnt;i++){
			server = arrObjectAtIndex(servers, i);
			users = srvAllUsers(server);
			cnt1 = arrObjectCount(users);
			for(j=0;j<cnt1;j++){
				user = arrObjectAtIndex(users, j);
				addLoginRecord(records, srvName(server), userUnixUid(user), userCifsName(user), userDomain(user), 1, 0);
			}
			arrFree(users);
		}
		arrFree(servers);
	}
	/* add stored users */
	users = udbAllRecords(cifsUserDb);
	cnt = arrObjectCount(users);
	for(i=0;i<cnt;i++){
		urec = arrObjectAtIndex(users, i);
		addLoginRecord(records, udbrecServer(urec), udbrecUnixUid(urec), udbrecCifsUser(urec), udbrecDomain(urec), 0, 1);
	}
	arrFree(users);
	arrSort(records);
	/* construct reply message */
	marshalError(reply, 0, 0);
	cnt = arrObjectCount(records);
	ipcAddInt(reply, cnt);
	for(i=0;i<cnt;i++){
		rec = arrObjectAtIndex(records, i);
		sprintf(buffer, "%d ", rec->unixUid);
		j = strlen(buffer);
		strncpy(buffer + j, rec->serverName, ELEMENTS(buffer) - j);
		buffer[ELEMENTS(buffer)-1] = 0;	/* always 0-terminate */
		ipcAddAscii(reply, buffer);	/* record ID = unix-UID + server name */
		ipcAddAscii(reply, rec->serverName);	/* 1st column: server */
		pw = getpwuid(rec->unixUid);			/* 2nd column: local user */
		if(pw != NULL){
			ipcAddAscii(reply, pw->pw_name);
		}else{
			sprintf(buffer, "#%d", rec->unixUid);
			ipcAddAscii(reply, buffer);
		}
		convUniToLocalAscii(rec->cifsName, buffer, sizeof(buffer));
		ipcAddAscii(reply, buffer);				/* 3rd column: remote user */
		ipcAddAscii(reply, rec->domain == NULL ? (btu8 *)"" : rec->domain);		/* 4th column: domain */
		buffer[0] = 0;
		if(rec->isConnected)
			strcat(buffer, "L");
		if(rec->isStored)
			strcat(buffer, "S");
		ipcAddAscii(reply, buffer);
	}
	arrFree(records);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static hashtable_t	*loginProcessesRunning = NULL;

static void	loginListLoginProc(void *obj, void *arg)
{
cifsCtx_t		*this = obj;
array_t			*srvList = arg;
guimanRes_t		*rval = arg;

	PROC_BEGIN;
	ipcSrvClientRetain(this->client);
/* allow sending of reply to GUI now, we don't want to reverse call orders */
	PROC_WAITFOR(elpCallCancelPrev(loginListLoginProc, this));
	if(loginProcessesRunning == NULL)
		loginProcessesRunning = htabNew(NULL, 4, 75);
/* Ignore duplicate clicks while the server list is retrieved. If the user
 * wants an other login selection dialog after the first one appeared, allow
 * it. The user may have lost the first behind other windows.
 */
	if(htabFindObjectEqualTo(loginProcessesRunning, this->client) == NULL){
		htabAddObject(loginProcessesRunning, this->client);
		PROC_WAITFOR(getServerList(1, this->uid, this));
		htabRemoveObjectEqualTo(loginProcessesRunning, this->client);
		PROC_WAITFOR(guimanRunSelection(this->client, "Login at Server", "Login", "Cancel", srvList, this));
		if(rval->err.class == 0 && rval->didSet){
			this->server = srvWithName(rval->file, 0);
			/* don't set gid in the following call because it's ignored anyway */
			PROC_WAITFOR(userAuthenticate(this->server, NULL, this->uid, 0, this));
			srvRelease(this->server);	/* now retained by user */
#if 0	/* errors are already displayed by userAuthenticate() */
			if(urval->err.class != 0)
				guimanDisplayError(this->client, &urval->err, NULL);
#endif
		}
	}
	ipcSrvClientRelease(this->client);
	PROC_END;
}

static void	loginListLogin(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
cifsCtx_t	*ctx;

	PROC_INIT(ctx, loginListLoginProc, NULL, cifsPool, poolFreeBlock);
	ctx->client = client;
	ctx->uid = uid;
	PROC_START(ctx, NULL);	/* start background operation, send reply immediately */
	marshalError(reply, 0, 0);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	loginListLogoutProc(void *obj, void *arg)
{
cifsCtx_t	*this = obj;
btint		parameter;
btu8		*rowId, *server;
btint		userUid;
server_t	*srv;
userRes_t	*rval = arg;

	PROC_BEGIN;
	parameter = ipcGetInt(this->rq);
	rowId = ipcGetAscii(this->rq);	/* consists of Unix uid + ' ' + server name */
	if(rowId == NULL){	/* no row selected */
		marshalError(this->reply, 0, 0);	/* just ignore action */
		PROC_RETURN(NULL);
	}
	userUid = atoi(rowId);
	if(this->uid != 0 && userUid != this->uid){	/* users may log out themselves, root may log out others */
		marshalError(this->reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOCotherlogin);
		PROC_RETURN(NULL);
	}
	server = strchr(rowId, ' ');
	if(server == NULL){
		marshalError(this->reply, ERRCLS_FRAMEWORK, ERR_FRWnosuchobj);
	}else{
		server++;
		srv = licLookupServer(server);
		if(srv == NULL){
			srv = lmbLookupServer(server);
		}
		if(srv == NULL){
			marshalError(this->reply, 0, 0);	/* simply ignore invalid logouts */
		}else{
			this->user = srvUserForUid(srv, userUid);
			if(this->user == NULL){
				marshalError(this->reply, 0, 0);	/* simply ignore invalid logouts */
			}else{
				PROC_WAITFOR(userDisconnect(this->user, this));
				if(rval->err.class != 0){
					marshalError(this->reply, rval->err.class, rval->err.code);
				}else{
					userInvalidate(this->user);
					marshalError(this->reply, 0, 0);
				}
			}
		}
	}
	mainModGuiUpdate(cifsLoginListObject);
	PROC_END;
}

static void	loginListLogout(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
cifsCtx_t	*ctx;

	PROC_INIT(ctx, loginListLogoutProc, caller, cifsPool, poolFreeBlock);
	ctx->uid = uid;
	ctx->reply = reply;
	ctx->rq = rq;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static btint	storeLogin(btint uid, btu8 *server);
static btint	delStoredLogin(btint uid, btu8 *server);

static void	loginListStore(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btint		parameter;
btu8		*rowId, *server;
btint		userUid, rval;

	parameter = ipcGetInt(rq);
	rowId = ipcGetAscii(rq);	/* consists of Unix uid + ' ' + server name */
	if(rowId == NULL){	/* no row selected */
		marshalError(reply, 0, 0);	/* ignore action */
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	userUid = atoi(rowId);
	if(uid != 0 && userUid != uid){	/* users may store themselves, root may store others */
		marshalError(reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOCotherlogin);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	server = strchr(rowId, ' ');
	if(server == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnosuchobj);
		mainModGuiUpdate(cifsLoginListObject);
	}else{
		server++;
		rval = storeLogin(userUid, server);
		marshalError(reply, rval == 0 ? 0 : ERRCLS_FRAMEWORK, rval);
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	loginListDelStored(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btint		parameter;
btu8		*rowId, *server;
btint		userUid, rval;

	parameter = ipcGetInt(rq);
	rowId = ipcGetAscii(rq);	/* consists of Unix uid + ' ' + server name */
	if(rowId == NULL){	/* no row selected */
		marshalError(reply, 0, 0);	/* ignore action */
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	userUid = atoi(rowId);
	if(uid != 0 && userUid != uid){	/* users may delete themselves, root may delete others */
		marshalError(reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOCotherlogin);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	server = strchr(rowId, ' ');
	if(server == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnosuchobj);
		mainModGuiUpdate(cifsLoginListObject);
	}else{
		server++;
		rval = delStoredLogin(userUid, server);
		marshalError(reply, rval == 0 ? 0 : ERRCLS_FRAMEWORK, rval);
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */


static guiClass_t	globalCfg = {
	MAIN_GUI_CLASS_PLAINCFG,
	"CIFS\nGeneral",
	NULL,
	NULL
};

static guiClass_t	serverCfg = {
	MAIN_GUI_CLASS_HIERARCHYCFG,
	"CIFS\nServers",
	NULL,
	NULL
};

#if defined USE_SSL
static guiClass_t	serverSslCfg = {
	MAIN_GUI_CLASS_HIERARCHYCFG,
	"CIFS\nSSL",
	NULL,
	NULL
};
#endif

static guiClass_t	licenseCfg = {
	MAIN_GUI_CLASS_PLAINCFG,
	"CIFS\nLicense",
	NULL,
	NULL
};

static guiDsourceMethod_t	loginListDsourceMethods[] = {
	loginListGet,
	loginListLogin,
	loginListLogout,
	loginListStore,
	loginListDelStored,
};

static guiClass_t	loginList = {
	MAIN_GUI_CLASS_LIST,
	"CIFS\nLogins",
	NULL,
	NULL
};

guiObject_t	*cifsLoginListObject;
static guiObject_t	*cifsGlobalCfgDsource;

static array_t	*frontendModuleNames;

/* ------------------------------------------------------------------------- */

static btu32	limits[] = {
	0, 1, 2, 5, 10, 20, 50, 100,
	200, 500, 1, 2, 5, 10, 20, 32000
};

static btu8	*ltypes[] = {
	"Demo",
	"Developer",
	"Student",
	"Educational Institution",
	"Business",
	"Individual",
	"Unknown",
	"Unknown",
	"Beta-Tester",
	"Non Profit Organisation"
};

static void	getMaxValues(btu8 lstring[8], btu32 *maxServers, btu32 *maxClients, btbool *isLocal)
{
btint	x;

	*isLocal = 0;
	if(lstring[0] != 1){
		x = lstring[3] / 16;
		if(x != 15){
			*maxServers = limits[x];
			if(x > 9)
				*isLocal = 1;
		}else{
			*maxServers = -1;
		}
		x = lstring[3] % 16;
		if(x != 15){
			*maxClients = limits[x];
		}else{
			*maxClients = -1;
		}
	}else{
		*maxServers = 255;
		*maxClients = lstring[3];
	}
}

static void	getLicenseValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btbool		getDefaults;
array_t		*entries;
btint		i, j;
guiEntry_t	*entry;
btu8		lstring[8], licenseId[20], *lname, buffer[64], number[20];
btbool		scountIsLocal = 0;
btu32		maxServers, maxClients;
bti16		expiry, ltype;
time_t		expDate;
struct tm	*tmp;

	marshalError(reply, 0, 0);
	getDefaults = ipcGetInt(rq);
	entries = licenseCfg.entries;
	licKeyIdentifier(lstring);
	for(j=i=0;i<8;i++){
		sprintf(licenseId + j, "%02X", lstring[i]);
		j += 2;
		if(i == 0 || i == 3)
			licenseId[j++] = '-';
	}
	ltype = lstring[4] / 16;
	if(ltype < 0 || ltype > sizeof(ltypes)/sizeof(ltypes[0])){
		lname = "Unknown";
	}else{
		lname = ltypes[ltype];
	}
	getMaxValues(lstring, &maxServers, &maxClients, &scountIsLocal);
	expiry = ((btu16)lstring[1] << 8) | lstring[2];

	for(i=0;i<arrObjectCount(entries);i++){
		entry = arrObjectAtIndex(entries, i);
		switch(i){
		case 0:	ipcAddAscii(reply, "");			break;
		case 1:	ipcAddAscii(reply, licenseId);	break;
		case 2:	ipcAddAscii(reply, lname);		break;
		case 3:
			if((btint)maxClients < 0)
				sprintf(buffer, "unlimited");
			else
				sprintf(buffer, "%d", maxClients);
			ipcAddAscii(reply, buffer);
			break;
		case 4:
			if((btint)maxServers < 0)
				sprintf(number, "unlimited");
			else
				sprintf(number, "%d", maxServers);
			if(scountIsLocal)
				sprintf(buffer, "%s (locally counted)", number);
			else
				sprintf(buffer, "%s", number);
			ipcAddAscii(reply, buffer);
			break;
		case 5:
			if(expiry < 0){
				sprintf(buffer, "%d minutes", -expiry);
			}else if(expiry > 0){
				expDate = 631148400 + expiry * 86400;
				tmp = localtime(&expDate);
				sprintf(buffer, "%04d-%02d-%02d", tmp->tm_year + 1900, tmp->tm_mon + 1, tmp->tm_mday);
			}else{
				sprintf(buffer, "never");
			}
			ipcAddAscii(reply, buffer);
			break;
		}
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	setLicenseValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btbool	isDirty;
btu8	*lstring, *p;
btu8	ldata[12], buffer[24];
btint	i, l;
array_t	*allServers;
btsize	srvCount;

	isDirty = ipcGetInt(rq);
	p = lstring = ipcGetAscii(rq);
	/* there's more to unmarshal, but we don't care */
	if(isDirty){
		if(!guimanAllowAccess(MAIN_GUI_PERM_AUTHORIZED, uid)){
			marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnotauth);
			goto doReturn;
		}
		allServers = licOurServers();
		srvCount = arrObjectCount(allServers);
		arrFree(allServers);
		if(srvCount > 0){
			marshalError(reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOCsrvconn);
			PROC_IMMEDIATE_RET(caller, NULL);
		}
		for(i=0;i<24;p++){
			if(*p == 0){
				marshalError(reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOClicense);
				goto doReturn;
			}
			if(!isalnum(*p))	/* skip any non-alphanumeric chars */
				continue;
			if(!isxdigit(*p)){
				marshalError(reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOClicense);
				goto doReturn;
			}
			buffer[i++] = *p;
		}
		for(i=0;i<12;i++){
			sscanf(&buffer[2 * i], "%02x", &l);
			ldata[i] = l;
		}
		if(licSetKey(ldata)){
			udbSetLicenseString(cifsUserDb, ldata);
			marshalError(reply, 0, 0);
		}else{
			marshalError(reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOClicense);
		}
	}
doReturn:
	mainModGuiUpdate(licenseDsource);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	getGlobalValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	*cfgPath[] = {CIFS_CONFIG_SECTION, NULL};
btu8	**cfgPathes[2]; /* = {cfgPath, NULL}; */

	cfgPathes[0] = cfgPath;
	cfgPathes[1] = NULL;
	guimanGetValues(cfgPathes, NULL, globalCfg.entries, rq, reply);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	setGlobalValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	*path[] = {CIFS_CONFIG_SECTION, NULL};

	guimanSetValues(0, path, globalCfg.entries, rq, reply, uid);
	mainModGuiUpdate(cifsGlobalCfgDsource);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	getServerValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	**cfgPathes[] = {NULL, NULL};
btu8	*hPath[] = {CIFS_CONFIG_SECTION, "servers", NULL};
btu8	**hPathes[2]; /* = {hPath, NULL}; */

	hPathes[0] = hPath;
	hPathes[1] = NULL;
	/* NULL terminate data section of frontendModuleNames array: */
	arrAddObject(frontendModuleNames, NULL);
	arrRemoveLastObject(frontendModuleNames);
	cfgPathes[0] = (btu8 **)arrContentArray(frontendModuleNames);
	guimanGetValues(cfgPathes, hPathes, serverCfg.entries, rq, reply);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	setServerValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	*path[] = {CIFS_CONFIG_SECTION, "servers", NULL};

	guimanSetValues(1, path, serverCfg.entries, rq, reply, uid);
	mainModGuiUpdate(cifsServerCfgDsource);
	PROC_IMMEDIATE_RET(caller, NULL);
}

#if defined USE_SSL
static void	getServerSslValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	**cfgPathes[] = {NULL, NULL};
btu8	*hPath[] = {CIFS_CONFIG_SECTION, "servers", NULL};
btu8	**hPathes[] = {hPath, NULL};

	/* NULL terminate data section of frontendModuleNames array: */
	arrAddObject(frontendModuleNames, NULL);
	arrRemoveLastObject(frontendModuleNames);
	cfgPathes[0] = (btu8 **)arrContentArray(frontendModuleNames);
	guimanGetValues(cfgPathes, hPathes, serverSslCfg.entries, rq, reply);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	setServerSslValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	*path[] = {CIFS_CONFIG_SECTION, "servers", NULL};

	guimanSetValues(1, path, serverSslCfg.entries, rq, reply, uid);
	mainModGuiUpdate(cifsServerCfgDsource);
	PROC_IMMEDIATE_RET(caller, NULL);
}
#endif

/* ------------------------------------------------------------------------- */

static guiDsourceMethod_t	globalDsourceMethods[] = {
	getGlobalValues,
	setGlobalValues,
};

static guiDsourceMethod_t	serverDsourceMethods[] = {
	getServerValues,
	setServerValues,
	listServers,
	addServer,
	removeServer,
};

#if defined USE_SSL
static guiDsourceMethod_t	serverSslDsourceMethods[] = {
	getServerSslValues,
	setServerSslValues,
	listServers,
	addServer,
	removeServer,
};
#endif

static guiDsourceMethod_t	licenseDsourceMethods[] = {
	getLicenseValues,
	setLicenseValues,
};

/* ------------------------------------------------------------------------- */
/* ---------------------------- stub functions: ---------------------------- */
/* ------------------------------------------------------------------------- */

static inline char	*ps(char *s)
{
	return s == NULL ? "(null string)" : s;
}

static void	stubUserLoginProc(void *obj, void *arg)
{
cifsCtx_t		*this = obj;
userRes_t		*urval = arg;
guimanRes_t		*grval = arg;
struct passwd	*pwent;
btint			gid, i;
ipcSrvClient_t	*client;
btError			res;
btuni			uncPath[256];

	PROC_BEGIN;
	gid = 0;
	if((pwent = getpwuid(this->uid)) != NULL)
		gid = pwent->pw_gid;
	PROC_WAITFOR(userAuthenticate(this->server, NULL, this->uid, gid, this));
	if(urval->err.class != 0)
		goto errorOccured;
	this->user = urval->user;
	userRetain(this->user);
	if(!srvIsUserlevelSecurity(this->server)){
		if(this->shrName != NULL && this->shrName[0] == 0)
			this->shrName = NULL;	/* treat empty name as NULL name */
		if(this->shrName == NULL){
			this->clients = ipcClientServicesForUid(mainModule->ipcModule, this->uid);
			for(this->index=0;this->index<arrObjectCount(this->clients);this->index++){
				client = arrObjectAtIndex(this->clients, this->index);
				if(ipcSrvClientIsInvalid(client))
					continue;
				PROC_WAITFOR(guimanAskData(client, "Login to Share: ", 1, this));
				if(!(grval->err.class == ERRCLS_FRAMEWORK && grval->err.code == ERR_FRWipcnotimpl)){
					if(grval->err.class == 0)
						this->shrName = grval->file;
					break;
				}
			}
			arrFree(this->clients);
		}
		if(this->shrName == NULL){	/* we still don't know the share */
			res.class = ERRCLS_CIFS_LOCAL;
			res.code = ERR_CIFS_LOCnosharegiven;
			urval = (void *)&res;
			goto errorOccured;
		}
		uncPath[0] = uncPath[1] = UNI_CHAR_BACKSLASH;
		convLocalAsciiToUni(srvName(this->server), &uncPath[2], ELEMENTS(uncPath)-3);
		i = uniStrLen(uncPath);
		uncPath[i++] = UNI_CHAR_BACKSLASH;
		convLocalAsciiToUni(this->shrName, &uncPath[i], ELEMENTS(uncPath)-i);
		this->share = shrForPath(uncPath);
		PROC_WAITFOR(shrConnect(this->share, this->user, this));
	}
errorOccured:
	srvRelease(this->server);	/* now retained by user */
	if(this->share != NULL)
		shrRelease(this->share);
	if(this->user != NULL)
		userRelease(this->user);
	marshalError(this->reply, urval->err.class, urval->err.code);
	PROC_END;
}

static void	stubUserLogin(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
cifsCtx_t	*ctx;
btu8		*srvName, *shrName;

	srvName = ipcGetAscii(rq);
	shrName = ipcGetAscii(rq);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "cifsmod: stubUserLogin(%s, %s)\n", ps(srvName), ps(shrName));
	if(srvName == NULL){	/* no server given */
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWipcparameters);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	PROC_INIT(ctx, stubUserLoginProc, caller, cifsPool, poolFreeBlock);
	ctx->uid = uid;
	ctx->reply = reply;
	ctx->server = srvWithName(srvName, 0);
	ctx->shrName = shrName;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	stubUserLogoutProc(void *obj, void *arg)
{
cifsCtx_t	*this = obj;
server_t	*srv;
userRes_t	*rval = arg;
btu8		*srvName = arg;

	PROC_BEGIN;
	if(srvName == NULL){	/* no server specified */
		marshalError(this->reply, ERRCLS_FRAMEWORK, ERR_FRWipcparameters);
		PROC_RETURN(NULL);
	}
	srv = licLookupServer(srvName);
	if(srv == NULL){
		srv = lmbLookupServer(srvName);
	}
	if(srv == NULL){
		marshalError(this->reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOCsrvnotconn);
	}else{
		this->user = srvUserForUid(srv, this->uid);
		if(this->user == NULL){
			marshalError(this->reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOCusrnotconn);
		}else{
			PROC_WAITFOR(userDisconnect(this->user, this));
			if(rval->err.class != 0){
				marshalError(this->reply, rval->err.class, rval->err.code);
			}else{
				userInvalidate(this->user);
				marshalError(this->reply, 0, 0);
			}
		}
	}
	mainModGuiUpdate(cifsLoginListObject);
	PROC_END;
}

static void	stubUserLogout(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
cifsCtx_t	*ctx;
btu8		*srvName;

	srvName = ipcGetAscii(rq);
	PROC_INIT(ctx, stubUserLogoutProc, caller, cifsPool, poolFreeBlock);
	ctx->uid = uid;
	ctx->reply = reply;
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "cifsmod: stubUserLogout(%s)\n", ps(srvName));
	PROC_START(ctx, srvName);
}

/* ------------------------------------------------------------------------- */

static void	stubGetLicense(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8		lstring[8], *lname, buffer[64];
btbool		scountIsLocal = 0;
btu32		maxServers, maxClients;
bti16		expiry, ltype;
time_t		expDate;
struct tm	*tmp;

	licKeyIdentifier(lstring);
	marshalError(reply, 0, 0);
	ipcAddBlock(reply, lstring, 8);
	
	ltype = lstring[4] / 16;
	if(ltype < 0 || ltype >= ELEMENTS(ltypes)){
		lname = "Unknown";
	}else{
		lname = ltypes[ltype];
	}
	ipcAddAscii(reply, lname);	/* license type in ASCII */
	getMaxValues(lstring, &maxServers, &maxClients, &scountIsLocal);
	ipcAddInt(reply, maxClients);
	ipcAddInt(reply, maxServers);
	ipcAddInt(reply, scountIsLocal);
	expiry = ((btu16)lstring[1] << 8) | lstring[2];
	if(expiry < 0){
		sprintf(buffer, "%d minutes", -expiry);
	}else if(expiry > 0){
		expDate = 631148400 + expiry * 86400;
		tmp = localtime(&expDate);
		sprintf(buffer, "%04d-%02d-%02d", tmp->tm_year + 1900, tmp->tm_mon + 1, tmp->tm_mday);
	}else{
		sprintf(buffer, "never");
	}
	ipcAddAscii(reply, buffer);	/* expiry in ASCII */
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	stubCurrentServers(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
array_t			*clients;
btint			i, j, count, scount;
licClientInfo_t	*c;

	marshalError(reply, 0, 0);
	clients = licAllClients();
	count = arrObjectCount(clients);
	ipcAddInt(reply, count);
	for(i=0;i<count;i++){
		c = arrObjectAtIndex(clients, i);
		ipcAddInt(reply, c->ipAddress);
		ipcAddAscii(reply, c->clName);
		scount = arrObjectCount(c->clServers);
		ipcAddInt(reply, scount);
		for(j=0;j<scount;j++){
			ipcAddAscii(reply, arrObjectAtIndex(c->clServers, j));
		}
	}
	arrFree(clients);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	stubLicenseLoad(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
	marshalError(reply, 0, 0);
	ipcAddInt(reply, licNumClients());
	ipcAddInt(reply, licNumServers());
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	stubSetLicense(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8	*licenseString;
array_t	*allServers;
btsize	srvCount;

	if(!guimanAllowAccess(MAIN_GUI_PERM_AUTHORIZED, uid)){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnotauth);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	if(ipcGetBlock(rq, &licenseString) != 12){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWipcparameters);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	allServers = licOurServers();
	srvCount = arrObjectCount(allServers);
	arrFree(allServers);
	if(srvCount > 0){
		marshalError(reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOCsrvconn);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	if(licSetKey(licenseString)){
		udbSetLicenseString(cifsUserDb, licenseString);
		mainModGuiUpdate(licenseDsource);
		marshalError(reply, 0, 0);
	}else{
		marshalError(reply, ERRCLS_CIFS_LOCAL, ERR_CIFS_LOClicense);
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static btint	storeLogin(btint uid, btu8 *server)
{
server_t		*srv;
cifsUser_t		*user;
cfgProperty_t	*path[8];
btint			numElements;
btbool			usersMayStore;

	numElements = mainModCfgPath(path, ELEMENTS(path));
	usersMayStore = cfgEnumValue(cfgQueryFromArray("usersMayStoreSessionData", numElements, path), cfgYesNoEnum, 1);
	if(uid != 0 && !usersMayStore)
		return ERR_FRWstoredisabled;
	srv = licLookupServer(server);
	if(srv == NULL)
		srv = lmbLookupServer(server);
	if(srv == NULL)
		return ERR_FRWnosuchobj;
	user = srvUserForUid(srv, uid);
	if(user == NULL)
		return ERR_FRWnosuchobj;
	udbAddUser(cifsUserDb, server, user);
	mainModGuiUpdate(cifsLoginListObject);
	return 0;
}

static btint	delStoredLogin(btint uid, btu8 *server)
{
cfgProperty_t	*path[8];
btint			numElements;
btbool			usersMayStore;

	numElements = mainModCfgPath(path, ELEMENTS(path));
	usersMayStore = cfgEnumValue(cfgQueryFromArray("usersMayStoreSessionData", numElements, path), cfgYesNoEnum, 1);
	if(uid != 0 && !usersMayStore)
		return ERR_FRWstoredisabled;
	udbRemoveUser(cifsUserDb, uid, server);
	mainModGuiUpdate(cifsLoginListObject);
	return 0;
}

static void	stubStoreLogin(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8	*server;
btbool	doDelete;
btint	rval;

	server = ipcGetAscii(rq);
	doDelete = ipcGetInt(rq);
	if(server == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWipcparameters);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	if(doDelete)
		rval = delStoredLogin(uid, server);
	else
		rval = storeLogin(uid, server);
	marshalError(reply, (rval != 0 ? ERRCLS_FRAMEWORK : 0), rval);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	uiWillTerminateProc(void *obj, void *arg)
{
cifsCtx_t	*this = obj;
array_t		*servers;
btint		i, cnt;
server_t	*srv;
guimanRes_t	*rval = arg;
btbool		doTerm, doLogout;
cifsUser_t	*user;

	PROC_BEGIN;
	servers = licOurServers();
	cnt = arrObjectCount(servers);
	for(i=0;i<cnt;i++){
		srv = arrObjectAtIndex(servers, i);
		if(srvUserForUid(srv, this->client->uid) != NULL)
			break;	/* user is logged in somewhere */
	}
	arrFree(servers);
	if(i < cnt){	/* if we are logged in somewhere */
		if(this->uid == MAIN_TERM_USRQUIT){
			PROC_WAITFOR(guimanRunAlert(this->client, "Log Out?", "Do you want to log out from all servers?", "Don\'t Quit", "No", "Yes", this));
			if(rval->err.class != 0){
				doTerm = doLogout = 1;
			}else{
				switch(rval->button){
					case 3:		doTerm = doLogout = 1;		break;
					case 2:		doTerm = 1; doLogout = 0;	break;
					default:	doTerm = doLogout = 0;		break;
				}
			}
		}else{
			doTerm = doLogout = 1;
		}
		if(doLogout){
			servers = licOurServers();
			cnt = arrObjectCount(servers);
			for(i=0;i<cnt;i++){
				srv = arrObjectAtIndex(servers, i);
				if((user = srvUserForUid(srv, this->client->uid)) != NULL){
					userDisconnect(user, NULL);
					userInvalidate(user);
				}
			}
			arrFree(servers);
		}
		PROC_RETURN((void *)(int)doTerm);
	}
	PROC_RETURN((void *)1);
	PROC_END;
}

static void	uiWillTerminate(struct ipcSrvClient *client, btint termReason, void *caller)
{
cifsCtx_t	*ctx;

	PROC_INIT(ctx, uiWillTerminateProc, caller, cifsPool, poolFreeBlock);
	ctx->uid = termReason;
	ctx->client = client;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */
/* --------------------------- module functions: --------------------------- */
/* ------------------------------------------------------------------------- */

ipcSrvStub_t	cifsFileIpcStubs[] = {
	stubUserLogin,
	stubUserLogout,
	stubGetLicense,
	stubCurrentServers,
	stubLicenseLoad,
	stubSetLicense,
	stubStoreLogin,
};

static struct{
	btu8	*name;
	void	*function;
} functionTable[] = {
	{BE_CONFIG_PATH,			cifsShareConfigPath},
	{BE_USER_AUTHENTICATE,		cifsUserAuthenticate},
	{BE_USER_RETAIN,			userRetain},
	{BE_USER_RELEASE,			userRelease},

	{BE_VFS_FOR_PATH,			shrForPath},
	{BE_VFS_RETAIN,				shrRetain},
	{BE_VFS_RELEASE,			shrRelease},
	{BE_VFS_CONNECT,			shrConnect},
	{BE_VFS_DISCONNECT,			shrDisconnect},
	{BE_VFS_ROOT_FILE,			cifsShareRootFile},
	{BE_VFS_STATFS,				shrGetAllocationInfo},
	{BE_VFS_PATH,				shrPath},
	{BE_VFS_CASE_SENSITIVE,		cifsShareIsCaseSensitive},
	{BE_VFS_OPEN_NOTIFICATION,	shrOpenNc},
	{BE_VFS_INVALIDATE_NOTIFICATION,	shrInvalidateNc},

	{BE_FILE_LOOKUP,			fileLookupChildWithName},
	{BE_FILE_NAME,				fileName},
	{BE_FILE_PARENT,			fileParent},
	{BE_FILE_RETAIN,			fileRetain},
	{BE_FILE_RELEASE,			fileRelease},
	{BE_FILE_GETATTR,			fileGetAttributes},
	{BE_FILE_SETATTR,			fileSetAttributes},
	{BE_FILE_READDIR,			fileReadDir},
	{BE_FILE_CREATE,			cifsFileCreate},
	{BE_FILE_REMOVE,			fileRemoveChild},
	{BE_FILE_RENAME,			fileRename},
	{BE_FILE_RW,				fileRW},
	{BE_FILE_HARDLINK,			fileHardLink},
	{BE_FILE_SYMLINK,			fileSymbolicLink},
	{BE_FILE_READLINK,			fileReadLink},
	{BE_FILE_OPENMODE,			fileOpenMode},
	{BE_FILE_SET_OPENMODE,		cifsFileSetOpenMode},
	{BE_FILE_OP_RUNNING,		cifsFileOperationsRunning},
};

static void	*getImp(btu8 *functionName)
{
btint	i;

	for(i=0;i<sizeof(functionTable)/sizeof(functionTable[0]);i++){
		if(strcmp(functionName, functionTable[i].name) == 0)
			return functionTable[i].function;
	}
	return NULL;
}

/* ------------------------------------------------------------------------- */

static void	moduleAdded(void *dummy, void *module)
{
module_t	*mod = module;
array_t		*arr, *(*feCfg)(void);
btint		i;

	if(mod->moduleType == MOD_TYPE_FRONTEND){
		arrAddObject(frontendModuleNames, mod->name);
		feCfg = mod->implementationOf(FE_CFGENTRIES_FOR_BACKEND);
		if(feCfg != NULL){
			arr = feCfg();
			for(i=0;i<arrObjectCount(arr);i++){
				arrAddObject(serverCfg.entries, arrObjectAtIndex(arr, i));
			}
			arrFree(arr);
		}
	}
}

/* ------------------------------------------------------------------------- */

static bttime	userdbParseInterval = 0;

static void	checkUserdb(void *obj, void *arg)
{
	if(cifsUserDb != NULL){
		if(udbParseIfChanged(cifsUserDb)){
			mainModGuiUpdate(NULL);
		}
	}
	if(userdbParseInterval != 0)
		elpCallAtCancelPrev(elpNow() + userdbParseInterval, checkUserdb, NULL);
}

/* ------------------------------------------------------------------------- */

void	cifsFile_init(void)
{
module_t	*mod, *m;
btu8		*udbPath;
btint		i;
array_t		*modules;

	cifsPool = poolNew(sizeof(cifsCtx_t), sizeof(cifsCtx_t));
	frontendModuleNames = arrNew(NULL, 8);
	udbPath = cfgStringValue(cfgQueryWithName("databaseFile", 1, cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.cfgFile))), CONST_USERDB_PATH);
	cifsUserDb = udbNew(udbPath);
	licInit();
	licSetKey(udbLicenseString(cifsUserDb));	/* must not be done when shares are connected! */
	mod = memCalloc(sizeof(*mod));
	mod->name = CIFSFILE_MODULE_NAME;
	mod->moduleType = MOD_TYPE_BACKEND;
	mod->implementationOf = getImp;
	mod->interfaceVersion = 0;
	mod->implementationVersion = MAINCONST_VERSION;
	mod->betaVersion = MAINCONST_BETA_VERSION;
	mod->ipcModule = ipcRegisterModule(gdata.ipcServer, mod->name, cifsFileIpcStubs, ELEMENTS(cifsFileIpcStubs), CIFS_IPC_PROTOCOL_VERSION);
	mod->sortPrecedence = cfgIntValue(cfgQueryWithName("sortPrecedence", 1, cfgQueryWithName(CIFSFILE_MODULE_NAME, 1, cfgFilePlist(gdata.cfgFile))), 2000);
	modRegister(mod);
	cifsFileModule = mod;
	mainModRegisterForUiTerminate(uiWillTerminate);
	ncAddObserver(modRegisterNotification(), moduleAdded, NULL);
	globalCfg.entries = arrNew(NULL, 16);
	globalCfg.buttons = arrNew(NULL, 4);
	for(i=0;i<ELEMENTS(globalCfgEntries);i++){
		arrAddObject(globalCfg.entries, &globalCfgEntries[i]);
	}
	for(i=0;i<ELEMENTS(buttonEntries);i++){
		arrAddObject(globalCfg.buttons, &buttonEntries[i]);
	}
	cifsGlobalCfgDsource = mainModRegisterDataSource(&cifsGlobalCfgDsource, &globalCfg, globalDsourceMethods, ELEMENTS(globalDsourceMethods));
	mainModAdvertizeConfig(cifsGlobalCfgDsource, 60000);

	serverCfg.entries = arrNew(NULL, 16);
	serverCfg.buttons = arrNew(NULL, 4);
	for(i=0;i<ELEMENTS(serverCfgEntries);i++){
		arrAddObject(serverCfg.entries, &serverCfgEntries[i]);
	}
	modules = modAllModules();
	for(i=0;i<arrObjectCount(modules);i++){
		m = arrObjectAtIndex(modules, i);
		moduleAdded(NULL, m);
	}
	arrFree(modules);
	for(i=0;i<ELEMENTS(buttonEntries);i++){
		arrAddObject(serverCfg.buttons, &buttonEntries[i]);
	}
	cifsServerCfgDsource = mainModRegisterDataSource(&cifsServerCfgDsource, &serverCfg, serverDsourceMethods, ELEMENTS(serverDsourceMethods));
	mainModAdvertizeConfig(cifsServerCfgDsource, 40000);

#if defined USE_SSL
	serverSslCfg.entries = arrNew(NULL, 16);
	serverSslCfg.buttons = arrNew(NULL, 4);
	for(i=0;i<ELEMENTS(serverSslEntries);i++){
		arrAddObject(serverSslCfg.entries, &serverSslEntries[i]);
	}
	for(i=0;i<ELEMENTS(buttonEntries);i++){
		arrAddObject(serverSslCfg.buttons, &buttonEntries[i]);
	}
	cifsServerSslDsource = mainModRegisterDataSource(&cifsServerSslDsource, &serverSslCfg, serverSslDsourceMethods, ELEMENTS(serverSslDsourceMethods));
	mainModAdvertizeConfig(cifsServerSslDsource, 50000);
#endif

	licenseCfg.entries = arrNew(NULL, 16);
	licenseCfg.buttons = arrNew(NULL, 4);
	for(i=0;i<ELEMENTS(licenseEntries);i++){
		arrAddObject(licenseCfg.entries, &licenseEntries[i]);
	}
	for(i=0;i<ELEMENTS(buttonEntries);i++){
		arrAddObject(licenseCfg.buttons, &buttonEntries[i]);
	}
	licenseDsource = mainModRegisterDataSource(&licenseDsource, &licenseCfg, licenseDsourceMethods, ELEMENTS(licenseDsourceMethods));
	mainModAdvertizeConfig(licenseDsource, 70000);

	loginList.entries = arrNew(NULL, 4);
	loginList.buttons = arrNew(NULL, 4);
	for(i=0;i<ELEMENTS(loginListColumns);i++){
		arrAddObject(loginList.entries, &loginListColumns[i]);
	}
	for(i=0;i<ELEMENTS(loginListButtons);i++){
		arrAddObject(loginList.buttons, &loginListButtons[i]);
	}
	cifsLoginListObject = mainModRegisterDataSource(&cifsLoginListObject, &loginList, loginListDsourceMethods, ELEMENTS(loginListDsourceMethods));
	mainModAdvertizeConfig(cifsLoginListObject, 20000);
	mainModRegisterList(CIFS_LIST_LOGINS, cifsLoginListObject->id);

	userdbParseInterval = cfgIntValue(cfgQueryWithName("databaseParseInterval", 1, cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.cfgFile))), 0);
	if(userdbParseInterval != 0)
		elpCallAtCancelPrev(elpNow() + userdbParseInterval, checkUserdb, NULL);
}

/* ------------------------------------------------------------------------- */

