/*
 * Name: dircache.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-03
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: dircache.c,v 1.3 1999/07/26 20:06:16 cs Exp $
 */

#include "stdheaders.h"
#include "dircache.h"
#include "file.h"
#include "process.h"
#include "user.h"
#include "blockpool.h"
#include "logging.h"
#include "unicode.h"
#include "evntloop.h"
#include "share.h"
#include "server.h"
#include "backend.h"


extern cacheInterface_t	dirCacheInterface;	/* forward declaration */

/* ------------------------------------------------------------------------- */
/* --------------------------- object life cycle --------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct dirCacheCtx{
    PROC_VARSET;
	dirCache_t		*cache;
	cifsUser_t		*user;
	btsize			startIndex;
	dirCallback_t	resCallback;
}dirCacheCtx_t;

static blockpool_t *cachePool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(cachePool, block);
}

/* ------------------------------------------------------------------------- */

dirCache_t	*dirCacheNew(void)
{
dirCache_t	*cache;
    
	if(cachePool == NULL)
		cachePool = poolNew(sizeof(dirCacheCtx_t), sizeof(dirCacheCtx_t));
    cache = memCalloc(sizeof(dirCache_t));
	cacheInit(&cache->general, &dirCacheInterface);
    return cache;
}

/* ------------------------------------------------------------------------- */

void	dirCacheFree(void *this)
{
dirCache_t *cache = this;
int			i;

	if(cache != NULL){
		if(cache->resumeKeySize > 0 && cache->general.owner != NULL)
			fileCloseReaddir(cache->general.owner, cache->general.forUser, cache->resumeKey, cache->resumeKeySize, NULL);
		for(i=0;i<DIRCACHE_SIZE;i++){
			if(cache->records[i].name != NULL){
				memFree(cache->records[i].name);
				cache->records[i].name = NULL;
			}
		}
		cacheDestroy(&cache->general);
		memFree(cache);
	}
}


/* ------------------------------------------------------------------------- */
/* ---------------------------- content access ----------------------------- */
/* ------------------------------------------------------------------------- */

static void	dirCacheInvalidate(void *obj)
{
dirCache_t	*cache = obj;

    if(cache->resumeKeySize > 0)		/* there is still an open search */
		fileCloseReaddir(cache->general.owner, cache->general.forUser, cache->resumeKey, cache->resumeKeySize, NULL);
    cache->headIndex = cache->tailIndex = cache->headPtr = cache->tailPtr = 0;
    cache->noMoreEntries = 0;
    cache->resumeKeySize = 0;				/* start from beginning! */
	cache->dirWasModified = 0;
}

/* ------------------------------------------------------------------------- */

static btbool	readdirCallback(void *context, dirRecord_t *rec)
{
dirCacheCtx_t	*this = context;
dirCache_t		*cache = this->cache;

    if(cache->records[cache->headPtr].name != NULL)
        memFree(cache->records[cache->headPtr].name);
    cache->records[cache->headPtr].name = uniStrDup(rec->name);
    cache->records[cache->headPtr].isDir = rec->isDir;
    cache->headPtr = (cache->headPtr + 1) % DIRCACHE_SIZE;
    cache->headIndex++;
    if(cache->tailPtr == cache->headPtr){
    	cache->tailPtr = (cache->tailPtr + 1) % DIRCACHE_SIZE;
        cache->tailIndex++;
    }
    return 1;
}

static void	dirCacheReadProc(void *obj, void *arg)
{
dirCacheCtx_t		*this = obj;
dirCache_t			*cache = this->cache;
beFileRes_t			res;
filePrivRes_t		*rval = arg;
btint				i, p;
bttime				age, maxAge;
btbool				noCallbackDone;

    PROC_BEGIN
    while(cache->general.opRunning > 0){	/* can't access right now! */
        PROC_WAITFOR_X(1, procCondWait(cache->general.opReady, this));
    }
	cacheStartOperation(cache);
	if(cache->headIndex == cache->tailIndex && cache->resumeKeySize != 0){	/* cache is empty but search is open */
        dirCacheInvalidate(cache);
	}
    if(userSmbUid(this->user) != cache->general.forUser){	/* an other user tries to access the cache */
        dirCacheInvalidate(cache);
        cache->general.forUser = userSmbUid(this->user);
    }
    if(this->startIndex < cache->tailIndex){	/* entries before cache range requested */
        dirCacheInvalidate(cache);
        if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
            logPrintf(LOGM_RARE, "file: dircache: recalibrating search to %d\n", this->startIndex);
    }
	if(this->startIndex == 0 && cache->dirWasModified){	/* need re-read */
		if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
			logPrintf(LOGM_RARE, "file: dircache: re-read because directory was modified\n");
		dirCacheInvalidate(cache);
	}else if(this->startIndex < cache->headIndex){	/* entries within cache range requested */
		maxAge = fileMaxDirCacheAge(cache->general.owner);
		age = elpNow() - cache->lastUpdated;
        if(age > maxAge || age < 0){	/* cache outdated */
            dirCacheInvalidate(cache);
			if(this->startIndex > 0 && LOG_MODE_IS_ACTIVE(LOGM_RARE))
                logPrintf(LOGM_RARE, "file: dircache: recalibrating to %d because outdated\n", this->startIndex);
        }
    }
retry:
	memset(&res, 0, sizeof(res));
    while(!cache->noMoreEntries && cache->headIndex <= this->startIndex){			/* advance cache to requested range */
		PROC_WAITFOR_X(2, fileReadDirUncached(cache->general.owner, this->user, cache->resumeKey, cache->resumeKeySize, DIRCACHE_SIZE/2-1, readdirCallback, this));
		cache->lastUpdated = elpNow();
		res.err = rval->err;
        if(rval->err.class == 0){
			if(rval->resumeKeySize > 0)
				memcpy(cache->resumeKey, rval->resumeKey, rval->resumeKeySize);
			cache->resumeKeySize = rval->resumeKeySize;
			if(rval->resumeKeySize == 0)
				cache->noMoreEntries = 1;
		}else{
            break;
		}
    }
    res.eof = cache->noMoreEntries;
	noCallbackDone = 1;
	p = (this->startIndex - cache->tailIndex + cache->tailPtr) % DIRCACHE_SIZE;
    for(i=this->startIndex;i<cache->headIndex;i++){		/* return data from cache */
		if(cache->records[p].name != NULL){
			noCallbackDone = 0;
			if(!this->resCallback(this->proc.caller, cache->records[p].name, i+1, cache->records[p].isDir)){
				res.eof = 0;
				break;
			}
		}
        p = (p + 1) % DIRCACHE_SIZE;
    }
	if(!res.eof && noCallbackDone && res.err.class == 0){	/* all entries were deleted... */
		this->startIndex = i;		/* continue after current window */
		goto retry;					/* do it again... */
	}
	cache->general.opRunning--;
    PROC_RETURN_AND_CLEANUP(&res,
		cacheEndOperation(cache, 0);
	);
    PROC_END
}

void	dirCacheRead(dirCache_t *cache, cifsUser_t *user, btint startIndex, dirCallback_t resCallback, void *caller)
{
dirCacheCtx_t	*ctx;

	PROC_INIT(ctx, dirCacheReadProc, caller, cachePool, poolFreeBlock);
    ctx->cache = cache;
	ctx->user = user;
	ctx->resCallback = resCallback;
	ctx->startIndex = startIndex;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	dirCacheFileAdded(dirCache_t *cache, btuni *file)
{
/* Don't do anything here. Since it's not defined where in the directory
 * a new file is inserted, we can pretend that it has been inserted before
 * the current position of the search, unless it's found later in the
 * search, of course. This may conflict with a later search, which returns
 * the inserted file in the currently cached area, but the directory may
 * have changed in the meantime. This is at least better than invalidating
 * the cache, which can change the index of all entries after re-reading
 * the contents.
 */
}

/* ------------------------------------------------------------------------- */

void	dirCacheFileRemoved(dirCache_t *cache, btuni *file)
{
btint	i;

/* Search the entry in the cache and mark it as deleted by setting the
 * file name to NULL. Do not invalidate the cache.
 */
	for(i = cache->tailPtr; i != cache->headPtr; i = (i+1) % DIRCACHE_SIZE){
		if(cache->records[i].name != NULL && uniStrCmp(file, cache->records[i].name, 1) == 0){	/* found entry */
			memFree(cache->records[i].name);
			cache->records[i].name = NULL;
			break;
		}
	}
}

/* ------------------------------------------------------------------------- */

void	dirCacheDirWasModified(dirCache_t *cache)
{
	cache->dirWasModified = 1;
}

/* ------------------------------------------------------------------------- */

static void	dummyWriteBack(void *cache, void *caller)
{
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

cacheInterface_t	dirCacheInterface = {
	dirCacheInvalidate,
	dirCacheInvalidate,
	dummyWriteBack,
	dirCacheFree,
};

/* ------------------------------------------------------------------------- */
