/*
 * Name: file.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-02
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: file.c,v 1.18 2002/06/03 21:58:18 cs Exp $
 */


#include "stdheaders.h"
#include "exception.h"
#include "memalloc.h"
#include "unicode.h"
#include "file.h"
#include "share.h"
#include "blockpool.h"
#include "smb.h"
#include "cifs.h"
#include "smbnumbers.h"
#include "errcodes.h"
#include "server.h"
#include "logging.h"
#include "cifsconst.h"
#include "process.h"
#include "dircache.h"
#include "datacache.h"
#include "user.h"
#include "main.h"
#include "fileattrdos.h"
#include "fileattrunix.h"
#include "backend.h"
#include "notification.h"
#include "licman.h"

/* ------------------------------------------------------------------------- */

typedef struct fileCtx{
    PROC_VARSET;
    cifsFile_t				*file;
    cifsFile_t				*newFile;
    cifsFile_t				*oldFile;
	cifsFile_t				*destFile;
	cifsFile_t				*srcFile;
    btuni					*name;
    btuni					*otherName;
	cifsUser_t				*user;
    btu16					uid;
    btu16					tid;
	btoffset				offset;
	btsize					len;
	btu8					*buffer;
	fileReaddirCallback_t	readdirCb;
	btu8					*resumeKey;
	btsize					resumeKeySize;
	btu32					cifsResumeKey;
	btu8					oplock;
	btbool					isDir;
	btbool					isLink;
	btbool					noLocking;
	btbool					writeThrough;
	btbool					makeReadonly;
	btbool					autoclose;
	btint					mode;
	btint					mask;
	btint					op;
	btint					retryCount;
	beFileRes_t				fres;
	filePrivRes_t			pres;
}fileCtx_t;

static blockpool_t *filePool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(filePool, block);
}

static void	doAutoclose(void *obj, void *dummy);
static void	fileDidOpen(cifsFile_t *file);

/* ------------------------------------------------------------------------- */
/* ------------------------- collection interfaces ------------------------- */
/* ------------------------------------------------------------------------- */

static btint	fileNameCompare(void *obj, void *obj2)
{
cifsFile_t	*this = obj, *other = obj2;

    return uniStrCmp(this->name, other->name, srvIsCaseSensitive(shrServer(this->share)));
}

static btu32	fileNameHashvalue(void *obj)
{
cifsFile_t	*this = obj;

    return uniStrHash(this->name, srvIsCaseSensitive(shrServer(this->share)));
}

/* ------------------------------------------------------------------------- */

collectionObjectI_t	fileByNameInterface = {
    fileNameCompare,
    fileNameHashvalue,
    NULL,	/* no retain/release by collections */
    NULL
};

/* ------------------------------------------------------------------------- */
/* ----------- file creation/destruction, relationship handling ------------ */
/* ------------------------------------------------------------------------- */

cifsFile_t	*fileNew(struct share *share, btuni *name)
{
cifsFile_t		*volatile this = NULL;
static btuni	emptyName[1] = {0};

    if(filePool == NULL)
        filePool = poolNew(sizeof(fileCtx_t), sizeof(fileCtx_t));
EXC_TRY
    this = memCalloc(sizeof(cifsFile_t));
	this->share = share;
	shrRetainFromFile(share);
    this->name = uniStrDup(name == NULL ? emptyName : name);
    this->openedBy = -1;
    this->fid = -1;
    this->refCount = 1;	/* new objects are created with reference count 1 */
    this->openSignal = procCondNew();
#if 0	/* we currently don't support the unix extension */
	/* We should allocate the attribute representation later because
	 * the server is not known (much less connected) when the root file is
	 * allocated.
	 */
	this->attributes = srvUseRequest(shrServer(share), SRV_USE_ATTR_UNIX) ? attrUnixNew() : attrDOSNew();
#endif
	this->attributes = attrDOSNew();	/* ### change call to attrDOSSetValues() if an other strategy is used */
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "file: creating ->%s<- at 0x%08x\n", uniToCstring(this->name), (int)this);
	if(name == NULL){	/* this is the root node */
		this->children = htabNew(&fileByNameInterface, 32, 75);	/* must be a directory */
	}
EXC_CATCH
	fileFree(this);
	EXC_RETHROW();
EXC_ENDTRY
    return this;
}

/* ------------------------------------------------------------------------- */

void	fileFree(cifsFile_t *file)
{
    if(file != NULL){
		if(file->openMode != FILE_CLOSED){
			fileSetOpenMode(file, -1, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 0, NULL);	/* close in background */
			if(file->refCount != 0)
				return;	/* close needs file until finished, will free then */
		}
		fileInvalidate(file);
		if(file->cache.any != NULL){
            cacheInvalidate(file->cache.any);
			cacheWriteBack(file->cache.any, NULL);
			if(file->refCount > 0)
				return;		/* ref count has changed, we'll free later... */
		}
        elpCancelCallAt(doAutoclose, file);	/* cancel any pending callbacks */
        if(file->cache.any != NULL){
			if(cacheOperationsRunning(file->cache.any))
				logPrintf(LOGM_ERROR, "+++file: freeing ->%s<- while cache running\n", file->name != NULL ? (char *)uniToCstring(file->name) : "<null>");
        	cacheSetOwner(file->cache.any, NULL);
        }
		fileUnregisterChild(file->parent, file, 1);
        if(file->share != NULL){
			shrReleaseFromFile(file->share);
			file->share = NULL;
		}
        if(file->children != NULL){
            htabFree(file->children);
            file->children = NULL;
        }
        if(file->name != NULL){
            if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
                logPrintf(LOGM_RESOURCE, "file: freeing ->%s<- at 0x%08x\n", uniToCstring(file->name), (int)file);
            memFree(file->name);
            file->name = NULL;
        }
		if(file->attributes != NULL){
			attrFree(file->attributes);
			file->attributes = NULL;
		}
        if(file->openSignal != NULL){
            procCondFree(file->openSignal);
			file->openSignal = NULL;
		}
        memFree(file);
    }
}

/* ------------------------------------------------------------------------- */

void	fileIsDirectory(cifsFile_t *file)
{
    if(file->children == NULL){
		file->children = htabNew(&fileByNameInterface, 32, 75);
		if(file->cache.any != NULL)
			logPrintf(LOGM_ERROR, "+++ file: file with cache changed file type: %s!!\n", uniToCstring(file->name));
    }
}

/* ------------------------------------------------------------------------- */

static int	fileDepth(cifsFile_t *file)
{
cifsFile_t	*p;
int			depth = 0;

	for(p=file->parent; p!=NULL; p=p->parent, depth++);
	return depth;
}

/* ------------------------------------------------------------------------- */

void	fileUnregisterChild(cifsFile_t *file, cifsFile_t *child, btbool bidirectional)
{
	if(file != NULL){
		if(file->children != NULL){
			htabRemoveObjectEqualTo(file->children, child);
		}else{
			logPrintf(LOGM_ERROR, "+++file: fileUnregisterChild(): parent is a regular file ->%s<-\n", file->name != NULL ? (char *)uniToCstring(file->name) : "<null>");
		}
	}
	if(bidirectional){
		if(child->parent != NULL)
			fileRelease(child->parent);
		child->parent = NULL;
	}
}

/* ------------------------------------------------------------------------- */

void	fileInvalidate(cifsFile_t *file)
{
btsize		i, l;
array_t		*allChildren;
btbool		doRelease;

	if(file->isInvalid)
		return;	/* already invalidated */
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE) && file->name != NULL)
		logPrintf(LOGM_RESOURCE, "file: invalidating ->%s<- at 0x%08x with refcnt=%d\n", uniToCstring(file->name), (int)file, file->refCount);
	file->isInvalid = 1;	/* avoid recursive invalidates */
	doRelease = file->refCount > 0;
	if(doRelease)
		fileRetain(file);
	if(file->children != NULL && htabObjectCount(file->children) > 0){
		allChildren = htabAllObjects(file->children);
		l = arrObjectCount(allChildren);
		for(i=0;i<l;i++){
			fileInvalidate(arrObjectAtIndex(allChildren, i));	/* don't release, that will be done by reference owner */
		}
		arrFree(allChildren);
	}
	if(file->openMode != FILE_CLOSED){	/* emergency close */
		/* close in background: */
		fileSetOpenMode(file, file->openedBy, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 0, NULL);
	}
	fileUnregisterChild(file->parent, file, 0);	/* unidirectional: only remove from parent list to avoid success of lookups */
	shrFileBecameInvalid(file->share, file);
	if(doRelease)
		fileRelease(file);
}

/* ------------------------------------------------------------------------- */
/* --------------------------- utility functions --------------------------- */
/* ------------------------------------------------------------------------- */

btuni	*fileBuildPath(cifsFile_t *file, void *append, btbool isUnicode, btuni *buffer, btsize bufSize)
{
cifsFile_t	*p;
btsize		l, wp = bufSize;

    buffer[--wp] = 0;
    if(append != NULL){
        if(isUnicode){
            l = uniStrLen(append);
            wp -= l;
            memcpy(&buffer[wp], append, 2 * l);
        }else{
            l = strlen(append);
            wp -= l;
            convLocalAsciiToUni(append, &buffer[wp], l + 1);
        }
		buffer[--wp] = UNI_CHAR_BACKSLASH;
    }
    for(p=file;p->parent!=NULL;p=p->parent){
        l = uniStrLen(p->name);
    	wp -= l;
		if(wp < 1)	/* emergency brake */
			return &buffer[bufSize - 1];
        memcpy(&buffer[wp], p->name, l * sizeof(btuni));
		buffer[--wp] = UNI_CHAR_BACKSLASH;
    }
    return &buffer[wp];
}

btbool	fileNameIsValid(btuni *name)
{
btint	i;

	for(i=0;name[i]!=0;i++){
		switch(name[i]){
			case '\"':	case '/':	case '\\':	case ':':
			case '|':	case '<':	case '>':	case '*':
			case '?':
			return 0;
		}
	}
	return 1;
}

/* ------------------------------------------------------------------------- */
/* ------------------- file attribute operations, lookup ------------------- */
/* ------------------------------------------------------------------------- */

static void	fileLookupChildWithNameProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
beFileRes_t		res, *rval = arg;
cifsFile_t		*f, *file = this->file;

    PROC_BEGIN
	fileRetain(file);
	this->newFile->parent = file;	/* needed for path creation */
	PROC_WAITFOR_X(1, attrMakeValid(this->newFile, this->user, this));
	this->newFile->parent = NULL;	/* that's it for now */
	if(rval->err.class != 0)
		goto errorOccured;
    f = htabFindObjectEqualTo(file->children, this->newFile);	/* check whether a parallel process succeeded */
    if(f != NULL){
        fileRelease(this->newFile);
        this->newFile = f;
		fileRetain(f);
    }else{
		licRegisterFile(file, this->newFile, &res.err);
		if(res.err.class != 0){
			rval = &res;
			goto errorOccured;
		}
    }
	fileRelease(file);			/* match the retain we made */
    res.err.class = res.err.code = 0;
    res.file = this->newFile;
    PROC_RETURN(&res);
errorOccured:
	fileRelease(file);			/* match the retain we made */
	fileRelease(this->newFile);	/* free the trial file object */
	PROC_RETURN(rval);
    PROC_END
}

void	fileLookupChildWithName(cifsFile_t *file, btuni *name, cifsUser_t *user, void *caller)
{
beFileRes_t	res;
cifsFile_t	f;
fileCtx_t	*ctx;
btuni		*pattern;

    if(file->children == NULL){	/* must be directory */
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
        PROC_IMMEDIATE_RET(caller, &res);
    }
	if(!fileNameIsValid(name)){
        res.err.class = ERRCLS_UNIX;
        res.err.code = ENOENT;
        PROC_IMMEDIATE_RET(caller, &res);
	}
	if((pattern = srvPreventCreationPattern(shrServer(file->share))) != NULL){
		if(uniWildcardMatch(name, pattern)){
			res.err.class = ERRCLS_UNIX;
			res.err.code = EACCES;
			PROC_IMMEDIATE_RET(caller, &res);
		}
	}
    res.err.class = res.err.code = 0;
	f.name = name;
	f.share = file->share;
	res.file = htabFindObjectEqualTo(file->children, &f);
    if(res.file != NULL){
		fileRetain(res.file);
		fileRetain(res.file);
		PROC_IMMEDIATE_RET_AND_CLEANUP(caller, &res,
			shrRemoveFromLookupCache(fileShare(res.file), res.file);
			fileRelease(res.file);
		);
    }
	PROC_INIT(ctx, fileLookupChildWithNameProc, caller, filePool, poolFreeBlock);
    ctx->file = file;
    ctx->tid = shrTid(file->share);
	ctx->user = user;
    ctx->newFile = fileNew(file->share, name);
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	fileGetAttributesProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
beFileRes_t		res, *rval = arg;
cifsFile_t		*file = this->file;
beFileAttr_t	attributes;

	PROC_BEGIN
	fileRetain(file);
	PROC_WAITFOR_X(1, attrMakeValid(file, this->user, this));
	if(rval->err.class != 0){
		fileInvalidate(file);		/* remove any caches */
		fileRelease(file);
		PROC_RETURN(rval);
	}
	attrGetAttributes(file, this->user, &attributes);
	res.attr = &attributes;
	fileRelease(file);
	res.err.class = res.err.code = 0;
	PROC_RETURN(&res);
	PROC_END
}

void	fileGetAttributes(cifsFile_t *file, cifsUser_t *user, void *caller)
{
fileCtx_t		*ctx;

	PROC_INIT(ctx, fileGetAttributesProc, caller, filePool, poolFreeBlock);
    ctx->file = file;
	ctx->user = user;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	fileSetAttributes(cifsFile_t *file, cifsUser_t *user, struct beFileSetAttr *setAttr, void *caller)
{
	if(file->openMode != FILE_CLOSED && srvCloseForSetattr(shrServer(file->share))){
		fileSetOpenMode(file, file->openedBy, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 0, NULL);	/* close in background */
	}
	attrSetAttributes(file, user, setAttr, caller);
}

/* ------------------------------------------------------------------------- */

void	fileReadLink(cifsFile_t *file, cifsUser_t *user, void *caller)
{
	attrReadLink(file, user, caller);
}


/* ------------------------------------------------------------------------- */
/* ------------------------- directory operations -------------------------- */
/* ------------------------------------------------------------------------- */

static btbool	fileReaddirResult(void *context, cifsResFile_t *rec)
{
fileCtx_t		*this = context;
dirRecord_t		dirRec;
btbool			fakeLookup = 1;
cifsFile_t		*newFile, dummyFile;
btError			err;

	this->cifsResumeKey = rec->resumeKey;
	switch(srvCaseMapping(shrServer(this->file->share))){
		case SRV_CASE_LOWER:	uniStrTolower(rec->name);	break;
		case SRV_CASE_UPPER:	uniStrToupper(rec->name);	break;
	}
	dirRec.name = rec->name;
	if(rec->infoLevel == SMB_FIND_FILE_UNIX)
		dirRec.isDir = rec->unixType == SMB_UNIX_FILETYPE_DIR;
	else
		dirRec.isDir = (rec->dosMode & CIFS_ATTR_DIR) != 0;

    if(rec->name[0] == UNI_CHAR_DOT){	/* special treatment for . and .. */
        if(rec->name[1] == 0){			/* name is '.' */
			fakeLookup = 0;
        }else if(rec->name[1] == UNI_CHAR_DOT && rec->name[2] == 0){	/* name is '..' */
			fakeLookup = 0;
        }
    }
	if(fakeLookup){	/* can have inum only for . and .. */
		dummyFile.name = rec->name;
		dummyFile.share = this->file->share;
		if((newFile = htabFindObjectEqualTo(this->file->children, &dummyFile)) == NULL){
			newFile = fileNew(this->file->share, rec->name);
			licRegisterFile(this->file, newFile, &err);
			if(err.class == 0){
				shrAddToLookupCache(newFile->share, newFile);
			}else{
				fileRelease(newFile);
				newFile = NULL;
			}
		}
		if(newFile != NULL){
			switch(rec->infoLevel){
			case SMB_INFO_STANDARD:
				attrDOSSetValues(newFile->attributes, newFile, this->uid, rec->dosMode, 
							attrTimeFromDateTime(rec->cdate, rec->ctime, newFile),	/* ctime */
							attrTimeFromDateTime(rec->wdate, rec->wtime, newFile),	/* wtime */
							attrTimeFromDateTime(rec->adate, rec->atime, newFile),	/* atime */
							rec->size, rec->allocSize);
				break;
			default:	/* ### does not handle SMB_FIND_FILE_UNIX correctly */
				attrDOSSetValues(newFile->attributes, newFile, this->uid, rec->dosMode, rec->chtime64, rec->wtime64, rec->atime64, rec->size, rec->allocSize);
				break;
			}
		}
	}
	return this->readdirCb(this->proc.caller, &dirRec);
}

static void	fileReadDirUncachedProc(void *obj, void *arg)
{
fileCtx_t			*this = obj;
cifsResReaddir_t	*rval = arg;
cifsFile_t			*file = this->file;
server_t			*server = shrServer(file->share);
btuni				*searchPath, pathBuffer[CIFS_MAX_PATHLEN];
smbRequest_t		*smb = arg;
filePrivRes_t		res;
btu32				resumeKey[2];
btint				infoLevel;

	PROC_BEGIN
	fileRetain(file);
retry:
	PROC_WAITFOR_X(1, srvWantRequest(shrServer(file->share), this));
#if 0	/* not currently implemented */
	if(srvUseRequest(server, SRV_USE_RDIR_UNIX)){
		infoLevel = SMB_FIND_FILE_UNIX;
	}else
#endif
	if(srvUseRequest(server, SRV_USE_RDIR_LM)){
		infoLevel = SMB_INFO_STANDARD;
	}else if(srvUseRequest(server, SRV_USE_RDIR_DIRINFO)){
		infoLevel = SMB_FIND_FILE_DIRECTORY_INFO;
	}else if(srvUseRequest(server, SRV_USE_RDIR_BOTHDIRINFO)){
		infoLevel = SMB_FIND_FILE_BOTH_DIRECTORY_INFO;
	}else{
		infoLevel = -1;
	}
	if(infoLevel >= 0){
		searchPath = fileBuildPath(this->file, "*", 0, pathBuffer, CIFS_MAX_PATHLEN);
		if(this->resumeKeySize == 0){
			PROC_WAITFOR_X(2, cifsFindFirst(smb, this, fileReaddirResult, this->uid, this->tid, searchPath, this->len, infoLevel, srvUseRequest(server, SRV_USE_RDIR_AUTOCLOSE), !srvUseRequest(server, SRV_USE_RDIR_IMPLICIT_RESUME)));
			resumeKey[0] = rval->searchId;
			resumeKey[1] = this->cifsResumeKey;
			if(rval->err.class == 0){
				this->resumeKey = (btu8 *)resumeKey;
				this->resumeKeySize = 8;
			}
		}else{
			PROC_WAITFOR_X(3, cifsFindNext(smb, this, fileReaddirResult, this->uid, this->tid, searchPath, this->len, *(btu32 *)this->resumeKey, infoLevel, srvUseRequest(server, SRV_USE_RDIR_AUTOCLOSE), !srvUseRequest(server, SRV_USE_RDIR_IMPLICIT_RESUME), ((btu32 *)this->resumeKey)[1]));
		}
		if(rval->err.class == ERRCLS_CIFS_SRV && rval->err.code == ERR_CIFS_SRVerror){				/* spurious NT error... */
			if(++this->retryCount < 3){	/* try up to 3 times */
				PROC_WAITFOR_X(4, elpCallAtCancelPrev(elpNow() + 50, fileReadDirUncachedProc, this));	/* wait some time... */
				goto retry;
			}
		}
		if(rval->err.class == 0 && !rval->eof && rval->entryCount > 0){
		/* NT Bug: may set entryCount=0 and forget to set eof=1 */
			res.resumeKeySize = this->resumeKeySize;
			res.resumeKey = this->resumeKey;
		}else{
			if(!srvUseRequest(server, SRV_USE_RDIR_AUTOCLOSE) && this->resumeKeySize != 0){
				fileCloseReaddir(file, this->uid, this->resumeKey, this->resumeKeySize, NULL);	/* close in background */
			}
			res.resumeKeySize = 0;	/* can't resume */
		}
	}else{
		searchPath = fileBuildPath(this->file, "*.*", 0, pathBuffer, CIFS_MAX_PATHLEN);
		PROC_WAITFOR_X(5, cifsSearch(smb, this, fileReaddirResult, this->uid, this->tid, searchPath, this->len, this->resumeKeySize, this->resumeKey));
		if(rval->entryCount == 0)	/* EOF: no more entries found */
			rval->resumeSize = 0;
		if(rval->resumeSize > 24 || rval->resumeSize < 0)
			rval->resumeSize = 0;
		res.resumeKeySize = rval->resumeSize;
		res.resumeKey = rval->resumeKey;
	}
	if(rval->err.class == ERRCLS_CIFS_DOS){
		if(rval->err.code == ERR_CIFS_DOSnofiles		/* translate "no more entries" to EOF */
			|| rval->err.code == ERR_CIFS_DOSbadfile){	/* translate "bad file" to EOF (AppleShareIP!) */
			rval->err.class = rval->err.code = 0;
			res.resumeKeySize = 0;		/* can't resume */
		}
	}
	res.err = rval->err;
	fileRelease(file);
	PROC_RETURN(&res);
	PROC_END
}

void	fileReadDirUncached(cifsFile_t *file, cifsUser_t *user, btu8 resumeKey[24], btsize resumeKeySize, btsize maxCount, fileReaddirCallback_t resCb, void *caller)
{
fileCtx_t	*ctx;

	PROC_INIT(ctx, fileReadDirUncachedProc, caller, filePool, poolFreeBlock);
    ctx->file = file;
	ctx->uid = userSmbUid(user);
	ctx->tid = shrTid(file->share);
	ctx->resumeKey = resumeKey;
	ctx->resumeKeySize = resumeKeySize;
	ctx->len = maxCount;
	ctx->readdirCb = resCb;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	fileCloseReaddirProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
smbRequest_t	*smb = arg;
cifsRes_t		*rval = arg;
beFileRes_t		res;

    PROC_BEGIN
	fileRetain(this->file);
    PROC_WAITFOR_X(1, srvWantRequest(shrServer(this->file->share), this));
    PROC_WAITFOR_X(2, cifsFindClose(smb, this, this->uid, this->tid, this->op));
	fileRelease(this->file);
	res.err = rval->err;
	PROC_RETURN(&res);
    PROC_END
}

void	fileCloseReaddir(cifsFile_t *file, btu16 cifsUid, btu8 resumeKey[24], btsize resumeKeySize, void *caller)
{
fileCtx_t	*ctx;
beFileRes_t	res;
server_t	*server = shrServer(file->share);

	if(resumeKeySize == 0 || !(srvUseRequest(server, SRV_USE_RDIR_UNIX) || srvUseRequest(server, SRV_USE_RDIR_LM) || srvUseRequest(server, SRV_USE_RDIR_DIRINFO) || srvUseRequest(server, SRV_USE_RDIR_BOTHDIRINFO))){
		res.err.code = res.err.class = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, fileCloseReaddirProc, caller, filePool, poolFreeBlock);
    ctx->file = file;
	ctx->uid = cifsUid;
	ctx->tid = shrTid(file->share);
	ctx->op = *(btu32 *)resumeKey;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	fileReadDir(cifsFile_t *file, cifsUser_t *user, btu32 startIndex, dirCallback_t resCb, void *caller)
{
beFileRes_t	res;

	if(file->children == NULL){		/* must be directory */
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	if(fileDepth(file) > licDepthLimit()){	/* too deep in hierarchy */
		res.err.class = ERRCLS_UNIX;
		res.err.code = ENAMETOOLONG;
        PROC_IMMEDIATE_RET(caller, &res);
	}
	cacheAcquire(CACHE_TYPE_DIR, file);
    dirCacheRead(file->cache.dir, user, startIndex, resCb, caller);
}

/* ### maybe open dir to get oplock breaks if files are created or deleted? */

/* ------------------------------------------------------------------------- */

static void	fileCreateUnixProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
smbRequest_t	*smb = arg;
cifsResFileOp_t	*rval = arg;
beFileRes_t		*frval = arg;
beFileSetAttr_t	*unixAttr = arg;
server_t		*server = shrServer(this->file->share);
btuni			*path, pathBuffer[CIFS_MAX_PATHLEN];
cifsFile_t		*new, *existing, f, *file = this->file;

    PROC_BEGIN
	/* check for ignore patterns: */
	if((path = srvPreventCreationPattern(server)) != NULL){
		if(uniWildcardMatch(this->name, path)){
			this->fres.err.class = ERRCLS_UNIX;
			this->fres.err.code = EACCES;
			PROC_RETURN(&this->fres);
		}
	}
	fileRetain(file);
	this->mode = attrDOSFromAttributes(unixAttr, this->user, server);	/* use this->mode to store dos-attributes */
	if(this->isLink)
		this->mode = (this->mode & ~srvLinkMask(server)) | srvLinkMode(server);
	this->len = unixAttr->setSize ? unixAttr->size : -1;	/* set size if required */
	f.name = this->name;
	f.share = file->share;
    if((existing = htabFindObjectEqualTo(file->children, &f)) != NULL && existing->openMode != FILE_CLOSED){
		if(this->isLink){
			this->fres.err.class = ERRCLS_CIFS_DOS;
			this->fres.err.code = ERR_CIFS_DOSfilexists;
		}else{
			this->fres.err.code = this->fres.err.class = 0; /* file must exist */
			this->fres.file = existing;
			fileRetain(existing);
		}
    }else{
		PROC_WAITFOR_X(1, srvWantRequest(server, this));
		path = fileBuildPath(file, this->name, 1, pathBuffer, CIFS_MAX_PATHLEN);
		if(this->isDir){
			PROC_WAITFOR_X(2, cifsMkdirCore(smb, this, this->uid, this->tid, path));
		}else{		/* the object to be created is not a directory and is not open */
			PROC_WAITFOR_X(3,
				if(srvUseRequest(server, SRV_USE_CREATE_OPENX) && !this->isLink){
					cifsOpen(smb, this, this->uid, this->tid, path, 0, this->len != -1, 1, this->mode, 0, this->len == -1 ? 0 : this->len);
					this->len = -1;	/* truncate was done, don't do it again */
				}else
					cifsCreateFileCore(smb, this, this->uid, this->tid, path, this->mode, 0);
			)
		}
		if(rval->err.class == 0 && !this->isDir){	/* we have an open file now */
			this->mode = rval->fid;		/* misuse mode to store file handle */
			this->oplock = rval->oplock;
		}else{
			this->mode = -1;
		}
		if(rval->err.class == 0 || (!srvUseRequest(server, SRV_USE_CREATE_OPENX) && !this->isLink)){
			/* ### maybe save lookup because we might already know attributes? */
			PROC_WAITFOR_X(4, fileLookupChildWithName(file, this->name, this->user, this));
			if(frval->err.class == 0){
				new = this->fres.file = frval->file;
				if(this->mode != -1 && !new->openLock && new->openMode == FILE_CLOSED){
					new->fid = this->mode;
					new->oplockLevel = this->oplock;
					new->openMode = FILE_OPEN_RW;
					new->openedBy = this->uid;
					new->doAutoclose = 1;	/* should be treated as closed file */
					fileDidOpen(new);
					this->mode = -1;
				}
			}else{
				this->fres.err = frval->err;
			}
			if(this->mode != -1){	/* file open but not associated to file cache */
				PROC_WAITFOR_X(5, srvWantRequest(shrServer(file->share), this));
				cifsClose(smb, NULL, this->uid, this->tid, this->mode, attrGetWtime(file));	/* close in background */
			}
		}else{
			this->fres.err = rval->err;
		}
	}
	if(!this->isDir && !this->isLink && this->len != -1 && this->fres.file != NULL){	/* set length on file */
		PROC_WAITFOR_X(6, fileTruncate(this->fres.file, this->len, this->user, this));
		if(frval->err.class != 0){
			fileRelease(this->fres.file);
			this->fres.err = frval->err;
			this->fres.file = NULL;
		}
	}
	if(srvUseRequest(server, SRV_USE_DIRDATE_EMUL)){
		attrTouch(file, this->user, NULL);	/* set directory modification date in background */
	}else{
		attrInvalidate(file);
	}
	if(file->cache.any != NULL && this->fres.file != NULL)
		dirCacheFileAdded(file->cache.any, fileName(this->fres.file));
	fileRelease(file);
    PROC_RETURN(&this->fres);
    PROC_END
}

void	fileCreateUnix(cifsFile_t *file, btuni *name, btbool isDir, btbool isLink, cifsUser_t *user, struct beFileSetAttr *attr, void *caller)
{
fileCtx_t	*ctx;
beFileRes_t	res;

    if(file->children == NULL){	/* must be directory */
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
        PROC_IMMEDIATE_RET(caller, &res);
    }
	if(!fileNameIsValid(name)){
        res.err.class = ERRCLS_UNIX;
        res.err.code = ENOENT;
        PROC_IMMEDIATE_RET(caller, &res);
	}
	if(fileDepth(file) > licDepthLimit()){	/* too deep in hierarchy */
		res.err.class = ERRCLS_UNIX;
		res.err.code = ENAMETOOLONG;
        PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, fileCreateUnixProc, caller, filePool, poolFreeBlock);
    ctx->file = file;
    ctx->tid = shrTid(file->share);
	ctx->user = user;
	ctx->uid = userSmbUid(user);
    ctx->isDir = isDir;
	ctx->isLink = isLink;
    ctx->name = name;
    PROC_START(ctx, attr);
}

/* ------------------------------------------------------------------------- */

static void	fileRemoveChildProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
smbRequest_t	*smb = arg;
cifsResFileOp_t	*rval = arg;
beFileRes_t		res;
btuni			*path, buffer[CIFS_MAX_PATHLEN];
cifsFile_t		f, *file = this->file;

    PROC_BEGIN
	fileRetain(file);
	f.name = this->name;
	f.share = file->share;
    if((this->oldFile = htabFindObjectEqualTo(file->children, &f)) != NULL){
		fileRetain(this->oldFile);
		while(this->oldFile->openLock){
			PROC_WAITFOR_X(1, procCondWait(this->oldFile->openSignal, this));
		}
		this->oldFile->openLock++;	/* lock by hand because we want to lock until end of function */
        if(this->oldFile->openMode != FILE_CLOSED){
            PROC_WAITFOR_X(2, fileSetOpenMode(this->oldFile, this->uid, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 1, this));
        }
    }
	if(this->oldFile == NULL || !attrValid(this->oldFile, this->user) || !attrGetWritable(this->oldFile, this->user)){
		PROC_WAITFOR_X(3, srvWantRequest(shrServer(file->share), this));
		path = fileBuildPath(file, this->name, 1, buffer, CIFS_MAX_PATHLEN);
		PROC_WAITFOR_X(4, cifsSetInformation(smb, this, this->uid, this->tid, path, 0, 0));	/* change to r/w before we attempt delete */
	}
    PROC_WAITFOR_X(5, srvWantRequest(shrServer(file->share), this));
    path = fileBuildPath(file, this->name, 1, buffer, CIFS_MAX_PATHLEN);
    PROC_WAITFOR_X(6, cifsRemoveCore(smb, this, this->uid, this->tid, path, this->isDir));
	if(srvUseRequest(shrServer(file->share), SRV_USE_DIRDATE_EMUL)){
		attrTouch(file, this->user, NULL);	/* set directory modification date in background */
/* One could think that changing the directory's modification date on
 * each deleted file costs performance. However, experiments show a
 * contra-intuitive behaviour: an 'rm -r' over a large directory runs
 * faster with 'attrTouch()' enabled than without it. It's therefore not
 * necessary (and not useful) to collect several touches of the same
 * directory into one operation.
 */
	}else{
		attrInvalidate(file);
	}
    if(file->cache.any != NULL)
		dirCacheFileRemoved(file->cache.any, this->name);
	if(this->oldFile != NULL){
		this->oldFile->openLock--;
		fileInvalidate(this->oldFile);		/* remove any caches */
		procCondBroadcast(this->oldFile->openSignal, NULL);
		fileRelease(this->oldFile);
	}
	fileRelease(file);
    res.err = rval->err;
    PROC_RETURN(&res);
    PROC_END
}

void	fileRemoveChild(cifsFile_t *file, btuni *name, btbool isDir, cifsUser_t *user, void *caller)
{
fileCtx_t	*ctx;
beFileRes_t	res;
    
	if(file->children == NULL){	/* not a directory */
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
        PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, fileRemoveChildProc, caller, filePool, poolFreeBlock);
    ctx->file = file;
    ctx->tid = shrTid(file->share);
	ctx->uid = userSmbUid(user);
	ctx->user = user;
    ctx->isDir = isDir;
    ctx->name = name;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	doRecursiveClose(cifsFile_t *file, fileCtx_t *ctx)
{
array_t	*allChildren;
btint	cnt, i;

	if(file->children == NULL){
		if(file->openMode != FILE_CLOSED){
			ctx->retryCount++;
			fileSetOpenMode(file, -1, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 1, ctx);
		}
	}else{
		if(htabObjectCount(file->children) > 0){
			allChildren = htabAllObjects(file->children);
			cnt = arrObjectCount(allChildren);
			for(i=0;i<cnt;i++)
				doRecursiveClose(arrObjectAtIndex(allChildren, i), ctx);
			arrFree(allChildren);
		}
	}
}

static void	fileRecursiveCloseProc(void *obj, void *arg)
{
fileCtx_t	*this = obj;

	PROC_BEGIN;
	PROC_WAITFOR_X(1,
		this->retryCount++;
		/* we increment the counter by one here and compensate this by a call
		 * from the eventloop because doRecursiveClose() might create no
		 * callback at all (if no files are open) or it might have all
		 * callbacks immediately instead of from the eventloop.
		 */
		elpCallCancelPrev(fileRecursiveCloseProc, this);
		doRecursiveClose(this->file, this);
	)
	/* we don't continue just once here, we get a callback for every closed
	 * file and from the eventloop. Ignore all but the last. 
	 */
	if(--this->retryCount > 0)
		return;	/* this was not the last callback */
	if(this->retryCount < 0){
		DEBUG_TRAP();
	}
	PROC_END;
}

static void	fileRecursiveClose(cifsFile_t *file, void *caller)
{
fileCtx_t	*ctx;
    
	PROC_INIT(ctx, fileRecursiveCloseProc, caller, filePool, poolFreeBlock);
    ctx->file = file;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	fileRenameProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
smbRequest_t	*smb = arg;
cifsResFileOp_t	*rval = arg;
beFileRes_t		res, *frval = arg;
cifsFile_t		f, *fromDir = this->file, *toDir = this->newFile;
btuni			*fromPath, *toPath, fromBuf[CIFS_MAX_PATHLEN], toBuf[CIFS_MAX_PATHLEN];
btuni			*fromName = this->name, *toName = this->otherName;
server_t		*server = shrServer(fromDir->share);

    PROC_BEGIN
	/* check for ignore patterns: */
	if((fromPath = srvPreventCreationPattern(server)) != NULL){
		if(uniWildcardMatch(this->otherName, fromPath)){
			this->fres.err.class = ERRCLS_UNIX;
			this->fres.err.code = EACCES;
			PROC_RETURN(&this->fres);
		}
	}
	fileRetain(fromDir);
	fileRetain(toDir);
	PROC_WAITFOR_X(1, fileLookupChildWithName(fromDir, fromName, this->user, this));
	if(frval->err.class != 0){
		fileRelease(toDir);
		fileRelease(fromDir);
		PROC_RETURN(frval);
	}
	this->srcFile = frval->file;
	f.name = toName;
	f.share = fromDir->share;
	if((this->destFile = htabFindObjectEqualTo(toDir->children, &f)) != NULL)
		fileRetain(this->destFile);
	while(this->srcFile->openLock || this->srcFile->needOpen || (this->destFile != NULL && this->destFile->openLock)){
		if(this->destFile != NULL && this->destFile->openLock){
			PROC_WAITFOR_X(2, procCondWait(this->destFile->openSignal, this));
		} 
		if(this->srcFile->openLock || this->srcFile->needOpen){
			PROC_WAITFOR_X(3, procCondWait(this->srcFile->openSignal, this));
		}
	}
/* Now the source file is not open-locked and not needOpen, the destination
 * file is not open-locked or does not exist. We grab both open locks.
 */
	this->srcFile->openLock++;
	if(this->destFile != NULL)
		this->destFile->openLock++;
	if(this->destFile != NULL && this->destFile->openMode != FILE_CLOSED){	/* emergency close */
		fileSetOpenMode(this->destFile, this->uid, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 1, NULL);
	}
	/* close source or all children of source if directory: */
	PROC_WAITFOR_X(5, fileRecursiveClose(this->srcFile, this));
	for(;;){
		PROC_WAITFOR_X(6, srvWantRequest(server, this));
		fromPath = fileBuildPath(fromDir, fromName, 1, fromBuf, CIFS_MAX_PATHLEN);
		toPath = fileBuildPath(toDir, toName, 1, toBuf, CIFS_MAX_PATHLEN);
		PROC_WAITFOR_X(7, cifsRename(smb, this, this->uid, this->tid, fromPath, toPath));
		if(this->makeReadonly || rval->err.class != ERRCLS_CIFS_DOS || (rval->err.code != ERR_CIFS_DOSdestexists && rval->err.code != ERR_CIFS_DOSnoaccess)){
			/* if error code is "destination exists", retry after deleting destination */
			/* win95 uses ERR_CIFS_DOSnoaccess if destination exists */
			break;
		}else{
			PROC_WAITFOR_X(8, srvWantRequest(server, this));
			toPath = fileBuildPath(toDir, toName, 1, toBuf, CIFS_MAX_PATHLEN);
			cifsSetInformation(smb, NULL, this->uid, this->tid, toPath, 0, 0);	/* change to r/w before we attempt delete, don't wait for it */
			PROC_WAITFOR_X(9, srvWantRequest(server, this));
			toPath = fileBuildPath(toDir, toName, 1, toBuf, CIFS_MAX_PATHLEN);
			PROC_WAITFOR_X(10, cifsRemoveCore(smb, this, this->uid, this->tid, toPath, 0));	/* try to delete destination and ignore result */
			/* we must wait for the delete to proceed because NT can return errors otherwise */
			this->makeReadonly = 1;	/* use this flag to avoid endless loop */
		}	/* retry after deleting destination file */
	}
    res.err = rval->err;
	if(srvUseRequest(server, SRV_USE_DIRDATE_EMUL)){
		attrTouch(fromDir, this->user, NULL);
		attrTouch(toDir, this->user, NULL);		/* set directory modification date in background */
	}else{
		attrInvalidate(fromDir);
		attrInvalidate(toDir);
	}
	if(rval->err.class == 0){	/* the rename was successful */
		if(fromDir->cache.any != NULL)
			dirCacheFileRemoved(fromDir->cache.any, fromName);
		if(toDir->cache.any != NULL)
			dirCacheFileAdded(toDir->cache.any, toName);
		fileUnregisterChild(fromDir, this->srcFile, 1);
		memFree(this->srcFile->name);
		this->srcFile->name = uniStrDup(toName);
		licRegisterFile(toDir, this->srcFile, &res.err);
		if(res.err.class != 0)
			fileInvalidate(this->srcFile);
	}else{	/* invalidate the local caches, require new lookup to supply up to date status */
		fileInvalidate(this->srcFile);
	}
	this->srcFile->openLock--;
	if(this->destFile != NULL){
		this->destFile->openLock--;
		fileInvalidate(this->destFile);
		procCondBroadcast(this->destFile->openSignal, NULL);
		fileRelease(this->destFile);
	}
	fileRelease(toDir);
	fileRelease(fromDir);
    PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(this->srcFile->openSignal, NULL);
		fileRelease(this->srcFile);
	);
    PROC_END
}

void	fileRename(cifsFile_t *fromDir, btuni *fromName, cifsFile_t *toDir, btuni *toName, cifsUser_t *user, void *caller)
{
fileCtx_t	*ctx;
beFileRes_t	res;
    
	if(fromDir->children == NULL || toDir->children == NULL){	/* any of the dirs is not a directory */
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
        PROC_IMMEDIATE_RET(caller, &res);
	}
	if(!fileNameIsValid(toName)){
        res.err.class = ERRCLS_UNIX;
        res.err.code = ENOENT;
        PROC_IMMEDIATE_RET(caller, &res);
	}
	if(fileDepth(toDir) > licDepthLimit()){	/* too deep in hierarchy */
		res.err.class = ERRCLS_UNIX;
		res.err.code = ENAMETOOLONG;
        PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, fileRenameProc, caller, filePool, poolFreeBlock);
    ctx->file = fromDir;
    ctx->newFile = toDir;
    ctx->tid = shrTid(fromDir->share);
	ctx->uid = userSmbUid(user);
	ctx->user = user;
    ctx->name = fromName;
    ctx->otherName = toName;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	fileHardLink(cifsFile_t *sourceFile, cifsFile_t *newDir, btuni *newName, cifsUser_t *user, void *caller)
{
beFileRes_t     res;

	res.err.class = ERRCLS_CIFS_DOS;                /* dummy implementation */
	res.err.code = ERR_CIFS_DOSnoaccess;
	PROC_IMMEDIATE_RET(caller, &res);
}

/* ------------------------------------------------------------------------- */

void	fileSymbolicLink(cifsFile_t *newDir, btuni *newName, btuni *path, cifsUser_t *user, struct beFileSetAttr *setAttr, void *caller)
{
	attrCreateLink(newDir, user, newName, path, setAttr, caller);
}

/* ------------------------------------------------------------------------- */
/* ---------------------- file read/write operations ----------------------- */
/* ------------------------------------------------------------------------- */

static void	doAutoclose(void *obj, void *dummy)
{
cifsFile_t	*file = obj;

	if(file->openMode != FILE_CLOSED && file->doAutoclose){
		if(file->openLock || file->needOpen){
			elpCallAtCancelPrev(elpNow() + 200, doAutoclose, file);	/* try again in 200ms */
			return;
		}
		fileSetOpenMode(file, file->openedBy, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 0, NULL);
	}
}

static void	fileUpdateAutoclose(cifsFile_t *file)
{
bttime	closeDelay;

	if(file->openMode != FILE_CLOSED && file->doAutoclose){
		closeDelay = srvFileCloseDelay(shrServer(file->share), file->oplockLevel);
		if(closeDelay <= 0)
			elpCancelCallAt(doAutoclose, file);
		else
			elpCallAtCancelPrev(elpNow() + closeDelay, doAutoclose, file);
	}else{
		elpCancelCallAt(doAutoclose, file);
	}
}

static void	oplockLevelDidChange(cifsFile_t *file)
{
server_t	*server = shrServer(file->share);

	if(file->cache.any != NULL){
		if(file->openMode == FILE_CLOSED){
			cacheInvalidate(file->cache.any);
		}else{
			if(file->children == NULL){
				dcacheSetReadaheadPages(file->cache.any, srvReadAheadPages(server, file->oplockLevel));
			}
		}
	}
}

static void	fileDidOpen(cifsFile_t *file)
{
	shrAddOpenFile(file->share, file);
	if(LOG_MODE_IS_ACTIVE(LOGM_OPLOCK))
		logPrintf(LOGM_OPLOCK, "file: fileDidOpen: granted oplock %d for ->%s<-\n", file->oplockLevel, uniToCstring(file->name));
	oplockLevelDidChange(file);
	fileUpdateAutoclose(file);
	shrFileOpenModeChanged(file->share, file);
}

static void	fileDidClose(cifsFile_t *file)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_OPLOCK))
		logPrintf(LOGM_OPLOCK, "file: fileDidClose: ->%s<-\n", uniToCstring(file->name));
	shrRemoveOpenFile(file->share, file);
	file->fid = -1;
	oplockLevelDidChange(file);
	fileUpdateAutoclose(file);
}

/* ------------------------------------------------------------------------- */

void	fileConnectionLost(cifsFile_t *file)
{
	file->openMode = FILE_CLOSED;
	fileDidClose(file);
	if(file->cache.any != NULL){
		cacheSetOwner(file->cache.any, NULL);
	}
	shrFileOpenModeChanged(file->share, file);
}

/* ------------------------------------------------------------------------- */

static void	fileSetOpenModeProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
filePrivRes_t	res;
smbRequest_t	*smb = arg;
btuni			*path, pathBuffer[CIFS_MAX_PATHLEN];
cifsResFileOp_t	*rval = arg;
cifsFile_t		*file = this->file;
server_t		*server = shrServer(file->share);

	PROC_BEGIN
	fileRetain(this->file);
	if(!this->noLocking)
		while(file->openLock){		/* wait until opening/closing has stabilized */
			PROC_WAITFOR_X(1, procCondWait(file->openSignal, this));
		}
	if(((file->openMode ^ this->mode) & this->mask) == 0 && (file->openedBy == this->uid || this->mode == FILE_CLOSED)){
		res.err.class = res.err.code = 0;
		res.didRW = 0;
		fileRelease(file);
		PROC_RETURN(&res);	/* nothing to be done... */
	}
	if(!this->noLocking){
		file->openLock++;	/* grab the open lock */
/* If one of our parent directories is being renamed, wait with opening the
 * file until the parent rename is ready. Closing does not need to wait.
 * We may do this while we own the open lock because fileRename() does never
 * try to open a file and thus can't deadlock with us */
 		if(this->mode != FILE_CLOSED){	/* file should open */
			for(this->newFile = file->parent; this->newFile != NULL; this->newFile = this->newFile->parent){
				if(this->newFile->openLock){
					PROC_WAITFOR_X(3, procCondWait(this->newFile->openSignal, this));
					this->newFile = file;	/* continue where we started... */
				}
			}
		/* we get here only if all parents are not open-locked (=in rename) */
		}
	}
	if(file->openMode != FILE_CLOSED){
		file->openMode = FILE_CLOSED;
		shrFileOpenModeChanged(file->share, file);
        PROC_WAITFOR_X(4, srvWantRequest(shrServer(file->share), this));
		if(this->mode != FILE_CLOSED){	/* in case of re-open, don't wait for close result */
			cifsClose(smb, NULL, file->openedBy, this->tid, file->fid, attrGetWtime(file));
		}else{
			PROC_WAITFOR_X(5, cifsClose(smb, this, file->openedBy, this->tid, file->fid, attrGetWtime(file)));
		}
		fileDidClose(file);
	}else if(this->mode != FILE_CLOSED){	/* if file was closed and should open, update autoclose */
		file->doAutoclose = this->autoclose;
	}
	if(this->mode != FILE_CLOSED){
		PROC_WAITFOR_X(6, srvWantRequest(server, this));
		path = fileBuildPath(file, NULL, 0, pathBuffer, CIFS_MAX_PATHLEN);
		if((this->op == FILE_OP_READ && srvUseRequest(server, SRV_USE_READ_OPENREAD)) || (this->op == FILE_OP_WRITE && srvUseRequest(server, SRV_USE_WRITE_OPENWRITE))){
			PROC_WAITFOR_X(7,
				if(this->op == FILE_OP_READ){
					cifsOpenRead(smb, this, this->uid, this->tid, path, this->offset, this->len, this->mode == FILE_OPEN_READ);
				}else{
					cifsOpenWrite(smb, this, this->uid, this->tid, path, this->buffer, this->offset, this->len, this->mode == FILE_OPEN_READ, this->writeThrough);
				}
			)
		}else{
			if(srvUseRequest(server, SRV_USE_OPEN_OPENX)){
				PROC_WAITFOR_X(8, cifsOpen(smb, this, this->uid, this->tid, path, this->mode == FILE_OPEN_READ, 0, 0, 0, 0, 0));
			}else{
				PROC_WAITFOR_X(9, cifsOpenCore(smb, this, this->uid, this->tid, path, this->mode == FILE_OPEN_READ));
			}
		}
		if(rval->didOpen){
    		file->fid = rval->fid;
			file->oplockLevel = rval->oplock;
    		file->openMode = this->mode;
    		file->openedBy = this->uid;
			fileDidOpen(file);
		}
		res.didRW = rval->didRW;
		res.data = rval->data;
		res.dataLen = rval->dataLen;
	}else{
		file->doAutoclose = 0;	/* closed files can't have autoclose attribute */
		res.didRW = 0;
	}
	res.err = rval->err;
	if(!this->noLocking)
		file->openLock--;
    PROC_RETURN_AND_CLEANUP(&res,
/* It's important to send the broadcast after returning. Otherwise concurrent
 * open/close have precedence over a read/write process that did this
 * open/close. It would be possible that files close that read processes need
 * open.
 */
		if(!this->noLocking)
			procCondBroadcast(file->openSignal, NULL);
		fileRelease(file);
	);
	PROC_END
}

void	fileSetOpenMode(cifsFile_t *file, btu16 cifsUid, btu8 desiredMode, btu8 relevantMask, btbool autoclose, int additionalOperation, btu8 *src, btoffset offset, btsize byteCount, btbool writeThrough, btbool noLocking, void *caller)
{
fileCtx_t	*ctx;

	PROC_INIT(ctx, fileSetOpenModeProc, caller, filePool, poolFreeBlock);
	ctx->file = file;
	ctx->tid = shrTid(file->share);
	ctx->uid = desiredMode == FILE_CLOSED ? file->openedBy : cifsUid;
	ctx->mode = desiredMode;
	ctx->mask = relevantMask;
	ctx->op = additionalOperation;
	ctx->buffer = src;
	ctx->offset = offset;
	ctx->len = byteCount;
	ctx->writeThrough = writeThrough;
	ctx->noLocking = noLocking;
	ctx->autoclose = autoclose;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	fileForceOpenWriteProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
beFileRes_t		*frval = arg;
filePrivRes_t	res, *rval = arg;
cifsFile_t		*file = this->file;

	PROC_BEGIN
	fileRetain(file);
	while(file->openLock || ((fileAttr_t *)file->attributes)->attrLocked){	/* wait until opening/closing/attr has stabilized */
		PROC_WAITFOR_X(1, procCondWait(file->openSignal, this));
	}
	if(file->openMode == FILE_OPEN_RW && file->openedBy == this->uid){
		res.err.class = res.err.code = 0;
		res.didRW = 0;
		fileRelease(file);
		PROC_RETURN(&res);
	}
	/* lock attributes and openLock at the same time to avoid deadlock! */
	((fileAttr_t *)file->attributes)->attrLocked++;
	file->openLock++;
/* If one of our parent directories is being renamed, wait with opening the
 * file until the parent rename is ready. Closing does not need to wait.
 * We may do this while we own the open lock because fileRename() does never
 * try to open a file and thus can't deadlock with us */
	for(this->newFile = file->parent; this->newFile != NULL; this->newFile = this->newFile->parent){
		if(this->newFile->openLock){
			PROC_WAITFOR_X(2, procCondWait(this->newFile->openSignal, this));
			this->newFile = file;	/* continue where we started... */
		}
	}
	/* we get here only if all parents are not open-locked (=in rename) */
	/* if the file is known to be write-protected: */
	if(attrValid(file, this->user) && !attrGetWritable(file, this->user)){	/* file is not writable */
	/* jump here if we did not know that file is write-protected but open failed: */
retry:
		if(file->openMode != FILE_CLOSED){
			fileSetOpenMode(file, this->uid, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 1, NULL);	/* don't wait for close */
		}
		PROC_WAITFOR_X(3, attrSetWritable(file, 1, this->user, 1, this));	/* this succeeds only after close */
		if(frval->err.code != 0){
			res.err = frval->err;
			res.didRW = 0;
			goto errorOccured;
		}
		this->makeReadonly = 1;
	}
	PROC_WAITFOR_X(4, fileSetOpenMode(file, this->uid, FILE_OPEN_RW, FILE_OPEN_RW, 1, this->op, this->buffer, this->offset, this->len, this->writeThrough, 1, this));
	this->pres = *rval;
	if(!this->makeReadonly && rval->err.class == ERRCLS_CIFS_DOS && rval->err.code == ERR_CIFS_DOSnoaccess)	/* first try, but access denied */
		goto retry;
	if(this->makeReadonly && !srvCloseForSetattr(shrServer(file->share))){	/* don't change back if impossible */
		PROC_WAITFOR_X(5, attrSetWritable(file, 0, this->user, 1, this));
		if(rval->err.code != 0 && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "file: error reverting file \"%s\" to read only: %s\n", uniToCstring(file->name), errDescription(&rval->err));
	}
	res = this->pres;
errorOccured:
	file->openLock--;
	((fileAttr_t *)file->attributes)->attrLocked--;
    PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(file->openSignal, NULL);	/* It's important to do this after returning! */
		fileRelease(file);
	);
	PROC_END
}

void	fileForceOpenWrite(cifsFile_t *file, cifsUser_t *user, int additionalOperation, btu8 *src, btoffset offset, btsize byteCount, btbool writeThrough, void *caller)
{
fileCtx_t	*ctx;

	PROC_INIT(ctx, fileForceOpenWriteProc, caller, filePool, poolFreeBlock);
	ctx->file = file;
	ctx->tid = shrTid(file->share);
	ctx->uid = userSmbUid(user);
	ctx->user = user;
	ctx->op = additionalOperation;
	ctx->buffer = src;
	ctx->offset = offset;
	ctx->len = byteCount;
	ctx->writeThrough = writeThrough;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	fileOplockBreakProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
cifsFile_t		*file = this->file;
smbRequest_t	*smb = arg;

	PROC_BEGIN
	fileRetain(file);
	this->op = file->oplockLevel;	/* store old oplockLevel */
	file->oplockLevel = this->oplock;
	oplockLevelDidChange(file);
	if(this->oplock < this->op && file->cache.any != NULL){	/* lock level becomes smaller */
		PROC_WAITFOR_X(1, cacheWriteBack(file->cache.any, this));
		if(this->oplock == CIFS_OPLOCK_NONE && file->cache.any != NULL)
			cacheInvalidate(this->file->cache.any);
	}
	if(file->openMode != FILE_CLOSED && this->op == CIFS_OPLOCK_BATCH){	/* batch oplock broken */
		if(LOG_MODE_IS_ACTIVE(LOGM_OPLOCK))
			logPrintf(LOGM_OPLOCK, "file: closing due to oplock break\n");
		PROC_WAITFOR_X(2, fileSetOpenMode(file, file->openedBy, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 1, 0, this));
	}else{
		PROC_WAITFOR_X(3, srvWantRequest(shrServer(this->file->share), this));
		if(file->openMode != FILE_CLOSED && this->op != CIFS_OPLOCK_L2)	/* don't send ack to break of level 2 oplock */
			cifsAckOplockBreak(smb, this->tid, file->fid, file->oplockLevel);
	}
	fileRelease(file);
	PROC_END
}

void	fileOplockBreak(cifsFile_t *file, btint mode)
{
fileCtx_t	*ctx;

	if(LOG_MODE_IS_ACTIVE(LOGM_OPLOCK))
		logPrintf(LOGM_OPLOCK, "file: %s: breaking oplock to %d\n", uniToCstring(file->name), mode);
	PROC_INIT(ctx, fileOplockBreakProc, NULL, filePool, poolFreeBlock);
    ctx->file = file;
    ctx->tid = shrTid(file->share);
	ctx->oplock = mode;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	fileRWUncachedProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
smbRequest_t	*smb = arg;
cifsResFileOp_t	*rval = arg;
filePrivRes_t	res, *prval = arg;
cifsFile_t		*file = this->file;
server_t		*server = shrServer(file->share);

	PROC_BEGIN
	fileRetain(file);
	file->needOpen++;
	PROC_WAITFOR_X(1,
		/* call the set-openmode function in any case (even if open mode is
		 * appropriate) since it waits for the lock */
		if(this->op == FILE_OP_WRITE || this->writeThrough)
			fileForceOpenWrite(file, this->user, this->op, this->buffer, this->offset, this->len, this->writeThrough, this);
		else
			fileSetOpenMode(file, this->uid, FILE_OPEN_READ, FILE_OPEN_READ, 1, this->op, NULL, this->offset, this->len, this->writeThrough, 0, this);
	)
	res.err = prval->err;
	if(prval->err.class == 0 && !prval->didRW){
		PROC_WAITFOR_X(2, srvWantRequest(server, this));
		PROC_WAITFOR_X(3,
			if(this->op == FILE_OP_READ && srvUseRequest(server, SRV_USE_READ_READX)){
				cifsRead(smb, this, this->uid, this->tid, this->file->fid, this->offset, this->len);
			}else if(this->op == FILE_OP_READ){
				cifsReadCore(smb, this, this->uid, this->tid, this->file->fid, this->offset, this->len);
			}else if(this->op == FILE_OP_WRITE && srvUseRequest(server, SRV_USE_WRITE_WRITEX)){
				cifsWrite(smb, this, this->uid, this->tid, this->file->fid, this->buffer, this->offset, this->len, this->writeThrough);
			}else{
				cifsWriteCore(smb, this, this->uid, this->tid, this->file->fid, this->buffer, this->offset, this->len);
			}
		)
		fileUpdateAutoclose(file);
		res.err = rval->err;
		res.data = rval->data;
		res.dataLen = rval->dataLen;
	}
    if(this->op == FILE_OP_WRITE){
        if(prval->err.class != 0){
            attrInvalidate(file);
        }else{
			file->lastWrite = elpNow();
			attrUpdateAfterWrite(file, this->offset, this->len);	/* if the result is OK, the number of bytes should be OK, too */
  		}
    }
	--file->needOpen;
	PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(file->openSignal, NULL);	/* someone may be waiting for needOpen */
		fileRelease(file);
	);
	PROC_END
}


void	fileRWUncached(cifsFile_t *file, int operation, btbool writeThrough, btu8 *src, btoffset offset, btsize len, cifsUser_t *user, void *caller)
{
fileCtx_t		*ctx;
filePrivRes_t	res;

	if(file->children != NULL){
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	if(len == 0){
		res.err.class = res.err.code = 0;
		res.data = NULL;
		res.dataLen = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, fileRWUncachedProc, caller, filePool, poolFreeBlock);
	ctx->file = file;
	ctx->tid = shrTid(file->share);
	ctx->uid = userSmbUid(user);
	ctx->user = user;
	ctx->op = operation;
	ctx->buffer = src;
	ctx->offset = offset;
	ctx->len = len;
	ctx->writeThrough = writeThrough;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	fileRW(cifsFile_t *file, int operation, btbool writeThrough, btu8 *buffer, btoffset offset, btsize len, cifsUser_t *user, void *caller)
{
beFileRes_t	res;

	if(file->children != NULL){		/* must not be directory */
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	if(len == 0){
		res.err.class = res.err.code = 0;
		res.dataLen = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	if(operation == FILE_OP_READ && attrValid(file, user) && offset >= attrFileSize(file)){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "file: shortcut in read(offset=%d) for ->%s<-\n", (int)offset, uniToCstring(file->name));
		res.err.class = res.err.code = 0;
		res.dataLen = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	cacheAcquire(CACHE_TYPE_DATA, file);
	if(srvUseWriteThrough(shrServer(file->share), file->oplockLevel))
		writeThrough = 1;
    dcacheRW(file->cache.data, operation, writeThrough, buffer, offset, len, user, caller);
}

/* ------------------------------------------------------------------------- */

static void	fileTruncateProc(void *obj, void *arg)
{
fileCtx_t		*this = obj;
cifsFile_t		*file = this->file;
smbRequest_t	*smb = arg;
beFileRes_t		res;
filePrivRes_t	*prval = arg;
cifsResFileOp_t	*rval = arg;

	PROC_BEGIN
	fileRetain(file);
	file->needOpen++;
	PROC_WAITFOR_X(1, fileForceOpenWrite(file, this->user, FILE_OP_NONE, NULL, 0, 0, 0, this));
	res.err = prval->err;
	if(rval->err.class == 0){	/* open succeeded */
		if(file->cache.any != NULL){
			PROC_WAITFOR_X(2, cacheWriteBack(file->cache.any, this));
		}
		fileUpdateAutoclose(file);
        PROC_WAITFOR_X(3, srvWantRequest(shrServer(file->share), this));
        PROC_WAITFOR_X(4, cifsWriteCore(smb, this, this->uid, this->tid, file->fid, (void *)&res, this->offset, 0));	/* &res is dummy */
		res.err = rval->err;
		if(file->cache.data != NULL)
			dcacheTruncate(file->cache.data, this->offset);
		attrTruncate(file, this->offset);
		file->lastWrite = elpNow();
	}
	--file->needOpen;
	PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(file->openSignal, NULL);	/* someone may be waiting for needOpen */
		fileRelease(file);
	);
	PROC_END
}

void	fileTruncate(cifsFile_t *file, btoffset newSize, cifsUser_t *user, void *caller)
{
fileCtx_t	*ctx;
beFileRes_t	res;

	if(file->children != NULL){
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, fileTruncateProc, caller, filePool, poolFreeBlock);
	ctx->file = file;
	ctx->tid = shrTid(file->share);
	ctx->uid = userSmbUid(user);
	ctx->user = user;
	ctx->offset = newSize;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

