/*
 * Name: fileattr.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-08
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: fileattr.c,v 1.4 2000/04/15 13:27:55 cs Exp $
 */

#include "stdheaders.h"
#include "fileattr.h"
#include "share.h"
#include "file.h"
#include "user.h"
#include "server.h"
#include "foundationconst.h"
#include "evntloop.h"
#include "backend.h"
#include "logging.h"


/* --------------------- Timezone and DST handling --------------------- */

typedef struct gmtOffsetEntry{
	btu32	gmtStart;
	btint	offset;
}gmtOffsetEntry_t;

/* estimated maximum table size:
 * time_t is 32 bits -> spans 137 years
 * 2 DST switches per year -> 274 entries
 */
static gmtOffsetEntry_t	gmtOffsetTable[300];
static btint			gmtOffsetTableEntryCount;
static btint			averageGmtOffset;

/* ------------------------------------------------------------------------- */

static inline btint leapDaysUpToYear(btint yearSince1900)
{
btint	yearBefore = yearSince1900 + 1899;

/* Leap years are every 4 years except every century unless it's a
 * multiple of 400.
 */
	return yearBefore/4 - yearBefore/100 + yearBefore/400;
}

/* ------------------------------------------------------------------------- */
/* This function is the real source of GMT offset wisdom. It asks libc. This,
 * of course, requires that the machine's resp. user's timezone is set up
 * correctly.
 * gmtOffset is defined by:   localTime = GMT + gmtOffset(GMT)
 */
static btint	gmtOffset(btu32 gmtVal)
{
struct tm	*local, gmt[1];
btint		difference;
time_t		theTime = gmtVal;

	gmt[0] = *gmtime(&theTime);
	local = localtime(&theTime);
 	difference = (local->tm_year - gmt->tm_year) * 365;
	difference += leapDaysUpToYear(local->tm_year) - leapDaysUpToYear(gmt->tm_year);
	difference += local->tm_yday - gmt->tm_yday;
	difference = difference * 24 + local->tm_hour - gmt->tm_hour;
	difference = difference * 60 + local->tm_min - gmt->tm_min;
	difference = difference * 60 + local->tm_sec - gmt->tm_sec;
	return difference;
}

/* ------------------------------------------------------------------------- */

static btu32	binarySearchForFirstDifferent(btu32 searchFrom, btu32 searchTo)
{
btu32	s, lastS = searchFrom;
btint	initialOffset = gmtOffset(searchFrom);

	for(;;){
		s = ((btu64)searchFrom + searchTo) / 2;
		if(s == lastS)
			return searchTo;
		lastS = s;
		if(gmtOffset(s) != initialOffset)
			searchTo = s;
		else
			searchFrom = s;
	}
}

static void	gmtOffsetTableInit(void)
{
btu32	s, nextS, start;
btint	i = 0, offset, nextOffset, offsetSum = 0;
time_t	forCtime;

	tzset();
	s = 0;
	offset = gmtOffset(s);
	gmtOffsetTable[i].gmtStart = s;
	gmtOffsetTable[i++].offset = offset;
	offsetSum += offset;
	for(;;){	/* do a rough linear search for DST switches */
		nextS = s + 86400 * 7;	/* search in terms of weeks */
		if(nextS < s)	/* wrap-around occured */
			break;
		nextOffset = gmtOffset(nextS);
		if(nextOffset != offset){
			start = binarySearchForFirstDifferent(s, nextS);
			gmtOffsetTable[i].gmtStart = start;
			gmtOffsetTable[i++].offset = nextOffset;
			offsetSum += nextOffset;
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
				forCtime = start;
				logPrintf(LOGM_DEBUG, "fileattr/gmtOffsetTableInit: GMT offset %d starts at %s", nextOffset, ctime(&forCtime));
			}
			if(i >= ELEMENTS(gmtOffsetTable))
				break;
		}
		offset = nextOffset;
		s = nextS;
	}
	gmtOffsetTableEntryCount = i;
	averageGmtOffset = offsetSum / gmtOffsetTableEntryCount;
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "fileattr/gmtOffsetTableInit: GMT table has %d entries (average offset is %d)\n", i, averageGmtOffset);
}

/* ------------------------------------------------------------------------- */

btint	attrLocalGmtOffset(btu32 gmtVal)
{
btint	s, lastS = 0, searchFrom = 0, searchTo = gmtOffsetTableEntryCount;

	for(;;){
		s = (searchFrom + searchTo) / 2;
		if(s == lastS)
			return gmtOffsetTable[searchFrom].offset;
		lastS = s;
		if(gmtOffsetTable[s].gmtStart <= gmtVal)
			searchFrom = s;
		else
			searchTo = s;
	}
}

/* ------------------------------------------------------------------------- */

static btu32	localFromGmt(btu32 gmtVal)
{
btu32	local;

	local = gmtVal + attrLocalGmtOffset(gmtVal);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_ATTR)){
		time_t		localT = local, gmtT = gmtVal;
		struct tm	localTm[1], *gmtTm;
		localTm[0] = *gmtime(&localT);
		gmtTm = gmtime(&gmtT);
		logPrintf(LOGM_DEBUG_ATTR, "fileattr localFromGmt(%04d-%02d-%02d %02d:%02d.%02d) = %04d-%02d-%02d %02d:%02d.%02d\n", gmtTm->tm_year+1900, gmtTm->tm_mon+1, gmtTm->tm_mday, gmtTm->tm_hour, gmtTm->tm_min, gmtTm->tm_sec, localTm->tm_year+1900, localTm->tm_mon+1, localTm->tm_mday, localTm->tm_hour, localTm->tm_min, localTm->tm_sec);
	}
	return local;
}

static btu32	gmtFromLocal(btu32 localVal)
{
btu32	gmt;

	gmt = localVal - attrLocalGmtOffset(localVal - averageGmtOffset);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_ATTR)){
		time_t	localT = localVal, gmtT = gmt;
		struct tm	localTm[1], *gmtTm;
		localTm[0] = *gmtime(&localT);
		gmtTm = gmtime(&gmtT);
		logPrintf(LOGM_DEBUG_ATTR, "fileattr gmtFromLocal(%04d-%02d-%02d %02d:%02d.%02d) = %04d-%02d-%02d %02d:%02d.%02d\n", localTm->tm_year+1900, localTm->tm_mon+1, localTm->tm_mday, localTm->tm_hour, localTm->tm_min, localTm->tm_sec, gmtTm->tm_year+1900, gmtTm->tm_mon+1, gmtTm->tm_mday, gmtTm->tm_hour, gmtTm->tm_min, gmtTm->tm_sec);
	}
	return gmt;
}

/* ------------------------------------------------------------------------- */
/* ----------------------- data conversion routines ------------------------ */
/* ------------------------------------------------------------------------- */


/* --------------- convert to internal representation --------------- */

inline bti64	attrTimeFromUnixTime(btu32 utime, cifsFile_t *file)
{
bti32	serverOffset = srvToLocalTimezoneOffset(shrServer(file->share));

    return ((bti64)gmtFromLocal(utime + serverOffset) + FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL) * (10LL * 1000LL * 1000LL);
}

/* this looks horrible, but mktime() does not work properly on NEXTSTEP */
static btu32	secondsSince1970(btint year, btint month, btint day, btint hour, btint minute, btint second)
{
static btint	monthBase[] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};
btint			days, seconds;

	seconds = (((hour * 60) + minute) * 60) + second;
	year += month / 12;
	month %= 12;
/* We take advantage of the fact that the only century inside the 32 bit range
 * of secondsSince1970() is the year 2000, which IS a leap year. We therefore
 * need NO century and no 4-century correction term in the leap year
 * computation.
 */
	days = (year - 70) * 365 + (year - 69)/4 + 1;
	days += monthBase[month];
	if(!(month > 1 && (year % 4) == 0)){	/* not >= March in leap year */
		days--;
	}
	days += day;
	return days * 86400 + seconds;
}

bti64	attrTimeFromDateTime(btu16 date, btu16 time, cifsFile_t *file)
{
btu32			utime;

	utime = secondsSince1970(
			(date >> 9) + 80,					/* year */
			((date >> 5) & 0xf) - 1,			/* month */
			(date & 0x1f) - 1,					/* day */
			time >> 11,							/* hour */
			(time >> 5) & 0x3f,					/* minute */
			(time & 0x1f) * 2					/* second */
		);
    return attrTimeFromUnixTime(utime, file);
}

bti64	attrTimeFromServerTime(bti64 serverTime, struct server *server)
{
/* ### This is currently only used by Unix extensions which are not
 * implemented completely. Check specification for timezone thing if cifsUnix
 * is implemented.
 */
	return serverTime + (bti64)(srvToLocalTimezoneOffset(server)) * (10LL * 1000LL * 1000LL);
}

bti64	attrTimeFromVfsTime(struct beFileTime *t)
{
    return ((bti64)t->seconds + FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL) * (10LL * 1000LL * 1000LL) + t->nseconds / 100;
}


/* --------------- convert to server representation --------------- */

btu32	attrUnixTimeFromTime(bti64 tm, cifsFile_t *file)
{
bti32	serverOffset = srvToLocalTimezoneOffset(shrServer(file->share));

    tm /= 10;							/* time is in microseconds now */
    tm -= FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL * 1000LL * 1000LL;	/* convert to 1970-based time */
    return localFromGmt(tm / 1000000) - serverOffset;
}

void	attrDateTimeFromTime(bti64 time, btu16 *timep, btu16 *datep, cifsFile_t *file)
{
struct tm	*tm;
time_t		uxTime;
    
	uxTime = attrUnixTimeFromTime(time, file);
    tm = gmtime(&uxTime);
    *timep = (tm->tm_sec >> 1) | (tm->tm_min << 5) | (tm->tm_hour << 11);
    *datep = tm->tm_mday | ((tm->tm_mon + 1) << 5) | ((tm->tm_year - 80) << 9);
}

bti64	attrServerTimeFromTime(bti64 serverTime, server_t *srvContext)
{
/* ### This is currently only used by Unix extensions which are not
 * implemented completely. Check specification for timezone thing if cifsUnix
 * is implemented.
 */
	return serverTime - (bti64)srvToLocalTimezoneOffset(srvContext) * (10LL * 1000LL * 1000LL);
}

/* ------------------- convert to client representation -------------------- */

void	attrFileVfsTime(struct beFileTime *vt, bti64 tm)
{
    tm -= FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL * 1000LL * 10000LL;	/* convert to 1970-based time */
    vt->seconds = tm / 10000000;
    vt->nseconds = (tm % 10000000) * 100;
}

/* ------------------------- attribute conversion -------------------------- */

static inline btint	attrMask(btint xmap)
{
	switch(xmap & ~SRV_EXMAP_INVERT){
		default:
		case SRV_EXMAP_ARCHIVE:	return CIFS_ATTR_ARCHIVE;
		case SRV_EXMAP_SYSTEM:	return CIFS_ATTR_SYSTEM;
		case SRV_EXMAP_HIDDEN:	return CIFS_ATTR_HIDDEN;
		case SRV_EXMAP_ON:
		case SRV_EXMAP_OFF:		return 0;
	}
}

btu32	attrUnixModeFromDOSAttr(btint dosMode, btint xmap)
{
btu32	mode = 0600;

    if(dosMode & CIFS_ATTR_READONLY)
        mode &= ~0200;
	if((dosMode & attrMask(xmap)) != 0 || (xmap & ~SRV_EXMAP_INVERT) == SRV_EXMAP_ON)
		mode |= 0100;	/* map bit to execute */
	if((xmap & SRV_EXMAP_INVERT) != 0)	/* invert execute bit on demand */
		mode ^= 0100;
	if((dosMode & CIFS_ATTR_DIR) != 0)	/* directories must have execute bit set */
		mode |= 0100;
	return mode;
}

btint	attrDOSAttrFromUnixMode(btu32 mode, btint oldDosMode, btint xmap)
{
btint	mask, dosMode;

	if((oldDosMode & CIFS_ATTR_DIR) != 0)
		xmap = SRV_EXMAP_ON;	/* exec flag is always on for directories */
	if((xmap & SRV_EXMAP_INVERT) != 0)	/* invert execute bit on demand */
		mode ^= 0100;
	dosMode = 0;
	mask = attrMask(xmap);
	if((mode & 0100) != 0)	/* execute bit */
		dosMode |= mask;
	if((mode & 0200) == 0)	/* write bit */
		dosMode |= CIFS_ATTR_READONLY;
	mask |= CIFS_ATTR_READONLY;
	return (oldDosMode & ~mask) | dosMode;
}

/* ------------------------------------------------------------------------- */
/* -------------------------- superclass methods --------------------------- */
/* ------------------------------------------------------------------------- */

void	attrInit(void *obj, fileAttrInterface_t *interface)
{
fileAttr_t	*attr = obj;

	attr->interface = interface;
	if(gmtOffsetTableEntryCount == 0){
		gmtOffsetTableInit();
	}
}

/* ------------------------------------------------------------------------- */

void	attrExit(void *obj)
{
fileAttr_t	*attr = obj;

	attr->interface = NULL;
}

/* ------------------------------------------------------------------------- */

btbool	attrGeneralValid(struct cifsFile *file, struct cifsUser *user)
{
bttime		age;
fileAttr_t	*attr = file->attributes;

	if(attr->updatedAt == 0)
		return 0;
	age = elpNow() - attr->updatedAt;
	return age < srvAttributesCacheTime(shrServer(fileShare(file))) && age >= 0;
}

/* ------------------------------------------------------------------------- */

void	attrGeneralInvalidate(struct cifsFile *file)
{
fileAttr_t	*attr = file->attributes;

	attr->updatedAt = 0;
}

/* ------------------------------------------------------------------------- */

void	attrGeneralTruncate(struct cifsFile *file, btoffset newSize)
{
fileAttr_t	*attr = file->attributes;

	if(attr->size != newSize){
		attr->size = newSize;
		/* the following is an approximation, but it should suffice */
		attr->allocSize = ((newSize / 32768) + 1) * 32768;
	}
	attr->wtime = elpNow64();
}

/* ------------------------------------------------------------------------- */

void	attrGeneralUpdateAfterWrite(struct cifsFile *file, btoffset offset, btsize bytesWritten)
{
btoffset	minSize = offset + bytesWritten;
fileAttr_t	*attr = file->attributes;

	if(attr->size < minSize){
		attr->size = minSize;
		/* the following is an approximation, but it should suffice */
		attr->allocSize = ((minSize / 32768) + 1) * 32768;
	}
	attr->wtime = elpNow64();
}

/* ------------------------------------------------------------------------- */

btint	attrDOSFromAttributes(struct beFileSetAttr *setAttr, struct cifsUser *userContext, struct server *serverContext)
{
	/* This function is only used to set the attributes of newly created
	 * files. It is thus sufficent to ignore the user context and use
	 * the user attributes.
	 */
	 return attrDOSAttrFromUnixMode(setAttr->setUnixMode ? setAttr->unixMode : 0666, 0, srvExecMapping(serverContext));
}

/* ------------------------------------------------------------------------- */
