/*
 * Name: nbnsclient.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-11-17
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nbnsclient.c,v 1.10 2002/04/05 17:53:49 cs Exp $
 */

#include "stdheaders.h"
#include "nbnsclient.h"
#include "process.h"
#include "main.h"
#include "cfgparse.h"
#include "blockpool.h"
#include "smbnumbers.h"
#include "logging.h"
#include "evntloop.h"
#include "system.h"
#include "util.h"
#include "errcodes.h"
#include "cifsmod.h"
#include "unicode.h"
#include "dgsocket.h"
#include "ipc.h"


#define	NBNS_PORT	137

typedef struct nbnsCtx{
	PROC_VARSET;
	btu8		*block;
	btsize		len;
	btu32		destIp;
	dgSocket_t	*sock;
	bttime		absMaxTime;
	bttime		timeout;
	btu16		xid;
	btbool		abortIfTimeout;
	btu8		buffer[580];
	ipcMsg_t	*msg;
}nbnsCtx_t;

static blockpool_t *nbnsPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(nbnsPool, block);
}

static btu16	xidCounter;

/* ------------------------------------------------------------------------- */

static void	nbnsRunQueryProc(void *obj, void *arg)
{
nbnsCtx_t			*this = obj;
btint				numEntries;
btu16				xid;
nbnsRes_t			res;
btError				*rval = arg;
dgSocketRx_t		*rxMsg = arg;
cfgProperty_t		*path[8];
bttime				nextTimeout;
static bttime		maxTime, initialTimeout;
static btint		bindNbnsPort;
static btu32		cfgChg = 0;

	PROC_BEGIN;
	if(CFG_DID_CHANGE(cfgChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		initialTimeout = cfgIntValue(cfgQueryFromArray("nbnsInitialTimeout", numEntries, path), 100);
		maxTime = cfgIntValue(cfgQueryFromArray("nbnsTotalTimeout", numEntries, path), 1000);
		bindNbnsPort = cfgIntValue(cfgQueryFromArray("bindNbnsPort", numEntries, path), NBNS_PORT);
	}
/* Try to bind to port 137. If samba is running, this will fail and we don't
 * care because samba should be configured as LMB and samba is not broken.
 * If samba is not running, we will succeed and thus work around broken
 * Windows 95 LMBs.
 */
	this->sock = dgSocketForPort(bindNbnsPort);
	this->xid = ((btu16)this->block[0] << 8) | this->block[1];
	this->timeout = initialTimeout;
	this->absMaxTime = elpNow() + maxTime;
	for(;;){
		if(this->destIp != 0xffffffff){
			PROC_WAITFOR_X(1, dgSocketSend(this, this->sock, this->destIp, NBNS_PORT, this->block, this->len));
		}else{
			PROC_WAITFOR_X(2, dgSocketBroadcast(this, this->sock, NBNS_PORT, this->block, this->len));
		}
		if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
			logPrintf(LOGM_NBNS, "nbns: sent request to 0x%08x: %02x %02x %02x %02x...\n", this->destIp, this->block[0], this->block[1], this->block[2], this->block[3]);
		if(rval->class != 0){
			res.err = *rval;
			goto doReturn;
		}
		nextTimeout = elpNow() + this->timeout;
		if(nextTimeout - this->absMaxTime >= -50){
			this->abortIfTimeout = 1;
			nextTimeout = this->absMaxTime;
		}
		elpCallAtCancelPrev(nextTimeout, nbnsRunQueryProc, this);	/* setup timeout */
		for(;;){
			PROC_WAITFOR_X(3, procCondWait(this->sock->msgReceivedSignal, this));
			if(rxMsg != NULL){	/* no timeout -- message arrived */
				if((rxMsg->data[2] & 0x80) == 0){	/* this is a request, not a response */
					if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
						logPrintf(LOGM_NBNS, "nbns: received request instead of reply, ignoring\n");
					continue;
				}
				xid = ((btu16)rxMsg->data[0] << 8) | rxMsg->data[1];
				if(xid != this->xid){	/* is it for us? */
					if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
						logPrintf(LOGM_NBNS, "nbns: received xid=%d instead of %d, ignoring\n", xid, this->xid);
					continue;
				}
				elpCancelCallAt(nbnsRunQueryProc, this);
			}else{				/* timeout occured */
				procCondRemoveWait(this->sock->msgReceivedSignal, this);
				break;
			}
			if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
				logPrintf(LOGM_NBNS, "nbns: received reply %02x %02x %02x %02x...\n", rxMsg->data[0], rxMsg->data[1], rxMsg->data[2], rxMsg->data[3]);
			elpCancelCallAt(nbnsRunQueryProc, this);
			memset(&res.err, 0, sizeof(res.err));
			res.data = rxMsg->data;
			res.len = rxMsg->len;
			res.ipAddress = rxMsg->remoteIp;
			goto doReturn;	/* reply arrived */
		}
		if(this->abortIfTimeout){
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOCtimeout;
			goto doReturn;
		}
		this->timeout *= 2;	/* ... and retry */
	}
doReturn:
	if(this->sock != NULL){
		dgSocketRelease(this->sock);
		this->sock = NULL;
	}
	PROC_RETURN(&res);
	PROC_END;
}

static void	nbnsRunQuery(btu8 *block, btsize len, btu32 destIp, void *caller)
{
nbnsCtx_t	*ctx;

    if(nbnsPool == NULL)
        nbnsPool = poolNew(sizeof(nbnsCtx_t), (btu8 *)&ctx->buffer - (btu8 *)ctx);
	PROC_INIT(ctx, nbnsRunQueryProc, caller, nbnsPool, poolFreeBlock);
	ctx->block = block;
	ctx->len = len;
	ctx->destIp = destIp;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

btsize	nbnsEncodeName(btu8 *dest, btsize bufSize, btu8 *name, btint tag)
{
btint			i, len, numEntries;
btu8			c, *s = dest;
cfgProperty_t	*path[8];
static btu32	scopeChg = 0;
static btu8		*scopeID;

	*dest++ = 32;
	for(i=0;i<15;i++){
		if(*name == 0)	/* name may be null-terminated string */
			break;
		c = asciiCharToupper(*name);
		name++;
		*dest++ = 'A' + ((c >> 4) & 0xf);
		*dest++ = 'A' + (c & 0xf);
	}
	for(;i<15;i++){		/* fill rest with spaces */
		*dest++ = 'A' + 2;
		*dest++ = 'A' + 0;
	}
	*dest++ = 'A' + ((tag >> 4) & 0xf);
	*dest++ = 'A' + (tag & 0xf);
	*dest++ = 0;
	bufSize -= 34;
	if(CFG_DID_CHANGE(scopeChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		scopeID = cfgStringValue(cfgQueryFromArray("scopeID", numEntries, path), NULL);
	}
	if(scopeID != NULL){
		for(i=len=0;bufSize > 0;i++,len++){
			switch(scopeID[i]){
			case 0:
				dest[-1] = len;
				if(len > 0)
					dest[len] = 0;
				goto outerBreak;
			case '.':
				dest[-1] = len;
				dest += len + 1;
				len = 0;
				bufSize--;
				break;
			default:
				dest[len] = scopeID[i];
				bufSize--;
				break;
			}
		}
	}
outerBreak:
    for(len = 1; *s != 0; s += *s + 1){
        len += *s + 1;
    }
    return len;
}

/* ------------------------------------------------------------------------- */

static btsize	skipName(btu8 *name, btsize remainingLen)
{
btu8	*p = name;
btint	len;

	for(;;){
		if(p - name > remainingLen)	/* limit to max. len */
			return remainingLen;
		len = *p++;
		if(len == 0 || len > 63)	/* end of name or pointer */
			return p - name;
		p += len;	/* skip this name */
	}
}

/* ------------------------------------------------------------------------- */

static void	nbnsResolveProc(void *obj, void *arg)
{
nbnsCtx_t	*this = obj;
nbnsRes_t	res, *rval = arg;
btint		rcode;
btbool		redirect;
btu8		*p;

	PROC_BEGIN;
	for(;;){
		PROC_WAITFOR_X(1, nbnsRunQuery(this->buffer, this->len, this->destIp, this));
		if(rval->err.class != 0){
			rval->ipAddress = 0;
			PROC_RETURN(rval);
		}
		if(rval->len < 23){	/* can't be a valid package */
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOCbogusdata;
			PROC_RETURN(&res);
		}
		rcode = rval->data[3] & 0xf;
		redirect = (rval->data[2] & 0x4) == 0;	/* authoritative answer == 0? */
		if(rcode != 0){	/* this was an error, report as not found */
			if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
				logPrintf(LOGM_NBNS, "nbns: nbnsResolve: error %d\n", rcode);
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOClookup;
			PROC_RETURN(&res);
		}
		if(redirect){	/* we are redirected to an other nameserver */
			if(this->abortIfTimeout){	/* misuse as counter: accept only one redirect */
				res.err.class = ERRCLS_CIFS_LOCAL;
				res.err.code = ERR_CIFS_LOClookup;
				PROC_RETURN(&res);
			}
			this->destIp = (btu32)(rval->data[rval->len - 4]) << 24;
			this->destIp |= (btu32)(rval->data[rval->len - 3]) << 16;
			this->destIp |= (btu32)(rval->data[rval->len - 2]) << 8;
			this->destIp |= rval->data[rval->len - 1];
			if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
				logPrintf(LOGM_NBNS, "nbns: nbnsResolve: redirect to ip=0x%08x\n", this->destIp);
			this->abortIfTimeout = 1;
		}else{	/* authoritative non-error answer */
			p = rval->data + 12;
			p += skipName(p, rval->len - 12);
			p += 10;	/* skip NB, IN, TTL and RDLENGTH */
			if(p - rval->data + 6 <= rval->len){
				p += 2;	/* skip flags field */
				res.ipAddress = (btu32)*p++ << 24;
				res.ipAddress |= (btu32)*p++ << 16;
				res.ipAddress |= (btu32)*p++ << 8;
				res.ipAddress |= *p++;
				if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
					logPrintf(LOGM_NBNS, "nbns: nbnsResolve: found ip=0x%08x\n", res.ipAddress);
				memset(&res.err, 0, sizeof(res.err));
				PROC_RETURN(&res);
			}else{	/* not enough data */
				res.err.class = ERRCLS_CIFS_LOCAL;
				res.err.code = ERR_CIFS_LOCbogusdata;
				PROC_RETURN(&res);
			}
		}
	}
	PROC_END;
}

void	nbnsResolve(btu8 *name, btint tag, btu32 nbnsIp, btbool broadcast, void *caller)
{
nbnsCtx_t	*ctx;
btu8		*p;

    if(nbnsPool == NULL)
        nbnsPool = poolNew(sizeof(nbnsCtx_t), (btu8 *)&ctx->buffer - (btu8 *)ctx);
	if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
		logPrintf(LOGM_NBNS, "nbnsclient: resolving %s <%02x>\n", name, tag);
	PROC_INIT(ctx, nbnsResolveProc, caller, nbnsPool, poolFreeBlock);
	ctx->destIp = nbnsIp;
	p = ctx->buffer;
	*p++ = xidCounter >> 8;
	*p++ = xidCounter++;
	*p++ = 1;	/* opcode = 0, flags = recursion desired */
	*p++ = broadcast ? 0x10 : 0;	/* rest of flags, rcode = 0 */
	*p++ = 0;
	*p++ = 1;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	p += nbnsEncodeName(p, 256, name, tag);
	*p++ = 0;
	*p++ = 0x20;
	*p++ = 0;
	*p++ = 1;
	ctx->len = p - ctx->buffer;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	nbnsNodeStatusProc(void *obj, void *arg)
{
nbnsCtx_t	*this = obj;
nbnsRes_t	*rval = arg;
btint		rcode, i, entries;
btu8		*p;

	PROC_BEGIN;
	PROC_WAITFOR_X(1, nbnsRunQuery(this->buffer, this->len, this->destIp, this));
	if(rval->err.class != 0){
		PROC_RETURN(rval);
	}
	if(rval->len < 23){	/* can't be a valid package */
		rval->err.class = ERRCLS_CIFS_LOCAL;
		rval->err.code = ERR_CIFS_LOCbogusdata;
		PROC_RETURN(rval);
	}
	rcode = rval->data[3] & 0xf;
	if(rcode != 0){	/* this was an error */
		if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
			logPrintf(LOGM_NBNS, "nbns: nbnsNodeStatus: error %d\n", rcode);
		rval->err.class = ERRCLS_CIFS_LOCAL;
		rval->err.code = ERR_CIFS_LOClookup;
		PROC_RETURN(rval);
	}
	p = rval->data + 12;
	p += skipName(p, rval->len - 12);
	p += 10;	/* skip NB, IN, TTL and RDLENGTH */
	entries = *p++;	/* number of NODE_NAME entries */
	for(i=0;i<entries;i++){
		if(p + 18 > rval->data + rval->len)	/* not an entire record */
			break;
		if(p[15] == 0x20){ /* first 0x20 is name */
			/* || p[15] == 0x00  don't use 0x00, don't know what it is... */
			for(i=14;i>=0;i--)
				if(p[i] != 0x20 && p[i] != 0)
					break;
			p[i+1] = 0;
			rval->name = p;
			PROC_RETURN(rval);
		}
		if(LOG_MODE_IS_ACTIVE(LOGM_NBNS)){
			rcode = p[15];
			p[15] = 0;
			logPrintf(LOGM_NBNS, "nbns: nbnsNodeStatus: skipping %s<%d>\n", p, rcode);
		}
		p += 18;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
		logPrintf(LOGM_NBNS, "nbns: nbnsNodeStatus: no valid 0x20 name found\n");
	rval->err.class = ERRCLS_CIFS_LOCAL;
	rval->err.code = ERR_CIFS_LOClookup;
	PROC_RETURN(rval);
	PROC_END;
}

void	nbnsNodeStatus(btu32 destIp, btu8 *name, btint tag, void *caller)
{
nbnsCtx_t	*ctx;
btu8		*p;

    if(nbnsPool == NULL)
        nbnsPool = poolNew(sizeof(nbnsCtx_t), (btu8 *)&ctx->buffer - (btu8 *)ctx);
	PROC_INIT(ctx, nbnsNodeStatusProc, caller, nbnsPool, poolFreeBlock);
	ctx->destIp = destIp;
	p = ctx->buffer;
	*p++ = xidCounter >> 8;
	*p++ = xidCounter++;
	*p++ = 0;
	*p++ = 0;	/* should be 0x10 for broadcast, but we should not broadcast anyway */
	*p++ = 0;
	*p++ = 1;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	p += nbnsEncodeName(p, 256, name, tag);
	*p++ = 0;
	*p++ = 0x21;
	*p++ = 0;
	*p++ = 1;
	ctx->len = p - ctx->buffer;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

typedef struct nsCache{
	btu8	*hostname;
	btu32	ipAddress;
	bttime	updatedAt;
}nsCache_t;

static hashtable_t	*nsCache = NULL;

static btint	cacheCompare(void *o1, void *o2)
{
nsCache_t	*this = o1, *other = o2;

	return utilStrCmp(this->hostname, other->hostname, 0);
}

static btu32	cacheHashVal(void *obj)
{
nsCache_t	*this = obj;

	return utilStrHash(this->hostname, 0);
}

static collectionObjectI_t	cacheInterface = {
	cacheCompare,
	cacheHashVal,
	NULL,
	NULL,
};

/* ------------------------------------------------------------------------- */

static bttime	maxCacheTime(void)
{
static btu32	timeChg = 0;
static bttime	maxTime;
btint			numEntries;
cfgProperty_t	*path[8];

	if(CFG_DID_CHANGE(timeChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		maxTime = cfgIntValue(cfgQueryFromArray("nbnsCacheTime", numEntries, path), 60*1000);
	}
	return maxTime;
}

static void	cacheEntryFree(nsCache_t *c)
{
	if(c != NULL){
		if(c->hostname != NULL){
			htabRemoveObjectEqualTo(nsCache, c);
			memFree(c->hostname);
		}
		memset(c, 0, sizeof(*c));
		memFree(c);
	}
}

static nsCache_t	*cacheEntryAdd(btu8 *name, btu32 ip)
{
nsCache_t	*volatile c;

	c = htabFindObjectEqualTo(nsCache, &name);
	if(c != NULL){
		c->ipAddress = ip;
		c->updatedAt = elpNow();
		return c;
	}
EXC_TRY;
	c = memCalloc(sizeof(*c));
	c->hostname = utilStrDup(name);
	c->ipAddress = ip;
	c->updatedAt = elpNow();
	htabAddObject(nsCache, c);
EXC_CATCH;
	cacheEntryFree(c);
	EXC_RETHROW();
EXC_ENDTRY;
	return c;
}

static void	cacheCheckOutdated(void)
{
array_t		*allEntries;
btint		i, cnt;
nsCache_t	*c;
bttime		now, age, maxTime = maxCacheTime();

	allEntries = htabAllObjects(nsCache);
	cnt = arrObjectCount(allEntries);
	now = elpNow();
	for(i=0;i<cnt;i++){
		c = arrObjectAtIndex(allEntries, i);
		age = now - c->updatedAt;
		if(age < 0 || age > maxTime){
			cacheEntryFree(c);
		}
	}
	arrFree(allEntries);
}

/* ------------------------------------------------------------------------- */

btu32	nbnsWinsIp(void)
{
btu8			*wins;
btint			numEntries;
cfgProperty_t	*path[8];
static btu32	cfgChg = 0;
static btu32	winsIp;

	if(CFG_DID_CHANGE(cfgChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		wins = cfgStringValue(cfgQueryFromArray("nbnsWinsIp", numEntries, path), "");
		if(wins == NULL || *wins == 0)
			winsIp = 0;
		else if(!utilResolveName(wins, &winsIp))
			winsIp = 0xffffffff;
		winsIp = ntohl(winsIp);
	}
	return winsIp;
}

/* ------------------------------------------------------------------------- */

static void	resolveNamePrimitive(void *arg, btint resultFd)
{
btu32	result[2];
btint	written, l;

	mainSignalHandlers(1);
	if((result[0] = utilResolveName(arg, &result[1])) != 0){
		result[1] = ntohl(result[1]);
	}
	for(written=0;written<8;written+=l){
		l = write(resultFd, (btu8 *)result + written, 8 - written);
		if(l <= 0)
			break;
	}
}

static void	nbnsLookupProc(void *obj, void *arg)
{
nbnsCtx_t		*this = obj;
nbnsRes_t		*rval = arg;
btu32			ip, *res;
btint			numEntries;
cfgProperty_t	*path[8];
static btu32	cfgChg = 0;
static btbool	tryNetbios, tryDns;

	if(CFG_DID_CHANGE(cfgChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		tryNetbios = cfgEnumValue(cfgQueryFromArray("lookupTryNetbios", numEntries, path), cfgYesNoEnum, 1);
		tryDns = cfgEnumValue(cfgQueryFromArray("lookupTryDns", numEntries, path), cfgYesNoEnum, 1);
	}
	PROC_BEGIN;
	if(tryNetbios){
		if((this->destIp = nbnsWinsIp()) != 0){
			PROC_WAITFOR_X(1, nbnsResolve(this->buffer, 0x20, this->destIp, 0, this));
			if(rval->err.class == 0){
				res = &rval->ipAddress;
				goto doReturn;
			}
		}
		/* try broadcast if wins failed */
		PROC_WAITFOR_X(2, nbnsResolve(this->buffer, 0x20, 0xffffffff, 1, this));
		if(rval->err.class == 0){
			res = &rval->ipAddress;
			goto doReturn;
		}
	}
	if(tryDns){
		PROC_WAITFOR_X(3, utilExecInBackground(resolveNamePrimitive, this->buffer, &this->msg, this));
		if(rval->err.class == 0){
			if(*(btu32 *)this->msg->buffer != 0){
				ip = ((btu32 *)this->msg->buffer)[1];
				res = &ip;
				ipcMsgFree(this->msg);
				goto doReturn;
			}
			ipcMsgFree(this->msg);
		}
	}
	res = NULL;
doReturn:
	if(res != NULL)
		cacheEntryAdd(this->buffer, *res);
	PROC_RETURN(res);	/* an error occured, but nobody cares what it was */
	PROC_END;
}

void	nbnsLookup(btu8 *name, void *caller)
{
nbnsCtx_t		*ctx;
nsCache_t		*c;
bttime			age;
static btint	lookupCount = 0;

    if(nbnsPool == NULL)
        nbnsPool = poolNew(sizeof(nbnsCtx_t), (btu8 *)&ctx->buffer - (btu8 *)ctx);
	if(nsCache == NULL)
		nsCache = htabNew(&cacheInterface, 32, 50);
	if(++lookupCount > 100){	/* clean cache every 100th call */
		lookupCount = 0;
		cacheCheckOutdated();
	}
	c = htabFindObjectEqualTo(nsCache, &name);
	if(c != NULL){
		age = elpNow() - c->updatedAt;
		if(age >= 0 && age < maxCacheTime()){	/* not outdated */
			PROC_IMMEDIATE_RET(caller, &c->ipAddress);
		}
	}
    PROC_INIT(ctx, nbnsLookupProc, caller, nbnsPool, poolFreeBlock);
	strncpy(ctx->buffer, name, sizeof(ctx->buffer));
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

