/*
 * Name: pslsslsocket.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-05-23
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: pslsslsocket.c,v 1.5 2001/04/05 13:20:29 cs Exp $
 */

#ifdef USE_SSL

#include "stdheaders.h"
#include "pslsslsocket.h"
#include "exception.h"
#include "memalloc.h"
#include "evntloop.h"
#include "logging.h"
#include "system.h"
#include "process.h"
#include "message.h"
#include "blockpool.h"
#include "util.h"
#include "server.h"
#include "errcodes.h"
#include <ssl.h>
#include <err.h>

/* ------------------------------------------------------------------------- */

EXC_DEFINE(excSslCertfile);
EXC_DEFINE(excSslKeyfile);
EXC_DEFINE(excSslNeedPassPhrase);

typedef struct pslCtx{
    PROC_VARSET;
    pslSslSocket_t	*psl;
	bttime			timeout;
}pslCtx_t;

static blockpool_t	*pslPool = NULL;

static void	freePslBlock(void *block)
{
    poolReturnBlock(pslPool, block);
}

static btbool	didInitializeSSL;

/* ------------------------------------------------------------------------- */

static void	pslConnectProc(void *obj, void *arg);

/* ------------------------------------------------------------------------- */

static inline void	initSSL(void)
{
	if(!didInitializeSSL){
		SSL_load_error_strings();
		SSLeay_add_ssl_algorithms();
		didInitializeSSL = 1;
	}
}

/* ------------------------------------------------------------------------- */

static btu8		*cbPassPhrase;
static btbool	askedForPassPhrase;

static int	passPhraseCallback(char *buffer, int bufferLen)
{
int	len;

	askedForPassPhrase = 1;
	if(cbPassPhrase != NULL){
		len = strlen(cbPassPhrase) + 1;
		if(len > bufferLen)
			len = bufferLen;
		memcpy(buffer, cbPassPhrase, len);
		return len - 1;
	}else{
		*buffer = 0;
		return 0;
	}
}

/* ------------------------------------------------------------------------- */

static RSA	*tempRsaCallback(SSL *ssl, int export)
{
static RSA	*rsa = NULL;
	
	if(rsa == NULL)
		rsa = RSA_generate_key(512, RSA_F4, NULL, NULL);
	return rsa;
}

/* ------------------------------------------------------------------------- */

static int	sslVerifyCallback(int ok, X509_STORE_CTX *ctx)
{
char	buffer[256];

	if(LOG_MODE_IS_ACTIVE(LOGM_INFO)){
		X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert), buffer, sizeof(buffer));
		if(ok){
			logPrintf(LOGM_INFO, "SSL: Certificate OK: %s\n", buffer);
		}else{
			switch (ctx->error){
			case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
				logPrintf(LOGM_INFO, "SSL: Cert error: CA not known: %s\n", buffer);
				break;
			case X509_V_ERR_CERT_NOT_YET_VALID:
				logPrintf(LOGM_INFO, "SSL: Cert error: Cert not yet valid: %s\n", buffer);
				break;
			case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
				logPrintf(LOGM_INFO, "SSL: Cert error: illegal \'not before\' field: %s\n", buffer);
				break;
			case X509_V_ERR_CERT_HAS_EXPIRED:
				logPrintf(LOGM_INFO, "SSL: Cert error: Cert expired: %s\n", buffer);
				break;
			case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
				logPrintf(LOGM_INFO, "SSL: Cert error: invalid \'not after\' field: %s\n", buffer);
				break;
			default:
				logPrintf(LOGM_INFO, "SSL: Cert error: unknown error %d in %s\n", ctx->error, buffer);
				break;
			}
		}
	}
	return ok;
}

/* ------------------------------------------------------------------------- */

static inline SSL_CTX	*allocateContext(server_t *server)
{
	switch(srvSslVersion(server)){
		case SRV_SSL_V2:	return SSL_CTX_new(SSLv2_method());
		case SRV_SSL_V3:	return SSL_CTX_new(SSLv3_method());
		default:
		case SRV_SSL_V23:	return SSL_CTX_new(SSLv23_method());
		case SRV_SSL_TLS1:	return SSL_CTX_new(TLSv1_method());
	}
}

/* ------------------------------------------------------------------------- */

btbool	pslSslSocketCheckPassPhrase(server_t *server, btu8 *passPhrase)
{
SSL_CTX	*sslContext;
btu8	*file;
btbool	passOK = 1;

	initSSL();
	if((sslContext = allocateContext(server)) == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "Error allocating SSL context: %s\n", ERR_error_string(ERR_get_error(), NULL));
		return 0;
	}
	cbPassPhrase = passPhrase;
	SSL_CTX_set_default_passwd_cb(sslContext, passPhraseCallback);
	if((file = srvSslKeyFile(server)) == NULL)
		file = srvSslCertFile(server);
	if(file != NULL)
		passOK = SSL_CTX_use_PrivateKey_file(sslContext, file, SSL_FILETYPE_PEM) != 0;
	SSL_CTX_free(sslContext);
	cbPassPhrase = NULL;
	return passOK;
}

/* ------------------------------------------------------------------------- */


static void	connectionLost(pslSslSocket_t *psl);

/* if we can't fulfill a send request within a given amount of time, there
 * must be something wrong and we'd better assume that the connection is
 * broken. The higher levels should try to reconnect anyway.
 */
static void	wantSendTimeout(void *obj, void *arg)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "pslsslsocket: send timed out, disconnecting\n");
	connectionLost(obj);
}

/* ------------------------------------------------------------------------- */

static void	pslSocketFree(pslSslSocket_t *psl)
{
	if(psl != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "pslsslsocket: freeing object at 0x%08x\n", (int)psl);
		elpCancelCallAt(wantSendTimeout, psl);
		if(psl->fd >= 0)
			elpCancelForFd(psl->fd, -1);
		if(psl->connectSignal != NULL){
			procCondFree(psl->connectSignal);
			psl->connectSignal = NULL;
		}
		if(psl->ssl != NULL){
			SSL_free(psl->ssl);
			psl->ssl = NULL;
		}
		if(psl->sslContext != NULL){
			SSL_CTX_free(psl->sslContext);
			psl->sslContext = NULL;
		}
		memFree(psl);
	}
}

/* ------------------------------------------------------------------------- */

pslSslSocket_t	*pslSslSocketNew(btu32 destIp, int port, int connectedSocket, struct server *server, btu8 *passPhrase)
{
pslSslSocket_t	*volatile psl = NULL;
btu8			*keyfile, *certfile, *caFile, *caPath;
int				err;

    if(pslPool == NULL)
        pslPool = poolNew(sizeof(pslCtx_t), sizeof(pslCtx_t));
EXC_TRY
	psl = memCalloc(sizeof(pslSslSocket_t));
	initSSL();
	if((psl->sslContext = allocateContext(server)) == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "Error allocating SSL context: %s\n", ERR_error_string(ERR_get_error(), NULL));
		EXC_THROW_MSG(excInternalError, "error allocating ssl context");
	}
	if(srvSslCompatibility(server)){
		SSL_CTX_set_options(psl->sslContext, SSL_OP_ALL);
	}
	if(srvSslCiphers(server) != NULL){
		SSL_CTX_set_cipher_list(psl->sslContext, srvSslCiphers(server));
	}
	
	if((caPath = srvSslCACertDir(server)) != NULL && *caPath == 0)
		caPath = NULL;	/* map empty string to NULL pointer */
	if((caFile = srvSslCACertFile(server)) != NULL && *caFile == 0)
		caFile = NULL;	/* map empty string to NULL pointer */
	if(caPath != NULL || caFile != NULL){
		if(!SSL_CTX_load_verify_locations(psl->sslContext, caFile, caPath)){
			err = ERR_get_error();
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsslsocket: error setting verify pathes: %s\n", ERR_error_string(err, NULL));
		}
	}
	SSL_CTX_set_tmp_rsa_callback(psl->sslContext, tempRsaCallback);
	if(srvSslRequireServerCert(server)){
		SSL_CTX_set_verify(psl->sslContext, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, sslVerifyCallback);
	}else{
		SSL_CTX_set_verify(psl->sslContext, SSL_VERIFY_NONE, sslVerifyCallback);
	}
	SSL_CTX_set_default_passwd_cb(psl->sslContext, passPhraseCallback);
	certfile = srvSslCertFile(server);
	if((keyfile = srvSslKeyFile(server)) == NULL)
		keyfile = certfile;
	if(certfile != NULL){
		if(!SSL_CTX_use_certificate_file(psl->sslContext, certfile, SSL_FILETYPE_PEM)){
			err = ERR_get_error();
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsslsocket: error reading certificate from file %s: %s\n", certfile, ERR_error_string(err, NULL));
			EXC_THROW_MSG(excSslCertfile, "error reading certificate file");
		}
		cbPassPhrase = passPhrase;
		askedForPassPhrase = 0;
		if(!SSL_CTX_use_PrivateKey_file(psl->sslContext, keyfile, SSL_FILETYPE_PEM)){
			if(askedForPassPhrase && passPhrase == NULL){	/* pass phrase required */
				EXC_THROW_MSG(excSslNeedPassPhrase, "error reading private key file");
			}
			err = ERR_get_error();
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsslsocket: error reading private key from file %s: %s\n", keyfile, ERR_error_string(err, NULL));
			EXC_THROW_MSG(excSslKeyfile, "error reading private key file");
		}
	}
	cbPassPhrase = NULL;
	psl->connectSignal = procCondNew();
	psl->cstatus = PSL_NOTCONNECTED;
	psl->destIp = destIp;
	psl->port = port;
	psl->fd = connectedSocket;
	psl->refCount = 1;
	psl->server = server;
EXC_CATCH
	pslSocketFree(psl);
	EXC_RETHROW();
EXC_ENDTRY
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "pslsslsocket: allocated new object at 0x%08x for fd=%d\n", (int)psl, connectedSocket);
	return psl;
}

/* ------------------------------------------------------------------------- */

static inline void	pslSocketRetain(void *obj)
{
pslSslSocket_t *psl = obj;

	psl->refCount++;
}

static inline void	pslSocketRelease(void *obj)
{
pslSslSocket_t *psl = obj;

	psl->refCount--;
    if(psl->refCount == 0){
        pslSocketFree(psl);
    }else if(psl->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

static void	pslSocketSetHigherLayer(void *obj, void *nextObj, struct psDownInterface *nextI)
{
pslSslSocket_t	*psl = obj;

	psl->higherI = nextI;
	psl->higherObj = nextObj;
}

static void	closeSocket(pslSslSocket_t *psl)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
		logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: closing SSL socket\n");
	if(psl->fd >= 0){
		elpCancelCallAt(wantSendTimeout, psl);
		elpCancelForFd(psl->fd, -1);
		psl->registeredForWrite = 0;
		close(psl->fd);
		psl->fd = -1;
	}
	if(psl->ssl != NULL){
		SSL_free(psl->ssl);
		psl->ssl = NULL;
	}
	psl->attemptImplicitRead = psl->attemptImplicitWrite = psl->attemptImplicitConnect = 0;
}

/* ------------------------------------------------------------------------- */

static void	connectionLost(pslSslSocket_t *psl)
{
	closeSocket(psl);
    psl->cstatus = PSL_NOTCONNECTED;
	if(psl->higherI != NULL)
		psl->higherI->connectStatusChanged(psl->higherObj, psl->cstatus);
	procCondBroadcast(psl->connectSignal, NULL);
}

/* ------------------------------------------------------------------------- */

static inline void	txReady(pslSslSocket_t *psl, btint errcode)
{
	msgRelease(psl->currentMsg);
	psl->currentMsg = NULL;
	if(psl->higherI != NULL)
		psl->higherI->txReady(psl->higherObj, errcode);
	if(psl->wantSendCount > 0){
		psl->wantSendCount--;
	}
}

/* ------------------------------------------------------------------------- */

static void	readFromSocket(pslSslSocket_t *psl)
{
btu8		buffer[8192];
btsize		bytesRead;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
		logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: readFromSocket()\n");
	do{
		bytesRead = SSL_read(psl->ssl, (char *)buffer, sizeof(buffer));
		if(bytesRead == 0){	/* socket closed by peer */
			if(LOG_MODE_IS_ACTIVE(LOGM_SMB_CONNECT))
				logPrintf(LOGM_SMB_CONNECT, "pslsslsocket: connection closed by server\n");
			connectionLost(psl);
		}else if(bytesRead < 0){
			if(BIO_sock_should_retry(bytesRead)){
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
					logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: readFromSocket(): reschedule\n");
				psl->attemptImplicitRead = 1;
				break;
			}
			if(LOG_MODE_IS_ACTIVE(LOGM_SMB_CONNECT))
				logPrintf(LOGM_SMB_CONNECT, "pslsslsocket: read error:  %s\n", ERR_error_string(ERR_get_error(), NULL));
			connectionLost(psl);
		}else if(psl->higherI != NULL){
			psl->higherI->rxMessage(psl->higherObj, buffer, bytesRead);
		}
		psl->attemptImplicitRead = 0;
	}while(bytesRead == sizeof(buffer));
}

/* ------------------------------------------------------------------------- */

static void	writeToSocket(pslSslSocket_t *psl)
{
btu8	*data;
btsize	len, written;
int		err;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
		logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: writeToSocket()\n");
	/* retrigger timeout on each chunk sent: */
	elpCallAtCancelPrev(elpNow() + srvRequestTimeout(psl->server), wantSendTimeout, psl);
	do{
		if(psl->currentMsg == NULL){
			if(psl->higherI != NULL)
				psl->currentMsg = psl->higherI->getTxMessage(psl->higherObj);
			if(psl->currentMsg == NULL){
				psl->wantSendCount = 0;
				break;	/* no message to send */
			}
			msgResetRead(psl->currentMsg);
		}
		msgPeekData(psl->currentMsg, &data, &len);
		written = SSL_write(psl->ssl, data, len);
		if(written <= 0){
			if(BIO_sock_should_retry(written)){
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
					logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: writeToSocket(): reschedule\n");
				psl->attemptImplicitWrite = 1;
				break;
			}
			err = ERR_get_error();
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsslsocket: write error: %s\n", ERR_error_string(err, NULL));
			txReady(psl, EPIPE);
			connectionLost(psl);
		}else{
			if(written >= len){		/* ready? */
				txReady(psl, 0);
			}else{
				msgReadData(psl->currentMsg, &data, written);	/* advance read pointer */
			}
		}
		psl->attemptImplicitWrite = 0;
	}while(psl->wantSendCount > 0 && written >= len);
}

/* ------------------------------------------------------------------------- */

static void	socketAccessible(void *obj, void *arg)
{
pslSslSocket_t	*psl = obj;
elpFile_t		*f = arg;
btbool			sockWritable, sockReadable, registerWrite;
int				err, rval;

	if(psl->ssl == NULL){	/* should never happen */
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "pslsslsocket: socketAccessible: called while no ssl allocated\n");
		return;
	}
	pslSocketRetain(psl);
	sockWritable = f->mode == ELP_FDMODE_WRITE;
	sockReadable = !sockWritable;
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
		logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: socketAccessible write=%d\n", sockWritable);
	if((SSL_want_read(psl->ssl) && sockReadable) || (SSL_want_write(psl->ssl) && sockWritable)){
		if(psl->attemptImplicitConnect){
			err = SSL_connect(psl->ssl);
			if(!BIO_sock_should_retry(err)){
				psl->attemptImplicitConnect = 0;
				rval = err < 0 ? ERR_get_error() : 0;
				if(psl->connectCtx != NULL)
					pslConnectProc(psl->connectCtx, &rval);	/* callback: connect proceeded */
			}else{
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
					logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: socketAccessible: retry connect\n");
			}
			sockWritable = sockReadable = 0;	/* never do read/write immediately after implicit connect */
		}
		if(psl->attemptImplicitRead && psl->ssl != NULL){
			readFromSocket(psl);
			sockReadable = 0;	/* everything that can be read has been read */
		}
		if(psl->attemptImplicitWrite && psl->ssl != NULL){
			writeToSocket(psl);
			sockWritable = 0;	/* everything that can be written has been written */
		}
	}
	if(sockReadable && psl->ssl != NULL){
		readFromSocket(psl);
	}
	if(sockWritable && psl->ssl != NULL){
		writeToSocket(psl);
	}
	if(psl->ssl != NULL){	/* socket has not already been closed */
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL)){
			if(SSL_want_write(psl->ssl))
				logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: SSL wants write\n");
			if(SSL_want_read(psl->ssl))
				logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: SSL wants read\n");
		}
		registerWrite = psl->wantSendCount != 0 || SSL_want_write(psl->ssl);
		if(psl->registeredForWrite != registerWrite){
			if(registerWrite){
				elpRegisterForFd(psl->fd, ELP_FDMODE_WRITE, socketAccessible, psl);
			}else{
				elpCancelCallAt(wantSendTimeout, psl);
				elpCancelForFd(psl->fd, ELP_FDMODE_WRITE);
			}
			psl->registeredForWrite = registerWrite;
		}
	}
	pslSocketRelease(psl);
}

/* ------------------------------------------------------------------------- */

static void	pslConnectProc(void *obj, void *arg)
{
pslCtx_t			*this = obj;
pslSslSocket_t		*psl = this->psl;
btError				res;
int					err, *rval = arg;
struct sockaddr_in	sockAddr;

	PROC_BEGIN;
	pslSocketRetain(psl);
    while(psl->cstatus == PSL_CONNECTING || psl->cstatus == PSL_DISCONNECTING){
		PROC_WAITFOR_X(1, procCondWait(psl->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(psl->cstatus == PSL_CONNECTED){	/* if someone connected meanwhile */
		pslSocketRelease(psl);
		res.class = res.code = 0;
		PROC_RETURN(&res);
	}
    psl->cstatus = PSL_CONNECTING;
	if(psl->ssl != NULL){
		SSL_free(psl->ssl);
	}
	psl->ssl = SSL_new(psl->sslContext);
	if(psl->ssl == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "Error allocating SSL structure: %s\n", ERR_error_string(ERR_get_error(), NULL));
		goto errorOccured;
	}
	if(psl->fd < 0){	/* if socket is not already connected */
		if((psl->fd = socket(AF_INET, SOCK_STREAM, 0)) < 0){
			res.class = ERRCLS_UNIX;
			res.code = errno;
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsslsocket: error creating socket: [%d] %s\n", errno, sysStrerror(errno));
			goto errorOccured;
		}
		if(sysSetNonblock(psl->fd) < 0){
			res.class = ERRCLS_UNIX;
			res.code = errno;
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsslsocket: error set nonblock: [%d] %s\n", errno, sysStrerror(errno));
			goto errorOccured;
		}
		memset(&sockAddr, 0, sizeof(sockAddr));
		sockAddr.sin_family = AF_INET;
		sockAddr.sin_port = htons(psl->port);
		sockAddr.sin_addr.s_addr = htonl(psl->destIp);
		if(connect(psl->fd, (struct sockaddr *)&sockAddr, sizeof(sockAddr)) != 0){
			if(errno == EINPROGRESS){
				elpCallAtCancelPrev(elpNow() + this->timeout, pslConnectProc, this);	/* setup timeout */
				PROC_WAITFOR_X(2, elpRegisterForFd(psl->fd, ELP_FDMODE_WRITE, pslConnectProc, this));
				elpCancelCallAt(pslConnectProc, this);
				elpCancelForFd(psl->fd, ELP_FDMODE_WRITE);
				if(rval == NULL){	/* connect timeout */
					res.class = ERRCLS_CIFS_LOCAL;
					res.code = ERR_CIFS_LOCtimeout;
					goto errorOccured;
				}
			}else{
				res.class = ERRCLS_UNIX;
				res.code = errno;
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "pslsslsocket: error connecting: [%d] %s\n", errno, sysStrerror(errno));
				goto errorOccured;
			}
		}
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
			logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: pslConnect: TCP connected\n");
	}
	SSL_set_fd(psl->ssl, psl->fd);
	elpRegisterForFd(psl->fd, ELP_FDMODE_READ, socketAccessible, psl);
	psl->registeredForWrite = 0;
	err = SSL_connect(psl->ssl);
	if(BIO_sock_should_retry(err)){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
			logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: pslConnect: SSL connect retry\n");
		psl->connectCtx = this;
		psl->attemptImplicitConnect = 1;
		elpCallAtCancelPrev(elpNow() + this->timeout, pslConnectProc, this);	/* setup timeout */
		elpRegisterForFd(psl->fd, ELP_FDMODE_WRITE, socketAccessible, psl);
		psl->registeredForWrite = 1;
		PROC_WAITFOR_X(3, ;);		/* wait for connect to proceed */
		psl->connectCtx = NULL;
		elpCancelCallAt(pslConnectProc, this);
		if(rval == NULL){	/* connect timeout */
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
				logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: SSL connect timed out\n");
			res.class = ERRCLS_CIFS_LOCAL;
			res.code = ERR_CIFS_LOCtimeout;
			goto errorOccured;
		}
		if(*rval != 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsslsocket: error connecting SSL: %s\n", ERR_error_string(*rval, NULL));
			res.class = ERRCLS_CIFS_LOCAL;
			res.code = ERR_CIFS_LOCsslerror;
			goto errorOccured;
		}
	}else{
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_SSL))
			logPrintf(LOGM_DEBUG_SSL, "pslsslsocket: pslConnect: no SSL connect retry\n");
		if(err <= 0){
			err = ERR_get_error();
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsslsocket: error connecting SSL: %s\n", ERR_error_string(err, NULL));
			res.class = ERRCLS_CIFS_LOCAL;
			res.code = ERR_CIFS_LOCsslerror;
			goto errorOccured;
		}
	}	
	if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
		logPrintf(LOGM_INFO, "SSL: connected with cipher: %s\n", SSL_get_cipher(psl->ssl));
    psl->cstatus = PSL_CONNECTED;
	psl->wantSendCount = 0;	/* there can't be pending writes just after connect */
	res.class = res.code = 0;
	goto doReturn;
errorOccured:
    psl->cstatus = PSL_NOTCONNECTED;
	if(psl->fd >= 0){
		closeSocket(psl);
	}
doReturn:
    PROC_RETURN_AND_CLEANUP(&res,
		if(psl->higherI != NULL)
			psl->higherI->connectStatusChanged(psl->higherObj, psl->cstatus);
		procCondBroadcast(psl->connectSignal, NULL);
		pslSocketRelease(psl);
	);
    PROC_END;
}

/* ------------------------------------------------------------------------- */

static void	pslDisconnectProc(void *obj, void *arg)
{
pslCtx_t		*this = obj;
pslSslSocket_t	*psl = this->psl;
btError			res;

	PROC_BEGIN;
	pslSocketRetain(psl);
    while(psl->cstatus == PSL_CONNECTING || psl->cstatus == PSL_DISCONNECTING){
		PROC_WAITFOR_X(1, procCondWait(psl->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(psl->cstatus == PSL_NOTCONNECTED){	/* if someone disconnected meanwhile */
		pslSocketRelease(psl);
		res.class = res.code = 0;
		PROC_RETURN(&res);
	}
    psl->cstatus = PSL_DISCONNECTING;
	closeSocket(psl);
    psl->cstatus = PSL_NOTCONNECTED;
	res.class = res.code = 0;
    PROC_RETURN_AND_CLEANUP(&res,
		if(psl->higherI != NULL)
			psl->higherI->connectStatusChanged(psl->higherObj, psl->cstatus);
		procCondBroadcast(psl->connectSignal, NULL);
		pslSocketRelease(psl);
	);
    PROC_END;
}

/* ------------------------------------------------------------------------- */

static void	pslSocketSetConnectStatus(void *obj, btint wantStatus, bttime timeout, void *caller)
{
pslSslSocket_t	*psl = obj;
pslCtx_t		*ctx;
btError			res;

	if(wantStatus == psl->cstatus){
		res.class = res.code = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}else{
		PROC_INIT(ctx, wantStatus == PSL_CONNECTED ? pslConnectProc : pslDisconnectProc, caller, pslPool, freePslBlock);
		ctx->psl = psl;
		ctx->timeout = timeout;
		PROC_START(ctx, NULL);
	}
}

/* ------------------------------------------------------------------------- */

static btint	pslSocketConnectStatus(void *obj)
{
pslSslSocket_t	*psl = obj;

	return psl->cstatus;
}

/* ------------------------------------------------------------------------- */

static void	pslSocketWantSend(void *obj)
{
pslSslSocket_t	*psl = obj;

	if(psl->cstatus == PSL_CONNECTED){
		if(psl->wantSendCount == 0 && !psl->registeredForWrite){
			elpCallAtCancelPrev(elpNow() + srvRequestTimeout(psl->server), wantSendTimeout, psl);
			elpRegisterForFd(psl->fd, ELP_FDMODE_WRITE, socketAccessible, psl);
			psl->registeredForWrite = 1;
		}
		psl->wantSendCount++;
	}
}

/* ------------------------------------------------------------------------- */

static void	pslSocketMaxInserts(void *obj, btsize *preInsert, btsize *postInsert)
{
	*preInsert = 0;
	*postInsert = 0;
}

/* ------------------------------------------------------------------------- */

psUpInterface_t	pslSslSocketUpInterface = {
	pslSocketSetHigherLayer,
	pslSocketRelease,
	pslSocketSetConnectStatus,
	pslSocketConnectStatus,
	pslSocketWantSend,
	pslSocketMaxInserts
};

#endif /* USE_SSL */
