/*
 * Name: user.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-28
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: user.c,v 1.16 2001/05/06 15:21:23 cs Exp $
 */


#include "stdheaders.h"
#include "memalloc.h"
#include "unicode.h"
#include "user.h"
#include "share.h"
#include "server.h"
#include "blockpool.h"
#include "cifs.h"
#include "smbnumbers.h"
#include "statemachine.h"
#include "smb.h"
#include "process.h"
#include "pwhash.h"
#include "logging.h"
#include "pslayer.h"
#include "userdb.h"
#include "main.h"
#include "cfgparse.h"
#include "uirep.h"
#include "errcodes.h"
#include "cifsconst.h"
#include "cifsmod.h"
#include "backend.h"
#include "licman.h"
#include "mainmodule.h"
#include "guimanager.h"
#include "ipcserver.h"
#include "util.h"

/* ------------------------------------------------------------------------- */

typedef struct userGuiLock{
	btint		unixUid;
	procCond_t	*guiSignal;
	btbool		guiIsLocked;
}userGuiLock_t;

typedef struct userPasswdCache{
	btuni			*share;			/* UNC path of share */
	btuni			*passwd;		/* clear text share password */
	btbool			allowUnencrypted;
}userPasswdCache_t;

/* ------------------------------------------------------------------------- */

typedef struct userCtx{
    PROC_VARSET;
    cifsUser_t		*user;
	server_t		*server;
	share_t			*share;
	btu32			unixUid;
	btu32			unixGid;
	userGuiLock_t	*lock;
	btbool			allowNullSession;
	btbool			allowAskUser;
	btbool			isNew;
	ipcSrvClient_t	*client;
	uirepAuthData_t	auth;
	btu8			*domain;
	btError			err;
}userCtx_t;

static blockpool_t	*userPool = NULL;

/* ------------------------------------------------------------------------- */

static void		pwCacheFree(void *obj)
{
userPasswdCache_t	*this = obj;

	if(this != NULL){
		if(this->share != NULL)
			memFree(this->share);
		if(this->passwd != NULL)
			memFree(this->passwd);
		memset(this, 0, sizeof(*this));
		memFree(this);
	}
}

static userPasswdCache_t	*pwCacheNew(btuni *share, btuni *passwd, btbool allowUnencrypted)
{
userPasswdCache_t	*volatile this;

EXC_TRY
	this = memCalloc(sizeof(userPasswdCache_t));
	this->share = uniStrDup(share);
	this->passwd = uniStrDup(passwd);
	this->allowUnencrypted = allowUnencrypted;
EXC_CATCH
	pwCacheFree(this);
	EXC_RETHROW();
EXC_ENDTRY
	return this;
}

static btint	pwCacheCompare(void *o1, void *o2)
{
userPasswdCache_t	*this = o1, *other = o2;

	return uniStrCmp(this->share, other->share, 0);
}

static btu32	pwCacheHashval(void *obj)
{
userPasswdCache_t	*this = obj;

	return uniStrHash(this->share, 0);
}

static collectionObjectI_t	pwCacheInterface = {
	pwCacheCompare,
	pwCacheHashval,
	NULL,
	pwCacheFree
};

/* ------------------------------------------------------------------------- */

cifsUser_t	*userNew(btu32 unixUid, btu32 unixGid, btuni *smbUsername, btuni *plainPasswd, btu8 *domain, btu8 passHashNt[16], btu8 passHashPreNt[16], btbool allowUnencrypted)
{
cifsUser_t	*volatile this = NULL;

    if(userPool == NULL)
        userPool = poolNew(sizeof(userCtx_t), sizeof(userCtx_t));
EXC_TRY
    this = memCalloc(sizeof(cifsUser_t));
	this->refCount = 1;
    this->connectSignal = procCondNew();
	this->sharePasswordCache = htabNew(&pwCacheInterface, 16, 50);
    this->unixUid = unixUid;
	this->unixGid = unixGid;
    this->smbUsername = uniStrDup(smbUsername);
	this->domain = utilStrDup(domain);
	if(plainPasswd != NULL){
		this->allowGuest = plainPasswd[0] == 0;			/* empty password indicates guest account */
		pwhNtHash(plainPasswd, uniStrLen(plainPasswd), this->pwHashNt);
		pwhPreNtHash(plainPasswd, uniStrLen(plainPasswd), this->pwHashPreNt);
		if(allowUnencrypted || plainPasswd[0] == 0)
			this->plainPasswd = uniStrDup(plainPasswd);
	}else{
		memcpy(this->pwHashNt, passHashNt, 16);
		memcpy(this->pwHashPreNt, passHashPreNt, 16);
	}
    this->smbUid = -1;
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "user: allocated new object at 0x%08x\n", (int)this);
EXC_CATCH
	userFree(this);
	EXC_RETHROW();
EXC_ENDTRY
    return this;
}

void	userFree(cifsUser_t *this)
{
    if(this != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "user: freeing object at 0x%08x\n", (int)this);
		if(this->server != NULL){
			srvRemoveUser(this->server, this);
			srvRelease(this->server);
			this->server = NULL;
		}
		if(this->sharePasswordCache != NULL)
			htabFree(this->sharePasswordCache);
        if(this->connectSignal != NULL)
            procCondFree(this->connectSignal);
        if(this->smbUsername != NULL)
            memFree(this->smbUsername);
		if(this->domain != NULL)
            memFree(this->domain);
        if(this->plainPasswd != NULL)
            memFree(this->plainPasswd);
		memset(this, 0, sizeof(*this));
        memFree(this);
    }
}

/* ------------------------------------------------------------------------- */

static void	freePoolBlock(void *block)
{
    poolReturnBlock(userPool, block);
}

/* ------------------------------------------------------------------------- */

static void	userConnectProc(void *obj, void *arg)
{
userCtx_t				*this = obj;
srvRes_t				*srvRes = arg;
userRes_t				res;
smbRequest_t			*smb = arg;
cifsResSessionSetup_t	*cifsRes = arg;
array_t					*allUsers;
btint					i, c, numEntries;
cifsUser_t				*u, *user = this->user;
btbool					pwIsEncrypted;
cfgProperty_t			*path[8];
static btu32			domainChg = 0;
static btu8				*domain;

    PROC_BEGIN;
	userRetain(user);
	while(user->cstatus == PSL_CONNECTING || user->cstatus == PSL_DISCONNECTING){
		PROC_WAITFOR_X(1, procCondWait(user->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(user->cstatus == PSL_CONNECTED){
		res.err.code = res.err.class = 0;
		userRelease(user);
		PROC_RETURN(&res);
	}
	user->cstatus = PSL_CONNECTING;
    if(srvConnectionStatus(user->server) != PSL_CONNECTED){
        PROC_WAITFOR_X(2, srvConnect(user->server, user->unixUid, this));
        if(srvRes->err.class != 0){
            res.err.class = srvRes->err.class;
            res.err.code = srvRes->err.code;
            user->cstatus = PSL_NOTCONNECTED;
            goto doReturn;
        }
    }
	if(!srvCanEncrypt(user->server) && user->plainPasswd == NULL){	/* we can't do with this server */
		user->cstatus = PSL_NOTCONNECTED;
		user->smbUid = -1;
		user->isGuest = 0;
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCpwencrypt;
		goto doReturn;
	}
	allUsers = srvAllUsers(user->server);
	c = arrObjectCount(allUsers);
	for(i=0;i<c;i++){	/* try to find this user name in list of connected users */
		u = arrObjectAtIndex(allUsers, i);
		if(u == user || uniStrCmp(u->smbUsername, user->smbUsername, 0) != 0)
			continue;
		if(u->cstatus == PSL_CONNECTED){	/* this SMB user is already connected */
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "user: SMB user %s is already connected.\n", uniToCstring(u->smbUsername));
			if(memcmp(u->pwHashNt, user->pwHashNt, sizeof(u->pwHashNt)) == 0){	/* same password */
				user->cstatus = PSL_CONNECTED;
				user->smbUid = u->smbUid;
				user->isGuest = u->isGuest;
				res.err.class = res.err.code = 0;
			}else{
				user->cstatus = PSL_NOTCONNECTED;
				res.err.class = ERRCLS_CIFS_DOS;
				res.err.code = ERR_CIFS_DOSnoaccess;
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "user: passwords for equally named users don\'t match.\n");
			}
			break;
		}
	}
	arrFree(allUsers);
	if(i < c)	/* loop exited with break, connected equal user found */
		goto doReturn;

	if(!srvUseRequest(user->server, SRV_USE_SESSION_SETUP)){
		res.err.class = res.err.code = 0;
		user->cstatus = PSL_CONNECTED;
		user->smbUid = 0;
		goto doReturn;
	}
	PROC_WAITFOR_X(3, srvWantRequest(user->server, this));
	pwIsEncrypted = srvCanEncrypt(user->server) && !(user->plainPasswd != NULL && user->plainPasswd[0] == 0);
	/* if encryption is possible and password is non-null, use it */
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "user: setting up session with %s password\n", pwIsEncrypted ? "encrypted" : "plain text");
	if((domain = user->domain) == NULL){	/* should not happen, but catch it anyway... */
		if(CFG_DID_CHANGE(domainChg)){
			numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
			domain = cfgStringValue(cfgQueryFromArray("domain", numEntries, path), "");
		}
	}
	PROC_WAITFOR_X(4, cifsSessionSetup(smb, this, user->unixUid, user->smbUsername, user->plainPasswd, user->pwHashPreNt, user->pwHashNt, pwIsEncrypted, domain));
	res.err = cifsRes->err;
	if(cifsRes->err.class == 0 && ((cifsRes->action & 1) == 0 || user->allowGuest)){
		srvSetSessionSetupResult(user->server, cifsRes->serverOS, cifsRes->serverLanMan);
		user->cstatus = PSL_CONNECTED;
		user->smbUid = cifsRes->uid;
		user->isGuest = (cifsRes->action & 1) != 0;
	}else{
		if(cifsRes->err.class == 0){	/* logged in as guest, but not accepted */
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "user: connect: did not accept guest login\n");
			user->smbUid = cifsRes->uid;
			PROC_WAITFOR_X(5, srvWantRequest(user->server, this));
			cifsLogoff(smb, NULL, user->smbUid);	/* disconnect in background */
			res.err.class = ERRCLS_CIFS_SRV;
			res.err.code = ERR_CIFS_SRVbadpw;
		}
		user->cstatus = PSL_NOTCONNECTED;
		user->smbUid = -1;
		user->isGuest = 0;
	}
doReturn:
    PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(user->connectSignal, NULL);
		userRelease(user);
	);
    PROC_END;
}

/* ------------------------------------------------------------------------- */

void	userConnect(cifsUser_t *this, void *caller)
{
userRes_t	res;
userCtx_t	*ctx;

    res.err.class = res.err.code = 0;
    if(this->cstatus == PSL_CONNECTED){
        PROC_IMMEDIATE_RET(caller, &res);
    }
    PROC_INIT(ctx, userConnectProc, caller, userPool, freePoolBlock);
    ctx->user = this;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	userDisconnectProc(void *obj, void *arg)
{
userCtx_t		*this = obj;
userRes_t		res;
smbRequest_t	*smb = arg;
cifsRes_t		*rval = arg;
array_t			*allUsers;
btint			i, c;
cifsUser_t		*u, *user = this->user;

    PROC_BEGIN;
	userRetain(user);
	while(user->cstatus == PSL_CONNECTING || user->cstatus == PSL_DISCONNECTING){
		PROC_WAITFOR_X(1, procCondWait(user->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(user->cstatus == PSL_NOTCONNECTED){
		res.err.code = res.err.class = 0;
		userRelease(user);
		PROC_RETURN(&res);
	}
	user->cstatus = PSL_DISCONNECTING;

	allUsers = srvAllUsers(user->server);
	c = arrObjectCount(allUsers);
	for(i=0;i<c;i++){	/* try to find this user name in list of connected users */
		u = arrObjectAtIndex(allUsers, i);
		if(u == user || uniStrCmp(u->smbUsername, user->smbUsername, 0) != 0)
			continue;
		if(u->cstatus == PSL_CONNECTED){	/* this SMB user is still connected */
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "user: won\'t disconnect because other user ->%s<- still connected.\n", uniToCstring(u->smbUsername));
			break;
		}
	}
	arrFree(allUsers);
	if(i < c){	/* other sitll connected user found */
		user->cstatus = PSL_NOTCONNECTED;
		res.err.code = res.err.class = 0;
	}else{		/* loop ran until end */
		PROC_WAITFOR_X(2, srvWantRequest(user->server, this));
		PROC_WAITFOR_X(3, cifsLogoff(smb, this, user->smbUid));
		user->cstatus = PSL_NOTCONNECTED;
		res.err = rval->err;
	}
    PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(user->connectSignal, NULL);
		userRelease(user);
	);
    PROC_END;
}

void	userDisconnect(cifsUser_t *this, void *caller)
{
userRes_t	res;
userCtx_t	*ctx;

    res.err.class = res.err.code = 0;
    if(this->cstatus == PSL_NOTCONNECTED){
        PROC_IMMEDIATE_RET(caller, &res);
    }
    PROC_INIT(ctx, userDisconnectProc, caller, userPool, freePoolBlock);
    ctx->user = this;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	userConnectionLost(cifsUser_t *this)
{
    this->cstatus = PSL_NOTCONNECTED;
    procCondBroadcast(this->connectSignal, NULL);
}

/* ------------------------------------------------------------------------- */

typedef struct userHoldoffInfo{
	btu8 	*srvName;
	btuni	*shareName;
	btint	unixUid;
}userHoldoffInfo_t;

static hashtable_t	*holdoff = NULL;

static btint	holdoffCompare(void *o1, void *o2)
{
userHoldoffInfo_t	*this = o1, *other = o2;
btint				c;
	
	if((this->srvName == NULL && other->srvName != NULL) || (this->shareName == NULL && other->shareName != NULL)){
		return 1;
	}
	if((this->srvName != NULL && other->srvName == NULL) || (this->shareName != NULL && other->shareName == NULL)){
		return -1;
	}
	if(this->srvName != NULL){
		if((c = utilStrCmp(this->srvName, other->srvName, 0)) != 0)
			return c;
	}
	if(this->shareName != NULL){
		if((c = uniStrCmp(this->shareName, other->shareName, 0)) != 0)
			return c;
	}
	return this->unixUid - other->unixUid;
}

static btu32	holdoffHashVal(void *obj)
{
userHoldoffInfo_t	*this = obj;
btu32				hash = this->unixUid;

	if(this->srvName != NULL)
		hash += utilStrHash(this->srvName, 0);
	if(this->shareName != NULL)
		hash += uniStrHash(this->shareName, 0);
	return hash;
}

static void	holdoffRelease(void *obj)
{
userHoldoffInfo_t	*this = obj;

	if(this->srvName != NULL){
		if(this->srvName != NULL){
			memFree(this->srvName);
			this->srvName = NULL;
		}
		if(this->shareName != NULL){
			memFree(this->shareName);
			this->shareName = NULL;
		}
	}
	memFree(this);
}

static collectionObjectI_t holdoffInterface = {
	holdoffCompare,
	holdoffHashVal,
	NULL,
	holdoffRelease
};

static void	userRemoveHoldoff(void *obj, void *arg)
{
	htabRemoveObjectEqualTo(holdoff, obj);
}

static void	userAddHoldoff(btu8 *srvName, btuni *shareName, btint unixUid)
{
userHoldoffInfo_t	h, *p;
btint				numEntries;
cfgProperty_t		*path[8];
static btu32		holdoffTimeChg = 0;
static bttime		holdoffTime;

	if(holdoff == NULL){
		holdoff = htabNew(&holdoffInterface, 16, 75);
	}
	h.srvName = srvName;
	h.shareName = shareName;
	h.unixUid = unixUid;
	p = htabFindObjectEqualTo(holdoff, &h);
	if(p == NULL){
		p = memCalloc(sizeof(userHoldoffInfo_t));
		if(srvName != NULL)
			p->srvName = utilStrDup(srvName);
		if(shareName != NULL)
			p->shareName = uniStrDup(shareName);
		p->unixUid = unixUid;
		htabAddObject(holdoff, p);
	}
	if(CFG_DID_CHANGE(holdoffTimeChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		holdoffTime = cfgIntValue(cfgQueryFromArray("holdoffTime", numEntries, path), 1000);
	}
	elpCallAtCancelPrev(elpNow() + holdoffTime, userRemoveHoldoff, p);
}

static btbool	userHoldoffIsActive(btu8 *srvName, btuni *shareName, btint unixUid)
{
userHoldoffInfo_t	h;

	if(holdoff == NULL){
		return 0;
	}
	h.srvName = srvName;
	h.shareName = shareName;
	h.unixUid = unixUid;
	return htabFindObjectEqualTo(holdoff, &h) != NULL;
}

/* ------------------------------------------------------------------------- */

static cifsUser_t	*userFromDb(server_t *server, btu32 unixUid, btu32 unixGid, btbool defaultDb, btError *err)
{
userRecord_t	*rec;
cifsUser_t		*user;

	err->class = err->code = 0;
	rec = udbQueryUser(cifsUserDb, unixUid, defaultDb ? (btu8 *)GUIMAN_DEFAULT_ENTRY : srvName(server));
	if(rec != NULL){	/* user found in database */
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: found user %s in database%s\n", uniToCstring(rec->cifsName), defaultDb ? " (default)" : "");
		user = userNew(rec->unixUid, unixGid, rec->cifsName, rec->plainPasswd, rec->domain, rec->ntHash, rec->preNtHash, rec->plainPasswd != NULL);
		licRegisterServer(server, user, err);
		if(err->class != 0){
			userInvalidate(user);
			return NULL;
		}
		return user;
	}
	return NULL;
}

/* ------------------------------------------------------------------------- */

static btint	userGuiLockCompare(void *o1, void *o2)
{
userGuiLock_t	*this = o1, *other = o2;

	return this->unixUid - other->unixUid;
}

static btu32	userGuiLockHashVal(void *obj)
{
userGuiLock_t	*this = obj;

	return this->unixUid;
}

static collectionObjectI_t	userGuiLockInterface = {
	userGuiLockCompare,
	userGuiLockHashVal,
	NULL,
	NULL
};

userGuiLock_t	*userGetGuiLock(btint unixUid)
{
static hashtable_t	*allLocks;
userGuiLock_t		*rval;

	if(allLocks == NULL)
		allLocks = htabNew(&userGuiLockInterface, 16, 75);
	rval = htabFindObjectEqualTo(allLocks, &unixUid);
	if(rval == NULL){
		rval = memCalloc(sizeof(userGuiLock_t));
		rval->unixUid = unixUid;
		rval->guiSignal = procCondNew();
		htabAddObject(allLocks, rval);
	}
	return rval;
}

/* ------------------------------------------------------------------------- */

static inline btbool	errorIsFatal(btError *err)
{
	switch(err->class){
	case ERRCLS_CIFS_DOS:
		switch(err->code){
		case ERR_CIFS_DOSnoaccess:
		case ERR_CIFS_DOSbadaccess:
			return 0;
		}
		break;
	case ERRCLS_CIFS_SRV:
		switch(err->code){
		case ERR_CIFS_SRVerror:
		case ERR_CIFS_SRVbadpw:
		case ERR_CIFS_SRVaccess:
		case ERR_CIFS_SRVsrverror:
		case ERR_CIFS_SRVbadpermits:
		case ERR_CIFS_SRVtoomanyuids:
			return 0;
		}
		break;
	case ERRCLS_CIFS_LOCAL:
		switch(err->code){	/* break on fatal errors */
		case ERR_CIFS_LOCpwencrypt:
			return 0;
		}
		break;
	}
	return 1;
}

/* ------------------------------------------------------------------------- */

#define	TRY_CONNECT(x)	\
		PROC_WAITFOR_X(x, userConnect(this->user, this));				\
		if(urval->err.class != 0){										\
			userInvalidate(this->user);									\
			this->user = NULL;											\
		}else{															\
			userRetain(this->user);	/* match release at return of function */\
			this->isNew = 1;											\
			res.err.class = res.err.code = 0;							\
			goto doReturn;												\
		}

static void	userLoginProc(void *obj, void *arg)
{
userCtx_t		*this = obj;
beVfsRes_t		res;
userRes_t		*urval = arg;
uirepRes_t		*pwrval = arg;
btuni			username[64], passwd[64];
btu8			*guestUser, *guestPw;
array_t			*allUsers;
btint			cnt, i, numEntries;
cifsUser_t		*tryUser;
struct passwd	*pwent;
cfgProperty_t	*path[8];
static btu32	domainChg = 0;
static btu8		*defaultDomain;

	PROC_BEGIN;
retry:
	this->user = srvUserForUid(this->server, this->unixUid);
	if(this->user != NULL){	/* user already cached at server */
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userAuthenticate: user cached at server\n");
		userRetain(this->user);
		PROC_WAITFOR_X(1, userConnect(this->user, this));
		res.err = urval->err;
		goto doReturn;
	}
	this->unixGid = 0;
	if((pwent = getpwuid(this->unixUid)) != NULL){
		this->unixGid = pwent->pw_gid;
		this->auth.username = utilStrDup(pwent->pw_name);
	}
	if(this->domain == NULL){
		if(CFG_DID_CHANGE(domainChg)){
			numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
			defaultDomain = cfgStringValue(cfgQueryFromArray("domain", numEntries, path), "");
		}
		this->domain = defaultDomain;
	}
	this->auth.domain = utilStrDup(this->domain);
	/* make sure server is connected; srvIsUserlevelSecurity() is not valid
	 * otherwise */
	PROC_WAITFOR_X(15, srvConnect(this->server, this->unixUid, this));
	/* if server is in share-level security mode, create dummy user object */
	if(!srvIsUserlevelSecurity(this->server)){
		convLocalAsciiToUni("<none>", username, 64);
		passwd[0] = 0;
		this->user = userNew(this->unixUid, this->unixGid, username, passwd, NULL, NULL, NULL, 1);
		licRegisterServer(this->server, this->user, &res.err);
		if(res.err.class != 0){
			userInvalidate(this->user);
			this->user = NULL;
		}else{
			TRY_CONNECT(14);
		}
	}
	/* try with stored credentials for server: */
	if((this->user = userFromDb(this->server, this->unixUid, this->unixGid, 0, &res.err)) != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userAuthenticate: trying with user from DB\n");
		TRY_CONNECT(2);
	}
	/* ### maybe check for userFromDb() error code here */
	/* try with default credentials of user: */
	if((this->user = userFromDb(this->server, this->unixUid, this->unixGid, 1, &res.err)) != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userAuthenticate: trying with default user from DB\n");
		TRY_CONNECT(3);
	}
	/* ### maybe check for userFromDb() error code here */
	/* check whether there's a guest account set up: */
	srvGuestCredentials(this->server, &guestUser, &guestPw);
	if(guestUser != NULL && guestUser[0] != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userAuthenticate: trying with guest credentials\n");
		convLocalAsciiToUni(guestUser, username, 64);
		passwd[0] = 0;
		if(guestPw != NULL)
			convLocalAsciiToUni(guestPw, passwd, 64);
		this->user = userNew(this->unixUid, this->unixGid, username, passwd, NULL, NULL, NULL, 1);
		licRegisterServer(this->server, this->user, &res.err);
		if(res.err.class != 0){
			userInvalidate(this->user);
			this->user = NULL;
			goto doReturn;
		}
		TRY_CONNECT(5);
	}
	/* try null session: */
	if(this->allowNullSession){
		this->user = srvNullSession(this->server);
		if(this->user != NULL){
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "user: userAuthenticate: trying with NULL session\n");
			userRetain(this->user);
			PROC_WAITFOR_X(7, userConnect(this->user, this));
			if(urval->err.class != 0){
				userRelease(this->user);
				this->user = NULL;
			}
			res.err = urval->err;
			goto doReturn;
		}
	}
	if(this->allowAskUser){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userAuthenticate: decided to ask user\n");
		if(this->auth.passwd != NULL){
			memset(this->auth.passwd, 0, strlen(this->auth.passwd));
			memFree(this->auth.passwd);
			this->auth.passwd = NULL;
		}
		for(;;){
			this->lock = userGetGuiLock(this->unixUid);
			if(this->lock->guiIsLocked){
				while(this->lock->guiIsLocked){
					if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
						logPrintf(LOGM_DEBUG, "user: userAuthenticate: waiting for UI lock\n");
					PROC_WAITFOR_X(8, procCondWait(this->lock->guiSignal, this));
				}
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "user: userAuthenticate: UI lock released\n");
				if(arg == NULL){	/* user cancelled */
					if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
						logPrintf(LOGM_DEBUG, "user: userAuthenticate: parallel GUI process cancelled, do cancel\n");
					res.err.class = ERRCLS_FRAMEWORK;
					res.err.code = ERR_FRWusrabort;
					goto doReturn;
				}
			}
/* Arrange for a callback from the eventloop. The following test whether the
 * user is alreayd connecting would not work reliably otherwise because a user
 * object from an unsuccessful connect could still be retained. */
			PROC_WAITFOR_X(9, elpCallCancelPrev(userLoginProc, this));
			/* if there's a connection attempt running, wait for it's result */
			if((this->user = srvUserForUid(this->server, this->unixUid)) != NULL){
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "user: userAuthenticate: there\'s a connect running, waiting for its results\n");
				userRetain(this->user);
				/* get return status of concurrent connect: */
				PROC_WAITFOR_X(10, userConnect(this->user, this));
				res.err = urval->err;
				goto doReturn;
			}
			if(userHoldoffIsActive(srvName(this->server), NULL, this->unixUid)){
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "user: userAuthenticate: automatic cancel due to holdoff\n");
				res.err.class = ERRCLS_FRAMEWORK;
				res.err.code = ERR_FRWusrabort;
				goto doReturn;
			}
			this->lock->guiIsLocked = 1;
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "user: userAuthenticate: really asking user now\n");
			PROC_WAITFOR_X(11, uirepGetCredentials(this->unixUid, srvName(this->server), &this->auth, this));
			this->lock->guiIsLocked = 0;
			if(pwrval->err.class != 0 || !pwrval->isOk){
				userAddHoldoff(srvName(this->server), NULL, this->unixUid);
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "user: userAuthenticate: user cancelled\n");
				procCondBroadcast(this->lock->guiSignal, NULL);
				res.err.class = ERRCLS_FRAMEWORK;
				res.err.code = ERR_FRWusrabort;
				goto doReturn;
			}
			this->client = pwrval->client;
			ipcSrvClientRetain(this->client);
			*username = *passwd = 0;
			if(this->auth.username != NULL)
				convLocalAsciiToUni(this->auth.username, username, ELEMENTS(username));
			if(this->auth.passwd != NULL){
				convLocalAsciiToUni(this->auth.passwd, passwd, ELEMENTS(passwd));
				memset(this->auth.passwd, 0, strlen(this->auth.passwd));
				memFree(this->auth.passwd);
				this->auth.passwd = NULL;
			}
			this->user = userNew(this->unixUid, this->unixGid, username, passwd, this->auth.domain, NULL, NULL, this->auth.allowUnencrypted);
			memset(passwd, 0, sizeof(passwd));
			licRegisterServer(this->server, this->user, &this->err);
			if(this->err.class != 0){
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "user: userAuthenticate: license limit exceeded\n");
				userInvalidate(this->user);
				this->user = NULL;
				procCondBroadcast(this->lock->guiSignal, NULL);
				goto errorTreatment;
			}
			if(this->auth.storePasswd)
				udbAddUser(cifsUserDb, srvName(this->server), this->user);
			if(this->auth.storeAsDefault)
				udbAddUser(cifsUserDb, GUIMAN_DEFAULT_ENTRY, this->user);
			procCondBroadcast(this->lock->guiSignal, (void *)1);
			TRY_CONNECT(12);
			if(urval->err.class != 0){
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "user: userAuthenticate: login failed\n");
				this->err = urval->err;
errorTreatment:
				PROC_WAITFOR_X(13, guimanDisplayError(this->client, &this->err, this));
				if(errorIsFatal(&this->err)){
					ipcSrvClientRelease(this->client);
					res.err = this->err;
					goto doReturn;
				}
			}
			ipcSrvClientRelease(this->client);
		}
	}
doReturn:
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "user: userAuthenticate: should return err %d/%d\n", res.err.class, res.err.code);
	if(res.err.class != 0 && this->unixUid == 0){	/* root should get access in any case */
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userAuthenticate: trying to map root access\n");
		allUsers = srvAllUsers(this->server);
		cnt = arrObjectCount(allUsers);
		for(i=0;i<cnt;i++){
			tryUser = arrObjectAtIndex(allUsers, i);
			if(userConnectionStatus(tryUser) == PSL_CONNECTED){
				this->unixUid = userUnixUid(tryUser);
				if(this->unixUid != 0){	/* don't get into loop... */
					arrFree(allUsers);
					goto retry;
				}
			}
		}
		for(i=0;i<cnt;i++){
			tryUser = arrObjectAtIndex(allUsers, i);
			this->unixUid = userUnixUid(tryUser);
			if(this->unixUid != 0){	/* try to map to non-root user */
				arrFree(allUsers);
				goto retry;
			}
			
		}
		arrFree(allUsers);
	}
	res.user = this->user;
	if(this->auth.username != NULL){
		memFree(this->auth.username);
		this->auth.username = NULL;
	}
	if(this->auth.passwd != NULL){
		memFree(this->auth.passwd);
		this->auth.passwd = NULL;
	}
	if(this->auth.domain != NULL){
		memFree(this->auth.domain);
		this->auth.domain = NULL;
	}
	PROC_RETURN_AND_CLEANUP(&res,
		if(this->user != NULL)
			userRelease(this->user);
	);
	PROC_END;
}

void	userAuthenticate(struct server *server, btu8 *domain, btint unixUid, btint unixGid, void *caller)
{
userCtx_t	*ctx;

    if(userPool == NULL)
        userPool = poolNew(sizeof(userCtx_t), sizeof(userCtx_t));
    PROC_INIT(ctx, userLoginProc, caller, userPool, freePoolBlock);
	ctx->server = server;
	ctx->unixUid = unixUid;
	ctx->allowNullSession = 0;
	ctx->allowAskUser = 1;
	ctx->domain = domain;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	userGetSharePasswordForShareProc(void *obj, void *arg)
{
userCtx_t			*this = obj;
cifsUser_t			*user = this->user;
userPasswdCache_t	*cache;
userRes_t			res;
uirepRes_t			*pwrval = arg;
btuni				passwd[64];

	PROC_BEGIN;
	userRetain(this->user);
	shrRetain(this->share);
	this->lock = userGetGuiLock(user->unixUid);
	if(this->lock->guiIsLocked){
		while(this->lock->guiIsLocked){
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "user: userGetSharePassword: waiting for UI lock\n");
			PROC_WAITFOR_X(1, procCondWait(this->lock->guiSignal, this));
		}
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userGetSharePassword: UI lock released\n");
	}

	if(userHoldoffIsActive(NULL, shrPath(this->share), user->unixUid)){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userGetSharePassword: automatic cancel due to holdoff\n");
		res.err.class = ERRCLS_FRAMEWORK;
		res.err.code = ERR_FRWusrabort;
		res.client = NULL;
		goto doReturn;
	}

	this->lock->guiIsLocked = 1;
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "user: userGetSharePassword: really asking user now\n");
	PROC_WAITFOR_X(2, uirepGetSharePassword(user->unixUid, shrPath(this->share), &this->auth, this));
	this->lock->guiIsLocked = 0;
	if(pwrval->err.class != 0 || !pwrval->isOk){
		userAddHoldoff(NULL, shrPath(this->share), user->unixUid);
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "user: userGetSharePassword: user cancelled\n");
		res.err.class = ERRCLS_FRAMEWORK;
		res.err.code = ERR_FRWusrabort;
		goto doReturn;
	}
	this->client = pwrval->client;
	ipcSrvClientRetain(this->client);
	*passwd = 0;
	if(this->auth.passwd != NULL)
		convLocalAsciiToUni(this->auth.passwd, passwd, ELEMENTS(passwd));
	cache = pwCacheNew(shrPath(this->share), passwd, this->auth.allowUnencrypted);
	res.passwd = passwd;
	res.allowUnencrypted = this->auth.allowUnencrypted;
	res.client = this->client;
	res.err.class = res.err.code = 0;
	htabAddObject(user->sharePasswordCache, cache);
doReturn:
	if(this->auth.passwd != NULL){
		memset(this->auth.passwd, 0, strlen(this->auth.passwd));
		memFree(this->auth.passwd);
		this->auth.passwd = NULL;
	}
	procCondBroadcast(this->lock->guiSignal, (void *)1);
	shrRelease(this->share);
	userRelease(this->user);
	PROC_RETURN_AND_CLEANUP(&res,
		if(this->client != NULL)
			ipcSrvClientRelease(this->client);
	);
	PROC_END;
}

void	userGetSharePasswordForShare(cifsUser_t *user, struct share *share, void *caller)
{
userRes_t			res;
btuni				*shareName;
userPasswdCache_t	*cache;
userCtx_t			*ctx;
btint				i, cnt;
array_t				*allUsers;
cifsUser_t			*u;

	shareName = shrPath(share);
	if((cache = htabFindObjectEqualTo(user->sharePasswordCache, &shareName)) != NULL){
		memset(&res, 0, sizeof(res));
		res.passwd = cache->passwd;
		res.allowUnencrypted = cache->allowUnencrypted;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	if(user->unixUid == 0){	/* try to map root access to any user who has a password */
		allUsers = srvAllUsers(user->server);
		cnt = arrObjectCount(allUsers);
		for(i=0;i<cnt;i++){
			u = arrObjectAtIndex(allUsers, i);
			if((cache = htabFindObjectEqualTo(u->sharePasswordCache, &shareName)) != NULL){
				memset(&res, 0, sizeof(res));
				res.passwd = cache->passwd;
				res.allowUnencrypted = cache->allowUnencrypted;
				PROC_IMMEDIATE_RET(caller, &res);
			}
		}
		arrFree(allUsers);
	}
	/* not cached, we have to ask the user... */
    PROC_INIT(ctx, userGetSharePasswordForShareProc, caller, userPool, freePoolBlock);
	ctx->user = user;
	ctx->share = share;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	userDeleteSharePasswordForShare(cifsUser_t *user, struct share *share)
{
btuni				*shareName;

	shareName = shrPath(share);
	htabRemoveObjectEqualTo(user->sharePasswordCache, &shareName);
}

/* ------------------------------------------------------------------------- */

btint	userUnixUidCompare(cifsUser_t *this, cifsUser_t *other)
{
    return this->unixUid - other->unixUid;
}

btu32	userUnixUidHashval(cifsUser_t *this)
{
    return this->unixUid;
}

void	userRetain(cifsUser_t *this)
{
    this->refCount++;
}

void	userRelease(cifsUser_t *this)
{
	if(this->refCount == 1 && this->server != NULL)	/* this release would free the object */
		userDisconnect(this, NULL);	/* don't wait for result */
	this->refCount--;
    if(this->refCount == 0){
        userFree(this);
    }else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

void	userInvalidate(cifsUser_t *user)
{
	if(!user->isInvalidated){
		user->isInvalidated = 1;
		userRelease(user);
	}
}

/* ------------------------------------------------------------------------- */

collectionObjectI_t	userByUnixUidInterface = {
    (btint (*)(void *obj, void *otherObj))			userUnixUidCompare,
    (btu32 (*)(void *obj))							userUnixUidHashval,
    NULL,
    NULL
};

/* ------------------------------------------------------------------------- */

