/*
 * Name: argparser.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-12-08
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: argparser.c,v 1.1.1.1 1999/04/12 13:56:37 cs Exp $
 */

#include "stdheaders.h"
#include "argparser.h"
#include "memalloc.h"
#include "hashtable.h"
#include "array.h"

typedef struct optDescriptor{
	btu8	name;			/* must be at first position for hashtable */
	btbool	isNumeric;
	btu8	*parameterName;	/* NULL if no parameter */
	btu8	*helpText;
	btu8	*value;			/* set during parsing, references argp[] */
}optDescriptor_t;

/* ------------------------------------------------------------------------- */

static btint	optCompare(void *obj, void *otherObj)
{
optDescriptor_t	*this = obj, *other = otherObj;

	return this->name - other->name;
}

static btu32	optHashVal(void *obj)
{
optDescriptor_t	*this = obj;

	return this->name;
}

static collectionObjectI_t	dummyInterface = {
	optCompare,
	optHashVal,
	NULL,
	NULL
};

static collectionObjectI_t	freeMeInterface = {
	optCompare,
	optHashVal,
	NULL,
	(void *)memFree
};

/* ------------------------------------------------------------------------- */

argParser_t	*argpNew(void)
{
argParser_t	*volatile ap;

EXC_TRY
	ap = memCalloc(sizeof(argParser_t));
	ap->optByName = htabNew(&dummyInterface, 8, 75);
	ap->optList = arrNew(&freeMeInterface, 4);
	ap->arguments = arrNew(&dummyInterface, 4);
EXC_CATCH
	argpFree(ap);
	EXC_RETHROW();
EXC_ENDTRY
	return ap;
}

/* ------------------------------------------------------------------------- */

void	argpFree(argParser_t *ap)
{
	if(ap != NULL){
		if(ap->optByName != NULL)
			htabFree(ap->optByName);
		if(ap->optList != NULL)
			arrFree(ap->optList);
		if(ap->arguments != NULL)
			ap->arguments = NULL;
		memset(ap, 0, sizeof(*ap));
		memFree(ap);
	}
}

/* ------------------------------------------------------------------------- */

void	argpAddOption(argParser_t *ap, btu8 name, btu8 *paramName, btbool isNumeric, btu8 *helpText)
{
optDescriptor_t	*od;

	od = memCalloc(sizeof(optDescriptor_t));
	od->name = name;
	od->parameterName = paramName;
	od->helpText = helpText;
	od->isNumeric = isNumeric;
	htabAddObject(ap->optByName, od);
	arrAddObject(ap->optList, od);
}

/* ------------------------------------------------------------------------- */

btbool	argpParse(argParser_t *ap, btint argc, char **argv, void *errFp)
{
btint			i;
btbool			noMoreFlags = 0;
char			name, *flagp;
optDescriptor_t	*opt;

	for(i=1;i<argc;i++){
		if(noMoreFlags || argv[i][0] != '-'){
			arrAddObject(ap->arguments, argv[i]);
		}else{	/* argv[i][0] must be '-' */
			if(argv[i][1] == 0){	/* single '-' */
				arrAddObject(ap->arguments, argv[i]);
			}else if(argv[i][1] == '-' && argv[i][2] == 0){	/* '--' = end of flags */
				noMoreFlags = 1;
			}else{
				flagp = &argv[i][1];
				for(;;){
					name = *flagp++;
					opt = htabFindObjectEqualTo(ap->optByName, &name);
					if(opt == NULL){
						fprintf(errFp, "option not supported: -%c\n", name);
						return 0;
					}
					if(opt->parameterName == NULL){	/* flag argument */
						opt->value = &opt->name;
						if(*flagp == 0)
							break;
					}else{	/* parameter argument */
						if(*flagp != 0){	/* immediate parameter */
							if(flagp != &argv[i][2]){	/* not the first flag in list */
								fprintf(errFp, "sequence of options ambiguous due to option \'-%c\' which requires a parameter\n", name);
								return 0;
							}
							opt->value = flagp;
						}else{
							if(++i >= argc){
								fprintf(errFp, "option \'-%c\' requires a %s parameter\n", name, opt->parameterName);
								return 0;
							}
							opt->value = argv[i];
						}
						if(opt->isNumeric && !(isdigit(opt->value[0]) || (opt->value[0] == '-' && isdigit(opt->value[1])))){
							fprintf(errFp, "option \'-%c\' requires a numeric parameter for %s\n", name, opt->parameterName);
							return 0;
						}
						break;
					}
				}
			}
		}
	}
	return 1;
}

/* ------------------------------------------------------------------------- */

void	argpPrintHelp(argParser_t *ap, void *outFd)
{
btint			i, j, cnt, maxPnameLen, len;
optDescriptor_t	*opt;
btu8			line[256], *hp, *pname;

	cnt = arrObjectCount(ap->optList);
	maxPnameLen = 1;
	for(i=0;i<cnt;i++){
		opt = arrObjectAtIndex(ap->optList, i);
		if(opt->helpText != NULL && opt->parameterName != NULL){
			len = strlen(opt->parameterName);
			if(len > maxPnameLen)
				maxPnameLen = len;
		}
	}
	for(i=0;i<cnt;i++){
		opt = arrObjectAtIndex(ap->optList, i);
		if(opt->helpText == NULL)
			continue;
		pname = opt->parameterName == NULL ? (btu8 *)"" : opt->parameterName;
		for(hp=opt->helpText;*hp!=0;){
			for(j=0; hp[j]!='\n' && hp[j]!=0 && j<255; j++)
				line[j] = hp[j];
			line[j] = 0;
			if(hp == opt->helpText)
				fprintf(outFd, "-%c %-*s %s\n", opt->name, maxPnameLen, pname, line);
			else
				fprintf(outFd, "   %-*s %s\n", maxPnameLen, "", line);
			hp += j;
			if(*hp != 0)	/* must have been linefeed */
				hp++;
		}
	}
}

/* ------------------------------------------------------------------------- */

btint	argpArgumentCount(argParser_t *ap)
{
	return arrObjectCount(ap->arguments);
}

/* ------------------------------------------------------------------------- */

btu8	*argpArgumentAtIndex(argParser_t *ap, btint index)
{
	return arrObjectAtIndex(ap->arguments, index);
}

/* ------------------------------------------------------------------------- */

btbool	argpOptionIsSet(argParser_t *ap, btu8 optName)
{
optDescriptor_t	*opt = htabFindObjectEqualTo(ap->optByName, &optName);

	if(opt == NULL)
		return 0;
	return opt->value != NULL;
}

/* ------------------------------------------------------------------------- */

btint	argpOptionIntVal(argParser_t *ap, btu8 optName, btbool *isValid)
{
optDescriptor_t	*opt = htabFindObjectEqualTo(ap->optByName, &optName);
char			*endp;

	if(isValid != NULL)
		*isValid = 0;
	if(opt == NULL)
		return 0;
	if(opt->parameterName == NULL || opt->value == NULL)
		return 0;
	if(isValid != NULL)
		*isValid = 1;
	return (btint)strtol(opt->value, &endp, 0);
}

/* ------------------------------------------------------------------------- */

btu8	*argpOptionStringVal(argParser_t *ap, btu8 optName)
{
optDescriptor_t	*opt = htabFindObjectEqualTo(ap->optByName, &optName);

	if(opt == NULL)
		return NULL;
	if(opt->parameterName == NULL || opt->value == NULL)
		return NULL;
	return opt->value;
}

/* ------------------------------------------------------------------------- */

#ifdef TEST

static void	printOpt(argParser_t *ap, btu8 opt, btbool hasParam, btbool isNumeric)
{
	if(argpOptionIsSet(ap, opt)){
		if(hasParam){
			if(isNumeric){
				printf("numeric option -%c = %d\n", opt, argpOptionIntVal(ap, opt, NULL));
			}else{
				printf("option -%c = %s\n", opt, argpOptionStringVal(ap, opt));
			}
		}else{
			printf("option -%c is set\n", opt);
		}
	}else{
		printf("option -%c is not set\n", opt);
	}
}

int	main(btint argc, char **argv)
{
argParser_t	*ap = argpNew();
btint		i, c;

	argpAddOption(ap, 'a', NULL, 0, "append to existing file");
	argpAddOption(ap, 'v', NULL, 0, "verbose mode");
	argpAddOption(ap, 'f', "<file>", 0, "write output to <file>\ninstead of stdout");
	argpAddOption(ap, 'p', "<performance>", 1, "numeric value describing performance");
	argpAddOption(ap, 'h', NULL, 0, "print help and exit");
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "type \'%s -h\' for help\n", argv[0]);
		exit(1);
	}
	if(argpOptionIsSet(ap, 'h')){
		printf("usage: %s [options]\n", argv[0]);
		printf("valid options are:\n");
		argpPrintHelp(ap, stdout);
		exit(0);
	}
	printOpt(ap, 'a', 0, 0);
	printOpt(ap, 'v', 0, 0);
	printOpt(ap, 'f', 1, 0);
	printOpt(ap, 'p', 1, 1);
	printOpt(ap, 'h', 0, 0);
	c = argpArgumentCount(ap);
	printf("%d arguments:\n", c);
	for(i=0;i<c;i++){
		printf("arg[%d] = %s\n", i, argpArgumentAtIndex(ap, i));
	}
	return 0;
}

#endif

/* ------------------------------------------------------------------------- */
