/*
 * Name: array.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-24
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: array.h,v 1.1.1.1 1999/04/12 13:56:37 cs Exp $
 */

#ifndef __array_H_INCLUDED__
#define __array_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This class implements the simplest collection of objects: an ordered array.
The elements in the array may implement the methods in 'collectionObjectI_t'.
If 'notifyAddedTo()' is non-null, it is called when the object is added
to the array. The parameter ('obj') is a pointer to the object itself.
If 'notifyRemovedFrom()' is non-null, it is called when the object is removed
from the array. 'compare()' is called by 'arrRemoveObjectEqualTo()',
'arrFindObjectEqualTo()' and 'arrIndexOfObjectEqualTo()' to find out whether
the objects are equal. The integer return value can be used for sorting. The
first parameter ('obj') is a pointer to the stored object and the second
parameter ('otherObj') is a pointer to the reference object.

The methods of this class are not documented in detail because their names
should describe the functionality sufficiently. If you are not sure about
a detail, look into the implementation. It's rather simple.
*/

/* ------------------------------------------------------------------------- */

/* the following data type is used by array and hashtable: */
typedef struct collectionObjectI{
    btint	(*compare)(void *obj, void *otherObj);	/* returns obj - otherObj for sorts and 0 if objects are equal for hashtables */
    btu32	(*hashValue)(void *obj);
    void	(*notifyAddedTo)(void *obj);
    void	(*notifyRemovedFrom)(void *obj);
}collectionObjectI_t;

typedef struct array{
    collectionObjectI_t	objectInterface;
    void				**table;
    btu32				tableSize;
    btu32				objectCount;
}array_t;

/* ------------------------------------------------------------------------- */

array_t	*arrNew(const collectionObjectI_t *contentObjectInterface, btu32 initialCapacity);
void	arrFree(array_t *this);
void	arrAddObject(array_t *this, void *obj);
void	arrRemoveObjectAtIndex(array_t *this, btsize index);
void	arrRemoveAllObjects(array_t *this);

/* these three methods require a valid 'compare()' function: */
btbool	arrRemoveObjectEqualTo(array_t *this, void *obj);
void	*arrFindObjectEqualTo(array_t *this, void *obj);
btsize	arrIndexOfObjectEqualTo(array_t *this, void *obj);
void	arrSort(array_t *arr);

/* ------------------------------------------------------------------------- */

static inline void	arrReplaceInterface(array_t *this, const collectionObjectI_t *interf)
{
	this->objectInterface = *interf;
}

static inline btu32	arrObjectCount(array_t *this)
{
    return this->objectCount;
}

static inline void	*arrObjectAtIndex(array_t *this, btu32 i)
{
    if(i >= this->objectCount)
        return NULL;
    return this->table[i];
}

static inline void **arrContentArray(array_t *this)
{
	return this->table;
}

static inline void	*arrRemoveLastObject(array_t *this)
{
void	*p = NULL;
    
    if(this->objectCount > 0){
        p = this->table[this->objectCount - 1];
        if(this->objectInterface.notifyRemovedFrom != NULL)
            this->objectInterface.notifyRemovedFrom(p);
        this->objectCount--;
    }
    return p;
}

/* ------------------------------------------------------------------------- */

#endif	/* __array_H_INCLUDED__ */

