/*
 * Name: blockpool.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-11-05
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: blockpool.h,v 1.1.1.1 1999/04/12 13:56:37 cs Exp $
 */

#ifndef __blockpool_H_INCLUDED__
#define __blockpool_H_INCLUDED__

#include "basictypes.h"
#include "array.h"
#include "memalloc.h"

/*
General Description:
This class implements a pool of storage blocks. It's used to optimize
frequently used memory allocation. Whereever the same size of memory blocks
is allocated and freed frequently, blockpools improve the performance.
The blockpool object knows about the size of the storage blocks and the
number of bytes within this size which must be cleared. When a new block
is needed, it's either taken from an array of unused blocks, or, if that array
is empty, it's allocated with 'memAlloc()'. Blocks that are not needed any
more must be returned with 'poolReturnBlock()'. It's obvious that you must
not return other blocks than those allocated by 'poolGetBlock()' to
'poolReturnBlock()'.
The function prototypes should be sufficient information to work with this
class.
*/

/* ------------------------------------------------------------------------- */

typedef struct blockpool{
    btsize	eltSize;
    btsize	clearSize;
    array_t	*freeBlocks;
}blockpool_t;

/* ------------------------------------------------------------------------- */

blockpool_t	*poolNew(btsize eltSize, btsize clearSize);
void	poolFree(blockpool_t *this);

/* ------------------------------------------------------------------------- */

static inline void	*poolGetBlock(blockpool_t *this)
{
void	*p = arrRemoveLastObject(this->freeBlocks);

    if(p == NULL){
        p = memAlloc(this->eltSize);
    }
    memset(p, 0, this->clearSize);
    return p;
}

/* ------------------------------------------------------------------------- */

static inline void	poolReturnBlock(blockpool_t *this, void *block)
{
    arrAddObject(this->freeBlocks, block);
}

/* ------------------------------------------------------------------------- */

#endif	/* __blockpool_H_INCLUDED__ */
