/*
 * Name: cfgparse.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-10
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: cfgparse.h,v 1.1.1.1 1999/04/12 13:56:39 cs Exp $
 */

#ifndef __cfgparse_H_INCLUDED__
#define __cfgparse_H_INCLUDED__

#include "basictypes.h"
#include "hashtable.h"

/*
General Description:
This module implements a "general purpose" configuration file parser. It's
modelled after NeXT's (now Apple's) property list format. Configuration
files consist of arrays, dictionaries and strings. These basic datatypes
may be combined after the following rules:
  - A configuration file is a dictionary. You may omit the sourrouding
    braces that are usually delimiting dictionaries.
  - A string variable is every string found in the file. Multiple words
    are parsed into multiple string variables. If you don't want that you
	have to quote the string. Within quoted strings you may escape special
	characters exactly as in C.
  - A dictionary variable is defined by putting key/value pairs into
    braces. The keys must be strings while the values may be any object
	(string, array or dictionary). Example:
		{key1 = value1; key2 = value2}
	You may omit the semicolon.
  - An array variable is defined by putting the values into parantheses.
    Again, values may be any valid object. Example:
	("value at index 0", "value at index 1", {key = "this is a dictionary"})
	You may omit the commas.
If you don't understand this at the first glance, have a look at the Sharity
configuration file.

The routines which read the parameters may convert the string to a number.
See the function documentation for details. If you don't understand how
to cascade the functions to read a parameter structure, have a look at
'srvReadConfig()' in "server.c".
*/

/* ------------------------------------------------------------------------- */

#define	CFG_PROPTYPE_STRING		0
#define	CFG_PROPTYPE_LIST		1
#define	CFG_PROPTYPE_DICT		2

typedef struct cfgProperty{
	btu8	*name;			/* name of the property, must be first element */
	int		contentType;	/* content type of property */
	union{
		btu8		*string;	/* string content */
		array_t		*list;
		hashtable_t	*dict;
	}		content;
	btu32	line;			/* where this property was defined */
	btu8	*file;			/* file name of config file */
}cfgProperty_t;
/* This structure is used to hold any object that may be part of the
 * configuration file.
 */
/* ------------------------------------------------------------------------- */

#define	plist_t	cfgProperty_t
/* property lists are properties as well; plist_t may be used synonymously. */

/* ------------------------------------------------------------------------- */

typedef struct cfgFile{
	btu8	*fileName;
	bti64	lastModify;
	plist_t	*plist;
}cfgFile_t;

/* ------------------------------------------------------------------------- */

extern btu32	cfgGlobalGenerationValue;

extern btu8	*cfgYesNoEnum[3];
/* This array contains the strings "no", "yes" and a NULL pointer. It may
 * be used together with 'cfgEnumValue()' to read boolean values.
 */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
cfgFile_t	*cfgFileNew(btu8 *fileName);
/* This function creates a new object representing the configuration file. */
/* ------------------------------------------------------------------------- */
void		cfgFileFree(cfgFile_t *this);
/* This function deletes the config file object from memory. */
/* ------------------------------------------------------------------------- */
btbool		cfgFileParseIfChanged(cfgFile_t *this);
/* This function re-parses the configuration file if it has changed. It
 * returns 1 if the file was parsed, 0 otherwise.
 */
/* ------------------------------------------------------------------------- */
plist_t		*cfgFilePlist(cfgFile_t *this);
/* This function returns a pointer to a dictionary representing the entire
 * configuration file. If the file could not be parsed, it returns NULL.
 */
/* ------------------------------------------------------------------------- */
btbool		cfgFileChanged(cfgFile_t *this);
/* This function returns 1 if the configuration file has changed since the
 * last parsing and 0 otherwise.
 */
/* ------------------------------------------------------------------------- */
static inline btu32		cfgGlobalGeneration(void){return cfgGlobalGenerationValue;}
/* This function returns a number that changes when any configuration
 * variable in any file changes. It can be used to validate configuration
 * caches.
 */
/* ------------------------------------------------------------------------- */
struct notificationCenter	*cfgChangedNotification(void);
/* This function returns a notification center that fires each time something
 * in one of the configuration files changes. It can be used to update
 * configuration caches or GUI representations of the configuration. The
 * argument of the notification is always NULL.
 */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* the following functions are used to read actual values from the parsed
 * configuration file representation */
/* ------------------------------------------------------------------------- */
cfgProperty_t	*cfgQueryWithName(btu8 *name, btint numLists, ...);
/* This function goes through the list of plist_t pointers passed in the
 * variable argument list to find an entry named 'name'. The number of
 * pointers passed must be given in 'numLists'. If no such entry is found in
 * any of the lists, NULL is returned. The returned value may be any property
 * and is returned as cfgProperty_t pointer.
 */
/* ------------------------------------------------------------------------- */
cfgProperty_t	*cfgQueryFromArray(btu8 *name, btint arrSize, plist_t **arr);
/* This function is analogous to cfgQueryWithName(), but it takes the
 * variable number of property lists to search in parameter 'arr'. The number
 * of entries in 'arr' is 'arrSize'.
 */
/* ------------------------------------------------------------------------- */
array_t	*cfgQueryList(plist_t *list);
/* This function requires a dictionary 'list' as parameter. It returns an
 * array of strings consisting of the names of the properties in the
 * dictionary. If 'list' was not a dictionary or if it was NULL, an empty
 * array is returned.
 */
/* ------------------------------------------------------------------------- */
cfgProperty_t	*cfgQueryAtIndex(btsize index, plist_t *plist);
/* This function expects either an array, a string or a NULL-pointer in
 * 'plist'. If 'plist' is an array, the object at index 'index' is returned,
 * provided that it exists. If it does not exist, NULL is returned. If a
 * string is passed in 'plist' and 'index' is 0, the string is returned. If
 * 'index' is not 0 for the string, NULL is returned. This feature allows
 * arrays with a single element to be written as that element. If 'plist' was
 * NULL, the function returns NULL.
 */
/* ------------------------------------------------------------------------- */
btsize			cfgQueryValueCount(cfgProperty_t *property);
/* This function returns: 0 for dictionary properties, 1 for string type
 * properties and the number of elements in the array for array type
 * properties.
 */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* The following functions are used to read values: */
/* ------------------------------------------------------------------------- */
btu32			cfgIntValue(cfgProperty_t *property, btu32 defaultValue);
/* This function reads a value as numeric value. If it starts with "0x",
 * it's interpreted as a hexadecimal number, if it starts with "0", it's
 * interpreted as an octal number and otherwise it's parsed as a decimal
 * number. If 'property' is NULL or is otherwise not usable, it returns
 * 'defaultValue'.
 */
/* ------------------------------------------------------------------------- */
btu8			*cfgStringValue(cfgProperty_t *property, btu8 *defaultValue);
/* This function reads 'property' as a string. If this is not possible (eg.
 * because it's NULL), it returns 'defaultValue'.
 */
/* ------------------------------------------------------------------------- */
int				cfgEnumValue(cfgProperty_t *property, btu8 **enumNames, int defaultValue);
/* This function interprets 'property' as the name of an enumeration element.
 * The string value of 'property' is compared to all elements of the string
 * array 'enumNames' up to the terminating NULL entry. The index of the
 * matching entry is returned. If no match is found or 'property' is NULL,
 * 'defaultValue' is returned.
 */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* The following functions are used to set configuration values. This is
 * used for persistent property lists that are serialized to files with
 * the unparse function. */
/* ------------------------------------------------------------------------- */
cfgProperty_t	*cfgBuildIntProperty(btu32 value);
/* Returns an integer property with value 'value'. */
/* ------------------------------------------------------------------------- */
cfgProperty_t	*cfgBuildStringProperty(btu8 *value);
/* Returns a string property with value 'value'. */
/* ------------------------------------------------------------------------- */
cfgProperty_t	*cfgBuildEnumProperty(btint value, btu8 **enumNames);
/* Returns an enum-property with value 'value'. 'enumNames' is the same
 * static string array as it is used for 'cfgEnumValue()'. If 'value' is
 * outside the valid range, an invalid enum-entry is created.
 */
/* ------------------------------------------------------------------------- */
cfgProperty_t	*cfgBuildArrayProperty(void);
/* Returns an empty array property. */
/* ------------------------------------------------------------------------- */
cfgProperty_t	*cfgBuildDictProperty(void);
/* Returns an empty dictionary property. */
/* ------------------------------------------------------------------------- */
void	cfgAppendToArrayProperty(cfgProperty_t *arr, cfgProperty_t *p);
/* Adds an entry at the next possible index to an array property. If 'arr' is
 * not an array, or if 'p' is NULL, nothing is done and the property is freed.
 */
/* ------------------------------------------------------------------------- */
void	cfgRemoveFromArrayProperty(cfgProperty_t *arr, btint index);
/* Removes the entry at index 'index' from an array property. All entries
 * following the deleted one are copied down one position in the array.
 * If 'arr' is not an array property, nothing is done.
 */
/* ------------------------------------------------------------------------- */
void	cfgSetProperty(btu8 *name, plist_t *plist, cfgProperty_t *value);
/* Sets the entry 'name' in 'plist' to the value 'value'. If an entry with
 * 'name' already exists, it is overwritten. If 'value' is NULL, the entry
 * is deleted from the property list.
 * An exception is raided if 'value' is already part of a property list or
 * if 'plist' is not a dictionary.
 */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
void	cfgUnparse(cfgFile_t *file);
/* Writes the configuration back to the file. */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

#define	CFG_DID_CHANGE(generationId)	(generationId == cfgGlobalGenerationValue ? 0 : ((generationId = cfgGlobalGenerationValue), 1))
/* This macro can be used to cache configuration variables in the following
 * way:
 * btint  getConfigValue(void)
 * {
 * static btint cachedVal;
 * static btu32 valChg = 0;
 *
 *     if(CFG_DID_CHANGE(valChg)){
 *         cachedVal = cfgQueryWithName("cfgval", 1, cfgFilePlist(gdata.cfgFile)), 0);
 *     }
 *     return cachedVal;
 * }
 */
/* ------------------------------------------------------------------------- */

#endif	/* __cfgparse_H_INCLUDED__ */
