/*
 * Name: evntloop.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-13
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: evntloop.c,v 1.2 2001/10/13 12:37:23 cs Exp $
 */

#include "stdheaders.h"
#include "evntloop.h"
#include "memalloc.h"
#include "logging.h"
#include "system.h"
#include "util.h"
#include "foundationconst.h"

/* ------------------------------------------------------------------------- */

#define	INITIAL_FDS		8
#define	INITIAL_TIMED	8
#define	INITIAL_CALL	8
#define	INITIAL_PID		8
#define	ELP_MAX_FDS		1024	/* for parameter sanity check only */
#define	POLL_TIME		10000

typedef struct elpTmInfo{
	btcallback			callback;
	void				*ref;
	bttime				when;
}elpTmInfo_t;

typedef struct elpCallInfo{
	btcallback			callback;
	void				*ref;
}elpCallInfo_t;

typedef struct elpPidInfo{
    btcallback			callback;
    void				*ref;
    int					pid;
}elpPidInfo_t;

/* ------------------------------------------------------------------------- */

EXC_DEFINE(elpExcBadArgument);
EXC_DEFINE(elpExcFdAlreadyUsed);

/* ------------------------------------------------------------------------- */

static elpCallInfo_t	*fdEntries[ELP_FDMODES];
static int				fdEntriesSize[ELP_FDMODES];
static fd_set			fdSets[ELP_FDMODES];

static elpTmInfo_t		*timedEntries;
static int				timedEntriesSize;

static elpCallInfo_t	*callEntries;
static int				callEntriesSize;

static elpPidInfo_t		*pidEntries;
static int				pidEntriesSize;

static btbool			elpRunning;
static btbool			childStatusChanged;

/* ------------------------------------------------------------------------- */

static void	childDied(int signo)
{
    childStatusChanged = 1;
}

/* ------------------------------------------------------------------------- */

void	elpInit(void)
{
int		i;

	for(i=0;i<ELP_FDMODES;i++){
		fdEntriesSize[i] = INITIAL_FDS;
		fdEntries[i] = memCalloc(sizeof(elpCallInfo_t) * fdEntriesSize[i]);
		FD_ZERO(&fdSets[i]);
	}
	timedEntriesSize = INITIAL_TIMED;
	timedEntries = memCalloc(sizeof(elpTmInfo_t) * timedEntriesSize);
	callEntriesSize = INITIAL_CALL;
	callEntries = memCalloc(sizeof(elpCallInfo_t) * callEntriesSize);
    pidEntriesSize = INITIAL_PID;
    pidEntries = memCalloc(sizeof(elpPidInfo_t) * pidEntriesSize);
    signal(SIGCHLD, childDied);
}

/* ------------------------------------------------------------------------- */

/* Under the assumption that the select() system call in the event loop
 * is the only place where Sharity can sleep, we can approximate the
 * current time with the current time of the event loop cycle. To avoid
 * querying the time for cycles where it's not needed, it's stored on the
 * first request in each cycle. Upon entry of select(), all valid-flags
 * are set to 0.
 */

static struct timeval	elpRawTime;
static btbool			elpRawTimeValid = 0;

static inline void	elpMakeRawTimeValid(void)
{
	if(!elpRawTimeValid){
		struct timezone	tz;
		gettimeofday(&elpRawTime, &tz);
		elpRawTimeValid = 1;
	}
}

/* ------------------------------------------------------------------------- */

bttime	elpCurrentTime;
btbool	elpCurrentTimeValid = 0;

void	elpMakeCurrentTimeValid(void)
{
bttime			now;

	elpMakeRawTimeValid();
	now = (unsigned long)elpRawTime.tv_sec * 1000U + (unsigned long)elpRawTime.tv_usec / 1000U;
	if(now == 0)	/* zero times have a special meaning at some places */
		now++;
	elpCurrentTime = now;
	elpCurrentTimeValid = 1;
}

/* ------------------------------------------------------------------------- */

bti64	elpCurrentTime64;
btbool	elpCurrentTime64Valid = 0;

void	elpMakeCurrentTime64Valid(void)
{
bti64			now;

	elpMakeRawTimeValid();
	now = ((bti64)elpRawTime.tv_sec + FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL) * (10LL * 1000LL * 1000LL) + elpRawTime.tv_usec * 10LL;
	elpCurrentTime64 = now;
	elpCurrentTime64Valid = 1;
}

/* ------------------------------------------------------------------------- */

static inline void	elpInvalidateTimeCaches(void)
{
	elpRawTimeValid = elpCurrentTimeValid = elpCurrentTime64Valid = 0;
}

/* ------------------------------------------------------------------------- */

void	elpRegisterForFd(int fd, int mode, btcallback callb, void *ref)
{
btsize	newSize;

	if(mode < ELP_FDMODE_READ || mode >= ELP_FDMODES || callb == NULL){
		EXC_THROW_MSG(elpExcBadArgument, "elpRegisterForFd");
	}
	if(fd < 0 || fd >= ELP_MAX_FDS){
		EXC_THROW_MSG(elpExcBadArgument, "elpRegisterForFd: fd");
	}
	if(fd >= fdEntriesSize[mode]){
		newSize = fdEntriesSize[mode];
		while(fd >= newSize){
			newSize *= 2;
		}
		fdEntries[mode] = memRealloc(fdEntries[mode], newSize * sizeof(elpCallInfo_t));
		memset(fdEntries[mode] + fdEntriesSize[mode], 0, (newSize - fdEntriesSize[mode]) * sizeof(elpCallInfo_t));
		fdEntriesSize[mode] = newSize;
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "elp: setting fdEntriesSize[%d]=%d\n", mode, newSize);
	}
	if(fdEntries[mode][fd].callback != NULL
				&& !(fdEntries[mode][fd].callback == callb
								&& fdEntries[mode][fd].ref == ref)){
		EXC_THROW(elpExcFdAlreadyUsed);
	}
	fdEntries[mode][fd].callback = callb;
	fdEntries[mode][fd].ref = ref;
	FD_SET(fd, &fdSets[mode]);
}

/* ------------------------------------------------------------------------- */

void	elpCancelForFd(int fd, int mode)
{
int		i;

	if(mode >= ELP_FDMODES || fd < 0){
		EXC_THROW_MSG(elpExcBadArgument, "elpCancelForFd");
	}
	if(mode >= 0){
		if(fd >= fdEntriesSize[mode]){
			return;	/* we don't have this fd */
		}
		if(fdEntries[mode][fd].callback != NULL){
			fdEntries[mode][fd].callback = NULL;
			FD_CLR(fd, &fdSets[mode]);
		}
	}else{
		for(i=0;i<ELP_FDMODES;i++){
			if(fd < fdEntriesSize[i]){
				if(fdEntries[i][fd].callback != NULL){
					fdEntries[i][fd].callback = NULL;
					FD_CLR(fd, &fdSets[i]);
				}
			}
		}
	}
}

/* ------------------------------------------------------------------------- */

static inline void	processChildren(void)
{
int				i, rval;
btcallback		cb;
void			*ref;

    childStatusChanged = 0;
	signal(SIGCHLD, childDied);	/* reschedule signal handler */
	for(i=0;i<pidEntriesSize;i++){
		if(pidEntries[i].callback != NULL){
			elpPid_t	pidRval;
			if(sysNonblockWait(&rval, pidEntries[i].pid) == pidEntries[i].pid){
				cb = pidEntries[i].callback;
				ref = pidEntries[i].ref;
				pidEntries[i].callback = NULL;
				pidRval.pid = pidEntries[i].pid;
				pidRval.rval = rval;
				cb(ref, &pidRval);
			}
		}
	}
}

/* ------------------------------------------------------------------------- */

void	elpRegisterForPid(int pid, btcallback cb, void *ref)
{
int		i, freeEntry = -1;

	if(cb == NULL){
		EXC_THROW_MSG(elpExcBadArgument, "elpRegisterForPid");
	}
	for(i=0;i<pidEntriesSize;i++){
		if(pidEntries[i].pid == pid && pidEntries[i].callback == cb && pidEntries[i].ref == ref){
			return;
		}
		if(freeEntry < 0 && pidEntries[i].callback == NULL){
			freeEntry = i;
		}
	}
	if((i = freeEntry) < 0){
		i = pidEntriesSize;
		pidEntries = memRealloc(pidEntries, 2 * pidEntriesSize * sizeof(elpPidInfo_t));
		memset(pidEntries + pidEntriesSize, 0, pidEntriesSize * sizeof(elpPidInfo_t));
		pidEntriesSize *= 2;
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "elp: setting pidEntriesSize=%d\n", pidEntriesSize);
	}
	pidEntries[i].callback = cb;
	pidEntries[i].ref = ref;
    pidEntries[i].pid = pid;
	/* child may have already died and we won't get a signal, so check now: */
	processChildren();
}

/* ------------------------------------------------------------------------- */

void	elpCancelForPid(int pid)
{
int		i;

	for(i=0;i<pidEntriesSize;i++){
		if(pidEntries[i].callback != NULL && pidEntries[i].pid == pid){
			pidEntries[i].callback = NULL;
		}
	}
}

/* ------------------------------------------------------------------------- */

btbool	elpCallAtCancelPrev(bttime when, btcallback callback, void *ref)
{
int		i, freeEntry = -1;
btbool	overwritten = 0;

	if(callback == NULL){
		EXC_THROW_MSG(elpExcBadArgument, "elpCallAtCancelPrev");
	}
	for(i=0;i<timedEntriesSize;i++){
		if(timedEntries[i].callback == callback && timedEntries[i].ref == ref){
			overwritten = 1;
			break;
		}
		if(freeEntry < 0 && timedEntries[i].callback == NULL)
			freeEntry = i;
	}
	if(i >= timedEntriesSize){
		if(freeEntry >= 0)
			i = freeEntry;
	}
	if(i >= timedEntriesSize){
		timedEntries = memRealloc(timedEntries, 2 * timedEntriesSize * sizeof(elpTmInfo_t));
		memset(timedEntries + timedEntriesSize, 0, timedEntriesSize * sizeof(elpTmInfo_t));
		timedEntriesSize *= 2;
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "elp: setting timedEntriesSize=%d\n", timedEntriesSize);
	}
	timedEntries[i].callback = callback;
	timedEntries[i].ref = ref;
	timedEntries[i].when = when;
	return overwritten;
}

/* ------------------------------------------------------------------------- */

void	elpCancelCallAt(btcallback callback, void *ref)
{
int		i;

	if(callback == NULL){
		EXC_THROW_MSG(elpExcBadArgument, "elpCancelCallAt");
	}
	for(i=0;i<timedEntriesSize;i++){
		if(timedEntries[i].callback == callback && timedEntries[i].ref == ref){
			timedEntries[i].callback = NULL;
		}
	}
}

/* ------------------------------------------------------------------------- */

void	elpCallCancelPrev(btcallback callback, void *ref)
{
int		i, freeEntry = -1;

	if(callback == NULL){
		EXC_THROW_MSG(elpExcBadArgument, "elpCallCancelPrev");
	}
	for(i=0;i<callEntriesSize;i++){
		if(callEntries[i].callback == callback && callEntries[i].ref == ref){
			return;
		}
		if(freeEntry < 0 && callEntries[i].callback == NULL){
			freeEntry = i;
		}
	}
	if((i = freeEntry) < 0){
		i = callEntriesSize;
		callEntries = memRealloc(callEntries, 2 * callEntriesSize * sizeof(elpCallInfo_t));
		memset(callEntries + callEntriesSize, 0, callEntriesSize * sizeof(elpCallInfo_t));
		callEntriesSize *= 2;
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "elp: setting callEntriesSize=%d\n", callEntriesSize);
	}
	callEntries[i].callback = callback;
	callEntries[i].ref = ref;
}

/* ------------------------------------------------------------------------- */

void	elpCancelCall(btcallback callback, void *ref)
{
int		i;

	for(i=0;i<callEntriesSize;i++){
		if(callEntries[i].callback == callback && callEntries[i].ref == ref){
			callEntries[i].callback = NULL;
		}
	}
}

/* ------------------------------------------------------------------------- */

void	elpTerminate(void)
{
	elpRunning = 0;
}

/* ------------------------------------------------------------------------- */

void	elpRun(void)
{
fd_set			localFdSets[ELP_FDMODES];
struct timeval	tv;
int				i, j, maxFd, selectErrno;
volatile int	nFd, runThrough = 0;
btcallback		cb;
void			*ref;
bttime			now, minTime;

	elpRunning = 1;
	while(elpRunning){
		tv.tv_sec = 0;
		tv.tv_usec = 0;
		for(i=0;i<callEntriesSize;i++)
			if(callEntries[i].callback != NULL)
				break;
		if(i >= callEntriesSize){	/* nothing to be called immediately */
			now = elpNow();
			minTime = now + POLL_TIME;
			for(i=0;i<timedEntriesSize;i++){
				if(timedEntries[i].callback != NULL && timedEntries[i].when - minTime < 0){
					minTime = timedEntries[i].when;
				}
			}
			minTime -= now - 10;	/* minTime is now time to wait */
			if(minTime < 0)
				minTime = 1;
			tv.tv_sec = minTime / 1000;
			tv.tv_usec = (minTime % 1000) * 1000;
		}
		if(i >= callEntriesSize || ++runThrough > 5){	/* we will actually sleep */
			runThrough = 0;
			for(maxFd=i=0;i<ELP_FDMODES;i++){
				memcpy(&localFdSets[i], &fdSets[i], sizeof(fdSets[i]));
				if(fdEntriesSize[i] > maxFd)
					maxFd = fdEntriesSize[i];
			}
			if(childStatusChanged)
				processChildren();
			elpInvalidateTimeCaches();
			nFd = select(maxFd, &localFdSets[ELP_FDMODE_READ], &localFdSets[ELP_FDMODE_WRITE], &localFdSets[ELP_FDMODE_EXCEPT], &tv);
		}else{
			nFd = 0;	/* no select done and therefore no FDs valid */
		}
        selectErrno = errno;
        if(childStatusChanged)
            processChildren();
		if(nFd < 0){
			if(selectErrno == EINTR)
				continue;
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "elp: select(): [%d] %s\n", selectErrno, sysStrerror(selectErrno));
			continue;
		}
		EXC_TRY		/* put an exception handler around execution part */
			if(nFd != 0){			/* file operation possible */
                elpFile_t	f;
				for(i=0;i<ELP_FDMODES;i++){
					for(j=0;j<fdEntriesSize[i];j++){
						if(fdEntries[i][j].callback != NULL && FD_ISSET(j, &localFdSets[i])){
                            f.mode = i;
                            f.fd = j;
							fdEntries[i][j].callback(fdEntries[i][j].ref, &f);
							if(--nFd <= 0)
								break;
						}
					}
				}
			}
			for(i=0;i<callEntriesSize;i++){
				if(callEntries[i].callback != NULL){
					cb = callEntries[i].callback;
					ref = callEntries[i].ref;
					callEntries[i].callback = NULL;
					cb(ref, NULL);
				}
			}
			now = elpNow();
			for(i=0;i<timedEntriesSize;i++){
				if(timedEntries[i].callback != NULL && timedEntries[i].when - now <= 0){
					cb = timedEntries[i].callback;
					ref = timedEntries[i].ref;
					timedEntries[i].callback = NULL;
					cb(ref, NULL);
				}
			}
		EXC_CATCH
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "elp: caught exception: %s (%s)\n", EXC_NAME, EXC_MSG);
		EXC_ENDTRY
	}
}

/* ------------------------------------------------------------------------- */

