/*
 * Name: notification.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-13
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: notification.c,v 1.2 1999/04/13 15:41:42 cs Exp $
 */

#include "stdheaders.h"
#include "notification.h"
#include "memalloc.h"
#include "exception.h"
#include "logging.h"

/* ------------------------------------------------------------------------- */

typedef struct notifyAction{
	btcallback	action;
	void		*argument;
}notifyAction_t;

/* ------------------------------------------------------------------------- */

static inline void	ncFree(notificationCenter_t *nc)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "notificationcenter: freeing object at 0x%08x\n", (int)nc);
	if(nc != NULL){
		if(nc->table != NULL)
			memFree(nc->table);
		memset(nc, 0, sizeof(*nc));
		memFree(nc);
	}
}

/* ------------------------------------------------------------------------- */

notificationCenter_t	*ncNew(btu32 initialCapacity)
{
notificationCenter_t	*volatile nc = NULL;

EXC_TRY;
	nc = memCalloc(sizeof(*nc));
	nc->tableSize = initialCapacity;
	nc->table = memCalloc(nc->tableSize * sizeof(notifyAction_t));
	nc->refCount = 1;
EXC_CATCH;
	ncFree(nc);
	EXC_RETHROW();
EXC_ENDTRY;
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "notificationcenter: allocated object at 0x%08x\n", (int)nc);
	return nc;
}

/* ------------------------------------------------------------------------- */

void	ncRetain(notificationCenter_t *nc)
{
	nc->refCount++;
}

void	ncRelease(notificationCenter_t *nc)
{
	if(--nc->refCount == 0){
		ncFree(nc);
	}else if(nc->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

void	ncAddObserver(notificationCenter_t *nc, btcallback action, void *arg)
{
btint	i, remaining, firstFree = -1;
btu32	newSize;

	remaining = nc->objectCount;
	for(i=0;i<nc->tableSize && remaining>0;i++){
		if(nc->table[i].action != NULL){
			remaining--;
			if(nc->table[i].action == action && nc->table[i].argument == arg){
				return;	/* nothing to be done, already registered */
			}
		}else if(firstFree < 0){
			firstFree = i;
		}
	}
	if(firstFree < 0){
		firstFree = i;
	}
	if(firstFree >= nc->tableSize){	/* enlarge table */
		newSize = nc->tableSize * 2;
		nc->table = memRealloc(nc->table, newSize * sizeof(notifyAction_t));
		memset(&nc->table[nc->tableSize], 0, nc->tableSize * sizeof(notifyAction_t));
		nc->tableSize = newSize;
        if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
            logPrintf(LOGM_RESOURCE, "notificationcenter: increasing to size %d\n", newSize);
	}
	nc->table[firstFree].action = action;
	nc->table[firstFree].argument = arg;
	nc->objectCount++;
}

/* ------------------------------------------------------------------------- */

void	ncRemoveObserver(notificationCenter_t *nc, btcallback action, void *arg)
{
btint	i, remaining;

	remaining = nc->objectCount;
	for(i=0;i<nc->tableSize && remaining>0;i++){
		if(nc->table[i].action != NULL){
			remaining--;
			if(nc->table[i].action == action && nc->table[i].argument == arg){
				nc->table[i].action = NULL;
				nc->objectCount--;
				break;
			}
		}
	}
}

/* ------------------------------------------------------------------------- */

void	ncPostNotification(notificationCenter_t *nc, void *data)
{
btint	i, remaining;

	remaining = nc->objectCount;
	for(i=0;i<nc->tableSize && remaining>0;i++){
		if(nc->table[i].action != NULL){
			remaining--;
			nc->table[i].action(nc->table[i].argument, data);
		}
	}
}

/* ------------------------------------------------------------------------- */
