/*
 * Name: process.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-11-11
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: process.h,v 1.3 2001/04/26 14:14:40 cs Exp $
 */

#ifndef __process_H_INCLUDED__
#define __process_H_INCLUDED__

#include "basictypes.h"
#include "fifo.h"

/*
General Description:
This header implements utilities to build a kind of multi-threading based
on an event-driven application model with non-blocking system calls. Each
"process" needs a context (which must contain a variable of type
procCtx_t as it's first element). The thread-switch is implemented by
setting up a continuation context in the context variable and then returning
from the function. The continuation immediately after the return is done
by the labels of a switch statement. This may be a quality-test for your
C-compiler (we support only gcc, anyway). Look at the definition of the
PROC_*() macros for details.
The second part of this header implements a semaphore class which is needed
to switch between quasi-"processes".
*/

/* ------------------------------------------------------------------------- */

typedef struct procCtx{
	fifoObject_t	fifoPtr;		/* we are subclass of fifoObject_t */
	btcallback		process;		/* function implementing the process */
	int				currentState;	/* where to continue */
	struct procCtx	*caller;		/* parent structure */
	void			(*free)(void *);/* cleanup of current process */
}procCtx_t;

typedef struct procCond{
    fifo_t		fifo;
}procCond_t;

/* ------------------------------------------------------------------------- */

#ifdef DEBUG_LOCKING
#	include "hashtable.h"
	extern hashtable_t	*procWaitingConds;
#endif

static inline procCond_t	*procCondNew(void)
{
	return (procCond_t *)fifoNew();
}

static inline void	procCondFree(procCond_t *this)
{
	if(!fifoIsEmpty((fifo_t *)this)){
		DEBUG_TRAP();
	}
	fifoFree((fifo_t *)this);
}

static inline void	procCondWait(procCond_t *this, void *proc)
{
    fifoAddObject(&this->fifo, proc);
#ifdef DEBUG_LOCKING
	if(procWaitingConds == NULL)
		procWaitingConds = htabNew(NULL, 16, 75);
	htabAddObject(procWaitingConds, proc);
#endif
}

static inline void	procCondBroadcast(procCond_t *this, void *val)
{
procCtx_t	*p, *q;

    p = (procCtx_t *)this->fifo.head;
    this->fifo.head = NULL;	/* clear the list */
    this->fifo.tail = &this->fifo.head;
    while(p != NULL){	/* walk through the entire list */
        q = (void *)p->fifoPtr.next;
        p->fifoPtr.next = NULL;
#ifdef DEBUG_LOCKING
		htabRemoveObjectEqualTo(procWaitingConds, p);
#endif
        p->process(p, val);
        p = q;
    }
}

static inline void	procCondSignal(procCond_t *this, void *val)
{
procCtx_t	*p;

    p = fifoGetObject(&this->fifo);
    if(p != NULL){
#ifdef DEBUG_LOCKING
		htabRemoveObjectEqualTo(procWaitingConds, p);
#endif
		p->process(p, val);
	}
}

static inline btbool	procCondWaiting(procCond_t *this)
{
    return !fifoIsEmpty(&this->fifo);
}

static inline void	procCondRemoveWait(procCond_t *cond, void *caller)
{
	fifoRemoveObject(&cond->fifo, caller);
}

/* ------------------------------------------------------------------------- */

#define	PROC_VARSET		procCtx_t	proc

#define	PROC_INIT(ctxvar, selffn, _caller, pool, freefn)	\
	(ctxvar) = poolGetBlock(pool);			\
	(ctxvar)->proc.process = (selffn);		\
	(ctxvar)->proc.caller = (_caller);		\
	(ctxvar)->proc.free = (freefn)

#define	PROC_START(ctxvar, arg)	\
	(ctxvar)->proc.process((ctxvar), arg)

#define	PROC_BEGIN	\
	switch(this->proc.currentState){	\
        default:

#define	PROC_WAITFOR_X(x, block)	\
    this->proc.currentState = (x);	\
    block;							\
    return;							\
    case (x):;

#define	PROC_WAITFOR(block)			PROC_WAITFOR_X(__LINE__, block)

#define	PROC_RETURN_AND_CLEANUP(val, block)	\
    if(this->proc.caller != NULL)								\
    	this->proc.caller->process(this->proc.caller, (val));	\
    block;														\
    if(this->proc.free != NULL)									\
        this->proc.free(this);									\
    return

#define	PROC_RETURN(val)			PROC_RETURN_AND_CLEANUP(val, ;)

#define	PROC_END	\
    }				\
    PROC_RETURN(NULL);	/* return part should never be reached */

#define	PROC_IMMEDIATE_RET(caller, val)	\
	if(caller != NULL)										\
		((procCtx_t *)(caller))->process(caller, (val));	\
	return

#define	PROC_IMMEDIATE_RET_AND_CLEANUP(caller, val, block)	\
	if(caller != NULL)										\
		((procCtx_t *)(caller))->process(caller, (val));	\
	block;													\
	return

/* ------------------------------------------------------------------------- */

#endif	/* __process_H_INCLUDED__ */
