/*
 * Name: unicode.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-22
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: unicode.h,v 1.4 2001/05/31 09:38:09 cs Exp $
 */

#ifndef __unicode_H_INCLUDED__
#define __unicode_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
General toolset for working with unicode strings: conversion from/to ASCII,
basic string functions. It also contains some operations on ASCII strings.
*/

#define	UNI_NUM_ENCODINGS	2	/* 2 encodings supported: one for server, one for client [must be <= 8] */
#define	UNI_ENCODING_CLIENT	0	/* index used for client encoding */
#define	UNI_ENCODING_SERVER	1	/* index used for server encoding */

typedef struct uniProperty{
	struct{
		/* some compilers can't do signed bitfields -> extra sign bit */
		unsigned	upcaseDiff			:11;
		unsigned	upcaseDiffNegative	:1;
		unsigned	lowcaseDiff	:11;
		unsigned	lowcaseDiffNegative	:1;
		unsigned	flags				:8;		/* 1<<encoding == 1 if ptr */
	}		info;
	union{
		btu8	*ptr;
		btu8	ascii[sizeof(btu8 *)];
	}		encodingRep[UNI_NUM_ENCODINGS];
}uniProperty_t;

/* ------------------------------------------------------------------------- */
/* --------------------------- unicode constants --------------------------- */
/* ------------------------------------------------------------------------- */

#define	UNI_CHAR_BACKSLASH		((btuni)'\\')
#define	UNI_CHAR_SLASH			((btuni)'/')
#define	UNI_CHAR_DOT			((btuni)'.')
#define	UNI_CHAR_QUESTION		((btuni)'?')
#define	UNI_CHAR_ASTERISK		((btuni)'*')

/* ------------------------------------------------------------------------- */
/* --------------------------- unicode parsing ----------------------------- */
/* ------------------------------------------------------------------------- */

btbool	uniParseMappingFile(btu8 *fileName, btint encoding, uniProperty_t *uniTable, btuni *uniErrorChar, btu8 **asciiErrorChar);
btbool	uniParseCaseConvertFile(btu8 *fileName, uniProperty_t *uniTable);

/* ------------------------------------------------------------------------- */
/* -------------------------- unicode operations --------------------------- */
/* ------------------------------------------------------------------------- */

btuni	*uniStrDup(btuni *s);
void	uniStrAddHash(btu32 *crcp, btuni *s, btbool isCaseSensitive);
btuni	uniCharToupper(btuni c);
btuni	uniCharTolower(btuni c);
btu8	asciiCharToupper(btu8 c);
btu8	asciiCharTolower(btu8 c);
void	uniStrCpy(btuni *dest, btuni *src, btsize arrSize);
void	asciiStrCpy(btu8 *dest, btu8 *src, btsize arrSize);

static inline btsize	uniStrLen(btuni *s)
{
btsize	l = 0;

	while(s[l])	l++;
	return l;
}

static inline btu32	uniStrHash(btuni *s, btbool isCaseSensitive)
{
btu32	crc = -1;

	uniStrAddHash(&crc, s, isCaseSensitive);
	return crc;
}

static inline int	uniStrCmp(btuni *s1, btuni *s2, btbool caseSensitive)
{
btuni	c1, c2;

	if(caseSensitive){
		while(*s1 != 0 && *s2 != 0){
			if(*s1 < *s2)	return -1;
			if(*s1 > *s2)	return 1;
			s1++; s2++;
		}
	}else{
		while(*s1 != 0 && *s2 != 0){
			c1 = uniCharToupper(*s1);
			c2 = uniCharToupper(*s2);
			if(c1 < c2)
				return -1;
			if(c1 > c2)
				return 1;
			s1++; s2++;
		}
	}
	if(*s2 != 0)
		return -1;
	if(*s1 != 0)
		return 1;
	return 0;
}

static inline void	uniStrToupper(btuni *s)
{
	while(*s != 0){
		*s = uniCharToupper(*s);
		s++;
	}
}

static inline void	uniStrTolower(btuni *s)
{
	while(*s != 0){
		*s = uniCharTolower(*s);
		s++;
	}
}

/* ------------------------------------------------------------------------- */
/* -------------------------- unicode conversions -------------------------- */
/* ------------------------------------------------------------------------- */

/* initialisation: */

void	uniInit(btu8 *caseConvertPath, btu8 *serverMapPath, btu8 *clientMapPath, btsize maxBlocks);
void	uniTrivialInit(void);

/* basic conversion operations: */

btuni	convCharAsciiToUni(btu8 **string, btint encoding);
btbool	convCharUniToAscii(btuni c, btu8 **string, btint encoding);

#define	convCharLocalAsciiToUni(s)		convCharAsciiToUni(s, UNI_ENCODING_CLIENT)
#define	convCharServerAsciiToUni(s)		convCharAsciiToUni(s, UNI_ENCODING_SERVER)
#define	convCharUniToLocalAscii(c, s)	convCharUniToAscii(c, s, UNI_ENCODING_CLIENT)
#define	convCharUniToServerAscii(c, s)	convCharUniToAscii(c, s, UNI_ENCODING_SERVER)

/* derived conversion operations: */
void	convAsciiToUni(btu8 *ascii, btuni *uni, btsize arrSize, btint encoding);
void	convUniToAscii(btuni *uni, btu8 *ascii, btsize arrSize, btint encoding);

#define	convLocalAsciiToUni(s, u, l)	convAsciiToUni(s, u, l, UNI_ENCODING_CLIENT)
#define	convServerAsciiToUni(s, u, l)	convAsciiToUni(s, u, l, UNI_ENCODING_SERVER)
#define	convUniToLocalAscii(u, s, l)	convUniToAscii(u, s, l, UNI_ENCODING_CLIENT)
#define	convUniToServerAscii(u, s, l)	convUniToAscii(u, s, l, UNI_ENCODING_SERVER)

/* intended for debug outputs: */
btu8	*uniToCstring(btuni *uni);

/* ------------------------------------------------------------------------- */
/* ----------------------- unicode wildcard matching ----------------------- */
/* ------------------------------------------------------------------------- */

btbool	uniWildcardMatch(btuni *text, btuni *pattern);
/* returns 1 if pattern matches text, 0 otherwise */

/* ------------------------------------------------------------------------- */

#endif	/* __unicode_H_INCLUDED__ */
