/*
 * Name: uniparse.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-03-09
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: uniparse.c,v 1.4 2001/05/31 09:38:09 cs Exp $
 */

/*
General Description:
This module contains the parser that is used to parse the unicode case
mapping file and the server and client code mapping files. 
*/

#include "stdheaders.h"
#include "logging.h"
#include "unicode.h"
#include "scanner.h"
#include "memalloc.h"
#include "system.h"
#include "util.h"

EXC_DEFINE(excUniParseError);

static btu8	*currentParserFile;

/* ------------------------------------------------------------------------- */

static btbool	uniParseKeywd(scanner_t *stream, btu8 *keywd)
{
int		token;
btu8	*string;

	token = scannerGetToken(stream, &string);
	if(token == SCAN_TOKEN_ALNUM){
		if(utilStrCmp(string, keywd, 0) == 0){
			return 1;
		}
	}
	scannerUngetToken(stream, token);
	return 0;
}

static btbool	uniParseOperator(scanner_t *stream, btu8 op)
{
int		token;
btu8	*value;

	if((token = scannerGetToken(stream, &value)) == SCAN_TOKEN_OPERATOR){
		if(value[0] == op && value[1] == 0)
			return 1;
	}
	scannerUngetToken(stream, token);
	return 0;
}

static btbool	uniParseOptionalOperator(scanner_t *stream, btu8 op)
{
int		token;
btu8	*value;

	if((token = scannerGetToken(stream, &value)) == SCAN_TOKEN_OPERATOR){
		if(value[0] == op && value[1] == 0)
			return 1;
	}
	scannerUngetToken(stream, token);
	return 1;
}

static btbool	uniParseUnival(scanner_t *stream, btuni *val)
{
int		token;
btu8	*stringval;

	token = scannerGetToken(stream, &stringval);
	if(token == SCAN_TOKEN_ALNUM){
		if(isdigit(stringval[0])){
			*val = strtol(stringval, (char **)&stringval, 0);
			return 1;
		}
	}
	scannerUngetToken(stream, token);
	return 0;
}

static btbool	uniParseEOF(scanner_t *stream)
{
int		token;
btu8	*value;

	if((token = scannerGetToken(stream, &value)) == SCAN_TOKEN_EOF){
		return 1;
	}
	scannerUngetToken(stream, token);
	return 0;
}

/* ------------------------------------------------------------------------- */

static btbool	uniParseCaseLine(scanner_t *stream, btuni *baseChar, btuni *upperChar, btuni *lowerChar)
{
	if(!uniParseUnival(stream, baseChar))
		return 0;
	if(!uniParseOperator(stream, '=')){
		EXC_THROW_MSG(excUniParseError, "\'=\' expected");
	}
	if(!uniParseOperator(stream, '(')){
		EXC_THROW_MSG(excUniParseError, "\'(\' expected");
	}
	if(!uniParseUnival(stream, upperChar)){
		EXC_THROW_MSG(excUniParseError, "char code for upper case version expected");
	}
	uniParseOptionalOperator(stream, ',');
	if(!uniParseUnival(stream, lowerChar)){
		EXC_THROW_MSG(excUniParseError, "char code for lower case version expected");
	}
	if(!uniParseOperator(stream, ')')){
		EXC_THROW_MSG(excUniParseError, "\')\' expected");
	}
	uniParseOptionalOperator(stream, ';');
	return 1;
}

static btbool	uniParseInt8List(scanner_t *stream, btu8 *mapping, btsize *mappingLen, btsize maxLen)
{
btuni	val;

	*mappingLen = 0;
	while(uniParseUnival(stream, &val)){
		if(maxLen > 0){
			*mapping++ = val;
			(*mappingLen)++;
			maxLen--;
		}else{
			if(LOG_MODE_IS_ACTIVE(LOGM_WARN))
				logPrintf(LOGM_WARN, "uniparse: excess mapping char ignored in line %d of file %s\n", scannerGetLineNumber(stream), currentParserFile);
		}
		uniParseOptionalOperator(stream, ',');
	}
	return 1;
}

static btbool	uniParseMapLine(scanner_t *stream, btuni *baseChar, btu8 *mapping, btsize *mappingLen, btint maxMappingLen)
{
btuni	mapVal;

	if(!uniParseUnival(stream, baseChar))
		return 0;
	if(!uniParseOperator(stream, '=')){
		EXC_THROW_MSG(excUniParseError, "\'=\' expected");
	}
	if(uniParseUnival(stream, &mapVal)){	/* single mapping value given */
		mapping[0] = mapVal;
		*mappingLen = 1;
		uniParseOptionalOperator(stream, ';');
	}else if(uniParseOperator(stream, '(')){	/* mapping sequence given */
		uniParseInt8List(stream, mapping, mappingLen, maxMappingLen);
		if(!uniParseOperator(stream, ')')){
			EXC_THROW_MSG(excUniParseError, "\')\' expected");
		}
		uniParseOptionalOperator(stream, ';');
	}else{
		EXC_THROW_MSG(excUniParseError, "integer character value or list of values in parentheses expected");
	}
	return 1;
}

static btbool	uniParseUniErrorChar(scanner_t *stream, btuni *errChar)
{
	if(!uniParseKeywd(stream, "uniErrorChar")){
		return 0;
	}
	uniParseOptionalOperator(stream, '=');
	if(!uniParseUnival(stream, errChar)){
		EXC_THROW_MSG(excUniParseError, "integer char code expected");
	}
	uniParseOptionalOperator(stream, ';');
	return 1;
}

static btbool	uniParseAsciiErrorChar(scanner_t *stream, btu8 *mapping, btsize *mappingLen, btsize maxLen)
{
btuni	val;

	if(!uniParseKeywd(stream, "asciiErrorChar")){
		return 0;
	}
	uniParseOptionalOperator(stream, '=');
	if(uniParseOperator(stream, '(')){
		uniParseInt8List(stream, mapping, mappingLen, maxLen);
		if(!uniParseOperator(stream, ')')){
			EXC_THROW_MSG(excUniParseError, "\')\' expected");
		}
		uniParseOptionalOperator(stream, ';');
	}else if(uniParseUnival(stream, &val)){
		mapping[0] = val;
		*mappingLen = 1;
		uniParseOptionalOperator(stream, ';');
	}else{
		EXC_THROW_MSG(excUniParseError, "integer character value or list of values in parentheses expected");
	}
	return 1;
}

/* ------------------------------------------------------------------------- */

static int	uniGetChar(FILE *fp)	/* need a function pointer, getc is a macro */
{
	return getc(fp);
}


btbool	uniParseCaseConvertFile(btu8 *volatile fileName, uniProperty_t *volatile uniTable)
{
FILE			*volatile fp = NULL;
scanner_t		*volatile stream = NULL;
volatile btbool	rval = 1;
btuni			baseChar, upperChar, lowerChar;
btint			d1, d2;

	currentParserFile = fileName;
	fp = fopen(fileName, "r");
	if(fp == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "error opening case conversion file %s: [%d] %s\n", fileName, errno, sysStrerror(errno));
		return 0;
	}
EXC_TRY
	stream = scannerNew(uniGetChar, fp);
	scannerSetAllowRem(stream, 1, 1, 1);	/* allow all kinds of remarks */
	while(uniParseCaseLine(stream, &baseChar, &upperChar, &lowerChar)){
		d1 = upperChar - baseChar;
		uniTable[baseChar].info.upcaseDiff = abs(d1);
		uniTable[baseChar].info.upcaseDiffNegative = d1 < 0;
		d2 = lowerChar - baseChar;
		uniTable[baseChar].info.lowcaseDiff = abs(d2);
		uniTable[baseChar].info.lowcaseDiffNegative = d2 < 0;
		if(abs(d1) > 0x7ff && LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "parsing file %s: upper case difference for char 0x%04x > 0x7ff\n", fileName, baseChar);
		}
		if(abs(d2) > 0x7ff && LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "parsing file %s: lower case difference for char 0x%04x > 0x7ff\n", fileName, baseChar);
		}
	}
	if(!uniParseEOF(stream)){
		EXC_THROW_MSG(excUniParseError, "End of file expected");
	}
EXC_CATCH
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
		if(EXC_NAME == excUniParseError)
			logPrintf(LOGM_ERROR, "parse error: %s in line %d of file %s\n", EXC_MSG, scannerGetLineNumber(stream), fileName);
		else
			logPrintf(LOGM_ERROR, "exception %s (%s) while parsing %s\n", EXC_NAME, EXC_MSG, currentParserFile);
	}
	rval = 0;
EXC_ENDTRY
	if(stream != NULL)
		scannerFree(stream);
	if(fp != NULL)
		fclose(fp);
	return rval;
}


btbool	uniParseMappingFile(btu8 *volatile fileName, btint encoding, uniProperty_t *volatile uniTable, btuni *uniErrorChar, btu8 **asciiErrorChar)
{
FILE			*volatile fp = NULL;
scanner_t		*volatile stream = NULL;
volatile btbool	rval = 1;
btuni			baseChar;
btu8			mapping[512];
btsize			mappingLen;

	currentParserFile = fileName;
	fp = fopen(fileName, "r");
	if(fp == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "error opening mapping file %s: [%d] %s\n", fileName, errno, sysStrerror(errno));
		return 0;
	}
EXC_TRY
	stream = scannerNew(uniGetChar, fp);
	scannerSetAllowRem(stream, 1, 1, 1);	/* allow all kinds of remarks */
	for(;;){
		if(uniParseMapLine(stream, &baseChar, mapping, &mappingLen, sizeof(mapping) - 1)){
			if(mappingLen <= sizeof(btu8 *)){
				uniTable[baseChar].encodingRep[encoding].ptr = NULL;
				memcpy(uniTable[baseChar].encodingRep[encoding].ascii, mapping, mappingLen);
			}else{
				mapping[mappingLen] = 0;	/* terminating 0 */
				uniTable[baseChar].info.flags |= 1 << encoding;
				uniTable[baseChar].encodingRep[encoding].ptr = utilStrDup(mapping);
			}
		}else if(uniParseUniErrorChar(stream, &baseChar)){
			*uniErrorChar = baseChar;
		}else if(uniParseAsciiErrorChar(stream, mapping, &mappingLen, sizeof(mapping) - 1)){
			mapping[mappingLen] = 0;	/* terminating 0 */
			*asciiErrorChar = utilStrDup(mapping);
		}else{
			break;
		}
	}
	if(!uniParseEOF(stream)){
		EXC_THROW_MSG(excUniParseError, "End of file expected");
	}
EXC_CATCH
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
		if(EXC_NAME == excUniParseError)
			logPrintf(LOGM_ERROR, "parse error: %s in line %d of file %s\n", EXC_MSG, scannerGetLineNumber(stream), fileName);
		else
			logPrintf(LOGM_ERROR, "exception %s (%s) while parsing %s\n", EXC_NAME, EXC_MSG, currentParserFile);
	}
	rval = 0;
EXC_ENDTRY
	if(stream != NULL)
		scannerFree(stream);
	if(fp != NULL)
		fclose(fp);
	return rval;
}

/* ------------------------------------------------------------------------- */

