/*
 * Name: childtable.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-30
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: childtable.c,v 1.2 1999/09/10 23:03:16 cs Exp $
 */


#include "stdheaders.h"
#include "childtable.h"
#include "blockalloc.h"
#include "exception.h"

/* ------------------------------------------------------------------------- */

#define	CHILDREN_PER_PAGE				(BLOCKALLOC_PAGESIZE / sizeof(btu32))
#define	INITIAL_CHILDREN_PER_PAGE		3

typedef struct childList{
	unsigned	isIndirect	: 1;
	unsigned	numEntries	:31;
	btu32		children[1];	/* actually more than 1 */
}childList_t;

/* ------------------------------------------------------------------------- */

btu32	ctabObjectCount(blockAllocator_t *ba, btu32 table)
{
childList_t	*list;

	if(table == 0)
		return 0;
	list = ballocBlockAddress(ba, table);
	return list->numEntries;
}

/* ------------------------------------------------------------------------- */

btu32	ctabObjectAtIndex(blockAllocator_t *ba, btu32 table, btu32 index)
{
childList_t	*list = ballocBlockAddress(ba, table);
btu32		*page;

	if(index >= list->numEntries)
		return 0;
	if(!list->isIndirect){
		return list->children[index];
	}
	page = ballocBlockAddress(ba, list->children[index/CHILDREN_PER_PAGE]);
	return page[index % CHILDREN_PER_PAGE];
}

/* ------------------------------------------------------------------------- */

static inline void	indirectMove(blockAllocator_t *ba, btu32 *pageList, btu32 startIndex, btu32 endIndex, btu32 insObject, btbool insert)
{
btint	startPage, endPage, page;
btint	pStartIndex, pEndIndex, size;
btu32	h, *hp = NULL, *data;

	startPage = startIndex / CHILDREN_PER_PAGE;
	endPage = (endIndex - 1) / CHILDREN_PER_PAGE + 1;
	for(page=startPage;page<endPage;page++){
		pStartIndex = startIndex - page * CHILDREN_PER_PAGE;
		if(pStartIndex < 0)
			pStartIndex = 0;
		pEndIndex = endIndex - page * CHILDREN_PER_PAGE;
		if(pEndIndex > CHILDREN_PER_PAGE)
			pEndIndex = CHILDREN_PER_PAGE;
		data = ballocBlockAddress(ba, pageList[page]);
		size = (pEndIndex - pStartIndex - 1) * sizeof(btu32);
		if(insert){
			h = data[pEndIndex - 1];
			memmove(&data[pStartIndex + 1], &data[pStartIndex], size);
			data[pStartIndex] = insObject;
			insObject = h;
		}else{
			if(hp)
				*hp = data[pStartIndex];
			memmove(&data[pStartIndex], &data[pStartIndex + 1], size);
			hp = &data[pEndIndex - 1];
		}
	}
}

/* ------------------------------------------------------------------------- */

void	ctabInsertObjectAtIndex(blockAllocator_t *ba, btu32 *tablep, btu32 index, btu32 object)
{
childList_t	*newList, *list;
btu32		size, newTable, *page, table = *tablep, newEntryCount;
btsize		dummy;

	if(object == 0){	/* don't allow NULL objects */
		EXC_THROW_MSG(excInternalError, "childtable.c: inserting null object");
	}
	if(table == 0){
		table = ballocAllocBlock(ba, (INITIAL_CHILDREN_PER_PAGE + 1) * sizeof(btu32), &dummy);	/* may trigger an exception! */
		list = ballocBlockAddress(ba, table);
		list->isIndirect = 0;
		list->numEntries = 0;
	}else
		list = ballocBlockAddress(ba, table);
	if(index > list->numEntries){	/* illegal insertion index */
		EXC_THROW_MSG(excInternalError, "childtable.c: inserting at illegal index");
	}
/* don't increment the numEntries entry directly, we must not increment if
 * an exception occurs during the following code. */
	newEntryCount = list->numEntries + 1;
	if(!list->isIndirect){
		size = ballocBlockSize(ba, table);
		if(newEntryCount > size/sizeof(btu32) - 1){	/* we must enlarge the list */
			if(size < BLOCKALLOC_PAGESIZE){	/* possible with direct list */
				newTable = ballocAllocBlock(ba, size * 2, &dummy);	/* may trigger an exception! */
				newList = ballocBlockAddress(ba, newTable);
				memcpy(newList, list, size);
				ballocFreeBlock(ba, table);
				table = newTable;
				list = newList;
			}else{	/* we must convert to indirect list */
				newTable = ballocAllocBlock(ba, BLOCKALLOC_PAGESIZE, &dummy);	/* may trigger an exception! */
				page = ballocBlockAddress(ba, newTable);
				memcpy(page, list->children, size - sizeof(btu32));
				list->isIndirect = 1;
				list->children[0] = newTable;
			}
		}
	}else{	/* is an indirect table */
		if(newEntryCount % CHILDREN_PER_PAGE == 1){	/* must start new page */
			newTable = ballocAllocBlock(ba, BLOCKALLOC_PAGESIZE, &dummy);	/* may trigger an exception! */
			list->children[newEntryCount / CHILDREN_PER_PAGE] = newTable;
		}
	}
	if(list->isIndirect){
		indirectMove(ba, list->children, index, newEntryCount, object, 1);
	}else{	/* straight table */
		memmove(&list->children[index + 1], &list->children[index], (list->numEntries - index) * sizeof(btu32));
		list->children[index] = object;
	}
	*tablep = table;	/* we may have relocated the table */
	list->numEntries = newEntryCount;
}

/* ------------------------------------------------------------------------- */

void	ctabRemoveObjectAtIndex(blockAllocator_t *ba, btu32 *tablep, btu32 index)
{
btu32		page, newTable, currentSize, table = *tablep;
childList_t	*newList, *list = ballocBlockAddress(ba, table);
btsize		dummy;

	if(list->isIndirect){
		indirectMove(ba, list->children, index, list->numEntries, 0, 0);
	}else{
		memmove(&list->children[index], &list->children[index + 1], (list->numEntries - 1 - index) * sizeof(btu32));
	}
	list->numEntries--;
	if(list->isIndirect){
		if(list->numEntries % CHILDREN_PER_PAGE == 0){	/* we have just lost a page */
			ballocFreeBlock(ba, list->children[list->numEntries / CHILDREN_PER_PAGE]);
		}else if(list->numEntries < CHILDREN_PER_PAGE){	/* would fit into the base page */
			page = list->children[0];
			memcpy(list->children, ballocBlockAddress(ba, page), list->numEntries * sizeof(btu32));
			list->isIndirect = 0;
			ballocFreeBlock(ba, page);
		}
	}else{
		if(list->numEntries == 0){	/* don't need any storage... */
			list->numEntries = 1;	/* entry count of 0 is illegal */
			ballocFreeBlock(ba, table);
			table = 0;
		}else{
			currentSize = ballocBlockSize(ba, table)/sizeof(btu32);
			if(list->numEntries < currentSize/3 && currentSize > INITIAL_CHILDREN_PER_PAGE){
				newTable = ballocAllocBlock(ba, currentSize/2 * sizeof(btu32), &dummy);
				newList = ballocBlockAddress(ba, newTable);
				memcpy(newList, list, (list->numEntries + 1) * sizeof(btu32));
				ballocFreeBlock(ba, table);
				table = newTable;
			}
		}
	}
	*tablep = table;
}

/* ------------------------------------------------------------------------- */

void	ctabRemoveAllObjects(blockAllocator_t *ba, btu32 *tablep)
{
childList_t	*list;
btu32		pages;
btint		i;
	
	if(*tablep != 0){
		list = ballocBlockAddress(ba, *tablep);
		if(list != NULL){
			if(list->isIndirect){
				pages = (list->numEntries - 1) / BLOCKALLOC_PAGESIZE + 1;
				for(i=0;i<pages;i++){
					ballocFreeBlock(ba, list->children[i]);
				}
			}
			ballocFreeBlock(ba, *tablep);
		}
		*tablep = 0;
	}
}


/* ------------------------------------------------------------------------- */
/* ------------------------- builtin test routines ------------------------- */
/* ------------------------------------------------------------------------- */

#ifdef TEST_CHILDTABLE

#include <math.h>
#include "test.h"

#define	MAX_ENTRIES		5000
#define	TABLES			4
#define	ALLOC_MEMORY	500000

typedef struct childTableSim{
	btint	numEntries;
	btu32	entries[MAX_ENTRIES];
}childTableSim_t;

childTableSim_t		sim[TABLES][1];
btu32				tables[TABLES];
blockAllocator_t	*ba;

inline btu32	simObjectCount(childTableSim_t *cs)
{
	return cs->numEntries;
}

inline btu32	simObjectAtIndex(childTableSim_t *cs, btu32 index)
{
	return cs->entries[index];
}

inline void	simInsertObjectAtIndex(childTableSim_t *cs, btu32 index, btu32 object)
{
	memmove(&cs->entries[index + 1], &cs->entries[index], (cs->numEntries - index) * sizeof(btu32));
	cs->entries[index] = object;
	cs->numEntries++;
}

inline void	simRemoveObjectAtIndex(childTableSim_t *cs, btu32 index)
{
	cs->numEntries--;
	memmove(&cs->entries[index], &cs->entries[index + 1], (cs->numEntries - index) * sizeof(btu32));
}

inline void	simRemoveAllObjects(childTableSim_t *cs)
{
	cs->numEntries = 0;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

void	verifyTables(void)
{
btint	table, count, i;
btbool	errorOccured = 0;
btu32	realVal, simVal;

	for(table=0;table<TABLES;table++){
		if(ctabObjectCount(ba, tables[table]) != simObjectCount(sim[table])){
			fprintf(stderr, "object counts differ!\n");
			abort();
		}
		count = simObjectCount(sim[table]);
		for(i=0;i<count;i++){
			realVal = ctabObjectAtIndex(ba, tables[table], i);
			simVal = simObjectAtIndex(sim[table], i);
			if(realVal != simVal){
				fprintf(stderr, "objects differ in table %d index %d of %d (0x%08x instead of 0x%08x)\n", table, i, count, realVal, simVal);
				errorOccured = 1;
			}
		}
	}
	if(errorOccured)
		abort();
}

void	insertObject(btint table)
{
btu32	object = 0;
btu32	index;

	if(ctabObjectCount(ba, tables[table]) != simObjectCount(sim[table])){
		fprintf(stderr, "object counts differ!\n");
		abort();
	}
	while(object == 0)
		object = random_btu32();
	index = random_int(simObjectCount(sim[table]) + 1);
	ctabInsertObjectAtIndex(ba, &tables[table], index, object);
	simInsertObjectAtIndex(sim[table], index, object);
}

void	removeObject(btint table)
{
btu32	index;

	if(ctabObjectCount(ba, tables[table]) != simObjectCount(sim[table])){
		fprintf(stderr, "object counts differ!\n");
		abort();
	}
	index = random_int(simObjectCount(sim[table]));
	ctabRemoveObjectAtIndex(ba, &tables[table], index);
	simRemoveObjectAtIndex(sim[table], index);
}

/* ------------------------------------------------------------------------- */

void	main(void)
{
btu32	blockSizes[] = {16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 0};
btint	table, i, rint;
btbool	doEnlarge = 1;

EXC_TRY
	random_init(12, 34, 56, 78);
	ba = ballocNew(pagePoolNew(ALLOC_MEMORY), blockSizes, 32);
	for(i=0;;i++){
		if(i%10000 == 0){
			printf("run %d\n", i);
		}
		if(random_int(100)){
			table = random_int(TABLES);
			if(simObjectCount(sim[table]) == 0 && !doEnlarge){
				printf("switching to enlarge in run %d\n", i);
				doEnlarge = 1;
			}else if(simObjectCount(sim[table]) >= MAX_ENTRIES && doEnlarge){
				printf("switching to shrinking in run %d\n", i);
				doEnlarge = 0;
			}
			if(doEnlarge)
				rint = random_int(2 * MAX_ENTRIES);
			else
				rint = random_int(2 * MAX_ENTRIES) - MAX_ENTRIES;
			if(rint >= (btint)simObjectCount(sim[table])){
				insertObject(table);
			}else{
				removeObject(table);
			}
		}else{
			verifyTables();
		}
	}
EXC_CATCH
	fprintf(stderr, "*** exception %s (%s)\n", EXC_NAME, EXC_MSG);
	abort();
EXC_ENDTRY
}
/* ------------------------------------------------------------------------- */

#endif /* TEST_CHILDTABLE */

/* ------------------------------------------------------------------------- */

