/*
 * Name: nfslookuptable.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-04
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfslutable.c,v 1.6 2001/10/30 14:56:03 cs Exp $
 */

/* General Description:
This subclass of nfsLookup implements the table-based lookup strategy. This
strategy stores the mapping between NFS file handles and file pathes in an
internal table. This should always work correctly (as far as this is possible
without insider knowledge of the backend's filesystem). The disadvantage is
that this table can become quite big if a lot of mappings must be stored.
Great care has been taken to reduce the memory requirements to a minimum and
to allow for efficient swapping.
*/

#include "stdheaders.h"
#include "nfslookup.h"
#include "blockalloc.h"
#include "hashtable.h"
#include "childtable.h"
#include "exception.h"
#include "process.h"
#include "blockpool.h"
#include "unicode.h"
#include "logging.h"
#include "module.h"
#include "backend.h"
#include "errcodes.h"
#include "nfs_prot.h"

/* ------------------------------------------------------------------------- */

typedef struct nfsfhInode{
	unsigned	useCount	:8;		/* how often this inode has been reused */
	unsigned	parent		:24;	/* index into inode table, 24 bit suffice */
	unsigned	childList	:32;	/* index into childList table */
	unsigned	nameIndex	:31;	/* index into name table */
	unsigned	isUnicode	:1;		/* 1 if name is in Unicode */
}nfsfhInode_t;


typedef struct nfsfhTable{
	nfsLookup_t				super;
	struct pagePool			*pool;			/* memory pool */
	struct blockAllocator	*inodes;		/* inode table */
	struct blockAllocator	*names;			/* table of inode names */
	struct blockAllocator	*childTables;	/* table of child tables */
	btu32					rootInode;		/* index of root inode */
	btu32					freeInodes[128];/* delayed free for inodes */
	btint					freeInodeI;		/* index into freeInodes[] */
	btint					shrIndex;		/* unique index of share */
}nfsfhLookupTable_t;

typedef struct nfsFh{
	btinum	inum;
	btint	shareIndex;
	btu8	useCount;
	btu8	dummy[32 - sizeof(btinum) - sizeof(btint) - sizeof(btu8) - sizeof(btu32)];
	btu32	mountId;
}nfsFh_t;

/* ------------------------------------------------------------------------- */

typedef struct findContext{
    PROC_VARSET;
	nfsfhLookupTable_t	*nt;
	void				*user;
	btu32				inum;
	void				*file;
}findContext_t;

static blockpool_t	*findPool = NULL;

static void	findCtxFree(void *block)
{
    poolReturnBlock(findPool, block);
}

extern nfsLookupInterface_t	nfsfhTableInterface;	/* forward declaration */
static void	ntFileDestroyFromInode(nfsfhLookupTable_t *nt, nfsfhInode_t *parentInode, btuni *name);

/* ------------------------------------------------------------------------- */

static void	nfsfhLookupTableFree(void *obj)
{
nfsfhLookupTable_t	*this = obj;

	if(this != NULL){
		if(this->childTables != NULL){
			ballocFree(this->childTables);
			this->childTables = NULL;
		}
		if(this->names != NULL){
			ballocFree(this->names);
			this->names = NULL;
		}
		if(this->inodes != NULL){
			ballocFree(this->inodes);
			this->inodes = NULL;
		}
		if(this->pool != NULL){
			pagePoolFree(this->pool);
			this->pool = NULL;
		}
		luExit(&this->super);
		memFree(this);
	}
}

/* ------------------------------------------------------------------------- */

nfsLookup_t	*nfsfhLookupTableNew(void *vfs, struct module *backend, struct cfgProperty **cfgPath, btint numEntries)
{
nfsfhLookupTable_t	*volatile this = NULL;
static btu32		inodeSizes[] = {sizeof(nfsfhInode_t), 0};
static btu32		nameSizes[] = {4, 8, 12, 16, 20, 32, 128, 256 * sizeof(btuni), 0};
static btu32		childTableSizes[] = {8, 32, 64, 128, 256, 512, 1024, 2048, 4096, BLOCKALLOC_PAGESIZE, 0};
nfsfhInode_t		*inode;
btsize				dummy;
static btint		shareCounter;

    if(findPool == NULL)
        findPool = poolNew(sizeof(findContext_t), sizeof(findContext_t));
EXC_TRY
	this = memCalloc(sizeof(nfsfhLookupTable_t));
	luInit(&this->super, &nfsfhTableInterface, vfs, backend, cfgPath, numEntries);
	this->pool = pagePoolNew();
	this->inodes = ballocNew(this->pool, inodeSizes, 24);
	this->names = ballocNew(this->pool, nameSizes, 31);
	this->childTables = ballocNew(this->pool, childTableSizes, 32);
	this->rootInode = ballocAllocBlock(this->inodes, sizeof(nfsfhInode_t), &dummy);
	this->shrIndex = shareCounter++;
	inode = ballocBlockAddress(this->inodes, this->rootInode);
	inode->parent = this->rootInode;	/* root is its own parent */
	inode->childList = 0;
	inode->nameIndex = 0;				/* root does not need a name */
EXC_CATCH
	nfsfhLookupTableFree(this);
	EXC_RETHROW();
EXC_ENDTRY
    return (void *)this;
}

/* ------------------------------------------------------------------------- */

static void	ntRootHandle(void *this, void *dest, btinum *inum, btu32 mountId)
{
nfsfhLookupTable_t	*nt = this;
nfsFh_t				*fh = dest;

	if(fh != NULL){
		memset(fh, 0, 32);
		fh->inum = nt->rootInode;
		fh->shareIndex = nt->shrIndex;
		fh->mountId = mountId;
	}
	if(inum != NULL){
		*inum = nt->rootInode;
	}
}

/* ------------------------------------------------------------------------- */

static btint	anyStrCompare(void *anyStr, btbool isUni, btsize anyStrBytes, btuni *refName, btbool caseSensitive)
{
btuni	name[NFS_MAXNAMLEN + 1], *up;
btint	i, l;
btu8	*cp;

	if(isUni){
		l = anyStrBytes/2;
		if(l > NFS_MAXNAMLEN)
			l = NFS_MAXNAMLEN;
		up = anyStr;
		for(i=0;i<l && up[i]!=0;i++){
			name[i] = up[i];
		}
		name[i] = 0;
	}else{
		l = anyStrBytes;
		if(l > NFS_MAXNAMLEN)
			l = NFS_MAXNAMLEN;
		cp = anyStr;
		for(i=0;i<l && cp[i]!=0;i++){
			name[i] = cp[i];
		}
		name[i] = 0;
	}
	return uniStrCmp(name, refName, caseSensitive);
}

/* ------------------------------------------------------------------------- */

static btu32	indexOfChild(nfsfhLookupTable_t *nt, btu32 childTableI, btuni *name, btbool caseSensitive, btbool *found)
{
btu32			lastI, i, minI = 0, toomuchI = ctabObjectCount(nt->childTables, childTableI);
btu32			childInum, childNameIndex;
nfsfhInode_t	*childInode;
int				cmp;

	if(toomuchI == 0){	/* if anything is too much */
		*found = 0;
		return 0;
	}
	lastI = -1;
	for(;;){
		i = (minI + toomuchI)/2;
		if(i == lastI){	/* the same we tried last time */
			*found = 0;
			return toomuchI;	/* insert before toomuchI */
		}
		lastI = i;
		childInum = ctabObjectAtIndex(nt->childTables, childTableI, i);
		childInode = ballocBlockAddress(nt->inodes, childInum);
		childNameIndex = childInode->nameIndex;
		cmp = anyStrCompare(ballocBlockAddress(nt->names, childNameIndex), childInode->isUnicode, ballocBlockSize(nt->names, childNameIndex), name, caseSensitive);
		if(cmp == 0){
			*found = 1;
			return i;
		}
		if(cmp > 0){	/* child > object */
			toomuchI = i;
		}else{			/* child < object */
			minI = i;
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	setInodeName(nfsfhLookupTable_t *nt, nfsfhInode_t *inode, btuni *name)
{
btint	i;
btbool	needUni = 0;
btsize	maxSize, len;
btu8	*p;

	if(inode->nameIndex != 0){
		ballocFreeBlock(nt->names, inode->nameIndex);
		inode->nameIndex = 0;
	}
	for(i=0;name[i] != 0;i++){
		if(name[i] >= 256)
			needUni = 1;
	}
	len = i;
	inode->isUnicode = needUni;
	if(needUni){
		inode->nameIndex = ballocAllocBlock(nt->names, len * 2, &maxSize);
		p = ballocBlockAddress(nt->names, inode->nameIndex);
		if(maxSize >= (len + 1) * 2){
			len += 1;	/* append the terminating 0 if possible */
		}
		memcpy(p, name, 2 * len);
	}else{
		inode->nameIndex = ballocAllocBlock(nt->names, len, &maxSize);
		p = ballocBlockAddress(nt->names, inode->nameIndex);
		for(i=0;i<len;i++){
			p[i] = name[i];
		}
		if(i < maxSize)
			p[i] = 0;	/* append the terminating 0 if possible */
	}
}

static void	getInodeName(nfsfhLookupTable_t *nt, nfsfhInode_t *inode, btuni *name, btsize arrSize)
{
btsize	len, i;
union{
	btuni	*u;
	btu8	*c;
	void	*any;
}		p;

	if((p.any = ballocBlockAddress(nt->names, inode->nameIndex)) == NULL){
		*name = 0;
		return;
	}
	len = ballocBlockSize(nt->names, inode->nameIndex);
	if(inode->isUnicode){
		for(i=0;i<len/2;i++){
			if(p.u[i] == 0 || i >= arrSize - 1)
				break;
			name[i] = p.u[i];
		}
	}else{
		for(i=0;i<len;i++){
			if(p.c[i] == 0 || i >= arrSize - 1)
				break;
			name[i] = p.c[i];
		}
	}
	name[i] = 0;
}

/* ------------------------------------------------------------------------- */

static inline void	*allocInode(nfsfhLookupTable_t *nt, btu32 *inump)
{
nfsfhInode_t	*inode;
btu32			inum;
btsize			dummy;

	inum = ballocAllocBlock(nt->inodes, sizeof(nfsfhInode_t), &dummy);
	inode = ballocBlockAddress(nt->inodes, inum);
	inode->useCount = inode->nameIndex;	/* has been stored there */
	inode->childList = 0;
	inode->nameIndex = 0;
	*inump = inum;
	return inode;
}

/* ------------------------------------------------------------------------- */

static btinum	ntInodeOfFile(nfsfhLookupTable_t *nt, nfsFh_t *parent, btuni *name)
{
nfsfhInode_t	*inode, *childInode;
btbool			found;
btu32			childI, childListI;
volatile btu32	childInum = 0;

	inode = ballocBlockAddress(nt->inodes, parent->inum);
	if(inode == NULL)
		return 0;
EXC_TRY
	childI = indexOfChild(nt, inode->childList, name, nt->super.isCaseSensitive, &found);
	if(!found){
		childInode = allocInode(nt, (btu32 *)&childInum);
		childInode->parent = parent->inum;
		setInodeName(nt, childInode, name);
		childListI = inode->childList;
		ctabInsertObjectAtIndex(nt->childTables, &childListI, childI, childInum);
		inode->childList = childListI;
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nfslutable: creating entry ->%s<- with inum=%d\n", uniToCstring(name), childInum);
	}else{
		childInum = ctabObjectAtIndex(nt->childTables, inode->childList, childI);
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nfslutable: found entry ->%s<- (inum=%d)\n", uniToCstring(name), childInum);
	}
EXC_CATCH
	if(EXC_NAME == memExcNomem){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfslutable: error creating entry: out of virtual memory\n");
		childInum = 0;
	}else{
		EXC_RETHROW();
	}
EXC_ENDTRY
	return childInum;
}

/* ------------------------------------------------------------------------- */

static void	ntHandleOfFile(void *this, void *parentHandle, btuni *name, void *destHandle, btinum *destInum)
{
nfsfhLookupTable_t	*nt = this;
btu32				inum = 0;
nfsfhInode_t		*inode;
nfsFh_t				*parent = parentHandle, *fh = destHandle;
static btuni		dot[2] = {'.', 0};
static btuni		dotdot[3] = {'.', '.', 0};
	
	if(name == NULL || memcmp(name, dot, sizeof(dot)) == 0){
		inum = parent->inum;
	}else if(memcmp(name, dotdot, sizeof(dotdot)) == 0){
		if((inode = ballocBlockAddress(nt->inodes, parent->inum)) != NULL)
			inum = inode->parent;
	}else{
		inum = ntInodeOfFile(this, parent, name);
	}
	if(destInum != NULL){
		*destInum = inum;
	}
	if(fh != NULL){
		memset(fh, 0, 32);
		fh->inum = inum;
		fh->shareIndex = nt->shrIndex;
		inode = ballocBlockAddress(nt->inodes, inum);
		if(inode != NULL){
			fh->useCount = inode->useCount;
		}
		fh->mountId = parent->mountId;
	}
}

/* ------------------------------------------------------------------------- */

static void	ntFileWithHandle(void *this, void *handle, void *user, btbool touchOpen, void *caller);

static void	ntFileWithHandleProc(void *obj, void *arg)
{
findContext_t		*this = obj;
nfsfhLookupTable_t	*nt = this->nt;
nfsfhInode_t		*inode, *parentInode;
luRes_t				res, *rval = arg;
beFileRes_t			*frval = arg;
btuni				name[NFS_MAXNAMLEN + 1];
nfsFh_t				fh;

	PROC_BEGIN;
	inode = ballocBlockAddress(nt->inodes, this->inum);
	if(inode == NULL || inode->nameIndex == 0){
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCstale;
		res.vfsFile = NULL;
		PROC_RETURN(&res);
	}
	parentInode = ballocBlockAddress(nt->inodes, inode->parent);
	if(parentInode == NULL){
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCstale;
		res.vfsFile = NULL;
		PROC_RETURN(&res);
	}
	memset(&fh, 0, sizeof(fh));
	fh.inum = inode->parent;
	fh.useCount = parentInode->useCount;
	fh.shareIndex = nt->shrIndex;
	PROC_WAITFOR_X(1, ntFileWithHandle(nt, &fh, this->user, 0, this));
	if(rval->err.class != 0){
		PROC_RETURN(rval);	/* pass on errors */
	}
	this->file = rval->vfsFile;
	inode = ballocBlockAddress(nt->inodes, this->inum);
	if(inode == NULL || inode->nameIndex == 0){	/* inode may have become stale */
		nt->super.backend.fileRelease(this->file);
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCstale;
		res.vfsFile = NULL;
		PROC_RETURN(&res);
	}
	getInodeName(nt, inode, name, NFS_MAXNAMLEN + 1);
	PROC_WAITFOR_X(2, nt->super.backend.fileLookup(this->file, name, this->user, this));
	nt->super.backend.fileRelease(this->file);
	res.err = frval->err;
	if(frval->err.class != 0){
		if(frval->err.class == ERRCLS_CIFS_DOS && (frval->err.code == ERR_CIFS_DOSbadpath || frval->err.code == ERR_CIFS_DOSbadfile)){
			/* the file disappeared and the handle is stale */
			inode = ballocBlockAddress(nt->inodes, this->inum);
			if(inode != NULL && inode->nameIndex != 0){
				getInodeName(nt, inode, name, NFS_MAXNAMLEN + 1);
				ntFileDestroyFromInode(nt, inode, name);	/* remove the stale entry from the database */
			}
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOCstale;
		}
	}else{
		luCacheFile(&nt->super, frval->file);
		if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
			logPrintf(LOGM_RARE, "nfslutable: looked up path (file=%s) for inum=%d\n", uniToCstring(nt->super.backend.fileName(frval->file)), this->inum);
	}
/* it's not necessary to touch the file in the open files LRU because if it
 * was already cached, we won't reach this point, and if it has not been
 * cached, it is new and will be at the tail of the LRU anyway.
 */
	res.vfsFile = frval->file;
	PROC_RETURN(&res);
	PROC_END;
}

static void	ntFileWithHandle(void *this, void *handle, void *user, btbool touchOpen, void *caller)
{
findContext_t		*ctx;
nfsfhLookupTable_t	*nt = this;
luRes_t				res;
nfsFh_t				*fh = handle;
luFileHandle_t		*lufh = NULL;

	if(fh->inum == nt->rootInode){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nfslutable: inum %d is root\n", ((nfsFh_t *)handle)->inum);
        res.err.class = res.err.code = 0;
        res.vfsFile = nt->super.rootFile;
		nt->super.backend.fileRetain(res.vfsFile);
        PROC_IMMEDIATE_RET(caller, &res);
	}
	lufh = (luFileHandle_t *)((btu8 *)handle - (btint)&lufh->nfsfh);
	lufh = htabFindObjectEqualTo(nt->super.filesByFh, lufh);
	if(lufh != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nfslutable: inum %d cached\n", ((nfsFh_t *)handle)->inum);
        res.err.class = res.err.code = 0;
        res.vfsFile = lufh->vfsFile;
		if(touchOpen && lufh->openFilesLru.prevnext != NULL){
			lruTouch(nt->super.openFilesLru, lufh);
		}
		lruTouch(nt->super.filesLru, lufh);
		nt->super.backend.fileRetain(res.vfsFile);
        PROC_IMMEDIATE_RET(caller, &res);
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
		logPrintf(LOGM_RARE, "nfslutable: inum %d not cached\n", ((nfsFh_t *)handle)->inum);
	PROC_INIT(ctx, ntFileWithHandleProc, caller, findPool, findCtxFree);
	ctx->nt = nt;
	ctx->inum = fh->inum;
	ctx->user = user;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	destroyInode(nfsfhLookupTable_t *nt, btu32 inum)
{
nfsfhInode_t	*inode, *freeMe;
btsize			i, len;
btu32			childListI;

	inode = ballocBlockAddress(nt->inodes, inum);
	if(inode != NULL){
		if(inode->nameIndex != 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
				btuni	inodeName[256];
				getInodeName(nt, inode, inodeName, 256);
				logPrintf(LOGM_DEBUG, "nfslutable: destroying inode %d (%s)\n", inum, uniToCstring(inodeName));
			}
			ballocFreeBlock(nt->names, inode->nameIndex);
			inode->nameIndex = 0;
		}
		inode->parent = nt->rootInode;
		if(inode->childList != 0){	/* inode still has children! */
			len = ctabObjectCount(nt->childTables, inode->childList);
			for(i=0;i<len;i++){
				destroyInode(nt, ctabObjectAtIndex(nt->childTables, inode->childList, i));
			}
			childListI = inode->childList;
			ctabRemoveAllObjects(nt->childTables, &childListI);
			inode->childList = childListI;
		}
		if(nt->freeInodes[nt->freeInodeI] != 0){
			freeMe = ballocBlockAddress(nt->inodes, nt->freeInodes[nt->freeInodeI]);
			freeMe->nameIndex = freeMe->useCount + 1;	/* store the use count there because original position is overwritten */
			ballocFreeBlock(nt->inodes, nt->freeInodes[nt->freeInodeI]);
		}
		nt->freeInodes[nt->freeInodeI] = inum;
		nt->freeInodeI = (nt->freeInodeI + 1) % ELEMENTS(nt->freeInodes);
	}
}

/* ------------------------------------------------------------------------- */

static void	ntFileRename(void *this, void *fromDirFh, btuni *fromName, void *toDirFh, btuni *toName)
{
nfsfhLookupTable_t	*nt = this;
nfsfhInode_t		*fromDirInode, *toDirInode, *inode;
btu32				fromI, toI, inum, toInum, childListI;
btbool				found;
nfsFh_t				*fromDir = fromDirFh, *toDir = toDirFh;

	fromDirInode = ballocBlockAddress(nt->inodes, fromDir->inum);
	toDirInode = ballocBlockAddress(nt->inodes, toDir->inum);
	if(fromDirInode == NULL || toDirInode == NULL){	/* should never happen, but do best recovery possible */
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nfslutable: parent dirs of rename do not exist\n");
		if(fromDirInode != NULL)
			ntFileDestroyFromInode(nt, fromDirInode, fromName);
		if(toDirInode != NULL)
			ntFileDestroyFromInode(nt, toDirInode, toName);
		return;
	}
EXC_TRY
	fromI = indexOfChild(nt, fromDirInode->childList, fromName, nt->super.isCaseSensitive, &found);
	if(found){
		inum = ctabObjectAtIndex(nt->childTables, fromDirInode->childList, fromI);
		childListI = fromDirInode->childList;
		ctabRemoveObjectAtIndex(nt->childTables, &childListI, fromI);
		fromDirInode->childList = childListI;
		inode = ballocBlockAddress(nt->inodes, inum);
	}else{
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nfslutable: rename: source ->%s<- does not exist\n", uniToCstring(fromName));
		inode = allocInode(nt, &inum);
	}
	setInodeName(nt, inode, toName);
	inode->parent = toDir->inum;
	toI = indexOfChild(nt, toDirInode->childList, toName, nt->super.isCaseSensitive, &found);
	if(found){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nfslutable: rename: destination ->%s<- exists!\n", uniToCstring(toName));
		toInum = ctabObjectAtIndex(nt->childTables, toDirInode->childList, toI);
		childListI = toDirInode->childList;
		ctabRemoveObjectAtIndex(nt->childTables, &childListI, toI);
		toDirInode->childList = childListI;
		destroyInode(nt, toInum);
	}
	childListI = toDirInode->childList;
	ctabInsertObjectAtIndex(nt->childTables, &childListI, toI, inum);
	toDirInode->childList = childListI;
EXC_CATCH
	if(EXC_NAME != memExcNomem){
		EXC_RETHROW();
	}else{
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfslutable: error renaming entry: out of virtual memory\n");
	}
EXC_ENDTRY
}

/* ------------------------------------------------------------------------- */

static void	ntFileDestroyFromInode(nfsfhLookupTable_t *nt, nfsfhInode_t *parentInode, btuni *name)
{
btu32	inodeI, inum, childListI;
btbool	found;

	if(parentInode == NULL)	/* parent is stale, nothing to be done */
		return;
EXC_TRY
	inodeI = indexOfChild(nt, parentInode->childList, name, nt->super.isCaseSensitive, &found);
	if(found){
		inum = ctabObjectAtIndex(nt->childTables, parentInode->childList, inodeI);
		childListI = parentInode->childList;
		ctabRemoveObjectAtIndex(nt->childTables, &childListI, inodeI);
		parentInode->childList = childListI;
		destroyInode(nt, inum);
	}
EXC_CATCH
	if(EXC_NAME != memExcNomem){
		EXC_RETHROW();
	}else{
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfslutable: error destroying entry: out of virtual memory\n");
	}
EXC_ENDTRY
}

static void	ntFileDestroy(void *this, void *parentHandle, btuni *name)
{
nfsfhLookupTable_t	*nt = this;
nfsfhInode_t		*dirInode;
nfsFh_t 			*parent = parentHandle;

	dirInode = ballocBlockAddress(nt->inodes, parent->inum);
	ntFileDestroyFromInode(nt, dirInode, name);
}

/* ------------------------------------------------------------------------- */

nfsLookupInterface_t	nfsfhTableInterface = {
	ntRootHandle,
	ntHandleOfFile,
	ntFileWithHandle,
	ntFileRename,
	ntFileDestroy,
	nfsfhLookupTableFree,
};

/* ------------------------------------------------------------------------- */

