/*
 * Name: nfs3.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 2001-10-28
 * Tabsize: 4
 * Copyright: (c) 2001 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfs3.c,v 1.4 2002/04/16 17:06:15 cs Exp $
 */

/*
General Description:
This is the NFS3 frontend module. It implements the frontend interface.
*/

#include "stdheaders.h"
#include "memalloc.h"
#include "hashtable.h"
#include "logging.h"
#include "system.h"
#include "sunrpc.h"
#include "nfs3_prot.h"
#include "nfs3xdr.h"
#include "nfs3print.h"
#include "blockpool.h"
#include "process.h"
#include "util.h"
#include "main.h"
#include "cfgparse.h"
#include "evntloop.h"
#include "nfs3fileadapter.h"
#include "nfslookup.h"	/* for nfsLookupStrategyNames */
#include "notification.h"
#include "frontend.h"
#include "backend.h"
#include "errcodes.h"
#include "unicode.h"
#include "module.h"
#include "mainmodule.h"
#include "../../uiLib/maininterface.h"
#include "mainconst.h"
#include "diskarb.h"

#define	NFS3_MODULE_NAME	"nfs3"

/* ------------------------------------------------------------------------- */

typedef struct backendImp{
	btint	(*vfsConfigPath)(void *obj, struct cfgProperty **path, btint maxEntries);
	void	(*vfsRetain)(void *obj);
	void	(*vfsRelease)(void *obj);
	btuni	*(*vfsPath)(void *obj);
	void	(*userAuthenticate)(void *vfs, btint unixUid, btint unixGid, void *caller);
	void	(*userRetain)(void *obj);
	void	(*userRelease)(void *obj);
}backendImp_t;

typedef struct nfs3{
	btu32					mountId;		/* unique ID for this mount, first entry for hashtable search */
	btu8					*mountpoint;
	btu8					*device;		/* how this entry is named in /etc/mnttab */
	backendImp_t			backend;
	struct module			*backendMod;
	void					*vfs;
	struct nfs3FileAdapter	*adapter;

	btint					refCount;
	btint					mountedBy;
	btbool					isFakeMode;
	btbool					isReadonly;
	btbool					isMounted;
}nfs3_t;

/* ------------------------------------------------------------------------- */

typedef struct nfs3Ctx{
    PROC_VARSET;
    nfs3_t			*nfsMp;
	srpcMsg_t		*msg;
	btint			retryCount;
	void			*user;
/* context is cleared only until here */
	bttime			startTime;
	nfs3Arg_t		argument;
/* reply context: */
	srpcRequestHeader_t	requestHdr;
    struct sockaddr_in	remoteAddr;
    int					raddrLen;
}nfs3Ctx_t;

static blockpool_t	*nfs3Pool = NULL;

static void	freeNfs3Block(void *block)
{
    poolReturnBlock(nfs3Pool, block);
}

/* ------------------------------------------------------------------------- */

btint	nfs3OperationsRunning = 0;
/* nfs3OperationsRunning is used for debugging purposes. In idle mode, it
 * should go back to 0. If not, an operation got stuck, probably due to an
 * internal deadlock. The state of this variable can be checked by means
 * of the debugger.
 */


static btint				nfs3Sock = -1;	/* socket for NFS RPC */
static struct sockaddr_in	nfs3SockAddr;
static struct hashtable		*runningRequests;
static btu16				mountCounter;
static btu16				incarnation;
static struct hashtable		*mountsById;

/* ------------------------------------------------------------------------- */

static void		nfs3MpRetain(void *obj);
static void		nfs3MpRelease(void *obj);
static void		nfs3MpFree(nfs3_t *mp);
static btint	nfs3ConfigPath(nfs3_t *mp, struct cfgProperty **path, btint maxEntries);

/* ------------------------------------------------------------------------- */
/* -------------------------- function dispatcher -------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct nfsProc {
	void	(*encodeResult)();	/* result marshalling routine */
	void	(*decodeArg)();		/* argument unmarshalling routine */
    char	*name;				/* function name */
    void	(*printArg)(void *);/* print argument for debug */
    void	(*printRes)(void *);/* print result for debug */
}nfsProc_t;

#define NFSPROC_ENTRY(resType, function, argType) {	\
	nfs3Encode##resType,		\
	nfs3Decode##argType,		\
	#function,					\
	(void *)nfs3Print##argType,	\
	(void *)nfs3Print##resType	\
}

static void	nfs3EncodeVoid(){}
static void	nfs3DecodeVoid(){}
static void	nfs3PrintVoid(){}

static nfsProc_t	dispatchTable[] = {
	NFSPROC_ENTRY(Void,				nfs3Null,			Void),
	NFSPROC_ENTRY(GetattrRes,		nfs3Getattr,		Fh),
	NFSPROC_ENTRY(SetattrRes,		nfs3Setattr,		SetattrArgs),
	NFSPROC_ENTRY(LookupRes,		nfs3Lookup,			Diropargs),
	NFSPROC_ENTRY(AccessRes,		nfs3Access,			AccessArgs),
	NFSPROC_ENTRY(ReadlinkRes,		nfs3Readlink,		Fh),
	NFSPROC_ENTRY(ReadRes,			nfs3Read,			ReadArgs),
	NFSPROC_ENTRY(WriteRes,			nfs3Write,			WriteArgs),
	NFSPROC_ENTRY(CreateRes,		nfs3Create,			CreateArgs),
	NFSPROC_ENTRY(MkdirRes,			nfs3Mkdir,			MkdirArgs),
	NFSPROC_ENTRY(SymlinkRes,		nfs3Symlink,		SymlinkArgs),
	NFSPROC_ENTRY(MknodRes,			nfs3Mknod,			MknodArgs),
	NFSPROC_ENTRY(RemoveRes,		nfs3Remove,			Diropargs),
	NFSPROC_ENTRY(RmdirRes,			nfs3Rmdir,			Diropargs),
	NFSPROC_ENTRY(RenameRes,		nfs3Rename,			RenameArgs),
	NFSPROC_ENTRY(LinkRes,			nfs3Link,			LinkArgs),
	NFSPROC_ENTRY(ReaddirRes,		nfs3Readdir,		ReaddirArgs),
	NFSPROC_ENTRY(ReaddirplusRes,	nfs3Readdirplus,	ReaddirplusArgs),
	NFSPROC_ENTRY(FsstatRes,		nfs3Fsstat,			Fh),
	NFSPROC_ENTRY(FsinfoRes,		nfs3Fsinfo,			Fh),
	NFSPROC_ENTRY(PathconfRes,		nfs3Pathconf,		Fh),
	NFSPROC_ENTRY(CommitRes,		nfs3Commit,			CommitArgs),
};

/* ------------------------------------------------------------------------- */

static void	*fakedReply3(nfs3Ctx_t *ctx, btbool fakeOK)
{
static nfs3Res_t	res;
btu8				fhBuffer[32];
btinum				inum = 0;

	memset(&res, 0, sizeof(res));
	if(fakeOK){
		switch(ctx->requestHdr.proc){
		case NFS3_PROC_GETATTR:
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG) || LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE))
				logPrintf(LOGM_DEBUG, "nfs3: faking getattr reply\n");
			if(ctx->nfsMp->adapter != NULL)
				inum = nfs3FaRootHandle(ctx->nfsMp->adapter, fhBuffer, ctx->nfsMp->mountId);
			if(ctx->nfsMp->adapter != NULL && ctx->argument.getattr.len == 32 && memcmp(fhBuffer, ctx->argument.getattr.data, sizeof(fhBuffer)) == 0){	/* getattr for root dir */
				res.getattr.attributes.type = NFS3_TYPE_DIR;
				res.getattr.attributes.mode = 0777;
				res.getattr.attributes.nlink = 2;
				res.getattr.attributes.size = 4096;
				res.getattr.attributes.used = 4096;
				res.getattr.attributes.fileid = inum;
			}else{
				fakeOK = 0;
			}
			break;
		case NFS3_PROC_FSSTAT:
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG) || LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE))
				logPrintf(LOGM_DEBUG, "nfs3: faking statfs reply\n");
			res.fsstat.tbytes = 1000000000;
			res.fsstat.fbytes = 1000000000;
			res.fsstat.abytes = 1000000000;
			res.fsstat.tfiles = 1000000;
			res.fsstat.ffiles = 1000000;
			res.fsstat.afiles = 1000000;
			res.fsstat.invarsec = 1;	/* also an estimate, assume relatively busy FS */
			break;
		case NFS3_PROC_FSINFO:
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG) || LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE))
				logPrintf(LOGM_DEBUG, "nfs3: faking fsinfo reply\n");
			/* Values copied from nfs3fileadapter.c / nfs3FSINFO() */
			res.fsinfo.rtmax = 16384;
			res.fsinfo.rtpref = 8192;
			res.fsinfo.rtmult = 4096;
			res.fsinfo.wtmax = 16384;
			res.fsinfo.wtpref = 8192;
			res.fsinfo.wtmult = 4096;
			res.fsinfo.dtpref = 4096;
			res.fsinfo.maxfilesize = (btu64)0xffffffff << 30;
			/* The following is usually provided by the backend. These are the values
			 * set by Sharity's CIFS backend. If the kernel needs these values at
			 * mount time, it will hopefully request them later again...
			 */
			res.fsinfo.time_delta.seconds = 2;
			res.fsinfo.time_delta.nseconds = 0;
			res.fsinfo.properties = NFS3_FEATURE_SYMLINK | NFS3_FEATURE_CANSETTIME | NFS3_FEATURE_HOMOGENEOUS;
			break;
		default:
			fakeOK = 0;
		}
	}
	if(!fakeOK){
		res.getattr.status = NFS3ERR_ACCES;
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG) || LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE))
			logPrintf(LOGM_DEBUG, "nfs3: faking error reply\n");
	}
	return &res;
}

/* ------------------------------------------------------------------------- */

static void	sendreply(nfs3Ctx_t *this)
{
	if(sendto(nfs3Sock, this->msg->data, this->msg->size, 0, (struct sockaddr *)&this->remoteAddr, this->raddrLen) < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3: error sending reply: [%d] %s\n", errno, sysStrerror(errno));
	}
}

static void	handleRequest3Proc(void *ctx, void *arg)
{
nfs3Ctx_t	*this = ctx;
nfsstat		*volatile res = arg;
beVfsRes_t	*rval = arg;

    PROC_BEGIN
	nfs3MpRetain(this->nfsMp);
    if(this->nfsMp->vfs == NULL || this->nfsMp->isFakeMode){
        res = fakedReply3(this, 1);
	}else{
		nfs3OperationsRunning++;
		for(this->retryCount = 0; this->retryCount < 2; this->retryCount++){
			/* try to authenticate user before we do anything */
			PROC_WAITFOR_X(1, this->nfsMp->backend.userAuthenticate(this->nfsMp->vfs, this->requestHdr.credUid, this->requestHdr.credGid, this));
			if(rval->err.class == ERRCLS_CIFS_LOCAL && rval->err.code == ERR_CIFS_LOCconnlost){
/* No use to retry if connecting the user failed, because there are two possibilities:
 * 1.: The user was already connected. In this case userAuthenticate() is a no-op.
 * 2.: The user was not connected. In this case the connection to the server has been lost
 *     before or has never been established; it's an initial connect attempt which should
 *     not be retried.
 * [Is this true? The server could be in connected state but the user not...]
 */
				res = NULL;
				break;
			}
			if(rval->err.class != 0){	/* error during authentication */
				res = NULL;
				break;
			}
			this->user = rval->user;
			this->nfsMp->backend.userRetain(this->user);
			PROC_WAITFOR_X(2, nfs3FaExecuteRequest(this->nfsMp->adapter, this->requestHdr.proc, &this->argument, this->user, this));
			this->nfsMp->backend.userRelease(this->user);
			if((btint)*res != -1)
				break;	/* retry if result == -1 which means: connection lost */
			/* get rid of stack: */
			PROC_WAITFOR_X(3, elpCallCancelPrev(handleRequest3Proc, this));
		}
		if(this->retryCount >= 2)	/* even retry failed */
			res = NULL;
		nfs3OperationsRunning--;
	}
    if(LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE)){
        if(res != NULL)
			dispatchTable[this->requestHdr.proc].printRes(res);
        logPrintf(LOGM_NFSTRACE, "*** nfs return[%d].%s\n", this->requestHdr.xid, res == NULL ? " (NULL)" : "");
    }
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
		bttime			duration;
		static bttime	maxDuration;
		duration = elpNow() - this->startTime;
		if(duration > maxDuration){
			maxDuration = duration;
			logPrintf(LOGM_DEBUG, "nfs3: maximum request duration: %d ms\n", maxDuration);
		}
	}
	if(res == NULL)
		res = fakedReply3(this, 0);
	/* we use the same srpcMsg object for call and reply data. This implies
	 * that the result *must not* reference any argument data.
	 */
	this->msg->size = this->msg->readIndex = 0;	/* reset message */
EXC_TRY
	srpcEncodeRpcHeader(this->msg, this->requestHdr.xid, 0, 0);
	dispatchTable[this->requestHdr.proc].encodeResult(this->msg, res);
EXC_CATCH
	this->msg->size = this->msg->readIndex = 0;	/* reset message */
	srpcEncodeRpcHeader(this->msg, this->requestHdr.xid, 0, 4);	/* GARBAGE_ARGS */
EXC_ENDTRY
	sendreply(this);	/* just send for the risk that we block or lose a packet */
	srpcMsgFree(this->msg);
	htabRemoveObjectEqualTo(runningRequests, this);
	nfs3MpRelease(this->nfsMp);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	sockReadable(void *dummyObj, void *dummyArg)
{
nfs3Ctx_t	*volatile ctx = NULL;
nfsProc_t	*proc;
btu32		mountId;

EXC_TRY
    PROC_INIT(ctx, handleRequest3Proc, NULL, nfs3Pool, freeNfs3Block);
	ctx->msg = srpcMsgNew();
	ctx->raddrLen = sizeof(ctx->remoteAddr);
	ctx->msg->size = recvfrom(nfs3Sock, ctx->msg->data, SRPC_MAX_BUFFER, 0, (struct sockaddr *)&ctx->remoteAddr, &ctx->raddrLen);
	if(ctx->msg->size < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3: error receiving: [%d] %s\n", errno, sysStrerror(errno));
		srpcMsgFree(ctx->msg);
		freeNfs3Block(ctx);
		return;
	}
	srpcDecodeRpcHeader(ctx->msg, &ctx->requestHdr);
	if(ctx->requestHdr.rpcvers != 2){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 1, 0);	/* RPC_MISMATCH */
	}else if(ctx->requestHdr.prog != NFS3_PROGRAM){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 1);	/* PROG_UNAVAIL */
	}else if(ctx->requestHdr.vers != NFS3_VERSION){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 2);	/* PROG_MISMATCH */
	}else if(ctx->requestHdr.proc > ELEMENTS(dispatchTable)){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 3);	/* PROC_UNAVAIL */
	}else if(ctx->requestHdr.credFlavor != AUTH_UNIX){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 1, 1);	/* AUTH_ERROR */
	}else{
		if(htabFindObjectEqualTo(runningRequests, ctx) != NULL){
			if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
				logPrintf(LOGM_RARE, "*** nfs3: detected RPC retry for xid=%d\n", ctx->requestHdr.xid);
			srpcMsgFree(ctx->msg);
			freeNfs3Block(ctx);
			return;	/* just ignore the message */
		}
/* By coincidence (?), all NFS3 argument structures have an nfs_fh3 as the first
 * argument. Use this coincidence to find the mountpoint related to the file handle.
 * This "coincidence" holds true at least for versions 2 and 3 of the NFS protocol.
 */
		proc = &dispatchTable[ctx->requestHdr.proc];
		proc->decodeArg(ctx->msg, &ctx->argument);
 		mountId = ((btu32 *)ctx->argument.getattr.data)[7];	/* this is a hack... */
		ctx->nfsMp = htabFindObjectEqualTo(mountsById, &mountId);
		if(ctx->nfsMp == NULL){	/* must be a stale mount, reply with NFSERR_STALE */
			nfs3Res_t	res;
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "nfs3: received request for stale mount id=0x%08x xid=%d\n", mountId, ctx->requestHdr.xid);
			ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
			memset(&res, 0, sizeof(res));
			res.getattr.status = NFS3ERR_STALE;
			proc->encodeResult(ctx->msg, &res);
		}else{
			htabAddObject(runningRequests, ctx);
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				ctx->startTime = elpNow();
			if(!ctx->nfsMp->isMounted && ctx->requestHdr.credUid == 0){	/* request during mount mapped to mounting UID */
				ctx->requestHdr.credUid = ctx->nfsMp->mountedBy;
			}
			if(LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE)){
				logPrintf(LOGM_NFSTRACE, "*** nfs3 command[%d]: %s\n", ctx->requestHdr.xid, proc->name);
				proc->printArg(&ctx->argument);
			}
			PROC_START(ctx, NULL);
			return;
		}
	}
EXC_CATCH
	if(EXC_ISX(srpcExcDecodeOverflow)){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3: error unmarshalling arguments for proc=%d xid=%d\n", ctx->requestHdr.proc, ctx->requestHdr.xid);
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 4);	/* GARBAGE_ARGS */
		/* reply is sent from cleanup code below */
	}
EXC_ENDTRY
	if(ctx != NULL){
		if(ctx->msg != NULL){
			sendreply(ctx);
			srpcMsgFree(ctx->msg);
		}
		freeNfs3Block(ctx);
	}
}

/* ------------------------------------------------------------------------- */

static btint	ctxXidCompare(void *obj, void *otherObj)
{
nfs3Ctx_t	*this = obj, *other = otherObj;

	return this->requestHdr.xid - other->requestHdr.xid;
}

static btu32	ctxHashval(void *obj)
{
nfs3Ctx_t	*ctx = obj;

	return ctx->requestHdr.xid ^ (ctx->requestHdr.xid >> 24);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------- mounting -------------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct mntCtx{
	PROC_VARSET;
	nfs3_t		*nfsMp;
	btint		mountingUid;
	btbool		readOnly;
}mntCtx_t;

static blockpool_t	*mnt3Pool = NULL;

static void	freeMntBlock(void *block)
{
    poolReturnBlock(mnt3Pool, block);
}

static procCond_t	*mntSignal = NULL;
static btbool		mntRunning = 0;

/* ------------------------------------------------------------------------- */

static void	setSockBuffer(int fd, btbool isTx, btsize size)
{
int		optname = isTx ? SO_SNDBUF : SO_RCVBUF;
int		bufSize, defaultSize = 0, optlen;

	optlen = sizeof(int);
	if(getsockopt(fd, SOL_SOCKET, optname, (char *)&defaultSize, &optlen) != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3: getsockopt(): [%d] %s\n", errno, sysStrerror(errno));
		return;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "nfs3: set %s buffer size: want=%d, default=%d\n", isTx ? "TX" : "RX", size, defaultSize);
	for(bufSize = size; bufSize > 8; bufSize -= bufSize / 8){
		if(setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&bufSize, sizeof(int)) == 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
				logPrintf(LOGM_INFO, "nfs3: set %s buffer to %d bytes\n", isTx ? "TX" : "RX", bufSize);
			return;
		}
		if(bufSize < defaultSize)	/* already less than default, does still not work */
			break;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "nfs3: setsockopt(%s, %d): [%d] %s\n", isTx ? "txbuf" : "rxbuf", bufSize, errno, sysStrerror(errno));
}

/* ------------------------------------------------------------------------- */

static collectionObjectI_t	contextByXidInterface = {
	ctxXidCompare,
	ctxHashval,
	NULL,
	NULL
};

static btint	createNfs3Socket(void)
{
cfgProperty_t	*cfgPath[4];
btint			rval, numEntries, wantPort, storedErrno;
btsize			socketBuffer;
btint			sockAddrlen = sizeof(nfs3SockAddr);

	if(nfs3Sock != -1)
		return 0;
	nfs3Sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	storedErrno = errno;
	errno = storedErrno;
    if(nfs3Sock < 0)
		goto errorOccured;
	numEntries = nfs3ConfigPath(NULL, cfgPath, sizeof(cfgPath)/sizeof(cfgPath[0]));
	socketBuffer = cfgIntValue(cfgQueryFromArray("nfsSockRxBuf", numEntries, cfgPath), 0x18000);
	if(socketBuffer > 0)	/* increase rx buffer to avoid nfs retries on writes */
		setSockBuffer(nfs3Sock, 0, socketBuffer);
	socketBuffer = cfgIntValue(cfgQueryFromArray("nfsSockTxBuf", numEntries, cfgPath), 0x18000);
	if(socketBuffer > 0)	/* allow changing the tx buffer of the socket, usually not needed */
		setSockBuffer(nfs3Sock, 1, socketBuffer);
	wantPort = cfgIntValue(cfgQueryFromArray("preferredPort", numEntries, cfgPath), 991);

    memset(&nfs3SockAddr, 0, sockAddrlen);
	nfs3SockAddr.sin_family = AF_INET;
	nfs3SockAddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	nfs3SockAddr.sin_port = htons(wantPort);

	rval = bind(nfs3Sock, (struct sockaddr *)&nfs3SockAddr, sockAddrlen);
	storedErrno = errno;
	if(rval < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3: can\'t bind to preferred port %d, trying free port ([%d] %s)\n", wantPort, storedErrno, sysStrerror(storedErrno));
		nfs3SockAddr.sin_port = 0;	/* if this port fails, try the any-port */
		if(bind(nfs3Sock, (struct sockaddr *)&nfs3SockAddr, sockAddrlen) < 0)
			goto errorOccured;
	}
    if(getsockname(nfs3Sock, (struct sockaddr *)&nfs3SockAddr, (void *)&sockAddrlen) < 0)
        goto errorOccured;
    elpRegisterForFd(nfs3Sock, ELP_FDMODE_READ, sockReadable, NULL);	/* nfs request handler */
	return 0;
errorOccured:
	if(nfs3Sock != -1){
		storedErrno = errno;
		close(nfs3Sock);
		nfs3Sock = -1;
		errno = storedErrno;
	}
	return -1;
}

/* ------------------------------------------------------------------------- */

static void	nfs3MountProc(void *obj, void *arg)
{
mntCtx_t			*this = obj;
nfs3_t				*mp = this->nfsMp;
elpPid_t			*rval = arg;
feRes_t				res;
btint				nfsTimeout, nfsRetransmit, len, numEntries, child, kernelCacheTime;
btint				nfsTransferSize;
btu8				rootFh[32];
char				nfsService[1024], *p;
btu8				*nfsHost;
cfgProperty_t		*cfgPath[10];

	PROC_BEGIN;
	nfs3MpRetain(mp);
	while(mntRunning){	/* restrict Sharity to one running mount/umount */
		PROC_WAITFOR_X(1, procCondWait(mntSignal, this));
	}
	mntRunning = 1;
	if(mp->isMounted){	/* already mounted */
		res.err.class = ERRCLS_UNIX;
		res.err.code = EBUSY;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3Mount: already mounted\n");
		goto doReturn;
	}
#if 0
/* The permission check must not be done in the server because it requires
 * the stat() system call. If stat() is called, it may be called on a
 * mounted directory served by Sharity, which results in a deadlock.
 * Since the cifsmount utility must be trusted to supply the correct
 * user id anyway, the check can equally well be done in that utility.
 */
	if(!mountPermissionsOk(mp, mountingUid)){
        errno = EPERM;	/* Not owner */
		goto errorOccured;
	}
#endif
	if(nfs3Sock == -1){
		res.err.class = ERRCLS_UNIX;
		res.err.code = EBADF;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfsMount: could not create socket during startup.\n");
		goto doReturn;
	}
	numEntries = nfs3ConfigPath(mp, cfgPath, sizeof(cfgPath)/sizeof(cfgPath[0]));
	nfsHost = cfgStringValue(cfgQueryFromArray("nfsHost", numEntries, cfgPath), "localhost");
	nfsTimeout = cfgIntValue(cfgQueryFromArray("nfsTimeout", numEntries, cfgPath), 6);
	nfsRetransmit = cfgIntValue(cfgQueryFromArray("nfsRetransmit", numEntries, cfgPath), 7);
	nfsTransferSize = cfgIntValue(cfgQueryFromArray("nfsTransferSize", numEntries, cfgPath), 8192);
	kernelCacheTime = cfgIntValue(cfgQueryFromArray("nfsKernelCacheTime", numEntries, cfgPath), 0);

	if(nfsTransferSize > 8192){		/* limit to useful range */
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3: nfsTransferSize %d out of range, limiting to 8192\n", nfsTransferSize);
		nfsTransferSize = 8192;
	}else if(nfsTransferSize < 1024){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3: nfsTransferSize %d out of range, setting to 1024\n", nfsTransferSize);
		nfsTransferSize = 1024;
	}
	/* Emergency brakes to avoid buffer overflow attacks: */
	if(strlen(nfsHost) > 500)
		nfsHost[500] = 0;
	if(strlen(mp->backendMod->name) > 500)
		mp->backendMod->name[500] = 0;
#ifdef HAVE_DISKARB
	if(strcmp(mp->backendMod->name, "cifsBrowse") == 0){
		strcpy(nfsService, "//.CIFS-Browsing/");
		len = strlen(nfsService);
		convUniToLocalAscii(mp->backend.vfsPath(mp->vfs), &nfsService[len], sizeof(nfsService)-len);
	}else{
		convUniToLocalAscii(mp->backend.vfsPath(mp->vfs), nfsService, sizeof(nfsService));
		for(p=nfsService;*p!=0;p++){
			if(*p == '\\')
				*p = '/';
		}
	}
#else
	sprintf(nfsService, "%s:(%s)", nfsHost, mp->backendMod->name);
	len = strlen(nfsService);
	convUniToLocalAscii(mp->backend.vfsPath(mp->vfs), &nfsService[len], sizeof(nfsService)-len);
#endif
	/* Replace whitespace in nfsService with underscore. At least Solaris can't
	 * tolerate whitespace in the mtab entry.
	 */
	for(p=nfsService;*p!=0;p++){
		if(*p == ' ' || *p == '\t')
			*p = '_';
	}
	if(mp->device != NULL)
		memFree(mp->device);
	mp->device = utilStrDup(nfsService);
	mp->isReadonly = this->readOnly;
	mp->mountedBy = this->mountingUid;
	nfs3FaRootHandle(mp->adapter, &rootFh, mp->mountId);
	/* we must fork because there is no non-blocking mount() */
    if((child = fork()) == 0){	/* we are the child */
		btint	storederr, mountRval;
		mainSignalHandlers(1);
		diskarbPreMountNotify(mp->mountpoint, mp->device);
		mountRval = sysMount(mp->mountpoint, &rootFh, nfs3Sock, &nfs3SockAddr, nfsHost, mp->device, nfsTimeout, nfsTransferSize, this->readOnly, nfsRetransmit, kernelCacheTime, 1);
		storederr = errno;
        if(mountRval < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "nfs3: sysMount(): [%d] %s\n", storederr, sysStrerror(storederr));
			_exit(storederr);
        }
        _exit(0);
    }
    PROC_WAITFOR_X(2, elpRegisterForPid(child, nfs3MountProc, this));
	diskarbPostMountNotify(mp->mountpoint, mp->device, rval->rval);
	res.err.class = rval->rval != 0 ? ERRCLS_UNIX : ERRCLS_OK;
    res.err.code = rval->rval;
	if(res.err.class == 0){
		mp->isMounted = 1;
	}
doReturn:
	mntRunning = 0;
	PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(mntSignal, NULL);
		nfs3MpRelease(mp);
	);
	PROC_END;
}

static void	nfs3Mount(void *obj, btint mountingUid, btbool readOnly, void *caller)
{
mntCtx_t	*ctx;

    PROC_INIT(ctx, nfs3MountProc, caller, mnt3Pool, freeMntBlock);
	ctx->nfsMp = obj;
	ctx->mountingUid = mountingUid;
	ctx->readOnly = readOnly;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	nfs3UmountProc(void *obj, void *arg)
{
mntCtx_t	*this = obj;
nfs3_t		*mp = this->nfsMp;
elpPid_t	*rval = arg;
btError		*errRval = arg;
btint		child;
feRes_t		res;

	PROC_BEGIN;
	nfs3MpRetain(mp);
	while(mntRunning){	/* serialize Sharity's mounts and umounts */
		PROC_WAITFOR_X(1, procCondWait(mntSignal, this));
	}
	mntRunning = 1;
	if(this->mountingUid != 0 && this->mountingUid != mp->mountedBy){	/* not permitted */
		res.err.class = ERRCLS_UNIX;
        res.err.code = EPERM;
		goto errorOccured;
	}
	if(!mp->isMounted){	/* not mounted */
		res.err.class = ERRCLS_UNIX;
		res.err.code = ENOENT;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfsUmount(): not mounted\n");
		goto errorOccured;
	}
	PROC_WAITFOR_X(2, diskarbPerformUnmount(mp->mountpoint, mp->device, this));
	if(errRval->class != 0){	/* diskarb failed or is dummy only, try conventional */
		if((child = fork()) == 0){	/* we are the child */
			btint	storederr, umountRval;
			mainSignalHandlers(1);
			umountRval = sysUmount(mp->mountpoint);
			storederr = errno;
			/* ignore ENOENT because this means that nothing is mounted */
			if(umountRval < 0 && storederr != ENOENT){
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "nfs3: sysUmount(): [%d] %s\n", storederr, sysStrerror(storederr));
				_exit(storederr);
			}
			_exit(0);
		}
		PROC_WAITFOR_X(3, elpRegisterForPid(child, nfs3UmountProc, this));
		res.err.class = rval->rval != 0 ? ERRCLS_UNIX : ERRCLS_OK;
		res.err.code = rval->rval;
	}else{
		res.err = *errRval;
	}
	if(res.err.class == 0)	/* unmount succeeded or nothing mounted */
		mp->isMounted = 0;
errorOccured:
	mntRunning = 0;
	PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(mntSignal, NULL);
		nfs3MpRelease(mp);
	);
	PROC_END;
}

static void	nfs3Umount(nfs3_t *mp, btint uid, void *caller)
{
mntCtx_t	*ctx;

    PROC_INIT(ctx, nfs3UmountProc, caller, mnt3Pool, freeMntBlock);
	ctx->nfsMp = mp;
	ctx->mountingUid = uid;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static inline btbool	nfs3MpIsMounted(nfs3_t *mp)
{
    return mp->isMounted;
}

static void	nfs3MpWasUnmounted(nfs3_t *mp)
{
	mp->isMounted = 0;
}

/* ------------------------------------------------------------------------- */
/* --------------------------- object allocation --------------------------- */
/* ------------------------------------------------------------------------- */

static void	nfs3MpRetain(void *obj)
{
nfs3_t	*this = obj;

    this->refCount++;
}

static void	nfs3MpRelease(void *obj)
{
nfs3_t	*this = obj;

    if(--this->refCount == 0){
        nfs3MpFree(this);
    }else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

static btint	nfs3MpCompare(void *mp1, void *mp2)
{
nfs3_t	*this = mp1, *other = mp2;

	return this->mountId - other->mountId;
}

static btu32	nfs3MpHashVal(void *mp)
{
nfs3_t	*this = mp;

	return this->mountId;
}

/* ------------------------------------------------------------------------- */

static nfs3_t	*nfs3MpNew(btu8 *mountpoint)
{
nfs3_t		*volatile mp = NULL;
btint		i;
    
EXC_TRY
    mp = memCalloc(sizeof(nfs3_t));
	mp->refCount = 1;
    mp->mountpoint = utilStrDup(mountpoint);
	i = 0;
	do{
		mp->mountId = ((btu32)incarnation << 16) | mountCounter++;
	}while(htabFindObjectEqualTo(mountsById, mp) != NULL && i++ < 65536);
	if(i >= 65536){
		EXC_THROW_MSG(excInternalError, "nfs3: maximum number of mountpoints exceeded");
	}
	htabAddObject(mountsById, mp);
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "nfs3Mp: allocated new object at 0x%08x with id=0x%08x\n", (int)mp, mp->mountId);
EXC_CATCH
	nfs3MpFree(mp);
	EXC_RETHROW();
EXC_ENDTRY
    return mp;
}

/* ------------------------------------------------------------------------- */

static void	nfs3MpFree(nfs3_t *mp)
{
    if(mp != NULL){
        if(nfs3MpIsMounted(mp)){
            EXC_THROW_MSG(excInternalError, "nfs3: freeing mounted nfs3Mp");
        }
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "nfs3Mp: freeing object at 0x%08x\n", (int)mp);
		htabRemoveObjectEqualTo(mountsById, mp);
		if(mp->adapter != NULL){
			nfs3FaRelease(mp->adapter);
			mp->adapter = NULL;
		}
		if(mp->vfs != NULL){
			mp->backend.vfsRelease(mp->vfs);
			mp->vfs = NULL;
			mp->backendMod = NULL;
		}
        if(mp->mountpoint != NULL){
            memFree(mp->mountpoint);
		}
		if(mp->device != NULL){
			memFree(mp->device);
		}
        memFree(mp);
    }
}

/* ------------------------------------------------------------------------- */

static void	nfs3MpSetFakemode(nfs3_t *this, btbool fakeMode)
{
	this->isFakeMode = fakeMode;
}

/* ------------------------------------------------------------------------- */

static void	nfs3MpSetBackend(nfs3_t *mp, struct module *backendMod, void *backendObj)
{
cfgProperty_t	*cfgPath[10];
btint			numEntries;

	if(mp->vfs != NULL)
		DEBUG_TRAP();
	mp->backendMod = backendMod;
	mp->vfs = backendObj;
	mp->backend.userAuthenticate = backendMod->implementationOf(BE_USER_AUTHENTICATE);
	mp->backend.userRetain = backendMod->implementationOf(BE_USER_RETAIN);
	mp->backend.userRelease = backendMod->implementationOf(BE_USER_RELEASE);
	mp->backend.vfsRetain = backendMod->implementationOf(BE_VFS_RETAIN);
	mp->backend.vfsRelease = backendMod->implementationOf(BE_VFS_RELEASE);
	mp->backend.vfsConfigPath = backendMod->implementationOf(BE_CONFIG_PATH);
	mp->backend.vfsPath = backendMod->implementationOf(BE_VFS_PATH);
	mp->backend.vfsRetain(mp->vfs);
	numEntries = nfs3ConfigPath(mp, cfgPath, sizeof(cfgPath)/sizeof(cfgPath[0]));
	mp->adapter = nfs3FaForBackend(backendMod, backendObj, cfgPath, numEntries);
}

/* ------------------------------------------------------------------------- */

static void	nfs3MpGetBackend(nfs3_t *mp, struct module **backendMod, void **backendObj)
{
	*backendMod = mp->backendMod;
	*backendObj = mp->vfs;
}

/* ------------------------------------------------------------------------- */

static btbool	nfs3MpIsReadonly(nfs3_t *mp)
{
	return mp->isReadonly;
}

/* ------------------------------------------------------------------------- */

static btu8	*nfs3MpMountpoint(nfs3_t *mp)
{
	return mp->mountpoint;
}

/* ------------------------------------------------------------------------- */

static btu8 *nfs3MpMtabDevice(nfs3_t *mp)
{
	if(mp->device == NULL)
		return "<none>";
	else
		return mp->device;
}

/* ------------------------------------------------------------------------- */

static btint	nfs3MpMountedBy(nfs3_t *mp)
{
	return mp->mountedBy;
}

/* ------------------------------------------------------------------------- */

static btint	nfs3ConfigPath(nfs3_t *mp, struct cfgProperty **path, btint maxEntries)
{
cfgProperty_t	*p;
btint			w = 0;

	if(mp != NULL)
		w = mp->backend.vfsConfigPath(mp->vfs, path, maxEntries);
	p = cfgQueryWithName(NFS3_MODULE_NAME, 1, cfgFilePlist(gdata.dbFile));
	if(p != NULL && w < maxEntries)
		path[w++] = p;
	p = cfgQueryWithName(NFS3_MODULE_NAME, 1, cfgFilePlist(gdata.cfgFile));
	if(p != NULL && w < maxEntries)
		path[w++] = p;
	return w;
}

/* ------------------------------------------------------------------------- */

static array_t	*nfs3BackendConfig(void)
{
array_t	*rval;

	rval = arrNew(NULL, 0);
	/* we use the NFS 2 configs, too. No need to have separate entries... */
	return rval;
}

/* ------------------------------------------------------------------------- */
/* ------------------------- module implementation ------------------------- */
/* ------------------------------------------------------------------------- */

static struct{
	btu8	*name;
	void	*function;
} functionTable[] = {
	{FE_MNT_NEW,			nfs3MpNew},
	{FE_MNT_RETAIN,			nfs3MpRetain},
	{FE_MNT_RELEASE,		nfs3MpRelease},
	{FE_MNT_SET_BACKEND,	nfs3MpSetBackend},
	{FE_MNT_GET_BACKEND,	nfs3MpGetBackend},
	{FE_MNT_SET_FAKEMODE,	nfs3MpSetFakemode},
	{FE_MNT_MOUNT,			nfs3Mount},
	{FE_MNT_UMOUNT,			nfs3Umount},
	{FE_MNT_IS_MOUNTED,		nfs3MpIsMounted},
	{FE_MNT_WAS_UNMOUNTED,	nfs3MpWasUnmounted},
	{FE_MNT_IS_READONLY,	nfs3MpIsReadonly},
	{FE_MNT_MOUNTPOINT,		nfs3MpMountpoint},
	{FE_MNT_MTAB_DEVICE,	nfs3MpMtabDevice},
	{FE_MNT_MOUNTED_BY,		nfs3MpMountedBy},
	{FE_CFGENTRIES_FOR_BACKEND, nfs3BackendConfig},
};

static void	*getImp(btu8 *functionName)
{
btint	i;

	for(i=0;i<sizeof(functionTable)/sizeof(functionTable[0]);i++){
		if(strcmp(functionName, functionTable[i].name) == 0)
			return functionTable[i].function;
	}
	return NULL;
}

static collectionObjectI_t	nfs3MpInterface = {
	nfs3MpCompare,
	nfs3MpHashVal,
	NULL,
	NULL
};

void	nfs3_init(void)
{
module_t	*mod;
btu32		t;
nfs3Ctx_t	*context = NULL;	/* only used for measuring sizes */

	t = time(NULL);
	incarnation = (btu16)t;	/* a somewhat unique ID of this daemon instance */
	mountCounter = (btu16)((t >> 8) ^ (t >> 16));	/* start counter at pseudo-random count */
	mountsById = htabNew(&nfs3MpInterface, 16, 50);
	runningRequests = htabNew(&contextByXidInterface, 16, 40);
    if(nfs3Pool == NULL)
        nfs3Pool = poolNew(sizeof(nfs3Ctx_t), (btu8 *)&context->argument - (btu8 *)context);
    if(mnt3Pool == NULL)
        mnt3Pool = poolNew(sizeof(mntCtx_t), sizeof(mntCtx_t));
	if(mntSignal == NULL)
		mntSignal = procCondNew();
	if(createNfs3Socket() < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs3: error creating NFS socket during startup: [%d] %s\n", errno, sysStrerror(errno));
	}
	mod = memCalloc(sizeof(*mod));
	mod->name = NFS3_MODULE_NAME;
	mod->moduleType = MOD_TYPE_FRONTEND;
	mod->implementationOf = getImp;
	mod->interfaceVersion = 0;
	mod->implementationVersion = MAINCONST_VERSION;
	mod->betaVersion = MAINCONST_BETA_VERSION;
	mod->sortPrecedence = cfgIntValue(cfgQueryWithName("sortPrecedence", 1, cfgQueryWithName(NFS3_MODULE_NAME, 1, cfgFilePlist(gdata.cfgFile))), 2000);
	modRegister(mod);
}

/* ------------------------------------------------------------------------- */
