/*
 * Name: nfs3fileadapter.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 2001-10-26
 * Tabsize: 4
 * Copyright: (c) 2001 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfs3fileadapter.c,v 1.7 2002/04/16 17:06:15 cs Exp $
 */

#include "stdheaders.h"
#include "nfs3_prot.h"
#include "nfs3fileadapter.h"
#include "memalloc.h"
#include "process.h"
#include "blockpool.h"
#include "errcodes.h"
#include "logging.h"
#include "unicode.h"
#include "nfslookup.h"
#include "module.h"
#include "backend.h"
#include "hashtable.h"
#include "cfgparse.h"

#define	NAME_ARRAY_SIZE	256
#define	PATH_ARRAY_SIZE	4096

#define	NFS3CONST_BLOCKSIZE	4096	/* plausible value... */

/* ------------------------------------------------------------------------- */

typedef struct backendImp{
	void	(*vfsRetain)(void *obj);
	void	(*vfsRelease)(void *obj);
	void	(*vfsStatfs)(void *obj, void *user, void *caller);
	void	(*vfsConnect)(void *obj, void *user, void *caller);

	void	(*fileRetain)(void *obj);
	void	(*fileRelease)(void *obj);
	void	(*fileLookup)(void *parent, btuni *name, void *user, void *caller);
	void	(*fileGetAttr)(void *obj, void *user, void *caller);
	void	(*fileSetAttr)(void *obj, void *user, beFileSetAttr_t *attr, void *caller);
	void	(*fileReaddir)(void *obj, void *user, btint startIndex, btbool (*resultCallback)(void *callerCtx, btuni *name, btu32 continuationIndex, btbool isDir), void *caller);
	void	(*fileCreate)(void *parent, btuni *name, btbool isDir, void *user, beFileSetAttr_t *attr, void *caller);
	void	(*fileRemove)(void *parent, btuni *name, btbool isDir, void *user, void *caller);
	void	(*fileRename)(void *fromDir, btuni *fromName, void *toDir, btuni *toName, void *user, void *caller);
	void	(*fileRW)(void *obj, btint operation, btbool writeThrough, btu8 *data, btoffset offset, btsize len, void *user, void *caller);
	void	(*fileHardLink)(void *from, void *toDir, btuni *name, void *user, void *caller);
	void	(*fileSymlink)(void *inDir, btuni *name, btuni *linkPath, void *user, beFileSetAttr_t *nfsAttr, void *caller);
	void	(*fileReadlink)(void *obj, void *user, void *caller);
	void	*(*fileParent)(void *obj);
}backendImp_t;

struct nfs3FileAdapter{
	void				*vfs;
	backendImp_t		backend;
	btint				refCount;
	btint				dirLinks;	/* number of links given to directories if not supplied by backend */
	btint				dirSize;	/* faked size for directories */
	struct nfsLookup	*lookupStrategy;
	btbool				changeMicrosecondFileTimes;
};

/* ------------------------------------------------------------------------- */

typedef struct nfs3Ctx{
    PROC_VARSET;
    nfs3FileAdapter_t	*a;
	void				*user;
	void				*file, *otherFile;
	struct entry3		*entryHead;
	struct entry3		**entryTail;
	btint				restLen;
	btu8				*data;
	btint				functionIndex;
	btbool				isDir;
	btbool				doFileLookup;
	nfs3Arg_t			*arg;
	nfs3Res_t			res;
	beFileSetAttr_t		setAttr;
	btu8				fhBuffer[32];
    btuni				uniname[NAME_ARRAY_SIZE];
    btuni				unipath[PATH_ARRAY_SIZE];
}nfs3Ctx_t;

/* ------------------------------------------------------------------------- */

static blockpool_t	*nfs3Pool = NULL;
static blockpool_t	*readdirEntries = NULL;

static inline void	freeEntries3(struct entry3 **headp)
{
struct entry3	*p, *next;
    
    if(*headp != NULL){
        for(p=*headp;p!=NULL;p=next){
			next = p->nextentry;
            memFree(p->name);
            poolReturnBlock(readdirEntries, p);
        }
        *headp = NULL;
    }
}

static void	nfs3CtxFree(void *block)
{
nfs3Ctx_t	*this = block;

	if(this->file != NULL){
		this->a->backend.fileRelease(this->file);
		this->file = NULL;
	}
	if(this->otherFile != NULL){
		this->a->backend.fileRelease(this->otherFile);
		this->otherFile = NULL;
	}
	freeEntries3(&this->entryHead);
    if(this->data != NULL)
        memFree(this->data);
    poolReturnBlock(nfs3Pool, block);
}

/* ------------------------------------------------------------------------- */

static btint	nfs3FaCompare(void *obj, void *otherObj)
{
nfs3FileAdapter_t	*this = obj, *other = otherObj;

	return (btu8 *)this->vfs - (btu8 *)other->vfs;
}

static btu32	nfs3FaHashVal(void *obj)
{
nfs3FileAdapter_t	*this = obj;

	return (btint)this->vfs ^ ((btint)this->vfs >> 8);
}

static collectionObjectI_t	fa3Interace = {
	nfs3FaCompare,
	nfs3FaHashVal,
	NULL,
	NULL
};

static hashtable_t	*allFa3 = NULL;

/* ------------------------------------------------------------------------- */

static void	nfs3FaFree(nfs3FileAdapter_t *fa)
{
	if(fa != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "nfs3FileAdapter: freeing object at 0x%08x\n", (int)fa);
		if(fa->vfs != NULL){
			htabRemoveObjectEqualTo(allFa3, fa);
			fa->backend.vfsRelease(fa->vfs);
			fa->vfs = NULL;
		}
		if(fa->lookupStrategy != NULL){
			luRelease(fa->lookupStrategy);
			fa->lookupStrategy = NULL;
		}
		memFree(fa);
	}
}

/* ------------------------------------------------------------------------- */

static nfs3FileAdapter_t	*nfs3FaNew(struct module *backend, void *vfs, struct cfgProperty **cfgPath, btint numEntries)
{
nfs3FileAdapter_t	*volatile fa;
nfs3Ctx_t			*context = NULL;	/* not really needed... */

    if(nfs3Pool == NULL)
        nfs3Pool = poolNew(sizeof(nfs3Ctx_t), (btu8 *)&context->setAttr - (btu8 *)context);
EXC_TRY
	fa = memCalloc(sizeof(nfs3FileAdapter_t));
	fa->lookupStrategy = luNew(vfs, backend, cfgPath, numEntries);
	fa->backend.vfsRetain = backend->implementationOf(BE_VFS_RETAIN);
	fa->backend.vfsRelease = backend->implementationOf(BE_VFS_RELEASE);
	fa->backend.vfsStatfs = backend->implementationOf(BE_VFS_STATFS);
	fa->backend.vfsConnect = backend->implementationOf(BE_VFS_CONNECT);
	
	fa->backend.fileRetain = backend->implementationOf(BE_FILE_RETAIN);
	fa->backend.fileRelease = backend->implementationOf(BE_FILE_RELEASE);
	fa->backend.fileLookup = backend->implementationOf(BE_FILE_LOOKUP);
	fa->backend.fileGetAttr = backend->implementationOf(BE_FILE_GETATTR);
	fa->backend.fileReaddir = backend->implementationOf(BE_FILE_READDIR);
	fa->backend.fileSetAttr = backend->implementationOf(BE_FILE_SETATTR);
	fa->backend.fileCreate = backend->implementationOf(BE_FILE_CREATE);
	fa->backend.fileRemove = backend->implementationOf(BE_FILE_REMOVE);
	fa->backend.fileRename = backend->implementationOf(BE_FILE_RENAME);
	fa->backend.fileRW = backend->implementationOf(BE_FILE_RW);
	fa->backend.fileHardLink = backend->implementationOf(BE_FILE_HARDLINK);
	fa->backend.fileSymlink = backend->implementationOf(BE_FILE_SYMLINK);
	fa->backend.fileReadlink = backend->implementationOf(BE_FILE_READLINK);
	fa->backend.fileParent = backend->implementationOf(BE_FILE_PARENT);

	fa->vfs = vfs;
	fa->backend.vfsRetain(fa->vfs);
	fa->dirLinks = cfgIntValue(cfgQueryFromArray("fakedDirLinks", numEntries, cfgPath), 2);
	fa->dirSize = cfgIntValue(cfgQueryFromArray("fakedDirSize", numEntries, cfgPath), 1024 * 128);
	fa->changeMicrosecondFileTimes = cfgEnumValue(cfgQueryFromArray("changeMicrosecondFileTimes", numEntries, cfgPath), cfgYesNoEnum, 0);
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "nfs3FileAdapter: allocated new object at 0x%08x\n", (int)fa);
EXC_CATCH
	nfs3FaFree(fa);
	EXC_RETHROW();
EXC_ENDTRY
	return fa;
}

nfs3FileAdapter_t	*nfs3FaForBackend(struct module *backend, void *vfs, struct cfgProperty **cfgPath, btint numEntries)
{
nfs3FileAdapter_t	*fa;

	if(allFa3 == NULL)
		allFa3 = htabNew(&fa3Interace, 8, 75);
	fa = htabFindObjectEqualTo(allFa3, &vfs);
	if(fa == NULL){
		fa = nfs3FaNew(backend, vfs, cfgPath, numEntries);
		htabAddObject(allFa3, fa);
	}
	nfs3FaRetain(fa);
	return fa;
}

void	nfs3FaRetain(nfs3FileAdapter_t *fa)
{
	fa->refCount++;
}

void	nfs3FaRelease(nfs3FileAdapter_t *fa)
{
	if(--fa->refCount == 0){
		nfs3FaFree(fa);
	}else if(fa->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

btinum	nfs3FaRootHandle(nfs3FileAdapter_t *fa, void *destHandle, btu32 mountId)
{
btinum	inum;

	luRootHandle(fa->lookupStrategy, destHandle, &inum, mountId);
	return inum;
}

/* ------------------------------------------------------------------------- */

static int	translateError3(btError *err)
{
    if(err->class != 0 && LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE)){
        logPrintf(LOGM_NFSTRACE, "nfs3: cifs error class=%d code=%d\n", err->class, err->code);
    }
    switch(err->class){
    case ERRCLS_OK:						return NFS3_OK;
    case ERRCLS_CIFS_DOS:
        switch(err->code){
		case ERR_CIFS_DOSbadfunc:		return NFS3ERR_INVAL;
		case ERR_CIFS_DOSbadfile:		return NFS3ERR_NOENT;
		case ERR_CIFS_DOSbadpath:		return NFS3ERR_NOENT;
		case ERR_CIFS_DOSnofids:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_DOSnoaccess:		return NFS3ERR_ACCES;
		case ERR_CIFS_DOSbadfid:		return NFS3ERR_INVAL;
		case ERR_CIFS_DOSbadmcb:		return NFS3ERR_INVAL;
		case ERR_CIFS_DOSnomem:			return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_DOSbadmem:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_DOSbadenv:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_DOSbadformat:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_DOSbadaccess:		return NFS3ERR_ACCES;
		case ERR_CIFS_DOSbaddata:		return NFS3ERR_FBIG;
		case ERR_CIFS_DOSbaddrive:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_DOSremcd:			return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_DOSdiffdevice:	return NFS3ERR_NODEV;
		case ERR_CIFS_DOSnofiles:		return NFS3ERR_NOSPC;
		case ERR_CIFS_DOSbadshare:		return NFS3ERR_ACCES;
		case ERR_CIFS_DOSlock:			return NFS3ERR_ACCES;
		case ERR_CIFS_DOSnotshared:		return NFS3ERR_ACCES;
		case ERR_CIFS_DOSfilexists:		return NFS3ERR_EXIST;
		case ERR_CIFS_DOSnotempty:		return NFS3ERR_NOTEMPTY;
            /* the following are heuristic values derived from smbfs */
		case 87:						return NFS3ERR_IO;
		case ERR_CIFS_DOSdestexists:	return NFS3ERR_EXIST;
        default:						return NFS3ERR_SERVERFAULT;
        }
        break;
    case ERRCLS_CIFS_SRV:
        switch(err->code){
		case ERR_CIFS_SRVerror:			return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVbadpw:			return NFS3ERR_ACCES;
		case ERR_CIFS_SRVaccess:		return NFS3ERR_ACCES;
		case ERR_CIFS_SRVinvnid:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVinvnetname:	return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVinvdevice:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVqfull:			return NFS3ERR_NOSPC;
		case ERR_CIFS_SRVqtoobig:		return NFS3ERR_NAMETOOLONG;
		case ERR_CIFS_SRVqeof:			return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVinvpfid:		return NFS3ERR_INVAL;
		case ERR_CIFS_SRVsmbcmd:		return NFS3ERR_INVAL;
		case ERR_CIFS_SRVsrverror:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVfilespecs:		return NFS3ERR_ACCES;
		case ERR_CIFS_SRVbadpermits:	return NFS3ERR_ACCES;
		case ERR_CIFS_SRVsetattrmode:	return NFS3ERR_ACCES;
		case ERR_CIFS_SRVpaused:		return NFS3ERR_ACCES;
		case ERR_CIFS_SRVmsgoff:		return NFS3ERR_ACCES;
		case ERR_CIFS_SRVnoroom:		return NFS3ERR_NOSPC;
		case ERR_CIFS_SRVrmuns:			return NFS3ERR_IO;
		case ERR_CIFS_SRVtimeout:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVnoresource:	return NFS3ERR_NOSPC;
		case ERR_CIFS_SRVtoomanyuids:	return NFS3ERR_ACCES;
		case ERR_CIFS_SRVbaduid:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVusempx:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVusestd:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVcontmpx:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_SRVnosupport:		return NFS3ERR_SERVERFAULT;
        default:						return NFS3ERR_SERVERFAULT;
        }
        break;
    case ERRCLS_CIFS_HARD:
        switch(err->code){
		case ERR_CIFS_HRDnowrite:		return NFS3ERR_ROFS;
		case ERR_CIFS_HRDbadunit:		return NFS3ERR_NXIO;
		case ERR_CIFS_HRDnotready:		return NFS3ERR_IO;
		case ERR_CIFS_HRDbadcmd:		return NFS3ERR_IO;
		case ERR_CIFS_HRDdata:			return NFS3ERR_IO;
		case ERR_CIFS_HRDbadreq:		return NFS3ERR_IO;
		case ERR_CIFS_HRDseek:			return NFS3ERR_IO;
		case ERR_CIFS_HRDbadmedia:		return NFS3ERR_IO;
		case ERR_CIFS_HRDbadsector:		return NFS3ERR_IO;
		case ERR_CIFS_HRDnopaper:		return NFS3ERR_IO;
		case ERR_CIFS_HRDwrite:			return NFS3ERR_IO;
		case ERR_CIFS_HRDread:			return NFS3ERR_IO;
		case ERR_CIFS_HRDgeneral:		return NFS3ERR_IO;
		case ERR_CIFS_HRDbadshare:		return NFS3ERR_IO;
		case ERR_CIFS_HRDlock:			return NFS3ERR_IO;
		case ERR_CIFS_HRDwrongdisk:		return NFS3ERR_IO;
		case ERR_CIFS_HRDFCBUnavail:	return NFS3ERR_IO;
		case ERR_CIFS_HRDsharebufexc:	return NFS3ERR_IO;
        default:						return NFS3ERR_IO;
        }
        break;
    case ERRCLS_FRAMEWORK:
        switch(err->code){
		case ERR_FRWmodulenotfound:		return NFS3ERR_SERVERFAULT;
		case ERR_FRWmoduletypeerr:		return NFS3ERR_SERVERFAULT;
		case ERR_FRWalreadymounted:		return NFS3ERR_EXIST;
		case ERR_FRWipcparameters:		return NFS3ERR_INVAL;
		case ERR_FRWipcversion:			return NFS3ERR_INVAL;
		case ERR_FRWnomem:				return NFS3ERR_SERVERFAULT;
		case ERR_FRWnosuchobj:			return NFS3ERR_SERVERFAULT;
		case ERR_FRWnosuchmethod:		return NFS3ERR_SERVERFAULT;
		case ERR_FRWumountperm:			return NFS3ERR_PERM;
		case ERR_FRWclientcomm:			return NFS3ERR_IO;
		case ERR_FRWnosuchmnt:			return NFS3ERR_NOENT;
		case ERR_FRWmpaccessperm:		return NFS3ERR_ACCES;
		case ERR_FRWautomounted:		return NFS3ERR_ACCES;
		case ERR_FRWipcnotimpl:			return NFS3ERR_INVAL;
		case ERR_FRWusrabort:			return NFS3ERR_ACCES;
        default:						return NFS3ERR_IO;
        }
        break;
    case ERRCLS_CIFS_LOCAL:
        switch(err->code){
		case ERR_CIFS_LOCconnlost:		return -1;
		case ERR_CIFS_LOCtimeout:		return NFS3ERR_IO;
		case ERR_CIFS_LOCbogusdata:		return NFS3ERR_INVAL;
		case ERR_CIFS_LOCstale:			return NFS3ERR_STALE;
		case ERR_CIFS_LOCpwencrypt:		return NFS3ERR_ACCES;
		case ERR_CIFS_LOClookup:		return NFS3ERR_NOENT;
		case ERR_CIFS_LOClicclients:	return NFS3ERR_PERM;
		case ERR_CIFS_LOClicservers:	return NFS3ERR_PERM;
		case ERR_CIFS_LOClicense:		return NFS3ERR_PERM;
		case ERR_CIFS_LOCsrvconn:		return NFS3ERR_SERVERFAULT;
		case ERR_CIFS_LOCwrongphrase:	return NFS3ERR_ACCES;
		case ERR_CIFS_LOCneedpassphrase: return NFS3ERR_ACCES;
		case ERR_CIFS_LOCnetbiosrefused: return NFS3ERR_ACCES;
		case ERR_CIFS_LOChavenossl:		return NFS3ERR_ACCES;
		case ERR_CIFS_LOCnosslserver:	return NFS3ERR_ACCES;
		case ERR_CIFS_LOCsslerror:		return NFS3ERR_IO;
		case ERR_CIFS_LOCnomountdir:	return NFS3ERR_IO;
		case ERR_CIFS_LOCnobrowser:		return NFS3ERR_ACCES;
		case ERR_CIFS_LOCotherlogin:	return NFS3ERR_ACCES;
        default:						return NFS3ERR_IO;
        }
        break;
    case ERRCLS_UNIX:
        switch(err->code){
		case 0:							return NFS3_OK;
		case EPERM:						return NFS3ERR_PERM;
		case ENOENT:					return NFS3ERR_NOENT;
		case ENXIO:						return NFS3ERR_NXIO;
		case ETXTBSY:
		case EWOULDBLOCK:
		case EACCES:					return NFS3ERR_ACCES;
#if EEXIST != ENOTEMPTY
		case EEXIST:					return NFS3ERR_EXIST;
#endif
		case EXDEV:						return NFS3ERR_XDEV;
		case ENODEV:					return NFS3ERR_NODEV;
		case ENOTDIR:					return NFS3ERR_NOTDIR;
		case EISDIR:					return NFS3ERR_ISDIR;
		case EINVAL:					return NFS3ERR_INVAL;
		case E2BIG:
		case EFBIG:						return NFS3ERR_FBIG;
		case ENOSPC:					return NFS3ERR_NOSPC;
		case EROFS:						return NFS3ERR_ROFS;
#ifdef EMLINK
		case EMLINK:					return NFS3ERR_MLINK;
#endif
		case ENAMETOOLONG:				return NFS3ERR_NAMETOOLONG;
		case ENOTEMPTY:					return NFS3ERR_NOTEMPTY;
#ifdef EDQUOT
		case EDQUOT:					return NFS3ERR_DQUOT;
#endif
#ifdef ETIMEDOUT
		case ETIMEDOUT:					return NFS3ERR_IO;
#endif
		default:						return NFS3ERR_ACCES;
        }
        break;
    default:					return NFS3ERR_ACCES;
    }
}

/* ------------------------------------------------------------------------- */

static void	translateReadAttributes3(nfs3FileAdapter_t *fa, struct fattr3 *nfsAttr, beFileAttr_t *fileAttr, btinum inum)
{
static btu32	microsecondDeviation = 0;

	nfsAttr->type = fileAttr->type;
	nfsAttr->mode = fileAttr->unixMode & 07777;
	if(fileAttr->numLinks != 0){
		nfsAttr->nlink = fileAttr->numLinks;
	}else if(fileAttr->type == BE_FILETYPE_DIR){
		nfsAttr->nlink = fa->dirLinks;
	}else{
		nfsAttr->nlink = 1;
	}
	nfsAttr->uid = fileAttr->uid;
	nfsAttr->gid = fileAttr->gid;
	if(fileAttr->type == BE_FILETYPE_DIR){
		nfsAttr->size = fileAttr->size == 0 ? fa->dirSize : fileAttr->size;
	}else{
		nfsAttr->size = fileAttr->size;
	}
	/* alloc size for directories should be at least 1... */
	if(fileAttr->type == BE_FILETYPE_DIR){
		nfsAttr->used = fileAttr->allocatedSize < 512 ? 512 : fileAttr->allocatedSize;
	}else{
		nfsAttr->used = nfsAttr->used;
	}
	if(fileAttr->type == BE_FILETYPE_BLK || fileAttr->type == BE_FILETYPE_CHR){
		nfsAttr->rdev.specdata1 = fileAttr->devMajor;
		nfsAttr->rdev.specdata2 = fileAttr->devMinor;
	}else{
		nfsAttr->rdev.specdata1 = nfsAttr->rdev.specdata2 = 0;
	}
	nfsAttr->fsid = 0;
	nfsAttr->fileid = inum;
	nfsAttr->atime.seconds = fileAttr->atime.seconds;
	nfsAttr->atime.nseconds = fileAttr->atime.nseconds;
	nfsAttr->mtime.seconds = fileAttr->mtime.seconds;
	nfsAttr->mtime.nseconds = fileAttr->mtime.nseconds;
	if(fa->changeMicrosecondFileTimes){
		nfsAttr->mtime.nseconds += microsecondDeviation * 1000;
		microsecondDeviation = (microsecondDeviation + 1) % 10000;
	}
	nfsAttr->ctime.seconds = fileAttr->ctime.seconds;
	nfsAttr->ctime.nseconds = fileAttr->ctime.nseconds;

#if defined SOLARIS2	/* Solaris 7 is not year 2038 safe.... ### verify for NFS3 */
	nfsAttr->atime.seconds &= 0x7fffffff;
	nfsAttr->mtime.seconds &= 0x7fffffff;
	nfsAttr->ctime.seconds &= 0x7fffffff;
#endif
}

static void	buildFattr(struct fattr3 *nfsAttr, nfs3FileAdapter_t *fa, void *handle, beFileAttr_t *fileAttr)
{
btinum	inum;

	luHandleOfFile(fa->lookupStrategy, handle, NULL, NULL, &inum);
	translateReadAttributes3(fa, nfsAttr, fileAttr, inum);
}

static void	buildWccData(struct wcc_data *wcc, nfs3FileAdapter_t *fa, void *handle, beFileAttr_t *fileAttr)
{
	buildFattr(&wcc->after.attributes, fa, handle, fileAttr);
	wcc->after.attributes_follow = 1;
	wcc->before.attributes_follow = 0;
}

static void	translateWriteAttributes3(beFileSetAttr_t *fileAttr, struct sattr3 *nfsAttr)
{
struct timeval	now[1];
struct timezone	tz[1];
btbool			timeValid = 0;

	fileAttr->setUnixMode = nfsAttr->mode.set;
	fileAttr->unixMode = nfsAttr->mode.val;
	fileAttr->setUid = nfsAttr->uid.set;
	fileAttr->uid = nfsAttr->uid.val;
	fileAttr->setGid = nfsAttr->gid.set;
	fileAttr->gid = nfsAttr->gid.val;
	fileAttr->setSize = nfsAttr->size.set;
	fileAttr->size = nfsAttr->size.val;
	if((fileAttr->setAtime = (nfsAttr->atime.mode != NFS3_TIMESET_DONT_CHANGE))){
		if(nfsAttr->atime.mode == NFS3_TIMESET_SET_TO_CLIENT_TIME){
			fileAttr->atime.seconds = nfsAttr->atime.time.seconds;
			fileAttr->atime.nseconds = nfsAttr->atime.time.nseconds;
		}else{
			if(!timeValid)
				gettimeofday(now, tz);
			fileAttr->atime.seconds = now->tv_sec;
			fileAttr->atime.nseconds = now->tv_usec * 1000;
		}
	}
	if((fileAttr->setMtime = (nfsAttr->mtime.mode != NFS3_TIMESET_DONT_CHANGE))){
		if(nfsAttr->mtime.mode == NFS3_TIMESET_SET_TO_CLIENT_TIME){
			fileAttr->mtime.seconds = nfsAttr->mtime.time.seconds;
			fileAttr->mtime.nseconds = nfsAttr->mtime.time.nseconds;
		}else{
			if(!timeValid)
				gettimeofday(now, tz);
			fileAttr->mtime.seconds = now->tv_sec;
			fileAttr->mtime.nseconds = now->tv_usec * 1000;
		}
	}
}

/* all structs in union have 'status' as first entry, take getattr as example */
#define	CHECK_FH(fh)	\
	if((fh).len != 32){									\
		this->res.getattr.status = NFS3ERR_BADHANDLE;	\
		goto doReturn;									\
	}

/* ------------------------------------------------------------------------- */

static void	nfs3NULL(void *context, void *arg)
{
nfs3Ctx_t	*this = context;

	PROC_BEGIN;
	PROC_END;
}

/* ------------------------------------------------------------------------- */

static void	nfs3GETATTR(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;

    PROC_BEGIN
	CHECK_FH(this->arg->getattr)
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->getattr.data, this->user, 0, this));
    if((this->res.getattr.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
    this->file = rval->vfsFile;
	PROC_WAITFOR_X(2, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((this->res.getattr.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	buildFattr(&this->res.getattr.attributes, this->a, this->arg->getattr.data, frval->attr);
doReturn:
    PROC_RETURN(&this->res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3SETATTR(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;

	PROC_BEGIN
	CHECK_FH(this->arg->setattr.object)
	if(this->a->backend.fileSetAttr == NULL){
		this->res.setattr.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->setattr.object.data, this->user, 0, this));
    if((this->res.setattr.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
    this->file = rval->vfsFile;
	translateWriteAttributes3(&this->setAttr, &this->arg->setattr.new_attributes);
    PROC_WAITFOR_X(2, this->a->backend.fileSetAttr(this->file, this->user, &this->setAttr, this));
    if((this->res.setattr.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((this->res.setattr.status = translateError3(&frval->err)) != NFS3_OK)
        goto doReturn;
	buildWccData(&this->res.setattr.obj_wcc, this->a, this->arg->setattr.object.data, frval->attr);
doReturn:
    PROC_RETURN(&this->res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3LOOKUP(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
void		*parentFile;
btinum		inum;

    PROC_BEGIN
	CHECK_FH(this->arg->lookup.dir)
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->lookup.dir.data, this->user, 0, this));
    if((this->res.lookup.status = translateError3(&rval->err)) != NFS3_OK)
        goto doReturn;
	this->file = rval->vfsFile;
	this->doFileLookup = 1;
	if(this->arg->lookup.name[0] == '.'){	/* special treatment for '.' and '..' */
		if(this->arg->lookup.name[1] == 0){
			this->doFileLookup = 0;	/* it's just the current file */
		}else if(this->arg->lookup.name[1] == '.' && this->arg->lookup.name[2] == 0){
			if((parentFile = this->a->backend.fileParent(this->file)) != NULL){
				this->a->backend.fileRetain(parentFile);
				this->a->backend.fileRelease(this->file);
				this->file = parentFile;
			}
			/* if parent was NULL, we must deal with the root handle */
			this->doFileLookup = 0;
		}
	}
	convLocalAsciiToUni(this->arg->lookup.name, this->uniname, NAME_ARRAY_SIZE);
	if(this->doFileLookup){
		PROC_WAITFOR_X(2, this->a->backend.fileLookup(this->file, this->uniname, this->user, this));
		if((this->res.lookup.status = translateError3(&frval->err)) != NFS3_OK)
			goto doReturn;
		luCacheFile(this->a->lookupStrategy, frval->file);
		this->a->backend.fileRelease(this->file);
		this->file = frval->file;
	}
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
	if((this->res.lookup.status = translateError3(&frval->err)) != NFS3_OK)
        goto doReturn;
	this->res.lookup.object.len = 32;
	this->res.lookup.object.data = this->fhBuffer;
	luHandleOfFile(this->a->lookupStrategy, this->arg->lookup.dir.data, this->uniname, this->fhBuffer, &inum);
	this->res.lookup.obj_attributes.attributes_follow = 1;
	translateReadAttributes3(this->a, &this->res.lookup.obj_attributes.attributes, frval->attr, inum);
doReturn:
    PROC_RETURN(&this->res.lookup);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3ACCESS(void *context, void *arg)
{
nfs3Ctx_t	*this = context;

	PROC_BEGIN
	this->res.access.access = this->arg->access.access;
	/* always assume all rights since this call is advisory only */
	PROC_RETURN(&this->res.access);
	PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3READLINK(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
char		path[PATH_ARRAY_SIZE];

	PROC_BEGIN
	CHECK_FH(this->arg->readlink)
	if(this->a->backend.fileReadlink == NULL){
		this->res.readlink.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->readlink.data, this->user, 0, this));
    if((this->res.readlink.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
    this->file = rval->vfsFile;
    PROC_WAITFOR_X(2, this->a->backend.fileReadlink(this->file, this->user, this));
    if((this->res.readlink.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
    convUniToLocalAscii(frval->path, path, PATH_ARRAY_SIZE);
    this->res.readlink.data = path;
doReturn:
    PROC_RETURN(&this->res.readlink);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3READ(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;

	PROC_BEGIN
	CHECK_FH(this->arg->read.file)
	if(this->a->backend.fileRW == NULL){
		this->res.read.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->read.file.data, this->user, 1, this));
    if((this->res.read.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
	this->file = rval->vfsFile;
    this->data = memAlloc(this->arg->read.count);	/* will be freed in context free */
    PROC_WAITFOR_X(2, this->a->backend.fileRW(this->file, BE_FILE_OP_READ, 1, this->data, this->arg->read.offset, this->arg->read.count, this->user, this));
    if((this->res.read.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
    this->res.read.data.len = frval->dataLen;
    this->res.read.data.data = this->data;
	this->res.read.count = frval->dataLen;	/* why this redundant variable?? */
	this->res.read.eof = frval->eof;
	/* we do not set file_attributes because this might require a round trip to the server */
doReturn:
    PROC_RETURN(&this->res.read);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3WRITE(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;

    PROC_BEGIN
	CHECK_FH(this->arg->write.file)
	if(this->a->backend.fileRW == NULL){
		this->res.write.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->write.file.data, this->user, 1, this));
    if((this->res.write.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
	this->file = rval->vfsFile;
	if(this->arg->write.data.len > this->arg->write.count)
		this->arg->write.data.len = this->arg->write.count;
    PROC_WAITFOR_X(2, this->a->backend.fileRW(this->file, BE_FILE_OP_WRITE, 1, this->arg->write.data.data, this->arg->write.offset, this->arg->write.data.len, this->user, this));
    if((this->res.write.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	/* we do not set any attributes because this might require a round trip to the server */
	this->res.write.count = frval->dataLen;
	this->res.write.committed = NFS3_STABLE_FILE_SYNC;	/* we have not implemented writeback */
	/* verf is implicitly zero */
doReturn:
	PROC_RETURN(&this->res.write);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3CREATE(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
btinum		inum;

	PROC_BEGIN
	CHECK_FH(this->arg->create.where.dir)
	if(this->a->backend.fileCreate == NULL || !(this->arg->create.mode == NFS3_CREATE_UNCHECKED || this->arg->create.mode == NFS3_CREATE_GUARDED)){
		/* we don't support exclusive mode for now... */
		this->res.create.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->create.where.dir.data, this->user, 1, this));
    if((this->res.create.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
    this->file = rval->vfsFile;
    convLocalAsciiToUni(this->arg->create.where.name, this->uniname, NAME_ARRAY_SIZE);
	translateWriteAttributes3(&this->setAttr, &this->arg->create.u.guarded_or_unchecked.obj_attributes);
	/* ### introduce "exclusive" flag in backend... */
    PROC_WAITFOR_X(2, this->a->backend.fileCreate(this->file, this->uniname, 0, this->user, &this->setAttr, this));
    if((this->res.create.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	this->a->backend.fileRelease(this->file);
    this->file = frval->file;
	luCacheFile(this->a->lookupStrategy, frval->file);
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((this->res.create.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	luHandleOfFile(this->a->lookupStrategy, this->arg->create.where.dir.data, this->uniname, this->fhBuffer, &inum);
	translateReadAttributes3(this->a, &this->res.create.obj_attributes.attributes, frval->attr, inum);
	this->res.create.obj_attributes.attributes_follow = 1;
	this->res.create.obj.handle_follows = 1;
	this->res.create.obj.handle.len = 32;
	this->res.create.obj.handle.data = this->fhBuffer;
doReturn:
    PROC_RETURN(&this->res.create);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3MKDIR(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
btinum		inum;

	PROC_BEGIN
	CHECK_FH(this->arg->mkdir.where.dir)
	if(this->a->backend.fileCreate == NULL){
		this->res.mkdir.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->create.where.dir.data, this->user, 1, this));
    if((this->res.mkdir.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
    this->file = rval->vfsFile;
    convLocalAsciiToUni(this->arg->mkdir.where.name, this->uniname, NAME_ARRAY_SIZE);
	translateWriteAttributes3(&this->setAttr, &this->arg->mkdir.attributes);
    PROC_WAITFOR_X(2, this->a->backend.fileCreate(this->file, this->uniname, 1, this->user, &this->setAttr, this));
    if((this->res.mkdir.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	this->a->backend.fileRelease(this->file);
    this->file = frval->file;
	luCacheFile(this->a->lookupStrategy, frval->file);
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((this->res.mkdir.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	luHandleOfFile(this->a->lookupStrategy, this->arg->create.where.dir.data, this->uniname, this->fhBuffer, &inum);
	this->res.mkdir.obj.handle_follows = 1;
	this->res.mkdir.obj.handle.len = 32;
	this->res.mkdir.obj.handle.data = this->fhBuffer;
	translateReadAttributes3(this->a, &this->res.mkdir.obj_attributes.attributes, frval->attr, inum);
	this->res.mkdir.obj_attributes.attributes_follow = 1;
doReturn:
    PROC_RETURN(&this->res.mkdir);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3SYMLINK(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
btinum		inum;

	PROC_BEGIN
	CHECK_FH(this->arg->symlink.where.dir)
	if(this->a->backend.fileSymlink == NULL){
		this->res.symlink.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->symlink.where.dir.data, this->user, 0, this));
    if((this->res.symlink.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
    this->file = rval->vfsFile;
    convLocalAsciiToUni(this->arg->symlink.where.name, this->uniname, NAME_ARRAY_SIZE);
    convLocalAsciiToUni(this->arg->symlink.symlink_data, this->unipath, PATH_ARRAY_SIZE);
	translateWriteAttributes3(&this->setAttr, &this->arg->symlink.symlink_attributes);
    PROC_WAITFOR_X(2, this->a->backend.fileSymlink(this->file, this->uniname, this->unipath, this->user, &this->setAttr, this));
    if((this->res.symlink.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	this->a->backend.fileRelease(this->file);
    this->file = frval->file;
	luCacheFile(this->a->lookupStrategy, frval->file);
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((this->res.symlink.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	luHandleOfFile(this->a->lookupStrategy, this->arg->symlink.where.dir.data, this->uniname, this->fhBuffer, &inum);
	this->res.symlink.obj.handle_follows = 1;
	this->res.symlink.obj.handle.len = 32;
	this->res.symlink.obj.handle.data = this->fhBuffer;
	translateReadAttributes3(this->a, &this->res.symlink.obj_attributes.attributes, frval->attr, inum);
	this->res.symlink.obj_attributes.attributes_follow = 1;
doReturn:
    PROC_RETURN(&this->res.symlink);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3MKNOD(void *context, void *arg)
{
nfs3Ctx_t	*this = context;

	PROC_BEGIN
	this->res.mknod.status = NFS3ERR_NOTSUPP;
	PROC_RETURN(&this->res.mknod);
	PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3RemoveAny(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;

	PROC_BEGIN
	CHECK_FH(this->arg->remove.dir)
	if(this->a->backend.fileRemove == NULL){
		this->res.remove.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->remove.dir.data, this->user, 0, this));
    if((this->res.remove.status = translateError3(&rval->err)) != NFS3_OK)
        goto doReturn;
	this->file = rval->vfsFile;
    convLocalAsciiToUni(this->arg->remove.name, this->uniname, NAME_ARRAY_SIZE);
    PROC_WAITFOR_X(2, this->a->backend.fileRemove(this->file, this->uniname, this->isDir, this->user, this));
	if((this->res.remove.status = translateError3(&frval->err)) != NFS3_OK)
        goto doReturn;
	luMarkStale(this->a->lookupStrategy, this->arg->remove.dir.data, this->uniname);
	/* don't provide dir attributes to avoid extra round trip to server */
doReturn:
    PROC_RETURN(&this->res.remove);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3REMOVE(void *context, void *arg)
{
nfs3Ctx_t	*this = context;

	this->isDir = 0;
    this->proc.process = nfs3RemoveAny;	/* overload process */
	nfs3RemoveAny(this, arg);
}

/* ------------------------------------------------------------------------- */

static void	nfs3RMDIR(void *context, void *arg)
{
nfs3Ctx_t	*this = context;

	this->isDir = 1;
    this->proc.process = nfs3RemoveAny;	/* overload process */
	nfs3RemoveAny(this, arg);
}

/* ------------------------------------------------------------------------- */

static void	nfs3RENAME(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;

	PROC_BEGIN
	CHECK_FH(this->arg->rename.from.dir)
	CHECK_FH(this->arg->rename.to.dir)
	if(this->a->backend.fileRename == NULL){
		this->res.rename.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->rename.from.dir.data, this->user, 0, this));
    if((this->res.rename.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
    this->file = rval->vfsFile;
	PROC_WAITFOR_X(2, luLookupFile(this->a->lookupStrategy, this->arg->rename.to.dir.data, this->user, 0, this));
    if((this->res.rename.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
	this->otherFile = rval->vfsFile;
	convLocalAsciiToUni(this->arg->rename.from.name, this->uniname, NAME_ARRAY_SIZE);
    convLocalAsciiToUni(this->arg->rename.to.name, this->unipath, PATH_ARRAY_SIZE);
    PROC_WAITFOR_X(3, this->a->backend.fileRename(this->file, this->uniname, this->otherFile, this->unipath, this->user, this));
	if((this->res.rename.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	luRename(this->a->lookupStrategy, this->arg->rename.from.dir.data, this->uniname, this->arg->rename.to.dir.data, this->unipath);
	/* don't return any dir attributes to avoid additional round-trips to server */
doReturn:
    PROC_RETURN(&this->res.rename);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3LINK(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;

	PROC_BEGIN
	CHECK_FH(this->arg->link.file)
	CHECK_FH(this->arg->link.link.dir)
	if(this->a->backend.fileHardLink == NULL){
		this->res.link.status = NFS3ERR_NOTSUPP;
		goto doReturn;
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->link.file.data, this->user, 0, this));
    if((this->res.link.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
	this->file = rval->vfsFile;
	PROC_WAITFOR_X(2, luLookupFile(this->a->lookupStrategy, this->arg->link.link.dir.data, this->user, 0, this));
    if((this->res.link.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
	this->otherFile = rval->vfsFile;
    convLocalAsciiToUni(this->arg->link.link.name, this->uniname, NAME_ARRAY_SIZE);
    PROC_WAITFOR_X(3, this->a->backend.fileHardLink(this->file, this->otherFile, this->uniname, this->user, this));
    if((this->res.link.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	this->a->backend.fileRelease(frval->file);
	/* return neither dir nor link attributes to avoid additional round trips */
doReturn:
    PROC_RETURN(&this->res.link);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static btbool	readdir3IterationCallback(void *context, btuni *fileName, btu32 continuationIndex, btbool isDir)
{
nfs3Ctx_t		*this = context;
char			name[NAME_ARRAY_SIZE];
btsize			slen, nlen, len;
struct entry3	*e;
btinum			inum;

    convUniToLocalAscii(fileName, name, NAME_ARRAY_SIZE);
    if(name[0] == '.' && (name[1] == 0 || (name[1] == '.' && name[2] == 0))){
        return 1;	/* skip '.' and '..' entries because we have already faked them */
    }
	slen = strlen(name);
    nlen = (slen/4 + 1) * 4;	/* rounded-up XDR length */
    len = nlen + sizeof(struct entry3);	/* length of this entry in XDR */
    if(len > this->restLen)
        return 0;						/* don't accept this entry, break */
    this->restLen -= len;
    e = poolGetBlock(readdirEntries);
    *(this->entryTail) = e;
	luHandleOfFile(this->a->lookupStrategy, this->arg->readdir.dir.data, fileName, NULL, &inum);
    e->fileid = inum;
    e->name = memAlloc(nlen);
    memcpy(e->name, name, slen + 1);
	e->cookie = continuationIndex + 2;	/* add an offset of 2 for faked '.' and '..' entries */
    this->entryTail = &(e->nextentry);
    return 1;
}

static void	nfs3READDIR(void *context, void *arg)
{
nfs3Ctx_t		*this = context;
luRes_t			*rval = arg;
beFileRes_t		*frval = arg;
btu32			startIndex;
struct entry3	*e;
btinum			inum;
btuni			fname[3];

	PROC_BEGIN
	CHECK_FH(this->arg->readdir.dir)
    if(readdirEntries == NULL)
        readdirEntries = poolNew(sizeof(struct entry3), sizeof(struct entry3));
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, this->arg->readdir.dir.data, this->user, 0, this));
    if((this->res.readdir.status = translateError3(&rval->err)) != NFS3_OK)
        goto doReturn;
    this->file = rval->vfsFile;
    startIndex = this->arg->readdir.cookie;
    this->restLen = this->arg->readdir.count;
    this->entryTail = &(this->entryHead);
/* fake '.' and '..' entries because they _must_ be correct and we can't rely on the server */
    if(startIndex == 0){	/* fake '.' entry */
        e = poolGetBlock(readdirEntries);
        *(this->entryTail) = e;
		fname[0] = '.';
		fname[1] = 0;
		luHandleOfFile(this->a->lookupStrategy, this->arg->readdir.dir.data, fname, NULL, &inum);
        e->fileid = inum;
        e->name = memAlloc(2);
        strcpy(e->name, ".");
		e->cookie = ++startIndex;
        this->entryTail = &(e->nextentry);
        this->restLen -= sizeof(struct entry3) + 4;	/* 4 is the rounded string length */
    }
    if(startIndex == 1){	/* fake '..' entry */
        e = poolGetBlock(readdirEntries);
        *(this->entryTail) = e;
		fname[0] = fname[1] = '.';
		fname[2] = 0;
		luHandleOfFile(this->a->lookupStrategy, this->arg->readdir.dir.data, fname, NULL, &inum);
        e->fileid = inum;
        e->name = memAlloc(3);
        strcpy(e->name, "..");
		e->cookie = ++startIndex;
        this->entryTail = &(e->nextentry);
        this->restLen -= sizeof(struct entry3) + 4;	/* 4 is the rounded string length */
    }
    startIndex -= 2;	/* correct startIndex by the . and .. entries */
    PROC_WAITFOR_X(2, this->a->backend.fileReaddir(this->file, this->user, startIndex, readdir3IterationCallback, this));
    if((this->res.readdir.status = translateError3(&frval->err)) != NFS3_OK)
		goto doReturn;
	this->res.readdir.entries = this->entryHead;
	this->res.readdir.eof = frval->eof;
	/* no dir attributes returned to avoid round-trip */
	/* cookieverf is implicitly 0 */
doReturn:
    PROC_RETURN(&this->res.readdir);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3READDIRPLUS(void *context, void *arg)
{
nfs3Ctx_t	*this = context;

	PROC_BEGIN
	this->res.readdirplus.status = NFS3ERR_NOTSUPP;
	/* We don't support readdirplus for the moment because it does not fit
	 * well with the backend concept. Although all files returned SHOULD
	 * be cached with valid attributes, the call semantics allows for a
	 * wait in the getattr() call in the backend. The readdir callback
	 * can thus not fetch the attributes easily */
	PROC_RETURN(&this->res.mknod);
	PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3FSSTAT(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
beVfsRes_t	*rval = arg;

    PROC_BEGIN
	CHECK_FH(this->arg->fsstat);
    PROC_WAITFOR_X(1, this->a->backend.vfsStatfs(this->a->vfs, this->user, this));
    if((this->res.fsstat.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
	this->res.fsstat.tbytes = (btu64)rval->totalAU * rval->bytesPerAU;
	this->res.fsstat.fbytes = (btu64)rval->availableAU * rval->bytesPerAU;
	this->res.fsstat.abytes = this->res.fsstat.fbytes;
	this->res.fsstat.tfiles = 1000000 + rval->totalAU / 2;	/* an estimate... */
	this->res.fsstat.ffiles = 1000000;
	this->res.fsstat.afiles = this->res.fsstat.ffiles;
	this->res.fsstat.invarsec = 1;	/* also an estimate, assume relatively busy FS */
	/* attributes are left blank */
doReturn:
    PROC_RETURN(&this->res.fsstat);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3FSINFO(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
beVfsRes_t	*rval = arg;

    PROC_BEGIN
	CHECK_FH(this->arg->fsinfo);
    PROC_WAITFOR_X(1, this->a->backend.vfsStatfs(this->a->vfs, this->user, this));
    if((this->res.fsinfo.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
	/* note that nfs3.c / fakedReply3() copies these values! */
	this->res.fsinfo.rtmax = 16384;	/* this is 4 cache pages with NT4 */
	this->res.fsinfo.rtpref = 8192;	/* 2 cache pages with NT4 */
	this->res.fsinfo.rtmult = 4096;	/* suggested multiple in cache pages for NT4 */
	this->res.fsinfo.wtmax = 16384;	/* same as above for write... */
	this->res.fsinfo.wtpref = 8192;
	this->res.fsinfo.wtmult = 4096;
	this->res.fsinfo.dtpref = 4096;	/* preferred readdir size */
	this->res.fsinfo.maxfilesize = (btu64)0xffffffff << 30;	/* virtually unlimited... */
	this->res.fsinfo.time_delta.seconds = rval->timeGranularity.seconds;
	this->res.fsinfo.time_delta.nseconds = rval->timeGranularity.nseconds;
	if(rval->features & BE_VFS_CAN_HARDLINK)
		this->res.fsinfo.properties |= NFS3_FEATURE_LINK;
	if(rval->features & BE_VFS_CAN_SYMLINK)
		this->res.fsinfo.properties |= NFS3_FEATURE_SYMLINK;
	if(rval->features & BE_VFS_CAN_SETTIMES)
		this->res.fsinfo.properties |= NFS3_FEATURE_CANSETTIME;
	this->res.fsinfo.properties |= NFS3_FEATURE_HOMOGENEOUS;	/* pathconf is implemented FS-wide */
	/* attributes are left blank */
doReturn:
    PROC_RETURN(&this->res.fsinfo);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3PATHCONF(void *context, void *arg)
{
nfs3Ctx_t	*this = context;
beVfsRes_t	*rval = arg;

    PROC_BEGIN
	CHECK_FH(this->arg->pathconf);
    PROC_WAITFOR_X(1, this->a->backend.vfsStatfs(this->a->vfs, this->user, this));
    if((this->res.pathconf.status = translateError3(&rval->err)) != NFS3_OK)
		goto doReturn;
	this->res.pathconf.linkmax = 65535;	/* reasonable default value */
	this->res.pathconf.name_max = NAME_ARRAY_SIZE - 1;
	this->res.pathconf.no_trunc = 0;	/* we silently truncate names */
	this->res.pathconf.chown_restricted = 1;	/* unix disallows chown for non-root */
	this->res.pathconf.case_insensitive = (rval->features & BE_VFS_IS_CASE_SENSITIVE) == 0;
	this->res.pathconf.case_preserving = (rval->features & BE_VFS_IS_CASE_PRESERVING) != 0;
	/* attributes are left blank */
doReturn:
    PROC_RETURN(&this->res.pathconf);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfs3COMMIT(void *context, void *arg)
{
nfs3Ctx_t	*this = context;

	PROC_BEGIN
	this->res.commit.status = NFS3_OK;
	/* return no attributes */
	/* verf is implicitly 0, consistent with implementation in nfs3WRITE() */
	PROC_RETURN(&this->res.commit);
	PROC_END
}

/* ------------------------------------------------------------------------- */

static btcallback	nfs3Funcs[] = {
	nfs3NULL,
	nfs3GETATTR,
	nfs3SETATTR,
	nfs3LOOKUP,
	nfs3ACCESS,
	nfs3READLINK,
	nfs3READ,
	nfs3WRITE,
	nfs3CREATE,
	nfs3MKDIR,
	nfs3SYMLINK,
	nfs3MKNOD,
	nfs3REMOVE,
	nfs3RMDIR,
	nfs3RENAME,
	nfs3LINK,
	nfs3READDIR,
	nfs3READDIRPLUS,
	nfs3FSSTAT,
	nfs3FSINFO,
	nfs3PATHCONF,
	nfs3COMMIT
};

/* ------------------------------------------------------------------------- */

static void	nfs3FaExecuteRequestProc(void *ctx, void *arg)
{
nfs3Ctx_t	*this = ctx;
beVfsRes_t	*rval = arg;

	PROC_BEGIN;
	PROC_WAITFOR_X(1, this->a->backend.vfsConnect(this->a->vfs, this->user, this));
	if((this->res.getattr.status = translateError3(&rval->err)) != NFS3_OK){
		PROC_RETURN(&this->res);
	}
    this->proc.currentState = 0;	/* overload current process */
	this->proc.process = nfs3Funcs[this->functionIndex];
	PROC_START(this, NULL);
	return;	/* as a consequence of the overload, terminate without callback */
    PROC_END
}


void	nfs3FaExecuteRequest(nfs3FileAdapter_t *fa, btint functionIndex, void *arg, void *user, void *caller)
{
nfs3Ctx_t	*ctx;

    PROC_INIT(ctx, nfs3FaExecuteRequestProc, caller, nfs3Pool, nfs3CtxFree);
	ctx->a = fa;
	ctx->arg = arg;
	ctx->user = user;
	ctx->functionIndex = functionIndex;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

