/*
 * Name: diskarb.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 2001-01-20
 * Tabsize: 4
 * Copyright: (c) 2001 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: diskarb.c,v 1.6 2002/04/05 17:53:50 cs Exp $
 */

#include "diskarb.h"
#include "process.h"
#include "errcodes.h"

#ifdef HAVE_DISKARB

#include "stdheaders.h"
#include <sys/mount.h>
#include <pthread.h>
#include <DiskArbitration/DiskArbitration.h>
#include "logging.h"
#include "evntloop.h"
#include "memalloc.h"
#include "system.h"
#include "blockpool.h"
#include "mainmodule.h"
#include "hashtable.h"
#include "util.h"

/*
Remarks:
This module contains functions only needed on Apple MacOS X and Darwin
systems. Darwin defines a global database for mounts managed by the
autodiskmount daemon. It is mandatory to communicate with this daemon
because it notifies all applications about unmount attempts, giving them
a chance to close all open files and release other resources they have
open on the mountpoint.

Unfortunately, autodiskmount is somewhat disk-centric. Its service consists
not only of managing the notifications, it also wants to do the mounting
and unmounting. There IS a function to inform it about mounts we established
with the mount() system call, but unmount() comes only in a package
consisting of notify + unmount + delete mountpoint directory.
*/

/* ------------------------------------------------------------------------- */

typedef struct diskarbCtx{
    PROC_VARSET;
	btu8	*mountpoint;
	btu8	*device;
	btu8	*s;		/* intermediate string storage */
	void	*info;	/* intermediate storage place */
}diskarbCtx_t;

static blockpool_t *diskarbPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(diskarbPool, block);
}

static procCond_t	*umountSignal;

typedef struct umountInfo{
	void	*timeoutThread;
	btu8	*diskId;
	btint	error;
	btint	dissenter;
}umountInfo_t;

static hashtable_t	*runningUmounts;

static btint	umountCompare(void *o1, void *o2)
{
btu8	*this = o1, *other = o2;

	return strcmp(this, other);
}

static btu32	umountHashVal(void *obj)
{
	return utilStrHash(obj, 1);
}

collectionObjectI_t	runningUmountsInterface = {
	umountCompare,
	umountHashVal,
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

static void	*diskarbThreadUnmountPreNotify0(void *diskId)
{
	DiskArbUnmountPreNotifyAck_async_auto(diskId, 0);	/* RPC is one-way */
	memFree(diskId);
	return NULL;
}

static void	*diskarbThreadUnmountPreNotifyErr(void *diskId)
{
	DiskArbUnmountPreNotifyAck_async_auto(diskId, EACCES);	/* RPC is one-way */
	memFree(diskId);
	return NULL;
}

static void	unmountPreNotification(DiskArbDiskIdentifier diskIdentifier, unsigned flags)
{
pthread_t	thread;
btint		err = 0;

	if(LOG_MODE_IS_ACTIVE(LOGM_DISKARB))
		logPrintf(LOGM_DISKARB, "diskarb: unmountPreNotification in pid=%d for %s\n", getpid(), diskIdentifier);
	if(htabFindObjectEqualTo(runningUmounts, diskIdentifier) == 0){	/* we did not initiate this umount */
		if(mainModUserForMtabDevice(diskIdentifier) == 0)	/* if root did the mount */
			err = EACCES;
	}
	if(pthread_create(&thread, NULL, err == 0 ? diskarbThreadUnmountPreNotify0 : diskarbThreadUnmountPreNotifyErr, utilStrDup(diskIdentifier)) == 0)
		pthread_detach(thread);
}

static void	*diskarbThreadEjectPreNotify(void *diskId)
{
	DiskArbEjectPreNotifyAck_async_auto(diskId, 0);	/* RPC is one-way */
	memFree(diskId);
	return NULL;
}

static void	ejectPreNotification(DiskArbDiskIdentifier diskIdentifier, unsigned flags)
{
pthread_t	thread;

	if(LOG_MODE_IS_ACTIVE(LOGM_DISKARB))
		logPrintf(LOGM_DISKARB, "diskarb: ejectPreNotification in pid=%d for %s\n", getpid(), diskIdentifier);
	if(pthread_create(&thread, NULL, diskarbThreadEjectPreNotify, utilStrDup(diskIdentifier)) == 0)
		pthread_detach(thread);
}

static void	ejectPostNotification(DiskArbDiskIdentifier diskIdentifier, int errorCode, pid_t dissenter)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DISKARB))
		logPrintf(LOGM_DISKARB, "diskarb: ejectPostNotification in pid=%d for %s, err=%d\n", getpid(), diskIdentifier, errorCode);
}

static btbool	unmountWasExpected = 0;

static void unmountPostNotification(DiskArbDiskIdentifier diskIdentifier, int errorCode, pid_t dissenter)
{
umountInfo_t	info;
btint			i, cnt;
struct statfs	*fs;

	if(LOG_MODE_IS_ACTIVE(LOGM_DISKARB))
		logPrintf(LOGM_DISKARB, "diskarb: unmountPostNotification in pid=%d for %s, err=%d\n", getpid(), diskIdentifier, errorCode);
	info.timeoutThread = NULL;
	info.diskId = diskIdentifier;
	info.error = errorCode;
	info.dissenter = dissenter;
	unmountWasExpected = 0;
	procCondBroadcast(umountSignal, &info);
	if(errorCode == 0 && !unmountWasExpected){
		/* check whether this has really been unmounted, autodiskmount often
		 * genereates false unmount notifications!
		 */
		cnt = getmntinfo(&fs, MNT_NOWAIT);
		for(i=0;i<cnt;i++){
			if(strcmp(fs[i].f_mntfromname, diskIdentifier) == 0 && strcmp(fs[i].f_fstypename, "nfs") == 0)
				break;	/* still mounted... */
		}
		if(i >= cnt)	/* if no break occured -> NOT still mounted */
			mainModNotifyUnmount(diskIdentifier);
	}
}

/* ------------------------------------------------------------------------- */

typedef struct diskarbThreadParam{
	mach_port_t	rxPort;
	btint		writeFd;
	btint		readFd;
}diskarbThreadParam_t;

typedef struct rxMsgInfo{
	mach_msg_header_t	*msg;	/* may be NULL */
}rxMsgInfo_t;

static void	*disarbEventThread(void *arg)
{
diskarbThreadParam_t	*param = arg;
mach_msg_header_t		rxBuffer[1];	/* too small to hold msg, we peek at it only */
rxMsgInfo_t				info[1];
btint					rval, len = 0;

	for(;;){
		rval = mach_msg(rxBuffer, MACH_RCV_MSG | MACH_RCV_LARGE, 0, sizeof(rxBuffer), param->rxPort, 0, MACH_PORT_NULL);
		switch(rval){
		case MACH_MSG_SUCCESS:
			len = sizeof(rxBuffer);
			info->msg = memAlloc(len);
			memcpy(info->msg, rxBuffer, len);
			break;
		case MACH_RCV_TOO_LARGE:
			len = rxBuffer->msgh_size + sizeof(mach_msg_trailer_t);
			info->msg = memAlloc(len);
			rval = mach_msg(info->msg, MACH_RCV_MSG | MACH_RCV_LARGE, 0, len, param->rxPort, 0, MACH_PORT_NULL);
			if(rval == MACH_MSG_SUCCESS)
				break;
			memFree(info->msg);
			/* fallthrough */
		default:
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "diskarb: mach_msg receive: error %d\n", rval);
			info->msg = NULL;
			sleep(1);	/* avoid error-loop consuming 100% CPU time */
			break;
		}
		if(info->msg != NULL){
			if(LOG_MODE_IS_ACTIVE(LOGM_DISKARB))
				logPrintf(LOGM_DISKARB, "diskarb: received mach msg of length %d\n", len);
			write(param->writeFd, info, sizeof(info));
		}
	}
	return NULL;
}

/* ------------------------------------------------------------------------- */

static void	handlePortMsg(void *obj, void *arg)
{
diskarbThreadParam_t		*param = obj;
rxMsgInfo_t					info[1];
btint						rval;
static btu8					*txBuffer[sizeof(mach_msg_empty_send_t) + 32];

	rval = read(param->readFd, info, sizeof(info));
	if(rval < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "diskarb: rx pipe: [%d] %s\n", errno, sysStrerror(errno));
	}else if(rval == 0){	/* broken pipe */
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "diskarb: rx pipe broken\n");
		elpCancelForFd(param->readFd, ELP_FDMODE_READ);
	}else if(rval != sizeof(info)){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "diskarb: rx pipe: received illegal size %d\n", rval);
	}else if(info->msg == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "diskarb: rx pipe: received NULL msg\n");
	}else{
		memset(txBuffer, 0, sizeof(mach_msg_header_t));
		DiskArbHandleMsg(info->msg, (mach_msg_header_t *)txBuffer);
		memFree(info->msg);
		mach_msg_send((mach_msg_header_t *)txBuffer);	/* ignore rval, we can't handle errors anyway */
		if(LOG_MODE_IS_ACTIVE(LOGM_DISKARB))
			logPrintf(LOGM_DISKARB, "diskarb: sent reply message\n");
	}
}

/* ------------------------------------------------------------------------- */
/* The following function is called in the daemon's main task to register for
 * general callbacks about mounting/unmounting activity. Most other functions
 * in this module are only called from a forked task to do the mount/unmount
 * in a background process.
 */
void	diskarbMainInit(void)
{
static diskarbThreadParam_t	param[1];	/* must be valid during life of program */
btint		pipeFds[2], rval;
pthread_t	thread;

	diskarbPool = poolNew(sizeof(diskarbCtx_t), sizeof(diskarbCtx_t));
	umountSignal = procCondNew();
	runningUmounts = htabNew(&runningUmountsInterface, 4, 75);
	if(pipe(pipeFds) != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "diskarb: creating pipe: [%d] %s\n", errno, sysStrerror(errno));
		return;
	}
	DiskArbRegisterCallback_UnmountPostNotification(unmountPostNotification);
	DiskArbRegisterCallback_UnmountPreNotification(unmountPreNotification);
	DiskArbRegisterCallback_EjectPreNotification(ejectPreNotification);
	DiskArbRegisterCallback_EjectPostNotification(ejectPostNotification);
	if(DiskArbStart(&param->rxPort) != KERN_SUCCESS){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "diskarb: initializing DiskArb framework failed\n");
		close(pipeFds[1]);
		close(pipeFds[0]);
		return;
	}
	param->writeFd = pipeFds[1];
	param->readFd = pipeFds[0];
	rval = pthread_create(&thread, NULL, disarbEventThread, param);
	if(rval != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "diskarb: pthread_create: [%d] %s\n", rval, sysStrerror(rval));
		close(pipeFds[1]);
		close(pipeFds[0]);
		return;
	}
	pthread_detach(thread);	/* could have been done in attr of pthread_create(), but this is easier */
	elpRegisterForFd(pipeFds[0], ELP_FDMODE_READ, handlePortMsg, param);
}

/* ------------------------------------------------------------------------- */

void	diskarbPreMountNotify(btu8 *mountpoint, btu8 *device)
{
	mkdir(mountpoint, 0755);	/* create mountpoint dir, ignore errors */
}

/* ------------------------------------------------------------------------- */

typedef struct postMountNotifyArgs{
	btu8	*mountpoint;
	btu8	*device;
}postMountNotifyArgs_t;

static void	*diskarbThreadPostMountNotify(void *arg)
{
postMountNotifyArgs_t	*a = arg;

	/* The "disappeared" message will fail, but send it anyway. A previous
	 * incarnation of Sharity may have crashed, leaving the stale entry
	 * in the table. */
	DiskArbDiskDisappearedPing_auto(a->device, kDiskArbUnmountOneFlag);	/* RPC is one-way */
	DiskArbDiskAppearedWithMountpointPing_auto(a->device, kDiskArbDiskAppearedNetworkDiskMask | kDiskArbDiskAppearedEjectableMask, a->mountpoint);	/* RPC is one-way */
	memFree(a->mountpoint);
	memFree(a->device);
	memFree(a);
	return NULL;
}

void	diskarbPostMountNotify(btu8 *mountpoint, btu8 *device, btint mountError)
{
pthread_t				thread;
btint					rval;
postMountNotifyArgs_t	*a;

	if(mountError != 0)
		return;
	a = memAlloc(sizeof(*a));
	a->device = utilStrDup(device);
	a->mountpoint = utilStrDup(mountpoint);
	/* fork a thread to avoid deadlocks when the mach msg buffer is full */
	rval = pthread_create(&thread, NULL, diskarbThreadPostMountNotify, a);
	if(rval != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "diskarb: pthread_create: [%d] %s\n", rval, sysStrerror(rval));
		memFree(a->mountpoint);
		memFree(a->device);
		memFree(a);
	}else{
		pthread_detach(thread);	/* could have been done in attr of pthread_create(), but this is easier */
	}
}

/* ------------------------------------------------------------------------- */

static void	*diskarbThreadUnmountRequest(void *device)
{
	DiskArbUnmountRequest_async_auto(device, kDiskArbUnmountOneFlag);	/* RPC is one-way */
	return NULL;
}

static void	*diskarbThreadDisappeared(void *device)
{
	DiskArbDiskDisappearedPing_auto(device, kDiskArbUnmountOneFlag);	/* RPC is one-way */
	memFree(device);
	return NULL;
}

static void	umountTimeoutOccured(void *obj, void *arg)
{
umountInfo_t	info;

	info.timeoutThread = obj;
	procCondBroadcast(umountSignal, &info);
}

static void	diskarbPerformUnmountProc(void *obj, void *arg)
{
diskarbCtx_t	*this = obj;
umountInfo_t	*info = arg;
btError			err;
pthread_t		thread;

	PROC_BEGIN;
	htabAddObject(runningUmounts, this->device);
	while(pthread_create(&thread, NULL, diskarbThreadUnmountRequest, this->device) != 0){
		/* retry after 500ms if no thread available */
		PROC_WAITFOR(elpCallAtCancelPrev(elpNow() + 500, diskarbPerformUnmountProc, this));
	}
	pthread_detach(thread);
	/* now wait for result of unmount attempt */
	elpCallAtCancelPrev(elpNow() + 20000, umountTimeoutOccured, this);
	for(;;){
		PROC_WAITFOR_X(1, procCondWait(umountSignal, this));
		if(info->timeoutThread == NULL){
			if(strcmp(info->diskId, this->device) == 0){
				elpCancelCallAt(umountTimeoutOccured, this);	/* forget timeout */
				unmountWasExpected = 1;
				break;
			}
		}else if(info->timeoutThread == this){
			break;
		}
	}
	this->info = info;	/* store because we may sleep */
	/* try diskDisappeared in any case -- this will fail if not unmounted */
	this->s = utilStrDup(this->device);
	while(pthread_create(&thread, NULL, diskarbThreadDisappeared, this->s) != 0){
		/* retry after 500ms if no thread available */
		PROC_WAITFOR(elpCallAtCancelPrev(elpNow() + 500, diskarbPerformUnmountProc, this));
	}
	pthread_detach(thread);
	info = this->info;
	if(info->timeoutThread != NULL){
		err.class = ERRCLS_UNIX;
		err.code = ETIMEDOUT;
	}else{
		if(info->error == ENOTEMPTY)
			info->error = 0;	/* ignore error if mountpoint dir can't be deleted */
		if(info->error < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))	/* log mac-style errors */
				logPrintf(LOGM_ERROR, "diskarb: cannot unmount because pid=%d reports error=%d\n", info->dissenter, info->error);
			info->error = EINVAL;	/* generic Unix error */
		}
		err.class = info->error != 0 ? ERRCLS_UNIX : ERRCLS_OK;
		err.code = info->error;
	}
	htabRemoveObjectEqualTo(runningUmounts, this->device);
	PROC_RETURN(&err);
	PROC_END;
}

void	diskarbPerformUnmount(btu8 *mountpoint, btu8 *device, void *caller)
{
diskarbCtx_t	*ctx;

	PROC_INIT(ctx, diskarbPerformUnmountProc, caller, diskarbPool, poolFreeBlock);
	ctx->mountpoint = mountpoint;
	ctx->device = device;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */
#else	/* not defined HAVE_DISKARB */
/* ------------------------------------------------------------------------- */

/* dummy functions to avoid #ifdefs in other moduels */

void	diskarbPreMountNotify(btu8 *mountpoint, btu8 *device)
{
}

void	diskarbPostMountNotify(btu8 *mountpoint, btu8 *device, btint mountError)
{
}

void	diskarbPerformUnmount(btu8 *mountpoint, btu8 *device, void *caller)
{
static btError	err = {ERRCLS_FRAMEWORK, ERR_FRWipcnotimpl};

	PROC_IMMEDIATE_RET(caller, &err);
}

void	diskarbMainInit(void)
{
}

#endif /* HAVE_DISKARB */
