/*
 * Name: ipcserver.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-11-27
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: ipcserver.c,v 1.6 2001/01/08 15:54:01 cs Exp $
 */


#include "stdheaders.h"
#include "ipcserver.h"
#include "ipc.h"
#include "exception.h"
#include "memalloc.h"
#include "logging.h"
#include "process.h"
#include "blockpool.h"
#include "evntloop.h"
#include "system.h"
#include "unicode.h"
#include "main.h"
#include "cfgparse.h"
#include "util.h"
#include "errcodes.h"
#include "mainconst.h"
#include "mainmodule.h"

/* ------------------------------------------------------------------------- */

typedef struct moduleInfo{
	btu8			*name;
	ipcSrvStub_t	*dispatchTable;
	btint			fidBase;
	btint			fidCount;
}moduleInfo_t;

static void	handleClientConnect(void *obj, void *arg);

/* ------------------------------------------------------------------------- */

ipcServer_t	*ipcSrvNew(btu8 *sockName)
{
ipcServer_t			*volatile this = NULL;
struct sockaddr_un	sockAddr;
int					rval, storederr, sockMode, ownerID;
int					volatile groupID = -1;
btu8				*sockOwner, *sockGroup;
struct passwd		*pwent;
struct group		*grent;
cfgProperty_t		*mainSection;

EXC_TRY
	this = memCalloc(sizeof(ipcServer_t));
	this->sockName = utilStrDup(sockName);
	this->moduleInfos = arrNew(NULL, 8);
	this->sockFd = socket(AF_UNIX, SOCK_STREAM, 0);
	if(this->sockFd < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "ipcserver: error creating socket: [%d] %s\n", errno, sysStrerror(errno));
		}
		EXC_THROW_MSG(excInternalError, "ipcserver: error creating socket");
	}
	if(sysSetNonblock(this->sockFd) < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "ipcserver: error setting socket nonblocking: [%d] %s\n", errno, sysStrerror(errno));
		}
		EXC_THROW_MSG(excInternalError, "ipcserver: error setting socket nonblocking");
	}
	if(strlen(this->sockName) >= sizeof(sockAddr.sun_path)){
		EXC_THROW_MSG(excInternalError, "ipcserver: socket path too long");
	}
	sockAddr.sun_family = AF_UNIX;
	strcpy(sockAddr.sun_path, this->sockName);
	unlink(this->sockName);	/* delete any existing sockets before we bind */
	rval = bind(this->sockFd, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
	storederr = errno;
	if(rval < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "ipcserver: error binding socket: [%d] %s\n", storederr, sysStrerror(storederr));
		}
		EXC_THROW_MSG(excInternalError, "ipcserver: cannot bind socket");
	}
	mainSection = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.cfgFile));
	sockMode = cfgIntValue(cfgQueryWithName("sockMode", 1, mainSection), 0666);
	sockOwner = cfgStringValue(cfgQueryWithName("sockOwner", 1, mainSection), "root");
	sockGroup = cfgStringValue(cfgQueryWithName("sockGroup", 1, mainSection), "");
	if((pwent = getpwnam(sockOwner)) != NULL){
		ownerID = pwent->pw_uid;
		groupID = pwent->pw_gid;
	}else{
		ownerID = atoi(sockOwner);
	}
	if((grent = getgrnam(sockGroup)) != NULL){
		groupID = grent->gr_gid;
	}else{
		if(*sockGroup != 0 && (isdigit(*sockGroup) || (*sockGroup == '-' && isdigit(sockGroup[1]))))
			groupID = atoi(sockGroup);
	}
	chown(this->sockName, ownerID, groupID);
	chmod(this->sockName, sockMode);
	if(listen(this->sockFd, 10) < 0){	/* a queue of 10 pending connects should suffice... */
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "ipcserver: error listening on socket: [%d] %s\n", errno, sysStrerror(errno));
		}
		EXC_THROW_MSG(excInternalError, "ipcserver: error listening on socket");
	}
	elpRegisterForFd(this->sockFd, ELP_FDMODE_READ, handleClientConnect, this);	/* wait for connect */
EXC_CATCH
	ipcSrvFree(this);
	EXC_RETHROW();
EXC_ENDTRY
	return this;
}

void	ipcSrvFree(ipcServer_t *this)
{
	if(this != NULL){
		if(this->sockFd >= 0){
			elpCancelForFd(this->sockFd, -1);
			close(this->sockFd);
		}
		if(this->sockName != NULL){
			unlink(this->sockName);
			memFree(this->sockName);
		}
		memFree(this);
	}
}

/* ------------------------------------------------------------------------- */

ipcModule_t	*ipcRegisterModule(ipcServer_t *srv, btu8 *name, ipcSrvStub_t *dispatchTable, btint numEntries, btint ipcProtVersion)
{
ipcModule_t	*mod;

	mod = memCalloc(sizeof(*mod));
	mod->name = name;
	mod->protocolVersion = ipcProtVersion;
	mod->server = srv;
	mod->fidBase = srv->nextFid;
	mod->fidCount = numEntries;
	srv->nextFid += numEntries;
	mod->dispatchTable = dispatchTable;
	mod->servingClients = arrNew(NULL, 2);
	arrAddObject(srv->moduleInfos, mod);
	return mod;
}

void	ipcUnregisterModule(ipcModule_t *mod)
{
	if(mod->server != NULL){
		arrRemoveObjectEqualTo(mod->server->moduleInfos, mod);
		mod->server = NULL;
		arrFree(mod->servingClients);
		mod->servingClients = NULL;
		memFree(mod);
	}else{
		DEBUG_TRAP();
	}
}

void	ipcRegisterClientService(ipcModule_t *mod, ipcSrvClient_t *client, btint uid)
{
	client->uid = uid;
	arrAddObject(mod->servingClients, client);
}

void	ipcUnregisterClientService(ipcModule_t *mod, ipcSrvClient_t *client)
{
btint	i, cnt;
array_t	*arr;

	if(mod != NULL){
		arrRemoveObjectEqualTo(mod->servingClients, client);
	}else{
		arr = client->server->moduleInfos;
		cnt = arrObjectCount(arr);
		for(i=0;i<cnt;i++){
			mod = arrObjectAtIndex(arr, i);
			arrRemoveObjectEqualTo(mod->servingClients, client);
		}
	}
}

/* ------------------------------------------------------------------------- */

static btint	srvClientCompare(void *o1, void *o2)
{
	return (btu8 *)o1 - (btu8 *)o2;
}

static btu32	srvClientHashVal(void *obj)
{
btu32	x = (btu32)obj;

	return x ^ (x>>4) ^ (x>>8) ^ (x>>12);
}

static collectionObjectI_t	srvClientInterface = {
	srvClientCompare,
	srvClientHashVal,
	ipcSrvClientRetain,
	ipcSrvClientRelease,
};

struct array *ipcClientServicesForUid(ipcModule_t *mod, btint uid)
{
ipcSrvClient_t	*client;
btint			i;
array_t			*rval;

	rval = arrNew(&srvClientInterface, 4);
	i = arrObjectCount(mod->servingClients);
	while(--i >= 0){
		client = arrObjectAtIndex(mod->servingClients, i);
		if(client->uid == uid && !client->isInvalid)
			arrAddObject(rval, client);
	}
	return rval;
}

/* ------------------------------------------------------------------------- */

typedef struct requestCtx{
	bti32		sid;
	procCond_t	*signal;
}requestCtx_t;

typedef struct ipcClientCtx{
	PROC_VARSET;
	ipcSrvClient_t	*client;
	ipcModule_t		*module;
	ipcMsg_t		*rxMsg;
	ipcMsg_t		*txMsg;
	requestCtx_t	rq;
}ipcClientCtx_t;

static blockpool_t	*clientPool = NULL;

static void	clientContextFree(void *block)
{
	poolReturnBlock(clientPool, block);
}

/* ------------------------------------------------------------------------- */

static btint	requestCtxCompare(void *obj, void *otherObj)
{
requestCtx_t	*this = obj, *other = otherObj;

	return this->sid - other->sid;
}

static btu32	requestCtxHashval(void *obj)
{
requestCtx_t	*ctx = obj;

	return ctx->sid;
}

static collectionObjectI_t	requestCtxInterface = {
	requestCtxCompare,
	requestCtxHashval,
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

static void	destroyTempSocket(ipcSrvClient_t *client)
{
	if(client->tempFd != -1){
		elpCancelForFd(client->tempFd, -1);
		close(client->tempFd);
		client->tempFd = -1;
	}
	if(client->tempSocket != NULL){
		unlink(client->tempSocket);
		memFree(client->tempSocket);
		client->tempSocket = NULL;
	}
}

/* ------------------------------------------------------------------------- */

static void	handleClientRequest(void *obj, void *arg);

static void	ipcSrvClientFree(ipcSrvClient_t *this)
{
array_t	*allRq;
btint	i, cnt;

	if(this != NULL){
		destroyTempSocket(this);
		ipcUnregisterClientService(NULL, this);
		if(this->server != NULL)
			this->server->numConnections--;
		if(this->requests != NULL){
			allRq = htabAllObjects(this->requests);
			cnt = arrObjectCount(allRq);
			for(i=0;i<cnt;i++){
				procCondSignal(((requestCtx_t *)arrObjectAtIndex(allRq, i))->signal, NULL);
			}
			arrFree(allRq);
			htabFree(this->requests);
		}
		if(this->rxBuffer != NULL)
			ipcMsgFree(this->rxBuffer);
		elpCancelForFd(this->fd, -1);
		close(this->fd);
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
			logPrintf(LOGM_DEBUG, "ipcserver: closing client connection\n");
		}
		if(this->txSignal != NULL)
			procCondFree(this->txSignal);
		if(this->rxSignal != NULL)
			procCondFree(this->rxSignal);
		memset(this, 0, sizeof(*this));
		memFree(this);
	}
}

static ipcSrvClient_t	*ipcSrvClientNew(ipcServer_t *server, int fd)
{
ipcSrvClient_t	*volatile this = NULL;

	if(clientPool == NULL)
		clientPool = poolNew(sizeof(ipcClientCtx_t), sizeof(ipcClientCtx_t));
EXC_TRY
	this = memCalloc(sizeof(ipcSrvClient_t));
	this->server = server;
	this->fd = fd;
	this->tempFd = -1;
	this->server->numConnections++;
	this->txSignal = procCondNew();
	this->rxSignal = procCondNew();
	if(sysSetNonblock(this->fd) < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "ipcserver: error setting client socket nonblocking: [%d] %s\n", errno, sysStrerror(errno));
		}
		EXC_THROW_MSG(excInternalError, "ipcserver: error setting client socket nonblocking");
	}
	this->refCount = 1;	/* create objects with refcount = 1 */
	elpRegisterForFd(this->fd, ELP_FDMODE_READ, handleClientRequest, this);
	this->rxBuffer = ipcMsgNew();
	this->requests = htabNew(&requestCtxInterface, 4, 75);
EXC_CATCH
	ipcSrvClientFree(this);
	EXC_RETHROW();
EXC_ENDTRY
	return this;
}

void	ipcSrvClientRetain(void *obj)
{
ipcSrvClient_t *this = obj;

	this->refCount++;
}

void	ipcSrvClientRelease(void *obj)
{
ipcSrvClient_t *this = obj;

	this->refCount--;
	if(this->refCount == 0){
		ipcSrvClientFree(this);
	}else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

static void	ipcSrvClientInvalidate(ipcSrvClient_t *this)
{
array_t	*allRq;
btint	i, cnt;
btbool	wasInvalid;

	wasInvalid = this->isInvalid;	/* should never be true, but be safe */
	this->isInvalid = 1;
	elpCancelForFd(this->fd, -1);
	/* empty out everything that's pending: */
	for(;;){
		if((cnt = htabObjectCount(this->requests)) == 0)
			break;
		allRq = htabAllObjects(this->requests);
		for(i=0;i<cnt;i++)
			procCondBroadcast(((requestCtx_t *)arrObjectAtIndex(allRq, i))->signal, NULL);
		arrFree(allRq);
		procCondBroadcast(this->txSignal, NULL);
		procCondBroadcast(this->rxSignal, NULL);
	}
	if(!wasInvalid)
		ipcSrvClientRelease(this);
}

/* ------------------------------------------------------------------------- */

static void	dumpMessage(btu8 *title, btu8 *data, btsize len)
{
int		i;
btu8	buffer[256];
int		lineLen = 4;

	while(len > 0){
		strcpy(buffer, title);
		for(i=0;i<lineLen && len>0;i++){
			sprintf(buffer + strlen(buffer), " %02x", *data);
			data++;
			len--;
		}
		lineLen = 16;	/* further lines will be 16 long */
		logPrintf(LOGM_UITRACE, "%s\n", buffer);
	}
}

/* ------------------------------------------------------------------------- */

static void	sendMsgProc(void *obj, void *arg)
{
ipcClientCtx_t	*this = obj;
ipcSrvClient_t	*client = this->client;
btsize			len;
btint			storederr;

	PROC_BEGIN;
	while(client->txRunning && !client->isInvalid){	/* tx is busy */
		PROC_WAITFOR_X(1, procCondWait(client->txSignal, this));
	}
	if(client->isInvalid)
		goto doReturn;
	client->txRunning = 1;
	if(LOG_MODE_IS_ACTIVE(LOGM_UITRACE)){
		dumpMessage("<- uiTx:", this->txMsg->buffer, this->txMsg->size);
	}
	elpRegisterForFd(client->fd, ELP_FDMODE_WRITE, sendMsgProc, this);
	this->txMsg->index = 0;
	while(this->txMsg->index < this->txMsg->size){
		PROC_WAITFOR_X(2, ;)	/* wait for tx ready */
		len = write(client->fd, this->txMsg->buffer + this->txMsg->index, this->txMsg->size - this->txMsg->index);
		storederr = errno;
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "ipcserver: wrote %d bytes to UI\n", len);
		if(len < 0){	/* an error occured */
			if(storederr != EWOULDBLOCK && storederr != EAGAIN && storederr != EINTR){		/* real error */
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "ipcserver: error writing: [%d] %s\n", storederr, sysStrerror(storederr));
				break;
			}
		}else if(len == 0){	/* connection lost */
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "ipcserver: broken pipe during write\n");
			break;	/* just finish, main dispatcher will detect this case */
		}else{		/* data has been sent */
			this->txMsg->index += len;
		}
	}
	elpCancelForFd(client->fd, ELP_FDMODE_WRITE);		
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
		logPrintf(LOGM_DEBUG, "ipcserver: ready sending message to UI\n");
	}
doReturn:
	ipcMsgFree(this->txMsg);
	client->txRunning = 0;
	procCondSignal(client->txSignal, NULL);
	PROC_END;
}

static void	sendMsg(ipcSrvClient_t *client, ipcMsg_t *msg, void *caller)
{
ipcClientCtx_t	*ctx;

	PROC_INIT(ctx, sendMsgProc, caller, clientPool, clientContextFree);
	ctx->client = client;
	ctx->txMsg = msg;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

typedef struct clientParam{
	bti32	pvers;
	bti32	sid;
	bti32	procId;
}clientParam_t;

/* ------------------------------------------------------------------------- */

static void	handleAuthConnect(void *obj, void *arg)
{
ipcSrvClient_t	*client = obj;
int				newFd;
char			buffer[256];	/* for dummy reads */

	newFd = accept(client->tempFd, NULL, NULL);
	if(newFd < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "ipcserver: error accepting on temp socket: [%d] %s\n", errno, sysStrerror(errno));
		}
		read(client->fd, buffer, sizeof(buffer));
		return;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "ipcserver: accepted client connection\n");
	destroyTempSocket(client);
	elpCancelForFd(client->fd, -1);
	close(client->fd);			/* break connection on authentication socket */
	client->fd = newFd;
	elpRegisterForFd(client->fd, ELP_FDMODE_READ, handleClientRequest, client);
	client->isAuthenticated = 1;
}

static void	clientAuthenticateProc(void *obj, void *arg)
{
ipcClientCtx_t		*this = obj;
ipcSrvClient_t		*client = this->client;
bti32				uid, gid;
ipcMsg_t			*volatile reply;
btsize				size;
btint				rval, storederr;
btError				err;
clientParam_t		*param = arg;
btu8				tmpSocketName[256];
struct sockaddr_un	sockAddr;
static btint		counter = 0;

	PROC_BEGIN
	memset(&err, 0, sizeof(err));
	ipcSrvClientRetain(client);
	uid = ipcGetInt(this->rxMsg);
	gid = ipcGetInt(this->rxMsg);
	this->txMsg = reply = ipcMsgNew();
	client->uid = uid;
	client->gid = gid;
	if(client->tempSocket == NULL){
		client->tempFd = socket(AF_UNIX, SOCK_STREAM, 0);
		if(client->tempFd < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "ipcserver: error creating temp socket: [%d] %s\n", errno, sysStrerror(errno));
			goto errorOccured;
		}
		if(sysSetNonblock(client->tempFd) < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "ipcserver: error setting temp socket nonblocking: [%d] %s\n", errno, sysStrerror(errno));
			goto errorOccured;
		}
		sprintf(tmpSocketName, CONST_TEMPSOCKET_PATH, (int)getpid(), counter++, uid);
		client->tempSocket = utilStrDup(tmpSocketName);
		if(strlen(client->tempSocket) >= sizeof(sockAddr.sun_path)){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "ipcserver: name of temp socket ->%s<- too long (limit=%d)\n", client->tempSocket, sizeof(sockAddr.sun_path));
			goto errorOccured;
		}
		sockAddr.sun_family = AF_UNIX;
		strcpy(sockAddr.sun_path, client->tempSocket);
		unlink(client->tempSocket);	/* delete any existing sockets before we bind */
		rval = bind(client->tempFd, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
		storederr = errno;
		if(rval < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "ipcserver: error binding temp socket: [%d] %s\n", storederr, sysStrerror(storederr));
			goto errorOccured;
		}
		if((rval = chmod(client->tempSocket, 0600)) < 0){
			logPrintf(LOGM_ERROR, "ipcserver: error in chmod on temp socket: [%d] %s\n", errno, sysStrerror(errno));
		}else if((rval = chown(client->tempSocket, client->uid, client->gid)) < 0){
			logPrintf(LOGM_ERROR, "ipcserver: error in chown on temp socket: [%d] %s\n", errno, sysStrerror(errno));
		}
		if(rval < 0)
			goto errorOccured;
		if(listen(client->tempFd, 2) < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "ipcserver: error listening on temp socket: [%d] %s\n", errno, sysStrerror(errno));
			goto errorOccured;
		}
		elpRegisterForFd(client->tempFd, ELP_FDMODE_READ, handleAuthConnect, client);	/* wait for connect */
	}
	goto doReturn;
errorOccured:
	destroyTempSocket(client);
	err.class = ERRCLS_FRAMEWORK;
	err.code = ERR_FRWclientcomm;
doReturn:
	ipcAddInt(reply, 0);	/* reserve for size */
	ipcAddInt(reply, 0);	/* protocol version */
	ipcAddInt(reply, param->sid);	/* sequence number */
	ipcAddInt(reply, -1);	/* proc ID = reply */
	ipcAddInt(reply, 0);	/* uid */
	ipcAddInt(reply, 0);	/* gid */
	ipcAddInt(reply, err.class);	/* error class */
	ipcAddInt(reply, err.code);	/* error code */
	ipcAddAscii(reply, errDescription(&err));
	if(err.class == 0)
		ipcAddAscii(reply, client->tempSocket);
	ipcMsgFree(this->rxMsg);
	this->rxMsg = NULL;
	size = reply->size;
	reply->size = 0;		/* hack to add total size at the beginning */
	ipcAddInt(reply, size);
	reply->size = size;
	PROC_WAITFOR_X(1, sendMsg(client, this->txMsg, this));
	ipcSrvClientRelease(this->client);	/* matches retain at the beginning */
	PROC_END
}

/* ------------------------------------------------------------------------- */

static void	clientRequestProc(void *obj, void *arg)
{
ipcClientCtx_t	*this = obj;
bti32			uid, gid;
ipcMsg_t		*volatile reply;
btsize			size;
clientParam_t	*param = arg;
btint			numModules, i;
ipcModule_t		*mod = NULL;

	PROC_BEGIN
	ipcSrvClientRetain(this->client);
	numModules = arrObjectCount(this->client->server->moduleInfos);
	for(i=0;i<numModules;i++){
		mod = arrObjectAtIndex(this->client->server->moduleInfos, i);
		if(param->procId >= mod->fidBase && param->procId < mod->fidBase + mod->fidCount)
			break;	/* this is our module! */
	}
	if(i >= numModules)
		mod = NULL;
	uid = ipcGetInt(this->rxMsg);
	gid = ipcGetInt(this->rxMsg);
	this->txMsg = reply = ipcMsgNew();
	ipcAddInt(reply, 0);	/* reserve for size */
	ipcAddInt(reply, mod != NULL ? mod->protocolVersion : 0);	/* protocol version */
	ipcAddInt(reply, param->sid);	/* sequence number */
	ipcAddInt(reply, -1);	/* proc ID = reply */
	ipcAddInt(reply, 0);	/* uid */
	ipcAddInt(reply, 0);	/* gid */
	if(mod == NULL){	/* no module found */
		ipcAddInt(reply, ERRCLS_FRAMEWORK);
		ipcAddInt(reply, ERR_FRWipcparameters);
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "ipcserver: cannot map request-id %d\n", param->procId);
	}else if(param->pvers != mod->protocolVersion){
		ipcAddInt(reply, ERRCLS_FRAMEWORK);
		ipcAddInt(reply, ERR_FRWipcversion);
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "ipcserver: client utility with protocol %d instead of %d\n", param->pvers, mod->protocolVersion);
	}else{
		i = param->procId - mod->fidBase;
		if(mod->dispatchTable[i] != NULL){
			PROC_WAITFOR_X(1, mod->dispatchTable[i](this->client, this->rxMsg, uid, gid, reply, this));
		}else{
			ipcAddInt(reply, ERRCLS_FRAMEWORK);
			ipcAddInt(reply, ERR_FRWipcparameters);
		}
	}
	ipcMsgFree(this->rxMsg);
	this->rxMsg = NULL;
	reply = this->txMsg;
	size = reply->size;
	reply->size = 0;		/* hack to add total size at the beginning */
	ipcAddInt(reply, size);
	reply->size = size;
	PROC_WAITFOR_X(2, sendMsg(this->client, this->txMsg, this));
	ipcSrvClientRelease(this->client);	/* matches retain at the beginning */
	PROC_END
}

/* ------------------------------------------------------------------------- */

static void	handleClientRequest(void *obj, void *arg)
{
ipcSrvClient_t	*this = obj;
bti32			maxRead, len;
ipcClientCtx_t	*ctx;
clientParam_t	param;
requestCtx_t	*rctx, rctxDummy;
ipcMsg_t		*msg;

	if(this->rxBuffer->index == 0){	/* nothing received until now */
		this->rxBuffer->size = 4;	/* read only the length information first */
	}
	for(;;){
		maxRead = this->rxBuffer->size - this->rxBuffer->index;
		if(maxRead <= 0){	/* NULL message, ignore */
			this->rxBuffer->size = this->rxBuffer->index = 0;
			break;
		}
		len = read(this->fd, this->rxBuffer->buffer + this->rxBuffer->index, maxRead);
		if(len == 0){		/* connection closed */
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "ipcserver: client closed socket\n");
			ipcSrvClientInvalidate(this);
			return;
		}else if(len < 0){	/* error occured */
			if(errno != EWOULDBLOCK && errno != EAGAIN && errno != EINTR){		/* real error */
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "ipcserver: error reading socket: [%d] %s\n", errno, sysStrerror(errno));
				ipcSrvClientInvalidate(this);
				return;
			}
			break;
		}
		this->rxBuffer->index += len;
		if(this->rxBuffer->index == this->rxBuffer->size){	/* whatever we received, it's complete */
			if(this->rxBuffer->index == 4){	/* it's just the length info */
				this->rxBuffer->index = 0;
				len = ipcGetInt(this->rxBuffer);
				this->rxBuffer->size = len;
				ipcMsgEnsureSize(this->rxBuffer, this->rxBuffer->size);
				continue;					/* try to read the rest */
			}
			if(LOG_MODE_IS_ACTIVE(LOGM_UITRACE)){
				dumpMessage("-> uiRx:", this->rxBuffer->buffer, this->rxBuffer->size);
			}
			this->rxBuffer->index = 4;		/* set to beginning of real message */
			param.pvers = ipcGetInt(this->rxBuffer);
			param.sid = ipcGetInt(this->rxBuffer);
			param.procId = ipcGetInt(this->rxBuffer);
			if(!this->isAuthenticated){			/* this must be an authentication request */
				PROC_INIT(ctx, clientAuthenticateProc, NULL, clientPool, clientContextFree);
				ctx->rxMsg = this->rxBuffer;
				this->rxBuffer = ipcMsgNew();
				ctx->client = this;
				PROC_START(ctx, &param);
			}else if(param.procId < 0){			/* this is a reply */
				rctxDummy.sid = param.sid;
				if((rctx = htabFindObjectEqualTo(this->requests, &rctxDummy)) != NULL){
					msg = this->rxBuffer;
					this->rxBuffer = ipcMsgNew();
					procCondSignal(rctx->signal, msg);
					procCondBroadcast(this->rxSignal, NULL);
				}else{
					if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
						logPrintf(LOGM_ERROR, "ipcserver: unexpected reply on SID=%d\n", param.sid);
				}
			}else{
				PROC_INIT(ctx, clientRequestProc, NULL, clientPool, clientContextFree);
				ctx->rxMsg = this->rxBuffer;
				this->rxBuffer = ipcMsgNew();
				ctx->client = this;
				PROC_START(ctx, &param);
			}
			break;	/* don't continue reading, there will most probably be nothing */
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	handleClientConnect(void *obj, void *arg)
{
ipcServer_t		*this = obj;
int				newFd;
char			buffer[256];	/* for dummy reads */

	newFd = accept(this->sockFd, NULL, NULL);
	if(newFd < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			logPrintf(LOGM_ERROR, "ipcserver: error accepting on socket: [%d] %s\n", errno, sysStrerror(errno));
		}
		read(this->sockFd, buffer, sizeof(buffer));
		return;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
		logPrintf(LOGM_DEBUG, "ipcserver: accepted client auth connection\n");
	}
EXC_TRY
	ipcSrvClientNew(this, newFd);	/* don't store client pointer, we don't need it */
EXC_CATCH	/* ignore errors here */
EXC_ENDTRY
}

/* ------------------------------------------------------------------------- */

static void	ipcSrvRPCProc(void *obj, void *arg)
{
ipcClientCtx_t	*this = obj;
ipcSrvClient_t	*client = this->client;
ipcMsg_t		*rval = arg, *txMsg = this->txMsg;
btsize			size;

	PROC_BEGIN;
	while(htabObjectCount(client->requests) > CONST_IPC_WINDOWSIZE && !client->isInvalid){
		PROC_WAITFOR_X(1, procCondWait(client->rxSignal, this));
	}
	if(client->isInvalid){
		ipcMsgFree(this->txMsg);
		PROC_RETURN(NULL);
	}
	this->rq.sid = client->sidCounter++;
	this->rq.signal = procCondNew();
	size = txMsg->size;
	txMsg->size = 0;		/* hack to add total size at the beginning */
	ipcAddInt(txMsg, size);
	ipcAddInt(txMsg, this->module->protocolVersion);	/* protocol version */
	ipcAddInt(txMsg, this->rq.sid);	/* sequence number */
	txMsg->size = size;
	PROC_WAITFOR_X(2, sendMsg(client, txMsg, this));
	htabAddObject(client->requests, &this->rq);
	if(!client->isInvalid){
		PROC_WAITFOR_X(3, procCondWait(this->rq.signal, this));	/* wait for reply */
	}else{
		rval = NULL;
	}
	htabRemoveObjectEqualTo(client->requests, &this->rq);
	procCondFree(this->rq.signal);
	PROC_RETURN_AND_CLEANUP(rval,
		if(rval != NULL)
			ipcMsgFree(rval);
	);
	PROC_END;
}

void	ipcSrvRPC(ipcModule_t *mod, ipcSrvClient_t *client, ipcMsg_t *txMsg, void *caller)
{
ipcClientCtx_t	*ctx;

	PROC_INIT(ctx, ipcSrvRPCProc, caller, clientPool, clientContextFree);
	ctx->client = client;
	ctx->txMsg = txMsg;
	ctx->module = mod;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

