/*
 * Name: ipcserver.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-11-27
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: ipcserver.h,v 1.2 1999/11/14 01:33:17 cs Exp $
 */

#ifndef __ipcserver_H_INCLUDED__
#define __ipcserver_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This module implements the 'ipcSrvClient' class, which represents a user
interface command connected to the daemon within the daemon and the
'ipcServer' class, which represents the daemon listening at the unix
domain socket for client (user interface command) connects. After creating
an 'ipcServer' object with 'ipcSrvNew()', and registering a module's stub
functions with 'ipcRegisterModule()' clients can connect to the server and
execute the stub functions until the server object is freed.
*/

/* ------------------------------------------------------------------------- */

struct ipcMsg;
struct hashtable;
struct ipcSrvClient;

typedef void	(*ipcSrvStub_t)(struct ipcSrvClient *client, struct ipcMsg *request, bti32 uid, bti32 gid, struct ipcMsg *reply, void *caller);

typedef struct ipcServer{
	btu8			*sockName;
	btint			sockFd;
	btint			numConnections;	/* current number of connections */
	btint			nextFid;		/* function-ID base for next module */
	struct array	*moduleInfos;
}ipcServer_t;
/* This is the internal representation of the listening socket in the daemon.
 * There's only one instance of this object in the daemon.
 */

typedef struct ipcModule{
	btu8			*name;
	ipcServer_t		*server;
	btint			protocolVersion;
	btint			fidBase;
	btint			fidCount;
	ipcSrvStub_t	*dispatchTable;
	struct array	*servingClients;
}ipcModule_t;
/* This structure represents the RPC function number space and dispatcher table
 * of a module. Separating the RPC functions for each module allows easier
 * integration of additional modules.
 */

typedef struct ipcSrvClient{
	bti32				refCount;
	btint				uid;		/* uid of client */
	btint				gid;		/* gid of client */
	ipcServer_t			*server;
	struct hashtable	*requests;	/* contexts of outstanding requests */
	int					fd;
	struct ipcMsg		*rxBuffer;
	struct procCond		*rxSignal;
	struct procCond		*txSignal;
	btu32				sidCounter;
	btbool				txRunning;
	btbool				isInvalid;
	btbool				isAuthenticated;	/* 1 if uid-field is valid */
	btu8				*tempSocket;		/* socket for authentication */
	btint				tempFd;				/* file descriptor of temp socket */
}ipcSrvClient_t;
/* This is the internal representation of a GUI or commandline application. */

/* ------------------------------------------------------------------------- */

ipcServer_t	*ipcSrvNew(btu8 *sockName);
/* This function creates a new socket for IPC communication.
 * Parameters: 'sockName' is the full path in the filesystem for the Unix
 *    domain socket.
 * Returns: A socket representation object.
 */
void	ipcSrvFree(ipcServer_t *this);
/* This function frees an object created with ipcSrvNew() and removes the
 * Unix domain socket from the filesystem.
 */
ipcModule_t	*ipcRegisterModule(ipcServer_t *srv, btu8 *name, ipcSrvStub_t *dispatchTable, btint numEntries, btint ipcProtVersion);
/* This function is used to get a new RPC function number space and to register
 * the RPC functions for a new module.
 * Parameters: 'srv' is the socket representation object obtained with
 *    ipcSrvNew(), 'name' is the module name, 'dispatchTable' is an array
 *    of function pointers to the implementations of the RPC functions,
 *    'numEntries' is the number of functions in the table, 'ipcProtVersion'
 *    is the version of the protocol. Whenever a parameter is changed or
 *    a function is added or removed, the version number should be changed to
 *    avoid misinterpretation of the marshalled data in the server or client.
 *    Servers and clients with different IPC version numbers can't communicate.
 * Returns: a representation of the RPC function space.
 */
void	ipcUnregisterModule(ipcModule_t *module);
/* This function frees a representation previously obtained with
 * ipcRegisterModule().
 */
void	ipcRegisterClientService(ipcModule_t *module, ipcSrvClient_t *client, btint uid);
/* This function informs the server (daemon) that the given client is willing
 * to serve requests. Usually there are only RPC calls from the GUI/commandline
 * application to the daemon. If an application is willing to server calls,
 * it must register (through a dedicated RPC function) with the server.
 * Parameters: 'module' is the module context for the RPC calls, 'client' is
 *    the representation of the client that is willing to serve requests and
 *    'uid' is the local Unix user ID of the user running the client
 *    application.
 */
void	ipcUnregisterClientService(ipcModule_t *module, ipcSrvClient_t *client);
/* This function informs the server that the given client won't serve requests
 * any more (probably because it will terminate).
 */
struct array *ipcClientServicesForUid(ipcModule_t *module, btint uid);
/* This function looks up GUI/commandline applications which are willing to
 * serve requests for the given module and which belong to the given user. The
 * usual application is for asking a particular user for the password on a
 * server. If more than one client is registered, they are returned in reverse
 * historical order, which means that the most recently registered client has
 * index 0 in the returned array.
 * Parameters: 'module' is the module context for which the client should
 *    serve and 'uid' is the local Unix user ID of the user that runs the
 *    client application,
 * Returns: An array with the matching clients. The caller is responsible for
 *    freeing the array.
 */
void	ipcSrvClientRetain(void *obj);
void	ipcSrvClientRelease(void *obj);
/* These two functions implement the reference counting for client
 * representations.
 * Parameters: 'obj' is a pointer to a ipcSrvClient_t object that should be
 *    retained or released.
 */
static inline btbool	ipcSrvClientIsInvalid(ipcSrvClient_t *c)
{
	return c->isInvalid;
}
/* This function returns whether the given client has become invalid because
 * it disconnected from the server. You should check for invalid clients
 * before you send it a request message. It's not necessary to check before
 * sending replies.
 */
void	ipcSrvRPC(ipcModule_t *mod, ipcSrvClient_t *client, struct ipcMsg *txMsg, void *caller);
/* This function sends txMsg to the persistently connected user interface tool
 * 'client', inserting message size, protocol ID, and SID automatically. It
 * then waits for the reply from the tool (recognized by SID) and returns the
 * received message. The received message is automatically freed.
 * Parameters: 'mod' is the module context, 'client' is the GUI/commandline
 *    application that should receive the message, 'txMsg' is the almost
 *    complete message to transmit (except the fields mentioned above, which
 *    should be passed as zero) and 'caller' is the caller's context.
 * Callback: ipcMsg_t with the received reply message or NULL if the client
 *    terminated.
 */
/* ------------------------------------------------------------------------- */

#endif	/* __ipcserver_H_INCLUDED__ */
