/*
 * Name: main.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-16
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: main.c,v 1.17 2002/04/05 17:53:50 cs Exp $
 */


#include "stdheaders.h"
#include "logging.h"
#include "evntloop.h"
#include "unicode.h"
#include "process.h"
#include "main.h"
#include "cfgparse.h"
#include "mainconst.h"
#include "ipcserver.h"
#include "ipc.h"
#include "blockpool.h"
#include "util.h"
#include "system.h"
#include "argparser.h"
#include "../uiLib/maininterface.h"	/* for MAIN_IPC_SOCKET_PATH */
#include "mainmodule.h"
#include "notification.h"
#include "help.h"
#include "diskarb.h"

/* ------------------------------------------------------------------------- */

globalData_t	gdata;

/* the following array must match the LOGM_* defines in logging.h */
static btu8	*loglevels[] = {
	"info",
	"error",
	"debug",
	"resource",
	"netbiosError",
	"netbiosDebug",
	"netbiosTrace",
	"nfsTrace",
	"rare",
	"cacheDebug",
	"cifsTrace",
	"oplock",
	"warn",
	"smbSequence",
	"debugAttributes",
	"debugSSL",
	"smbConnect",
	"uiTrace",
	"nbnsTrace",
	"diskarb",
	NULL
};

/* ------------------------------------------------------------------------- */

static bttime	cfgParseInterval = 0;
static btu8		*pidFile = NULL;

/* ------------------------------------------------------------------------- */

static void	sharityTerminate(btbool doCoredump)
{
	ncPostNotification(gdata.terminateNotification, (void *)(btint)doCoredump);
	if(pidFile != NULL){
		unlink(pidFile);	/* clean up before exiting */
	}
	if(doCoredump)
		abort();
	exit(1);
}

/* ------------------------------------------------------------------------- */

typedef struct mainCtx{
	PROC_VARSET;
}mainCtx_t;

static btbool		nonFatalSignalReceived = 0;
static mainCtx_t	termContext;

static void	gotNonFatalSignalProc(void *obj, void *arg)
{
mainCtx_t	*this = obj;

	PROC_BEGIN;
	if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
		logPrintf(LOGM_INFO, "main: received non-fatal signal, trying to unmount cleanly\n");
	PROC_WAITFOR(elpCallCancelPrev(gotNonFatalSignalProc, this));
/* call from eventloop because signal may have interrupted almost anything */
	PROC_WAITFOR(mainModUmountAll(this));
	sharityTerminate(0);
	PROC_END;
}

static void	gotNonFatalSignal(int signr)
{
	if(nonFatalSignalReceived){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "main: non-fatal signal already in progress, ignoring\n");
		return;
	}
	nonFatalSignalReceived = 1;
	termContext.proc.process = gotNonFatalSignalProc;
	termContext.proc.caller = NULL;
	termContext.proc.free = NULL;
	gotNonFatalSignalProc(&termContext, NULL);
}

/* ------------------------------------------------------------------------- */

static void	gotFatalSignal(int signr)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "*** main: received fatal signal %d, trying emergency unmount\n", signr);
	sharityTerminate(1);
}

/* ------------------------------------------------------------------------- */

static void	gotExitSignal(int signr)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "*** main: received termination signal %d, exiting\n", signr);
	sharityTerminate(0);
}

/* ------------------------------------------------------------------------- */

static void	checkCfgFile(void *obj, void *arg)
{
plist_t	*mainPlist, *levels;
btsize	lcount, i;

	if(cfgFileParseIfChanged(gdata.cfgFile) || obj != NULL){
		mainPlist = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.cfgFile));
		levels = cfgQueryWithName("logLevels", 1, mainPlist);
		if(levels != NULL){
			lcount = cfgQueryValueCount(levels);
			logModeMask = 0;
			for(i=0;i<lcount;i++){
				LOG_SET_MODE(cfgEnumValue(cfgQueryAtIndex(i, levels), loglevels, 0));
			}
		}
		cfgParseInterval = cfgIntValue(cfgQueryWithName("cfgParseInterval", 1, mainPlist), cfgParseInterval);
	}
	if(gdata.dbFile != NULL)
		cfgFileParseIfChanged(gdata.dbFile);
	if(cfgParseInterval != 0)
		elpCallAtCancelPrev(elpNow() + cfgParseInterval, checkCfgFile, NULL);
}

/* ------------------------------------------------------------------------- */

static void	reparseConfig(int signal)
{
	checkCfgFile((void *)1, NULL);
}

/* ------------------------------------------------------------------------- */

static void	restoreMounts(void *dummy1, void *dummy2)
{
cfgProperty_t	*cfg, *db, *p, *entry;
btint			i, cnt, uid;
btu8			*frontend, *backend, *mountpoint, *backendSpec;
btbool			readonly, isAutomount;
btuni			mountedObj[1024];
struct stat		statbuf[1];

	cfg = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.cfgFile));
	db = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.dbFile));
	p = cfgQueryWithName("automount", 2, db, cfg);
	cnt = cfgQueryValueCount(p);
	for(i=0;i<cnt;i++){
		entry = cfgQueryAtIndex(i, p);
		frontend = cfgStringValue(cfgQueryWithName("frontend", 1, entry), NULL);
		backend = cfgStringValue(cfgQueryWithName("backend", 1, entry), NULL);
		mountpoint = cfgStringValue(cfgQueryWithName("mountpoint", 1, entry), NULL);
		backendSpec = cfgStringValue(cfgQueryWithName("backendSpec", 1, entry), NULL);
		uid = cfgIntValue(cfgQueryWithName("owner", 1, entry), 0);
		readonly = cfgEnumValue(cfgQueryWithName("readonly", 1, entry), cfgYesNoEnum, 0);
		isAutomount = cfgEnumValue(cfgQueryWithName("isAutomount", 1, entry), cfgYesNoEnum, 0);
		if(backend == NULL || mountpoint == NULL || backendSpec == NULL)
			continue;	/* no valid entry */
		convLocalAsciiToUni(backendSpec, mountedObj, ELEMENTS(mountedObj));
#ifndef HAVE_DISKARB
		/* See the comment in backends/cifs/cifsbrowse.c for why we don't create
		 * the mount point directory when we have diskarb
		 */
		/* The mountpoint directory may have been deleted if we try to restore
		 * an automount. We therefore create the mountpoint directory (only
		 * the last component) on demand.
		 */
		if(stat(mountpoint, statbuf) < 0 && errno == ENOENT){
			if(mkdir(mountpoint, 0777) == 0)
				chown(mountpoint, uid, -1);
		}
#endif
		/* perform mount in background: */
		mainModMount(uid, frontend, mountpoint, backend, mountedObj, readonly, 1, isAutomount, NULL);
	}
}

/* ------------------------------------------------------------------------- */

static void	printVersion(char *name)
{
	if(MAINCONST_BETA_VERSION != 0)
		fprintf(stderr, "%s: daemon version %d.%db%d started\n", name, MAINCONST_VERSION / 1000, MAINCONST_VERSION % 1000, MAINCONST_BETA_VERSION);
	else
		fprintf(stderr, "%s: daemon version %d.%d started\n", name, MAINCONST_VERSION / 1000, MAINCONST_VERSION % 1000);
	if(strlen(CONST_LOCALISATION_TEXT) > 0){
		fprintf(stderr, "%s\n", CONST_LOCALISATION_TEXT);
	}
}

/* ------------------------------------------------------------------------- */

void	mainSignalHandlers(btbool removeThem)
{
void	*fatalHandler;
void	*nonFatalHandler;
void	*exitHandler;

	if(removeThem){
		exitHandler = fatalHandler = nonFatalHandler = SIG_DFL;
	}else{
		fatalHandler = gotFatalSignal;
		nonFatalHandler = gotNonFatalSignal;
		exitHandler = gotExitSignal;
	}
	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, nonFatalHandler);
	signal(SIGQUIT, nonFatalHandler);
	signal(SIGILL, fatalHandler);
/*	signal(SIGIOT, fatalHandler); leads to recursive signal on abort */
	signal(SIGTRAP, SIG_IGN);
#ifdef SIGEMT
	signal(SIGEMT, fatalHandler);
#endif
	signal(SIGFPE, fatalHandler);
	signal(SIGBUS, fatalHandler);
	signal(SIGSEGV, fatalHandler);
#ifdef SIGSYS
	signal(SIGSYS, fatalHandler);
#endif
	signal(SIGPIPE, SIG_IGN);		/* ignore pipe signals */
	signal(SIGTERM, exitHandler);
	signal(SIGUSR1, SIG_IGN);
	signal(SIGUSR2, SIG_IGN);
}

/* ------------------------------------------------------------------------- */

/* statically linked modules: */
void	mainMod_init(void);
void	dummyFs_init(void);
void	plainFile_init(void);
void	nfs2_init(void);
void	nfs3_init(void);
void	cifsFile_init(void);
void	cifsBrowse_init(void);


int	main(int argc, char **argv)
{
btbool			goBackground = 1;
btint			cnt, i;
btu8			*cfgFile = CONST_CONFIG_PATH, *dbPath, *helpFile, *mod, *err;
FILE			*fp;
argParser_t		*ap;
cfgProperty_t	*mainSection, *modArray;

#if 0	/* enable this on Openstep to debug malloc problems */
	malloc_debug(14);
#endif
	openlog("sharity", LOG_PID, LOG_DAEMON);
	logDestination = LOG_DEST_STDERR;
	uniTrivialInit();
	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");
	argpAddOption(ap, 'v', NULL, 0, "Print version and exit");
	argpAddOption(ap, 'd', NULL, 0, "Debug mode. Don\'t go to background.");
	argpAddOption(ap, 'f', "<config-file>", 0, "Use the given config file");
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", argv[0]);
		exit(1);
	}
	if(argpOptionIsSet(ap, 'h')){
		printf("usage: %s [<options>]\n", argv[0]);
		printf("valid options are:\n");
		argpPrintHelp(ap, stdout);
		exit(0);
	}
	if(argpOptionIsSet(ap, 'v')){
		printVersion(argv[0]);
		exit(0);
	}
	if(argpArgumentCount(ap) > 0)
		fprintf(stderr, "Warning: excess arguments ignored\n");
	if(argpOptionIsSet(ap, 'd')){
		goBackground = 0;			/* don't go to background in debug mode */
		logModeMask = 0xffffffff;	/* set maximum debug level until initialized */
	}
	if(argpOptionIsSet(ap, 'f'))
		cfgFile = argpOptionStringVal(ap, 'f');
	LOG_SET_MODE(LOGM_ERROR);
	LOG_SET_MODE(LOGM_WARN);
	elpInit();
	gdata.cfgFile = cfgFileNew(cfgFile);
	checkCfgFile((void *)1, NULL);
	mainSection = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.cfgFile));
	gdata.terminateNotification = ncNew(4);
	sysSetCorefileLimit(cfgIntValue(cfgQueryWithName("corefileLimit", 1, mainSection), 0) * 1024 * 1024);
	helpFile = cfgStringValue(cfgQueryWithName("helpFile", 1, mainSection), CONST_HELPFILE_PATH);
	helpParseFile(helpFile);
	dbPath = cfgStringValue(cfgQueryWithName("configDbFile", 1, mainSection), CONST_DBFILE_PATH);
	gdata.dbFile = cfgFileNew(dbPath);
	gdata.ipcServer = ipcSrvNew(MAIN_IPC_SOCKET_PATH);
	mainMod_init();
#if 0	/* don't link test backend into final product */
	dummyFs_init();
#endif
	cifsFile_init();
	cifsBrowse_init();
	nfs2_init();
	if(sysHaveNfs3())
		nfs3_init();
	modArray = cfgQueryWithName("modules", 1, mainSection);
	cnt = cfgQueryValueCount(modArray);
	for(i=0;i<cnt;i++){
		mod = cfgStringValue(cfgQueryAtIndex(i, modArray), NULL);
		if(mod != NULL){
			if((err = sysLoadModule(mod)) != NULL){
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "main: error loading module %s:  %s\n", mod, err);
			}
		}
	}

	pidFile = cfgStringValue(cfgQueryWithName("pidFile", 1, mainSection), NULL);

/* call restoreMounts() after initialisation */
	elpCallAtCancelPrev(elpNow() + 1000, restoreMounts, NULL);
/* become immune to several signals, ignore others */
	mainSignalHandlers(0);	
	if(goBackground){
		logDestination = LOG_DEST_SYSLOG;
		sysDetach();
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_INFO)){
		if(MAINCONST_BETA_VERSION != 0)
			logPrintf(LOGM_INFO, "Sharity daemon version %d.%db%d started\n", MAINCONST_VERSION / 1000, MAINCONST_VERSION % 1000, MAINCONST_BETA_VERSION);
		else
			logPrintf(LOGM_INFO, "Sharity daemon version %d.%d started\n", MAINCONST_VERSION / 1000, MAINCONST_VERSION % 1000);
	}
	if(pidFile != NULL){
		unlink(pidFile);	/* try to remove old pid files */
		fp = fopen(pidFile, "w");
		if(fp == NULL){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "main: cannot create pid-file %s: [%d] %s\n", pidFile, errno, sysStrerror(errno));
		}else{
			fprintf(fp, "%d\n", (int)getpid());
			fclose(fp);
		}
	}
	signal(SIGHUP, reparseConfig);
	diskarbMainInit();
	elpRun();
	return 0;
}

/* ------------------------------------------------------------------------- */
