/*
 * Name: mainmodule.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-10
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: mainmodule.c,v 1.19 2002/05/31 16:31:14 cs Exp $
 */

#include "stdheaders.h"
#include "module.h"
#include "hashtable.h"
#include "array.h"
#include "memalloc.h"
#include "ipc.h"
#include "ipcserver.h"
#include "unicode.h"
#include "blockpool.h"
#include "process.h"
#include "backend.h"
#include "frontend.h"
#include "errcodes.h"
#include "util.h"
#include "logging.h"
#include "cfgparse.h"
#include "main.h"
#include "notification.h"
#include "mainconst.h"
#include "../uiLib/maininterface.h"
#include "mainmodule.h"
#include "guimanager.h"
#include "evntloop.h"
#include "system.h"
#include "help.h"

#define	MAIN_MODULE_NAME	"main"

/* ------------------------------------------------------------------------- */

typedef struct frontendInterface{
	module_t	*forModule;
	void		*(*mntNew)(btu8 *mountpoint);
	void		(*mntRelease)(void *obj);
	void		(*mntSetBackend)(void *obj, struct module *backendMod, void *backendObj);
	void		(*mntMount)(void *obj, btint uid, btbool readOnly, void *caller);
	void		(*mntUmount)(void *obj, btint uid, void *caller);
	void		(*mntSetFakemode)(void *obj, btbool fakeMode);
	void		(*mntWasUnmounted)(void *obj);
	btbool		(*mntIsReadonly)(void *obj);
	btu8		*(*mntMtabDevice)(void *obj);
}frontendInterface_t;

typedef struct backendInterface{
	module_t	*forModule;
	void		*(*vfsForPath)(btuni *path);
	void		(*vfsConnect)(void *obj, void *user, void *caller);
	void		(*vfsRelease)(void *obj);
	void		(*userAuthenticate)(void *vfs, btint unixUid, btint unixGid, void *caller);
	btuni 		*(*vfsPath)(void *obj);
}backendInterface_t;

/* ------------------------------------------------------------------------- */

module_t	*mainModule;	/* pointer to our module description */

typedef struct mountInfo{
	btu8		*mountpoint;
	btu8		*device;
	btint		mountedBy;
	module_t	*frontendMod;
	void		*frontendObj;
	module_t	*backendMod;
	void		*backendObj;
	btbool		isAutomounted;
}mountInfo_t;

static hashtable_t	*activeMounts = NULL;

static btint	activeMountCompare(void *o1, void *o2)
{
mountInfo_t	*this = o1, *other = o2;

	return strcmp(this->mountpoint, other->mountpoint);
}

static btu32	activeMountHashVal(void *obj)
{
mountInfo_t	*this = obj;

	return utilStrHash(this->mountpoint, 1);
}

static collectionObjectI_t	activeMountsInterface = {
	activeMountCompare,
	activeMountHashVal,
	NULL,
	NULL
};

static hashtable_t	*mountsByDevice = NULL;

static btint	mountsByDeviceCompare(void *o1, void *o2)
{
mountInfo_t	*this = o1, *other = o2;

	return strcmp(this->device, other->device);
}

static btu32	mountsByDeviceHashVal(void *obj)
{
mountInfo_t	*this = obj;

	return utilStrHash(this->device, 1);
}

static collectionObjectI_t	mountsByDeviceInterface = {
	mountsByDeviceCompare,
	mountsByDeviceHashVal,
	NULL,
	NULL
};

static guiObject_t	*mountListObject;

/* ------------------------------------------------------------------------- */

struct notificationCenter	*mainModMountNotification(void)
{
static notificationCenter_t	*nc = NULL;

	if(nc == NULL)
		nc = ncNew(4);
	return nc;
}

struct notificationCenter	*mainModUmountNotification(void)
{
static notificationCenter_t	*nc = NULL;

	if(nc == NULL)
		nc = ncNew(4);
	return nc;
}

/* ------------------------------------------------------------------------- */

btint	mainModCfgPath(struct cfgProperty **path, btint maxEntries)
{
cfgProperty_t	*p;
btint			i = 0;

	p = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.dbFile));
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	p = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.cfgFile));
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	return i;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------ gui actions ------------------------------ */
/* ------------------------------------------------------------------------- */

static btint	guiObjectCompare(void *o1, void *o2)
{
guiObject_t	*this = o1, *other = o2;

	return this->id - other->id;
}

static btu32	guiObjectHashVal(void *obj)
{
guiObject_t	*this = obj;

	return this->id;
}

static collectionObjectI_t	guiObjectInterface = {
	guiObjectCompare,
	guiObjectHashVal,
	NULL,
	NULL
};

static hashtable_t	*allGuiObjects = NULL;
static array_t		*advertizedObjects = NULL;

guiObject_t	*mainModRegisterDataSource(void *dsource, guiClass_t *class, guiDsourceMethod_t *methods, btint numMethods)
{
guiObject_t		*go;
static btint	idCnt = 1;	/* 0 is reserved, e.g. for wildcards */

	if(allGuiObjects == NULL)
		allGuiObjects = htabNew(&guiObjectInterface, 16, 75);
	go = memCalloc(sizeof(guiObject_t));
	go->id = idCnt++;
	if(idCnt == 0)	/* wrap-around is unlikely, but take care of it */
		idCnt++;
	go->class = class;
	go->dsource = dsource;
	go->methods = methods;
	go->numMethods = numMethods;
	htabAddObject(allGuiObjects, go);
	return go;
}

void	mainModUnregisterDataSource(guiObject_t *guiObj)
{
	htabRemoveObjectEqualTo(allGuiObjects, guiObj);
	memFree(guiObj);
}

static guiObject_t	*mainModFindDataSourceFromId(btint id)
{
	if(allGuiObjects == NULL)
		return NULL;
	return htabFindObjectEqualTo(allGuiObjects, &id);
}

typedef struct advertizeInfo{
	btint	precedence;
	btint	objectId;
}advertizeInfo_t;

static btint	aiCompare(void *o1, void *o2)
{
advertizeInfo_t	*this = o1, *other = o2;

	return this->precedence - other->precedence;
}

static collectionObjectI_t	aiInterface = {
	aiCompare,
	NULL,
	NULL,
	NULL
};

void	mainModAdvertizeConfig(guiObject_t *guiObj, btint precedence)
{
advertizeInfo_t	*ai;

	if(advertizedObjects == NULL)
		advertizedObjects = arrNew(&aiInterface, 16);
	ai = memCalloc(sizeof(*ai));
	ai->precedence = precedence;
	ai->objectId = guiObj->id;
	arrAddObject(advertizedObjects, ai);
	arrSort(advertizedObjects);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct mainCtx{
	PROC_VARSET;
	btint			uid;
	btint			fd;
	btint			index;
	ipcMsg_t		*reply;
	ipcMsg_t		*request;
	module_t		*backendModule;
	module_t		*frontendModule;
	backendInterface_t	*backendI;
	frontendInterface_t	*frontendI;
	void			*backendObj;
	void			*mp;
	mountInfo_t		*minfo;
	btuni			*backendPath;
	btu8			*mountpoint;
	btbool			readOnly;
	btbool			fakeMode;
	btbool			isAutomount;
	ipcSrvClient_t	*client;
	void			*panelData;
	btError			err;
	array_t			*all;
}mainCtx_t;

static blockpool_t *mainPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(mainPool, block);
}

/* ------------------------------------------------------------------------- */

static hashtable_t	*moduleInterfaces = NULL;

static btint	interfaceCompare(void *o1, void *o2)
{
backendInterface_t	*this = o1, *other = o2;

	return (btu8 *)this->forModule - (btu8 *)other->forModule;
}

static btu32	interfaceHashVal(void *o)
{
backendInterface_t	*this = o;
btint				x = (btint)this->forModule;

	return x ^ (x >> 6) ^ (x >> 12);
}

static collectionObjectI_t	interfaceI = {
	interfaceCompare,
	interfaceHashVal,
	NULL,
	NULL
};

static void	*findInterface(module_t *mod)
{
	return htabFindObjectEqualTo(moduleInterfaces, &mod);
}

/* ------------------------------------------------------------------------- */

static inline char	*ps(char *s)
{
	return s == NULL ? "(null string)" : s;
}

/* ------------------------------------------------------------------------- */

static void	doUpdates(void *obj, void *arg)
{
btint			i, j, cnt;
array_t			*all, *arr, *clients;
guiObject_t		*go;
ipcSrvClient_t	*client;
ipcMsg_t		*msg;

	arr = arrNew(NULL, 16);
	all = htabAllObjects(allGuiObjects);
	cnt = arrObjectCount(all);
	for(i=0;i<cnt;i++){
		go = arrObjectAtIndex(all, i);
		if(go->needUpdate){
			arrAddObject(arr, go);
			go->needUpdate = 0;
		}
	}
	arrFree(all);
	clients = mainModule->ipcModule->servingClients;
	for(j=0;j<arrObjectCount(clients);j++){
		client = arrObjectAtIndex(clients, j);
		if(ipcSrvClientIsInvalid(client))
			continue;
		msg = ipcMsgNew();
		ipcAddInt(msg, 0);	/* space for size */
		ipcAddInt(msg, 0);	/* space for protocol version */
		ipcAddInt(msg, 0);	/* space for SID */
		ipcAddInt(msg, MAIN_IPCCFN_NOTIFY_UPDATE);	/* procedure ID */
		ipcAddInt(msg, 0);	/* uid */
		ipcAddInt(msg, 0);	/* gid */
		cnt = arrObjectCount(arr);
		ipcAddInt(msg, cnt);
		for(i=0;i<cnt;i++){
			go = arrObjectAtIndex(arr, i);
			ipcAddInt(msg, go->id);
		}
		ipcSrvRPC(mainModule->ipcModule, client, msg, NULL);
		/* run in background and ignore result, we don't care */
	}
	arrFree(arr);
}

/* ------------------------------------------------------------------------- */

void	mainModGuiUpdate(guiObject_t *go)
{
array_t			*all;
btint			i, cnt;
static btint	invocationCount = 0;

	if(go != NULL){
		go->needUpdate = 1;
	}else{
		all = htabAllObjects(allGuiObjects);
		cnt = arrObjectCount(all);
		for(i=0;i<cnt;i++){
			go = arrObjectAtIndex(all, i);
			go->needUpdate = 1;
		}
		arrFree(all);
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: update for go=0x%08x\n", (int)go);
	elpCallAtCancelPrev(elpNow() + 20, doUpdates, (void *)(invocationCount++/100));
/* This invocation count hack ensures that if updates are forced frequently,
 * at least every 100th update will be sent to the clients (and not wait for
 * more updates).
 */
}

/* ------------------------------------------------------------------------- */

typedef struct{
	btError	err;
	btu8	*path;
}npRes_t;

static void	normalizePathPrimitive(void *arg, btint resultFd)
{
btu8	realPath[4 + MAXPATHLEN];
btint	totalLen, written, l;

	mainSignalHandlers(1);
	*(btu32 *)realPath = 0;	/* prepend error number */
	if(sysRealpath(arg, realPath + 4) == NULL){
		*(btu32 *)realPath = errno;
		totalLen = 4;
	}else{
		totalLen = 4 + strlen(realPath + 4);
	}
	for(written=0;written<totalLen;written+=l){
		l = write(resultFd, realPath + written, totalLen - written);
		if(l <= 0)
			break;
	}
}

static void	normalizePathProc(void *obj, void *arg)
{
mainCtx_t	*this = obj;
btError		*err = arg;
npRes_t		res;

	PROC_BEGIN;
	PROC_WAITFOR(utilExecInBackground(normalizePathPrimitive, this->mountpoint, &this->reply, this));
	res.err = *err;
	if(err->class == 0){
		ipcMsgEnsureSize(this->reply, this->reply->size + 1);
		this->reply->buffer[this->reply->size] = 0;	/* 0-terminate string */
		res.err.code = *(btu32 *)this->reply->buffer;
		if(res.err.code != 0)
			res.err.class = ERRCLS_UNIX;
		res.path = this->reply->buffer + 4;
		PROC_RETURN_AND_CLEANUP(&res,
			ipcMsgFree(this->reply);
		);
	}else{
		PROC_RETURN(&res);
	}
	PROC_END;
}

static void	normalizePath(btu8 *path, void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, normalizePathProc, caller, mainPool, poolFreeBlock);
	ctx->mountpoint = path;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	checkMountPermissionsProc(void *obj, void *arg)
{
mainCtx_t		*this = obj;
btint			pid;
btError			res;
struct stat		sbuf;
elpPid_t		*prval = arg;
btu8			*p;

	PROC_BEGIN;
	pid = fork();
	if(pid == -1){
		res.class = ERRCLS_UNIX;
		res.code = errno;
		PROC_RETURN(&res);
	}
	if(pid == 0){	/* this is the child */
		mainSignalHandlers(1);
		if(stat(this->mountpoint, &sbuf) == -1){
			if(errno != ENOENT)
				exit(errno);
			/* try with parent: */
			if((p = strrchr(this->mountpoint, '/')) != NULL){
				*p = 0;
				if(stat(this->mountpoint, &sbuf) == -1)
					exit(errno);
			}
		}
		exit(sbuf.st_uid == this->uid ? 0 : EACCES);
	}else{
		PROC_WAITFOR(elpRegisterForPid(pid, checkMountPermissionsProc, this));
		if(prval->rval == 0){
			res.class = res.code = 0;
		}else{
			res.class = ERRCLS_UNIX;
			res.code = prval->rval;
		}
		PROC_RETURN(&res);
	}
	PROC_END;
}

static void	checkMountPermissions(btu8 *path, btint uid, void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, checkMountPermissionsProc, caller, mainPool, poolFreeBlock);
	ctx->mountpoint = path;
	ctx->uid = uid;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	mainModMountProc(void *obj, void *arg)
{
mainCtx_t		*this = obj;
btError			res, *crval = arg;
beVfsRes_t		*rval = arg;
feRes_t			*frval = arg;
mountInfo_t		*minfo;
npRes_t			*nrval = arg;
struct passwd	*pwent;
btint			gid;

	PROC_BEGIN;
	if(this->backendModule == NULL || this->backendI == NULL || this->frontendModule == NULL || this->frontendI == NULL){
		res.class = ERRCLS_FRAMEWORK;
		res.code = ERR_FRWmodulenotfound;
		goto doReturn;
	}
	if(this->backendModule->moduleType != MOD_TYPE_BACKEND || this->frontendModule->moduleType != MOD_TYPE_FRONTEND){
		res.class = ERRCLS_FRAMEWORK;
		res.code = ERR_FRWmoduletypeerr;
		goto doReturn;
	}
	/* Get a representation for backendPath now because we did not copy
	 * the string and it might become invalid during PROC_WAITs
	 */
	this->backendObj = this->backendI->vfsForPath(this->backendPath);
	/* Do the login and vfsConnect before we check the mountpoint. This allows
	 * users to use cifsmount to supply the share password for already mounted
	 * shares. The mount will fail (because the mountpoint is already used),
	 * but the user will gain access.
	 */
	if(!this->fakeMode){
		gid = 0;
		if((pwent = getpwuid(this->uid)) != NULL)
			gid = pwent->pw_gid;
		PROC_WAITFOR(this->backendI->userAuthenticate(this->backendObj, this->uid, gid, this));
		if(rval->err.class != 0){
			res = rval->err;
			goto doReturn;
		}
		PROC_WAITFOR(this->backendI->vfsConnect(this->backendObj, rval->user, this));
		if(rval->err.class != 0){
			res = rval->err;
			goto doReturn;
		}
	}
	if(!this->isAutomount){	/* automount is trusted to provide normalized pathes */
		PROC_WAITFOR(normalizePath(this->mountpoint, this));
		if(nrval->err.class != 0){
			res = nrval->err;
			goto doReturn;
		}
		memFree(this->mountpoint);
		this->mountpoint = utilStrDup(nrval->path);
	}
	if(htabFindObjectEqualTo(activeMounts, &this->mountpoint) != NULL){	/* already something mounted */
		res.class = ERRCLS_FRAMEWORK;
		res.code = ERR_FRWalreadymounted;
		goto doReturn;
	}
	if(!this->isAutomount && this->uid != 0){	/* automount and root are trusted */
		PROC_WAITFOR(checkMountPermissions(this->mountpoint, this->uid, this));
		if(crval->class != 0){
			res = *crval;
			goto doReturn;
		}
	}
	this->mp = this->frontendI->mntNew(this->mountpoint);
	this->frontendI->mntSetBackend(this->mp, this->backendModule, this->backendObj);
	this->frontendI->mntSetFakemode(this->mp, this->fakeMode);
	PROC_WAITFOR(this->frontendI->mntMount(this->mp, this->uid, this->readOnly, this));
	if(frval->err.class != 0){
		this->frontendI->mntRelease(this->mp);
		res = frval->err;
		goto doReturn;
	}
	res = frval->err;
	this->frontendI->mntSetFakemode(this->mp, 0);	/* don't fake after mounting */
	minfo = memCalloc(sizeof(mountInfo_t));
	minfo->mountpoint = this->mountpoint;
	this->mountpoint = NULL;	/* avoid freeing at return */
	minfo->device = this->frontendI->mntMtabDevice(this->mp);
	minfo->mountedBy = this->uid;
	minfo->frontendMod = this->frontendModule;
	minfo->frontendObj = this->mp;
	minfo->backendMod = this->backendModule;
	minfo->backendObj = this->backendObj;
	minfo->isAutomounted = this->isAutomount;
	htabAddObject(activeMounts, minfo);
	htabAddObject(mountsByDevice, minfo);
	ncPostNotification(mainModMountNotification(), minfo->mountpoint);
	mainModGuiUpdate(mountListObject);
doReturn:
	if(this->backendObj != NULL)
		this->backendI->vfsRelease(this->backendObj);	/* should now be retained by frontend if succeeded */
	if(this->mountpoint != NULL)
		memFree(this->mountpoint);
	if(res.class == ERRCLS_UNIX){
		if(res.code == ENOENT){
/* translate Unix error "no such file or directory" to ERR_FRWnomountpoint */
			res.class = ERRCLS_FRAMEWORK;
			res.code = ERR_FRWnomountpoint;
		}
		if(res.code == EACCES){
			res.class = ERRCLS_FRAMEWORK;
			res.code = ERR_FRWnotownerofmp;
		}
	}
	PROC_RETURN(&res);
	PROC_END;
}

void	mainModMount(btint uid, btu8 *frontendModule, btu8 *mountpoint, btu8 *backendModule, btuni *mountedPath, btbool readOnly, btbool fakeMode, btbool isAutomount, void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, mainModMountProc, caller, mainPool, poolFreeBlock);
	ctx->uid = uid;
	ctx->backendModule = modFind(backendModule);
	ctx->backendI = findInterface(ctx->backendModule);
	ctx->backendPath = mountedPath;
	if(frontendModule == NULL)
		ctx->frontendModule = arrObjectAtIndex(modAllFrontends(), 0);
	else
		ctx->frontendModule = modFind(frontendModule);
	ctx->frontendI = findInterface(ctx->frontendModule);
	ctx->mountpoint = utilStrDup(mountpoint);
	ctx->readOnly = readOnly;
	ctx->fakeMode = fakeMode;
	ctx->isAutomount = isAutomount;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

btbool	mainModMountExists(btu8 *mountpoint)
{
	return htabFindObjectEqualTo(activeMounts, &mountpoint) != NULL;
}

/* ------------------------------------------------------------------------- */

static void	minfoFree(mountInfo_t *minfo, frontendInterface_t *fi)
{
	htabRemoveObjectEqualTo(activeMounts, minfo);
	htabRemoveObjectEqualTo(mountsByDevice, minfo);
	fi->mntRelease(minfo->frontendObj);
	ncPostNotification(mainModUmountNotification(), minfo->mountpoint);
	memFree(minfo->mountpoint);
	memFree(minfo);
	mainModGuiUpdate(mountListObject);
}

static void	mainModUmountProc(void *obj, void *arg)
{
mainCtx_t	*this = obj;
mountInfo_t	*minfo;
btError		res;
feRes_t		*rval = arg;
npRes_t		*nrval = arg;

	PROC_BEGIN;
	minfo = htabFindObjectEqualTo(activeMounts, &this->mountpoint);
	if(minfo == NULL){	/* not mounted */
		/* try to normalize and look it up then: */
		PROC_WAITFOR(normalizePath(this->mountpoint, this));
		if(nrval->err.class != 0){
			PROC_RETURN(&nrval->err);
		}
		minfo = htabFindObjectEqualTo(activeMounts, &nrval->path);
		if(minfo == NULL){	/* still not found */
			res.class = ERRCLS_FRAMEWORK;
			res.code = ERR_FRWnosuchmnt;
			PROC_RETURN(&res);
		}
	}
	if(minfo->mountedBy != this->uid && this->uid != 0){
		res.class = ERRCLS_FRAMEWORK;
		res.code = ERR_FRWumountperm;
		PROC_RETURN(&res);
	}
	this->minfo = minfo;
	this->frontendModule = minfo->frontendMod;
	this->frontendI = findInterface(this->frontendModule);
	this->mp = minfo->frontendObj;
	if(this->fakeMode){
		this->frontendI->mntSetFakemode(this->mp, 1);
	}
	PROC_WAITFOR(this->frontendI->mntUmount(this->mp, this->uid, this));
	if(rval->err.class != 0){
		this->frontendI->mntSetFakemode(this->mp, 0);
		PROC_RETURN(&rval->err);
	}
	minfoFree(this->minfo, this->frontendI);
	res.class = res.code = 0;
	PROC_RETURN(&res);
	PROC_END;
}

void	mainModUmount(btint uid, btu8 *mountpoint, btbool fakeMode, void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, mainModUmountProc, caller, mainPool, poolFreeBlock);
	ctx->uid = uid;
	ctx->mountpoint = mountpoint;
	ctx->fakeMode = fakeMode;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	mainModNotifyUnmount(btu8 *device)
{
mountInfo_t			*minfo, dummy;
frontendInterface_t	*fi;

	dummy.device = device;
	if((minfo = htabFindObjectEqualTo(mountsByDevice, &dummy)) != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_DISKARB))
			logPrintf(LOGM_DISKARB, "mainmodule: unmount notify for mp=%s\n", minfo->mountpoint);
		fi = findInterface(minfo->frontendMod);
		fi->mntWasUnmounted(minfo->frontendObj);
		minfoFree(minfo, fi);
	}else{
		if(LOG_MODE_IS_ACTIVE(LOGM_DISKARB))
			logPrintf(LOGM_DISKARB, "mainmodule: mount for device %s not found\n", device);
	}
}

/* ------------------------------------------------------------------------- */

btint	mainModUserForMtabDevice(btu8 *device)
{
mountInfo_t			*minfo, dummy;

	dummy.device = device;
	if((minfo = htabFindObjectEqualTo(mountsByDevice, &dummy)) != NULL){
		return minfo->mountedBy;
	}else{
		return -1;
	}
}

/* ------------------------------------------------------------------------- */

static void	marshalError(ipcMsg_t *reply, btint class, btint code)
{
btError	err;

	ipcAddInt(reply, class);
	ipcAddInt(reply, code);
	err.class = class;
	err.code = code;
	ipcAddAscii(reply, errDescription(&err));
}

/* ------------------------------------------------------------------------- */
/* ----------------------------- ipc functions ----------------------------- */
/* ------------------------------------------------------------------------- */

static void	listModules(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
array_t		*allModules = modAllModules();
module_t	*mod;
btint		i, cnt;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: listModules(%d)\n", uid);
	marshalError(reply, 0, 0);
	cnt = arrObjectCount(allModules);
	ipcAddInt(reply, cnt);
	for(i=0;i<cnt;i++){
		mod = arrObjectAtIndex(allModules, i);
		ipcAddInt(reply, mod->moduleType);
		ipcAddAscii(reply, mod->name);
		ipcAddInt(reply, mod->interfaceVersion);
		ipcAddInt(reply, mod->implementationVersion);
		ipcAddInt(reply, mod->betaVersion);
		if(mod->ipcModule != NULL){
			ipcAddInt(reply, mod->ipcModule->fidBase);
			ipcAddInt(reply, mod->ipcModule->fidCount);
		}else{
			ipcAddInt(reply, 0);
			ipcAddInt(reply, 0);
		}
	}
	arrFree(allModules);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	slashChars(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8	*s, slash[16], backsl[16];

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: slashChars(%d)\n", uid);
	marshalError(reply, 0, 0);
	s = slash;
	convCharUniToLocalAscii(UNI_CHAR_SLASH, &s);
	*s++ = 0;
	ipcAddAscii(reply, slash);
	s = backsl;
	convCharUniToLocalAscii(UNI_CHAR_BACKSLASH, &s);
	*s++ = 0;
	ipcAddAscii(reply, backsl);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	mountAnyProc(void *obj, void *arg)
{
mainCtx_t	*this = obj;
btu8		*frontendMod, *frontendPath, *backendMod, *backendPath;
btbool		readOnly, fakeMode;
btError		*rval = arg;
btuni		vfsPath[2048];

	PROC_BEGIN;
	frontendMod = ipcGetAscii(this->request);
	frontendPath = ipcGetAscii(this->request);
	backendMod = ipcGetAscii(this->request);
	backendPath = ipcGetAscii(this->request);
	readOnly = ipcGetInt(this->request);
	fakeMode = ipcGetInt(this->request);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: mountAny(%d, %s, %s, %s, %s, %d, %d)\n", this->uid, ps(frontendMod), ps(frontendPath), ps(backendMod), ps(backendPath), readOnly, fakeMode);
	if(frontendPath == NULL || backendMod == NULL || backendPath == NULL){
		marshalError(this->reply, ERRCLS_FRAMEWORK, ERR_FRWipcparameters);
	}else{
		convLocalAsciiToUni(backendPath, vfsPath, ELEMENTS(vfsPath));
		PROC_WAITFOR(mainModMount(this->uid, frontendMod, frontendPath, backendMod, vfsPath, readOnly, fakeMode, 0, this));
		marshalError(this->reply, rval->class, rval->code);
	}
	PROC_RETURN(NULL);
	PROC_END;
}

static void	mountAny(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, mountAnyProc, caller, mainPool, poolFreeBlock);
	ctx->reply = reply;
	ctx->request = rq;
	ctx->uid = uid;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	umountAnyProc(void *obj, void *arg)
{
mainCtx_t	*this = obj;
btu8		*mountpoint;
btError		*rval = arg;
array_t		*allMounts;
btint		i, cnt;
mountInfo_t	*mi;

	PROC_BEGIN;
	mountpoint = ipcGetAscii(this->request);
	this->fakeMode = ipcGetInt(this->request);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: umountAny(%d, %s, %d)\n", this->uid, ps(mountpoint), this->fakeMode);
	if(mountpoint == NULL){	/* NULL means unmount all mounts */
		allMounts = htabAllObjects(activeMounts);
		cnt = arrObjectCount(allMounts);
		this->all = arrNew(NULL, cnt + 1);
		for(i=0;i<cnt;i++){
			mi = arrObjectAtIndex(allMounts, i);
			arrAddObject(this->all, utilStrDup(mi->mountpoint));
		}
		arrFree(allMounts);
		for(this->index=0; this->index<arrObjectCount(this->all); this->index++){
			PROC_WAITFOR(mainModUmount(this->uid, arrObjectAtIndex(this->all, this->index), this->fakeMode, this));
			if(rval->class != 0)
				this->err = *rval;
		}
		cnt = arrObjectCount(this->all);
		for(i=0;i<cnt;i++)
			memFree(arrObjectAtIndex(this->all, i));
		arrFree(this->all);
		marshalError(this->reply, this->err.class, this->err.code);
	}else{
		PROC_WAITFOR(mainModUmount(this->uid, mountpoint, this->fakeMode, this));
		marshalError(this->reply, rval->class, rval->code);
	}
	PROC_RETURN(NULL);
	PROC_END;
}

static void	umountAny(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, umountAnyProc, caller, mainPool, poolFreeBlock);
	ctx->reply = reply;
	ctx->request = rq;
	ctx->uid = uid;
	PROC_START(ctx, NULL);
}


/* ------------------------------------------------------------------------- */

static btint	storeMount(btint uid, btu8 *storeMountpoint)
{
btu8				*mountpoint, buffer[1024];
mountInfo_t			*mi;
cfgProperty_t		*p, *parent;
btint				i, cnt, numElements;
btuni				*path;
btbool				readOnly, usersMayStore;
frontendInterface_t	*frontendI;
backendInterface_t	*backendI;
cfgProperty_t		*cfgPath[8];

	mi = htabFindObjectEqualTo(activeMounts, &storeMountpoint);
	if(mi == NULL)
		return ERR_FRWnosuchobj;
	if(uid != 0 && mi->mountedBy != uid)
		return ERR_FRWmpaccessperm;
	numElements = mainModCfgPath(cfgPath, ELEMENTS(cfgPath));
	usersMayStore = cfgEnumValue(cfgQueryFromArray("usersMayStoreSessionData", numElements, cfgPath), cfgYesNoEnum, 1);
	if(uid != 0 && !usersMayStore)
		return ERR_FRWstoredisabled;
	parent = cfgFilePlist(gdata.dbFile);
	if((p = cfgQueryWithName(MAINMOD_CFGNAME, 1, parent)) == NULL){
		p = cfgBuildDictProperty();
		cfgSetProperty(MAINMOD_CFGNAME, parent, p);
	}
	parent = p;
	if((p = cfgQueryWithName("automount", 1, parent)) == NULL){
		p = cfgBuildArrayProperty();
		cfgSetProperty("automount", parent, p);
	}
	parent = p;
	cnt = cfgQueryValueCount(parent);
	for(i=0;i<cnt;i++){
		p = cfgQueryAtIndex(i, parent);
		mountpoint = cfgStringValue(cfgQueryWithName("mountpoint", 1, p), NULL);
		if(mountpoint != NULL && strcmp(storeMountpoint, mountpoint) == 0){
			cfgRemoveFromArrayProperty(parent, i);
			i--;
			cnt--;
		}
	}
	p = cfgBuildDictProperty();
	cfgAppendToArrayProperty(parent, p);
	frontendI = findInterface(mi->frontendMod);
	backendI = findInterface(mi->backendMod);
	cfgSetProperty("mountpoint", p, cfgBuildStringProperty(mi->mountpoint));
	cfgSetProperty("frontend", p, cfgBuildStringProperty(mi->frontendMod->name));
	cfgSetProperty("backend", p, cfgBuildStringProperty(mi->backendMod->name));
	if(backendI->vfsPath != NULL){
		path = backendI->vfsPath(mi->backendObj);
		convUniToLocalAscii(path, buffer, ELEMENTS(buffer));
	}else{
		strcpy(buffer, "unknown");
	}
	cfgSetProperty("backendSpec", p, cfgBuildStringProperty(buffer));
	cfgSetProperty("owner", p, cfgBuildIntProperty(mi->mountedBy));
	if(frontendI->mntIsReadonly != NULL){
		readOnly = frontendI->mntIsReadonly(mi->frontendObj);
	}else{
		readOnly = 0;
	}
	cfgSetProperty("readOnly", p, cfgBuildEnumProperty(readOnly, cfgYesNoEnum));
	cfgSetProperty("isAutomount", p, cfgBuildEnumProperty(mi->isAutomounted, cfgYesNoEnum));
	cfgUnparse(gdata.dbFile);
	mainModGuiUpdate(mountListObject);
	return 0;
}

static btint	delStoredMount(btint uid, btu8 *delMountpoint)
{
btu8			*mountpoint;
btbool			found = 0, usersMayStore;
btint			i, cnt, owner, numElements;
cfgProperty_t	*parent, *p;
cfgProperty_t	*path[8];

	numElements = mainModCfgPath(path, ELEMENTS(path));
	usersMayStore = cfgEnumValue(cfgQueryFromArray("usersMayStoreSessionData", numElements, path), cfgYesNoEnum, 1);
	if(uid != 0 && !usersMayStore)
		return ERR_FRWstoredisabled;
	parent = cfgFilePlist(gdata.dbFile);
	if((p = cfgQueryWithName(MAINMOD_CFGNAME, 1, parent)) == NULL)
		return ERR_FRWnosuchobj;
	parent = p;
	if((p = cfgQueryWithName("automount", 1, parent)) == NULL)
		return ERR_FRWnosuchobj;
	parent = p;
	cnt = cfgQueryValueCount(parent);
	for(i=0;i<cnt;i++){
		p = cfgQueryAtIndex(i, parent);
		mountpoint = cfgStringValue(cfgQueryWithName("mountpoint", 1, p), NULL);
		owner = cfgIntValue(cfgQueryWithName("owner", 1, p), 0);
		if(mountpoint != NULL && strcmp(delMountpoint, mountpoint) == 0){
			if(uid != 0 && owner != uid)
				return ERR_FRWmpaccessperm;
			cfgRemoveFromArrayProperty(parent, i);
			i--;
			cnt--;
			found = 1;
		}
	}
	if(!found)
		return ERR_FRWnosuchobj;
	cfgUnparse(gdata.dbFile);
	mainModGuiUpdate(mountListObject);
	return 0;
}

/* ------------------------------------------------------------------------- */

static void	storeMountRpc(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8	*mountpoint;
btbool	doDelete;
btint	rval;

	mountpoint = ipcGetAscii(rq);
	doDelete = ipcGetInt(rq);
	if(mountpoint == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWipcparameters);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	if(doDelete)
		rval = delStoredMount(uid, mountpoint);
	else
		rval = storeMount(uid, mountpoint);
	marshalError(reply, (rval != 0 ? ERRCLS_FRAMEWORK : 0), rval);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

typedef struct listDescriptor{
	btu8	*shortName;
	btint	objectId;
}listDescriptor_t;

static btint	listDescriptorCompare(void *o1, void *o2)
{
listDescriptor_t	*this = o1, *other = o2;

	return utilStrCmp(this->shortName, other->shortName, 1);
}

static btu32	listDescriptorHash(void *obj)
{
listDescriptor_t	*ld = obj;

	return utilStrHash(ld->shortName, 1);
}

static collectionObjectI_t	listDescriptorInterface = {
	listDescriptorCompare,
	listDescriptorHash,
	NULL,
	NULL,
};

static hashtable_t	*listsByName;

/* ------------------------------------------------------------------------- */

void	mainModRegisterList(btu8 *shortName, btint objectId)
{
listDescriptor_t	*ld;

	ld = memCalloc(sizeof(*ld));
	ld->shortName = utilStrDup(shortName);
	ld->objectId = objectId;
	htabAddObject(listsByName, ld);
}

/* ------------------------------------------------------------------------- */

static void	lookupList(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8				*shortName;
listDescriptor_t	*ld;

	shortName = ipcGetAscii(rq);
	if(shortName == NULL || (ld = htabFindObjectEqualTo(listsByName, &shortName)) == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnosuchobj);
	}else{
		marshalError(reply, 0, 0);
		ipcAddInt(reply, ld->objectId);
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	helpForKey(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8				*helpKey;

	helpKey = ipcGetAscii(rq);
	if(helpKey == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnosuchobj);
	}else{
		marshalError(reply, 0, 0);
		ipcAddAscii(reply, helpTextForItem(helpKey));
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static void	registerCb(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8		*modName;
module_t	*module;

	modName = ipcGetAscii(rq);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: registerCb(%d, %s)\n", uid, ps(modName));
	module = modFind(modName);
	if(module == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWmodulenotfound);
	}else{
		if(module->ipcModule != NULL){
			ipcRegisterClientService(module->ipcModule, client, uid);
			marshalError(reply, 0, 0);
		}else{
			marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWmodulenotfound);
		}
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	unregisterCb(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8		*modName;
module_t	*module;

	modName = ipcGetAscii(rq);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: unregisterCb(%d, %s)\n", uid, ps(modName));
	module = modFind(modName);
	if(module == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWmodulenotfound);
	}else{
		if(module->ipcModule != NULL){
			ipcUnregisterClientService(module->ipcModule, client);
			marshalError(reply, 0, 0);
		}else{
			marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWmodulenotfound);
		}
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static array_t	*uiTerminateFunctions = NULL;

void	mainModRegisterForUiTerminate(mainModUiTermFunction_t fn)
{
	if(uiTerminateFunctions == NULL)
		uiTerminateFunctions = arrNew(NULL, 4);
	arrRemoveObjectEqualTo(uiTerminateFunctions, fn);
	arrAddObject(uiTerminateFunctions, fn);
}

void	mainModUnregisterForUiTerminate(mainModUiTermFunction_t fn)
{
	arrRemoveObjectEqualTo(uiTerminateFunctions, fn);
}

/* ------------------------------------------------------------------------- */

static void	uiWillTerminateProc(void *obj, void *arg)
{
mainCtx_t				*this = obj;
mainModUiTermFunction_t	fn;
btbool					allowTerminate = (btbool)(int)arg;

	PROC_BEGIN;
	this->fd = ipcGetInt(this->request);	/* termReason */
	for(this->index=0; this->index < arrObjectCount(uiTerminateFunctions); this->index++){
		fn = arrObjectAtIndex(uiTerminateFunctions, this->index);
		PROC_WAITFOR(fn(this->client, this->fd, this));
		if(!allowTerminate)
			break;
	}
	marshalError(this->reply, 0, 0);
	if(this->index < arrObjectCount(uiTerminateFunctions)){	/* we did break */
		ipcAddInt(this->reply, 0);	/* don't allow termination */
	}else{
		ipcAddInt(this->reply, 1);	/* allow termination */
	}
	PROC_END;
}

static void	uiWillTerminate(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, uiWillTerminateProc, caller, mainPool, poolFreeBlock);
	ctx->reply = reply;
	ctx->request = rq;
	ctx->uid = uid;
	ctx->client = client;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	execDsourceMethod(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btint		methodIndex, dsourceId;
guiObject_t	*go;


	dsourceId = ipcGetInt(rq);
	methodIndex = ipcGetInt(rq);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: execDsourceMethod(%d, id=%d, method=%d)\n", uid, dsourceId, methodIndex);
	go = mainModFindDataSourceFromId(dsourceId);
	if(go == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnosuchobj);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	if(methodIndex >= go->numMethods || methodIndex < 0){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnosuchmethod);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	go->methods[methodIndex](client, uid, gid, go->dsource, rq, reply, caller);
}

/* ------------------------------------------------------------------------- */

static void	getCfgValue(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8			*cfgPath, *p;
btbool			dbOnly;
array_t			*levels;
btint			cnt, i;
cfgProperty_t	*cfg, *db;

	cfgPath = ipcGetAscii(rq);
	dbOnly = ipcGetInt(rq);
	levels = arrNew(NULL, 8);
	for(;;){
		arrAddObject(levels, cfgPath);
		p = strchr(cfgPath, '/');
		if(p == NULL)
			break;
		*p = 0;
		cfgPath = p + 1;
	}
	cfg = cfgFilePlist(gdata.cfgFile);
	db = cfgFilePlist(gdata.dbFile);
	cnt = arrObjectCount(levels);
	for(i=0;i<cnt;i++){
		cfg = cfgQueryWithName(arrObjectAtIndex(levels, i), 1, cfg);
		db = cfgQueryWithName(arrObjectAtIndex(levels, i), 1, db);
	}
	arrFree(levels);
	if(db != NULL || dbOnly)
		p = cfgStringValue(db, NULL);
	else
		p = cfgStringValue(cfg, NULL);
	marshalError(reply, 0, 0);
	ipcAddAscii(reply, p);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	setCfgValue(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8			*cfgPath, *p, *value;
array_t			*levels;
btint			cnt, i;
cfgProperty_t	*db, *newProp;

	cfgPath = ipcGetAscii(rq);
	value = ipcGetAscii(rq);
	levels = arrNew(NULL, 8);
	for(;;){
		arrAddObject(levels, cfgPath);
		p = strchr(cfgPath, '/');
		if(p == NULL)
			break;
		*p = 0;
		cfgPath = p + 1;
	}
	db = cfgFilePlist(gdata.dbFile);
	cnt = arrObjectCount(levels);
	for(i=0;i<cnt-1;i++){
		newProp = cfgQueryWithName(arrObjectAtIndex(levels, i), 1, db);
		if(newProp == NULL){
			newProp = cfgBuildDictProperty();
			cfgSetProperty(arrObjectAtIndex(levels, i), db, newProp);
		}
		db = newProp;
	}
	newProp = cfgBuildStringProperty(value);
	cfgSetProperty(arrObjectAtIndex(levels, i), db, newProp);
	arrFree(levels);
	cfgUnparse(gdata.dbFile);
	marshalError(reply, 0, 0);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	delCfgValue(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btu8			*cfgPath, *p;
array_t			*levels;
btint			cnt, i;
cfgProperty_t	*db, *newProp;

	cfgPath = ipcGetAscii(rq);
	levels = arrNew(NULL, 8);
	for(;;){
		arrAddObject(levels, cfgPath);
		p = strchr(cfgPath, '/');
		if(p == NULL)
			break;
		*p = 0;
		cfgPath = p + 1;
	}
	db = cfgFilePlist(gdata.dbFile);
	cnt = arrObjectCount(levels);
	for(i=0;i<cnt-1;i++){
		newProp = cfgQueryWithName(arrObjectAtIndex(levels, i), 1, db);
		if(newProp == NULL){
			goto doReturn;	/* not found, no need to delete anything */
		}
		db = newProp;
	}
	cfgSetProperty(arrObjectAtIndex(levels, i), db, NULL);
	arrFree(levels);
	cfgUnparse(gdata.dbFile);
doReturn:
	marshalError(reply, 0, 0);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	marshalColumn(ipcMsg_t *reply, guiColumn_t *column)
{
	ipcAddAscii(reply, column->title);
	ipcAddInt(reply, column->alignment);
	ipcAddInt(reply, column->widthPercent);
	ipcAddAscii(reply, helpTextForItem(column->helpItem));
}

static void	marshalEntry(ipcMsg_t *reply, guiEntry_t *entry)
{
guiTypedesEnum_t	*enumDes;
guiTypedesNumeric_t	*numDes;
btint				j, items;

	ipcAddInt(reply, entry->type);
	ipcAddInt(reply, entry->flags);
	ipcAddInt(reply, entry->accessPermissions);
	ipcAddAscii(reply, entry->title);
	ipcAddAscii(reply, helpTextForItem(entry->helpItem));
	switch(entry->type){
	case MAIN_GUI_ENTRY_STRING:
	case MAIN_GUI_ENTRY_PASSWORD:
	case MAIN_GUI_ENTRY_FILE:
	case MAIN_GUI_ENTRY_BOOL:
	case MAIN_GUI_ENTRY_SEPARATOR:
		/* no additional data necessary */
		break;
	case MAIN_GUI_ENTRY_ENUM:
		enumDes = entry->typeDescriptor;
		for(items=0; enumDes->externalNames[items]!=NULL; items++);
		ipcAddInt(reply, items);
		for(j=0;j<items;j++){
			ipcAddAscii(reply, enumDes->externalNames[j]);
		}
		break;
	case MAIN_GUI_ENTRY_NUMBER:
		numDes = entry->typeDescriptor;
		ipcAddInt(reply, numDes->fractionDivider);
		ipcAddInt(reply, numDes->fractionMultiplier);
		ipcAddInt(reply, numDes->increment);
		ipcAddInt(reply, numDes->minVal);
		ipcAddInt(reply, numDes->maxVal);
		ipcAddAscii(reply, numDes->unit);
		break;
	}
}

static void	getObjectDescription(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
btint				dsourceId;
guiObject_t			*go;
guiButton_t			*button;
btint				i, cnt;

	dsourceId = ipcGetInt(rq);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: getObjectDescription(%d, id=%d)\n", uid, dsourceId);
	go = mainModFindDataSourceFromId(dsourceId);
	if(go == NULL){
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnosuchobj);
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	marshalError(reply, 0, 0);
	ipcAddAscii(reply, go->class->name);
	ipcAddInt(reply, go->class->type);
	cnt = arrObjectCount(go->class->entries);
	ipcAddInt(reply, cnt);
	for(i=0;i<cnt;i++){
		if(go->class->type == MAIN_GUI_CLASS_LIST){
			marshalColumn(reply, arrObjectAtIndex(go->class->entries, i));
		}else{
			marshalEntry(reply, arrObjectAtIndex(go->class->entries, i));
		}
	}
	cnt = arrObjectCount(go->class->buttons);
	ipcAddInt(reply, cnt);
	for(i=0;i<cnt;i++){
		button = arrObjectAtIndex(go->class->buttons, i);
		ipcAddAscii(reply, button->title);
		ipcAddInt(reply, button->function);
		ipcAddInt(reply, button->parameter);
		ipcAddInt(reply, button->flags);
		ipcAddAscii(reply, helpTextForItem(button->helpItem));
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	getConfigObjects(ipcSrvClient_t *client, ipcMsg_t *rq, bti32 uid, bti32 gid, ipcMsg_t *reply, void *caller)
{
advertizeInfo_t	*ai;
btint			i, cnt;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "mainmodule: getConfigObjects(%d)\n", uid);
	marshalError(reply, 0, 0);
	if(advertizedObjects != NULL){
		cnt = arrObjectCount(advertizedObjects);
		ipcAddInt(reply, cnt);
		for(i=0;i<cnt;i++){
			ai = arrObjectAtIndex(advertizedObjects, i);
			ipcAddInt(reply, ai->objectId);
		}
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */
/* ------------------------ the GUI list of mounts ------------------------- */
/* ------------------------------------------------------------------------- */

guiColumn_t	mountListColumns[] = {
	/*  name			align	 width	help */
	{	"Mountpoint",		0,		30,	"mainMountListColumnMp"		},
	{	"Owner",			0,		15,	"mainMountListColumnOwner"	},
	{	"Module",			0,		15,	"mainMountListColumnModule"	},
	{	"Mounted Object",	0,		30,	"mainMountListColumnObject"	},
	{	"State",			0,		10,	"mainMountListColumnState"	},
};

static guiButton_t	mountListButtons[] = {
	{	"Mount...",			MAIN_GUI_METHOD_MOUNT,
		0,	/* parameter */	0,	/* flags */
		"mainMountListButtonMount",
	},
	{	"Unmount",			MAIN_GUI_METHOD_UMOUNT,
		0,	/* parameter */	0,	/* flags */
		"mainMountListButtonUnmount",
	},
	{	"Store",			MAIN_GUI_METHOD_STOREMOUNT,
		0,	/* parameter */	0,	/* flags */
		"mainMountListButtonStore",
	},
	{	"Delete Stored",	MAIN_GUI_METHOD_DELSTORED,
		0,	/* parameter */	0,	/* flags */
		"mainMountListButtonDelStored",
	},
};

/* ------------------------------------------------------------------------- */

typedef struct{
	btu8	*mp;
	btu8	*frontend;
	btu8	*backend;
	btu8	*objectSpec;
	btint	mountUid;
	btbool	isReadonly;
	btbool	isAutomounted;
	btbool	isMounted;
	btbool	isStored;
}mpRecord_t;

static void	addMpRecord(array_t *arr, btu8 *mp, btu8 *frontend, btu8 *obSpec, btu8 *backend, btint mountUid, btbool isReadonly, btbool isAutomounted, btbool isMounted, btbool isStored)
{
btint		cnt, i;
mpRecord_t	*rec;

	if(mp == NULL || obSpec == NULL)
		return;	/* ignore invalid records */
	cnt = arrObjectCount(arr);
	for(i=0;i<cnt;i++){
		rec = arrObjectAtIndex(arr, i);
		if(strcmp(rec->mp, mp) == 0 && strcmp(rec->frontend, frontend) == 0 && strcmp(rec->backend, backend) == 0 && strcmp(rec->objectSpec, obSpec) == 0){
			rec->isStored |= isStored;
			rec->isMounted |= isMounted;
			rec->isAutomounted |= isAutomounted;
			rec->isReadonly |= isReadonly;
			return;
		}
	}
	rec = memCalloc(sizeof(mpRecord_t));
	rec->mp = mp;
	rec->frontend = frontend;
	rec->objectSpec = utilStrDup(obSpec);
	rec->backend = backend;
	rec->mountUid = mountUid;
	rec->isReadonly = isReadonly;
	rec->isAutomounted = isAutomounted;
	rec->isMounted = isMounted;
	rec->isStored = isStored;
	arrAddObject(arr, rec);
}

static void	mountRecordFree(void *rec)
{
mpRecord_t	*mrec = rec;

	memFree(mrec->objectSpec);
	memFree(mrec);
}

static btint	mountRecordCompare(void *obj, void *otherObj)
{
mpRecord_t	*this = obj, *other = otherObj;
btint		rval;

	rval = strcmp(this->backend, other->backend);
	if(rval != 0)
		return rval;
	rval = strcmp(this->mp, other->mp);
	if(rval != 0)
		return rval;
	rval = strcmp(this->objectSpec, other->objectSpec);
	return rval;
}

static collectionObjectI_t	mountRecordInterface = {
	mountRecordCompare,	/* used for arrSort */
	NULL,
	NULL,
	mountRecordFree
};

/* ------------------------------------------------------------------------- */

static void	mountListGet(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8				buffer[256], *frontend, *backend, *mountpoint, *backendSpec;
btint				i, cnt, storedUid;
array_t				*all, *records;
mountInfo_t			*mi;
struct passwd		*pw;
backendInterface_t	*backendI;
frontendInterface_t	*frontendI;
btuni				*path;
btbool				isReadonly;
cfgProperty_t		*cfg, *db, *automounts, *entry;
mpRecord_t			*mrec;

	marshalError(reply, 0, 0);
	records = arrNew(&mountRecordInterface, 8);
	all = htabAllObjects(activeMounts);
	cnt = arrObjectCount(all);
	for(i=0;i<cnt;i++){
		mi = arrObjectAtIndex(all, i);
		frontendI = findInterface(mi->frontendMod);
		isReadonly = frontendI->mntIsReadonly(mi->frontendObj);
		backendI = findInterface(mi->backendMod);
		path = backendI->vfsPath(mi->backendObj);
		convUniToLocalAscii(path, buffer, ELEMENTS(buffer));
		addMpRecord(records, mi->mountpoint, mi->frontendMod->name, buffer, mi->backendMod->name, mi->mountedBy, isReadonly, mi->isAutomounted, 1, 0);
	}
	arrFree(all);
	cfg = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.cfgFile));
	db = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.dbFile));
	automounts = cfgQueryWithName("automount", 2, db, cfg);
	cnt = cfgQueryValueCount(automounts);
	for(i=0;i<cnt;i++){
		entry = cfgQueryAtIndex(i, automounts);
		frontend = cfgStringValue(cfgQueryWithName("frontend", 1, entry), NULL);
		backend = cfgStringValue(cfgQueryWithName("backend", 1, entry), NULL);
		mountpoint = cfgStringValue(cfgQueryWithName("mountpoint", 1, entry), NULL);
		backendSpec = cfgStringValue(cfgQueryWithName("backendSpec", 1, entry), NULL);
		storedUid = cfgIntValue(cfgQueryWithName("owner", 1, entry), 0);
		isReadonly = cfgEnumValue(cfgQueryWithName("readonly", 1, entry), cfgYesNoEnum, 0);
		if(frontend == NULL || backend == NULL || mountpoint == NULL || backendSpec == NULL)
			continue;	/* no valid entry */
		addMpRecord(records, mountpoint, frontend, backendSpec, backend, storedUid, isReadonly, 0, 0, 1);
	}
	arrSort(records);
	cnt = arrObjectCount(records);
	ipcAddInt(reply, cnt);
	for(i=0;i<cnt;i++){
		mrec = arrObjectAtIndex(records, i);
		ipcAddAscii(reply, mrec->mp);	/* row id */
		ipcAddAscii(reply, mrec->mp);	/* 1st column */
		pw = getpwuid(mrec->mountUid);	/* 2nd column: owner */
		if(pw != NULL){
			ipcAddAscii(reply, pw->pw_name);
		}else{
			sprintf(buffer, "#%d", mrec->mountUid);
			ipcAddAscii(reply, buffer);
		}
		ipcAddAscii(reply, mrec->backend);	/* backend module */
		ipcAddAscii(reply, mrec->objectSpec);
		buffer[0] = 0;
		if(mrec->isMounted)
			strcat(buffer, mrec->isAutomounted ? "A" : "M");
		if(mrec->isStored)
			strcat(buffer, "S");
		if(mrec->isReadonly)
			strcat(buffer, "R");
		ipcAddAscii(reply, buffer);
	}
	arrFree(records);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static guiTypedesEnum_t	enumdesFrontend = {
	NULL,	/* will be set during initialisation */
	NULL,	/* will be set during initialisation */
};
static guiTypedesEnum_t	enumdesBackend = {
	NULL,	/* will be set during initialisation */
	NULL,	/* will be set during initialisation */
};

typedef struct mountPanelData{
	btu8	*mountpoint;
	btint	frontend;
	btu8	*mountedObj;
	btint	backend;
	btbool	readOnly;
}mountPanelData_t;


static array_t	*frontendNames;
static array_t	*backendNames;

#define	PANEL_OFFSET(entry)	(void *)&(((mountPanelData_t *)NULL)->entry)

#define	ALLOW_NULL	0
#define	FLG_0		0

static guiEntry_t	mountDialogEntries[] = {
	{	MAIN_GUI_ENTRY_FILE,	"Mountpoint",
		NULL,					"mainMountDialogMountpoint",
		MAIN_GUI_PERM_ANYBODY,	FLG_0, ALLOW_NULL,	PANEL_OFFSET(mountpoint),
	},
	{	MAIN_GUI_ENTRY_ENUM,	"Frontend Module",
		&enumdesFrontend,		"mainMountDialogFrontend",
		MAIN_GUI_PERM_ANYBODY,	FLG_0, ALLOW_NULL,	PANEL_OFFSET(frontend),
	},
	{	MAIN_GUI_ENTRY_STRING,	"Mounted Object",
		NULL,					"mainMountDialogObject",
		MAIN_GUI_PERM_ANYBODY,	FLG_0, ALLOW_NULL,	PANEL_OFFSET(mountedObj),
	},
	{	MAIN_GUI_ENTRY_ENUM,	"Backend Module",
		&enumdesBackend,		"mainMountDialogBackend",
		MAIN_GUI_PERM_ANYBODY,	FLG_0, ALLOW_NULL,	PANEL_OFFSET(backend),
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Mount read-only",
		NULL,					"mainMountDialogReadonly",
		MAIN_GUI_PERM_ANYBODY,	FLG_0, ALLOW_NULL,	PANEL_OFFSET(readOnly),
	},
};

static guiButton_t	mountDialogButtons[] = {
	{	"Cancel",			MAIN_GUI_METHOD_CANCEL,
		0,	/* parameter */	MAIN_GUI_BUTTON_CLOSE_DIALOG | MAIN_GUI_BUTTON_EQUALS_CLOSEBOX,
		"mainMountDialogButtonCancel",
	},
	{	"OK",				MAIN_GUI_METHOD_SET,
		0,	/* parameter */	MAIN_GUI_BUTTON_CLOSE_DIALOG | MAIN_GUI_BUTTON_DEFAULT,
		"mainMountDialogButtonOk",
	},
};

static guiClass_t	mountDialog = {
	MAIN_GUI_CLASS_PLAINCFG,
	"Mount",
	NULL,
	NULL,
};

static void	mountListMountProc(void *obj, void *arg)
{
mainCtx_t			*this = obj;
btint				i, j, cnt;
btuni				buffer[512];
btError				err, *rval = arg;
guimanRes_t			*grval = arg;
mountPanelData_t	*panel;
btu8				*mountpoint, *frontend, *backend;
cfgProperty_t		*cfg, *db, *automounts, *entry;

	PROC_BEGIN;
	ipcSrvClientRetain(this->client);
	if(mountDialog.entries == NULL){
		mountDialog.entries = arrNew(NULL, 8);
		for(i=0;i<ELEMENTS(mountDialogEntries);i++)
			arrAddObject(mountDialog.entries, &mountDialogEntries[i]);
	}
	if(mountDialog.buttons == NULL){
		mountDialog.buttons = arrNew(NULL, 4);
		for(i=0;i<ELEMENTS(mountDialogButtons);i++)
			arrAddObject(mountDialog.buttons, &mountDialogButtons[i]);
	}
	panel = this->panelData = memCalloc(sizeof(mountPanelData_t));
	ipcGetInt(this->request);	/* ignore parameter */
	mountpoint = ipcGetAscii(this->request);
	/* row selected AND not already mounted? */
	if(mountpoint != NULL && htabFindObjectEqualTo(activeMounts, &mountpoint) == NULL){
		/* try to pre-initialize panel with stored data */
		cfg = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.cfgFile));
		db = cfgQueryWithName(MAINMOD_CFGNAME, 1, cfgFilePlist(gdata.dbFile));
		automounts = cfgQueryWithName("automount", 2, db, cfg);
		cnt = cfgQueryValueCount(automounts);
		for(i=0;i<cnt;i++){
			entry = cfgQueryAtIndex(i, automounts);
			if(strcmp(mountpoint, cfgStringValue(cfgQueryWithName("mountpoint", 1, entry), "")) == 0){
				panel->mountpoint = utilStrDup(mountpoint);
				panel->mountedObj = utilStrDup(cfgStringValue(cfgQueryWithName("backendSpec", 1, entry), NULL));
				frontend = cfgStringValue(cfgQueryWithName("frontend", 1, entry), "");
				for(j=0;j<arrObjectCount(frontendNames);j++){
					if(strcmp(frontend, arrObjectAtIndex(frontendNames, j)) == 0){
						panel->frontend = j;
						break;
					}
				}
				backend = cfgStringValue(cfgQueryWithName("backend", 1, entry), "");
				for(j=0;j<arrObjectCount(backendNames);j++){
					if(strcmp(backend, arrObjectAtIndex(backendNames, j)) == 0){
						panel->backend = j;
						break;
					}
				}
				panel->readOnly = cfgEnumValue(cfgQueryWithName("readonly", 1, entry), cfgYesNoEnum, 0);
				break;
			}
		}
	}
	if(panel->mountpoint == NULL){	/* not pre-initialized */
		panel->mountpoint = utilStrDup("");
		panel->mountedObj = utilStrDup("");
	}
	PROC_WAITFOR(guimanRunDialog(&mountDialog, this->client, this->panelData, 1, 0, this));
	if(grval->err.class != 0){
		err = grval->err;
		goto doReturn;
	}
	if(!grval->didSet){
		err.class = err.code = 0;
		goto doReturn;
	}
	panel = this->panelData;
	convLocalAsciiToUni(panel->mountedObj, buffer, ELEMENTS(buffer));
	PROC_WAITFOR(mainModMount(this->uid, arrObjectAtIndex(frontendNames, panel->frontend), panel->mountpoint, arrObjectAtIndex(backendNames, panel->backend), buffer, panel->readOnly, 0, 0, this));
	panel = this->panelData;
	if(panel->mountpoint != NULL)
		memFree(panel->mountpoint);
	if(panel->mountedObj != NULL)
		memFree(panel->mountedObj);
	memFree(this->panelData);
	err = *rval;
doReturn:
	if(err.class != 0){
		guimanDisplayError(this->client, &err, NULL);
	}
	ipcSrvClientRelease(this->client);
	PROC_END;
}

static void	mountListMount(ipcSrvClient_t *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, mountListMountProc, NULL, mainPool, poolFreeBlock);
	ctx->reply = reply;
	ctx->request = rq;
	ctx->uid = uid;
	ctx->client = client;
	PROC_START(ctx, NULL);
	marshalError(reply, 0, 0);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	moduleAppeared(void *dummy, void *module)
{
module_t	*mod;
array_t		*frontendsByIndex, *backendsByIndex;
btint		i;

	frontendsByIndex = modAllFrontends();
	backendsByIndex = modAllBackends();
	if(frontendNames == NULL)
		frontendNames = arrNew(NULL, 8);
	if(backendNames == NULL)
		backendNames = arrNew(NULL, 8);
	arrRemoveAllObjects(frontendNames);
	arrRemoveAllObjects(backendNames);
	for(i=0;i<arrObjectCount(frontendsByIndex);i++){
		mod = arrObjectAtIndex(frontendsByIndex, i);
		arrAddObject(frontendNames, mod->name);
	}
	for(i=0;i<arrObjectCount(backendsByIndex);i++){
		mod = arrObjectAtIndex(backendsByIndex, i);
		arrAddObject(backendNames, mod->name);
	}
	arrAddObject(frontendNames, NULL);
	arrRemoveLastObject(frontendNames);	/* terminate content-arr with NULL pointer */
	arrAddObject(backendNames, NULL);
	arrRemoveLastObject(backendNames);	/* terminate content-arr with NULL pointer */
	enumdesFrontend.externalNames = (btu8 **)arrContentArray(frontendNames);
	enumdesFrontend.internalNames = (btu8 **)arrContentArray(frontendNames);
	enumdesBackend.externalNames = (btu8 **)arrContentArray(backendNames);
	enumdesBackend.internalNames = (btu8 **)arrContentArray(backendNames);
}

/* ------------------------------------------------------------------------- */

static void	mountListUmountProc(void *obj, void *arg)
{
mainCtx_t	*this = obj;
btError		*rval = arg;

	PROC_BEGIN;
	PROC_WAITFOR(mainModUmount(this->uid, this->mountpoint, 0, this));
	marshalError(this->reply, rval->class, rval->code);
	PROC_END;
}

static void	mountListUmount(ipcSrvClient_t *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btint		parameter;
mainCtx_t	*ctx;

	parameter = ipcGetInt(rq);
	PROC_INIT(ctx, mountListUmountProc, caller, mainPool, poolFreeBlock);
	ctx->reply = reply;
	ctx->request = rq;
	ctx->uid = uid;
	ctx->mountpoint = ipcGetAscii(rq);
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	mountListStore(ipcSrvClient_t *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btint	parameter, rval;
btu8	*rowId;

	parameter = ipcGetInt(rq);
	rowId = ipcGetAscii(rq);
	if(rowId == NULL){	/* no row selected */
		marshalError(reply, 0, 0);	/* ignore action */
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	rval = storeMount(uid, rowId);
	marshalError(reply, (rval != 0 ? ERRCLS_FRAMEWORK : 0), rval);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	mountListDelStored(ipcSrvClient_t *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btint			parameter;
btu8			*rowId;
btint			rval;

	parameter = ipcGetInt(rq);
	rowId = ipcGetAscii(rq);
	if(rowId == NULL){	/* no row selected */
		marshalError(reply, 0, 0);	/* ignore action */
		PROC_IMMEDIATE_RET(caller, NULL);
	}
	rval = delStoredMount(uid, rowId);
	marshalError(reply, (rval != 0 ? ERRCLS_FRAMEWORK : 0), rval);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static guiDsourceMethod_t	mountListMethods[] = {
	mountListGet,
	mountListMount,
	mountListUmount,
	mountListStore,
	mountListDelStored,
};

static guiClass_t	mountListDescription = {
	MAIN_GUI_CLASS_LIST,
	"Sharity\nMounts",
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

ipcSrvStub_t	mainIpcStubs[] = {
	listModules,
	slashChars,
	mountAny,
	registerCb,
	unregisterCb,
	getObjectDescription,
	getConfigObjects,
	execDsourceMethod,
	getCfgValue,
	setCfgValue,
	delCfgValue,
	umountAny,
	storeMountRpc,
	lookupList,
	helpForKey,
	uiWillTerminate,
};

/* ------------------------------------------------------------------------- */

static void	moduleAdded(void *dummy, void *module)
{
module_t			*mod = module;
backendInterface_t	*b;
frontendInterface_t	*f;

	if(mod->moduleType == MOD_TYPE_BACKEND){
		b = memCalloc(sizeof(*b));
		b->forModule = mod;
		b->vfsForPath = mod->implementationOf(BE_VFS_FOR_PATH);
		b->vfsConnect = mod->implementationOf(BE_VFS_CONNECT);
		b->vfsRelease = mod->implementationOf(BE_VFS_RELEASE);
		b->userAuthenticate = mod->implementationOf(BE_USER_AUTHENTICATE);
		b->vfsPath = mod->implementationOf(BE_VFS_PATH);
		htabAddObject(moduleInterfaces, b);
	}else if(mod->moduleType == MOD_TYPE_FRONTEND){
		f = memCalloc(sizeof(*f));
		f->forModule = mod;
		f->mntNew = mod->implementationOf(FE_MNT_NEW);
		f->mntRelease = mod->implementationOf(FE_MNT_RELEASE);
		f->mntSetBackend = mod->implementationOf(FE_MNT_SET_BACKEND);
		f->mntMount = mod->implementationOf(FE_MNT_MOUNT);
		f->mntUmount = mod->implementationOf(FE_MNT_UMOUNT);
		f->mntSetFakemode = mod->implementationOf(FE_MNT_SET_FAKEMODE);
		f->mntWasUnmounted = mod->implementationOf(FE_MNT_WAS_UNMOUNTED);
		f->mntIsReadonly = mod->implementationOf(FE_MNT_IS_READONLY);
		f->mntMtabDevice = mod->implementationOf(FE_MNT_MTAB_DEVICE);
		htabAddObject(moduleInterfaces, f);
	}
}

/* ------------------------------------------------------------------------- */
/* -------------------------- config description --------------------------- */
/* ------------------------------------------------------------------------- */

#define	PERM_NORMAL	MAIN_GUI_PERM_AUTHORIZED
#define	FLG_RST	MAIN_GUI_ENTRY_FLG_NEEDS_RESTART

static guiObject_t	*mainCfgDsource;

static guiEntry_t	mainCfgEntries[] = {
	{	MAIN_GUI_ENTRY_FILE,	"Server Side Character Map",
		NULL,					"mainConfigServerMap",
		PERM_NORMAL,			FLG_RST, ALLOW_NULL,	"serverCharMapFile",
	},
	{	MAIN_GUI_ENTRY_FILE,	"Client Side Character Map",
		NULL,					"mainConfigClientMap",
		PERM_NORMAL,			FLG_RST, ALLOW_NULL,	"clientCharMapFile",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Users may change Configuration",
		NULL,					"mainConfigUsersMayConfigure",
		MAIN_GUI_PERM_ROOT,		FLG_0, ALLOW_NULL,	"usersMayConfigure",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Users may store Passwords and Mounts",
		NULL,					"mainConfigUsersMayStoreSessionData",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"usersMayStoreSessionData",
	},
};

static guiButton_t	mainButtonEntries[] = {
	{	"Defaults",			MAIN_GUI_METHOD_GET,
		1,	/* parameter */	0,	/* flags */
		"mainConfigButtonDefaults",
	},
	{	"Revert",			MAIN_GUI_METHOD_GET,
		0,	/* parameter */	MAIN_GUI_BUTTON_ACTIVE_IF_DIRTY,	/* flags */
		"mainConfigButtonRevert",
	},
	{	"OK",				MAIN_GUI_METHOD_SET,
		0,	/* parameter */	MAIN_GUI_BUTTON_ACTIVE_IF_DIRTY,	/* flags */
		"mainConfigButtonOk",
	},
};

/* ------------------------------------------------------------------------- */

static void	getValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller);
static void	setValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller);


static guiDsourceMethod_t	mainCfgMethods[] = {
	getValues,
	setValues,
};

static guiClass_t	mainCfgDescription = {
	MAIN_GUI_CLASS_PLAINCFG,
	"Sharity\nGeneral",
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

static void	getValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	*cfgPath[] = {MAINMOD_CFGNAME, NULL};
btu8	**cfgPathes[2]; /* = {cfgPath, NULL};*/

	cfgPathes[0] = cfgPath;
	cfgPathes[1] = NULL;
	guimanGetValues(cfgPathes, NULL, mainCfgDescription.entries, rq, reply);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	setValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	*path[] = {MAINMOD_CFGNAME, NULL};

	guimanSetValues(0, path, mainCfgDescription.entries, rq, reply, uid);
	mainModGuiUpdate(mainCfgDsource);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	reloadGui(void *dummyObj, void *dummy)
{
	mainModGuiUpdate(NULL);
}

/* ------------------------------------------------------------------------- */

static void	mainModUmountAllProc(void *obj, void *arg)
{
mainCtx_t	*this = obj;
mountInfo_t	*minfo;

	PROC_BEGIN;
	if(activeMounts != NULL){
		this->all = htabAllObjects(activeMounts);
		for(this->index=0;this->index<arrObjectCount(this->all);this->index++){
			minfo = arrObjectAtIndex(this->all, this->index);
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "mainmod: umountAll: going to unmount %s\n", minfo->mountpoint);
			PROC_WAITFOR(mainModUmount(0, minfo->mountpoint, 1, this));
		}
		arrFree(this->all);
	}
	PROC_END;
}

void	mainModUmountAll(void *caller)
{
mainCtx_t	*ctx;

	PROC_INIT(ctx, mainModUmountAllProc, caller, mainPool, poolFreeBlock);
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

/* last chance to unmount what's left */
static void	terminateHook(void *dummyObj, void *dummyArg)
{
array_t			*allMounts;
btint			i, cnt, rval, storederr;
mountInfo_t		*minfo;

	if(activeMounts != NULL){
		allMounts = htabAllObjects(activeMounts);
		cnt = arrObjectCount(allMounts);
		for(i=0;i<cnt;i++){
			minfo = arrObjectAtIndex(allMounts, i);
/* panic unmount: use low-level utility and wait for result */
			rval = sysUmount(minfo->mountpoint);
			storederr = errno;
			if(rval != 0){
				logPrintf(LOGM_ERROR, "panic unmount: cannot unmount %s: [%d] %s\n", minfo->mountpoint, storederr, sysStrerror(storederr));
			}
		}
		arrFree(allMounts);
	}
}


/* ------------------------------------------------------------------------- */
/* ------------------------- module implementation ------------------------- */
/* ------------------------------------------------------------------------- */

static void	*haveNoImp(btu8 *name)
{
	return NULL;
}

void	mainMod_init(void)
{
module_t		*mod;
btu8			*cfgCasePath, *cfgServerMapPath, *cfgClientMapPath;
cfgProperty_t	*cfgPath[4];
btint			numEntries, i;

	if(mainPool == NULL)
		mainPool = poolNew(sizeof(mainCtx_t), sizeof(mainCtx_t));
	listsByName = htabNew(&listDescriptorInterface, 8, 75);
	moduleInterfaces = htabNew(&interfaceI, 8, 75);
	activeMounts = htabNew(&activeMountsInterface, 16, 75);
	mountsByDevice = htabNew(&mountsByDeviceInterface, 16, 75);
	mod = memCalloc(sizeof(*mod));
	mod->name = MAIN_MODULE_NAME;
	mod->moduleType = MOD_TYPE_FRAMEWORK;
	mod->implementationOf = haveNoImp;
	mod->ipcModule = ipcRegisterModule(gdata.ipcServer, mod->name, mainIpcStubs, ELEMENTS(mainIpcStubs), MAIN_IPC_PROTOCOL_VERSION);
	mod->interfaceVersion = 0;
	mod->implementationVersion = MAINCONST_VERSION;
	mod->betaVersion = MAINCONST_BETA_VERSION;
	mainModule = mod;
	modRegister(mod);
	moduleAppeared(NULL, NULL);	/* initialize arrays */
	ncAddObserver(modRegisterNotification(), moduleAppeared, NULL);

	numEntries = mainModCfgPath(cfgPath, ELEMENTS(cfgPath));
	ncAddObserver(modRegisterNotification(), moduleAdded, NULL);

	cfgCasePath = cfgStringValue(cfgQueryFromArray("caseConvertFile", numEntries, cfgPath), NULL);
	cfgServerMapPath = cfgStringValue(cfgQueryFromArray("serverCharMapFile", numEntries, cfgPath), NULL);
	cfgClientMapPath = cfgStringValue(cfgQueryFromArray("clientCharMapFile", numEntries, cfgPath), NULL);
	uniInit(cfgCasePath, cfgServerMapPath, cfgClientMapPath, cfgIntValue(cfgQueryFromArray("uniTableCompressBlocks", numEntries, cfgPath), CONST_UNI_TABLE_COMPRESS));


	mainCfgDescription.entries = arrNew(NULL, 8);
	mainCfgDescription.buttons = arrNew(NULL, 8);
	for(i=0;i<ELEMENTS(mainCfgEntries);i++){
		arrAddObject(mainCfgDescription.entries, &mainCfgEntries[i]);
	}
	for(i=0;i<ELEMENTS(mainButtonEntries);i++){
		arrAddObject(mainCfgDescription.buttons, &mainButtonEntries[i]);
	}
	mainCfgDsource = mainModRegisterDataSource(&mainCfgDsource, &mainCfgDescription, mainCfgMethods, ELEMENTS(mainCfgMethods));
	mainModAdvertizeConfig(mainCfgDsource, 100000);

	mountListDescription.entries = arrNew(NULL, 8);
	mountListDescription.buttons = arrNew(NULL, 8);
	for(i=0;i<ELEMENTS(mountListColumns);i++){
		arrAddObject(mountListDescription.entries, &mountListColumns[i]);
	}
	for(i=0;i<ELEMENTS(mountListButtons);i++){
		arrAddObject(mountListDescription.buttons, &mountListButtons[i]);
	}
	mountListObject = mainModRegisterDataSource(&mountListObject, &mountListDescription, mountListMethods, ELEMENTS(mountListMethods));
	mainModAdvertizeConfig(mountListObject, 10000);
	mainModRegisterList(MAIN_LIST_MOUNTS, mountListObject->id);
	ncAddObserver(cfgChangedNotification(), reloadGui, NULL);
	ncAddObserver(gdata.terminateNotification, terminateHook, NULL);
}

/* ------------------------------------------------------------------------- */
