/*
 * Name: mainmodule.h
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-10
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: mainmodule.h,v 1.4 2001/10/15 22:53:09 cs Exp $
 */

#ifndef __mainmodule_H_INCLUDED__
#define __mainmodule_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
The main module implements the general framework functionality: Generic
functions for mounting and unmounting, functions to communicate with a
generic user interface etc.
*/

/* ------------------------------------------------------------------------- */
/* ------------------------------ Data Types ------------------------------- */
/* ------------------------------------------------------------------------- */

#define	MAINMOD_CFGNAME		"main"	/* name of config section */

struct ipcMsg;
struct cfgProperty;
struct ipcSrvClient;

typedef struct guiButton{
	btu8	*title;		/* what's printed inside the button */
	btint	function;	/* depends on container object, method index by default */
	btint	parameter;	/* function dependent */
	btu32	flags;		/* see uiLib/maininterface.h for defines */
	btu8	*helpItem;
}guiButton_t;
/* This structure describes a button in the generic user interface. When the
 * button is clicked, the GUI object's method with index 'function' is called.
 * The well known methods with index 0 and 1 are treated a bit different: They
 * set or retrieve the GUI object's data values. 'parameter' is passed as an
 * argument to all other methods.
 */

typedef struct guiTypedesNumeric{
	btint	fractionDivider;		/* display = value / fractionDivider */
	btint	fractionMultiplier;		/* display = value * fractionMultiplier */
	btint	increment;				/* in value units */
	btint	minVal;					/* in value units */
	btint	maxVal;					/* in value units */
	btu8	*unit;					/* what's displayed to the right of the entry field */
}guiTypedesNumeric_t;
/* This structure specifies the behaviour of numeric entry fields in the user
 * interface.'fractionDivider' and 'fractionMultiplier' are there to avoid the
 * need for floating point arithmetic in the daemon.
 */

typedef struct guiTypedesEnum{
	btu8	**externalNames;
	btu8	**internalNames;
}guiTypedesEnum_t;
/* This structure specifies a popup menu entry in the GUI. 'externalNames' is a
 * (NULL terminated) array of strings that are displayed on the menu's buttons,
 * 'internalNames' are the corresponding names used internally in the config
 * and db file.
 */

typedef struct guiEntry{
	btint	type;		/* see uiLib/maininterface.h for defines */
	btu8	*title;		/* what's displayed to the left of the entry */
	void	*typeDescriptor;	/* guiTypedesEnum_t for popups, guiTypedesNumeric_t for numbers */
	btu8	*helpItem;
	btint	accessPermissions;	/* see uiLib/mainInterface.h MAIN_GUI_PERM* */
	btu32	flags;				/* see uiLib/mainInterface.h MAIN_GUI_ENTRY_FLG* */
	btbool	mayBeNull;			/* whether an emtpy value should be stored in the file */
	void	*extraInfo;			/* may be used by data source implementation */
}guiEntry_t;
/* This structure specifies an input field in the GUI. All the configurations
 * that act directly on the configuration database through the functions
 * defined in guimanager.c expect the configuration property's name in
 * 'extraInfo'. guimanRunDialog(), also defined in guimanager.c, expects the
 * offset in bytes of the data value in the result structure.
 */

typedef struct guiColumn{
	btu8	*title;			/* the column's title */
	btint	alignment;		/* not currently used, should be 0 */
	btint	widthPercent;	/* width of this column in percent of the total width */
	btu8	*helpItem;
}guiColumn_t;
/* This structure defines a column in a list view of the GUI. */

typedef struct guiClass{
	btint			type;		/* determines which set of dsourceMethods is available, see uiLib/maininterface.h for defines */
	btu8			*name;		/* name of this cfg for GUI */
	struct array	*entries;	/* entries in panel, or columns of list */
	struct array	*buttons;	/* buttons on this object */
}guiClass_t;
/* This structure defines an entire list view, configuration panel or dialog
 * panel in the GUI. The 'type' determines the "class" of the object, which in
 * turn defines the set of datasource methods used. List views call other
 * methods to retrieve their content than dialogs, of course. 'entries'
 * describes the content of the GUI object, either an array of guiEntry_t for
 * configuration or dialog panels, or an array of guiColumn_t for list views.
 * 'buttons' is always an array of guiButton_t with the buttons that should
 * appear at the bottom of the object.
 */

typedef void	(*guiDsourceMethod_t)(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, struct ipcMsg *rq, struct ipcMsg *reply, void *caller);
/* This is the prototype for GUI data source methods. They are called from the
 * GUI in the daemon. It's the data source method's responsibility to unmarshal
 * parameters and marshal results.
 */

typedef struct guiObject{
	btint				id;
	guiClass_t			*class;
	void				*dsource;
	btint				numMethods;
	guiDsourceMethod_t	*methods;
	btbool				needUpdate;
}guiObject_t;
/* This is the internal reprensentation of an active (probably visible) user
 * interface component.
 */

typedef void	(*mainModUiTermFunction_t)(struct ipcSrvClient *client, btint termReason, void *caller);
/* This function data type is defined for a callback function used to tell
 * modules that a user interface application is going to quit.
 * Parameters: 'client' is the UI application representation for the quitting
 *    UI application, 'termReason' is the reason for termination (as defined
 *    in uiLib/maininterface.h in defines MAIN_TERM*) and 'caller' is the
 *    caller's context.
 * Callback: The callback argument is a boolean casted to a pointer. It is
 *    1, if the module allows termination and 0, if the module objects the
 *    termination of the UI application.
 */

/* ------------------------------------------------------------------------- */

extern struct module	*mainModule;	/* pointer to our module description */

/* ------------------------------------------------------------------------- */
/* ------------------------------- Functions ------------------------------- */
/* ------------------------------------------------------------------------- */
btint	mainModCfgPath(struct cfgProperty **path, btint maxEntries);
/* This function is used to get the search path for configuration data in
 * the main module and framework.
 */
void	mainModMount(btint uid, btu8 *frontendModule, btu8 *mountpoint, btu8 *backendModule, btuni *mountedPath, btbool readOnly, btbool fakeMode, btbool isAutomount, void *caller);
/* This function establishes a mount with a given frontend and backend. It
 * performs all necessary access permission checks and transforms the mount
 * path into it's canonical representation with all symbolic links resolved.
 * Parameters: 'uid' is the local Unix user ID of the user that establishes the
 *    mount, 'frontendModule' is the name of the frontend module that should
 *    be used, 'mountpoint' is the mount path, 'backendModule' is the name of
 *    the backend module to be used, 'mountedPath' is the backend specific
 *    description of the object to be mounted, 'readOnly' specifies whether
 *    the mount should be performed in read-only mode, 'fakeMode' specifies
 *    whether the mount should be established in fake-mode (only during mount,
 *    fake mode will be turned off after successfully mounting), 'isAutomount'
 *    is for informational purposes only, it tells the user that this mount
 *    was caused by an automounter and 'caller' is the caller's context for
 *    callbacks.
 * Callback: btError
 */
void	mainModUmount(btint uid, btu8 *mountpoint, btbool fakeMode, void *caller);
/* This function unmounts what is mounted at 'mountpoint'.
 * Parameters: 'uid' is the local Unix user ID of the user who attempts the
 *    unmount, 'mountpoint' is the path (in the local filesystem) where the
 *    object is currently mounted, 'fakeMode' specifies whether the unmount
 *    should be performed in fake-mode and 'caller' is the caller's context.
 * Callback: btError
 */
btbool	mainModMountExists(btu8 *mountpoint);
/* Call this function if you want to know whether a specific path is a
 * mountpoint of Sharity. Returns 1 if Sharity has mounted on the path,
 * 0 otherwise.
 */
void	mainModNotifyUnmount(btu8 *device);
/* This function informs the main module that a mount has been unmounted
 * bypassing Sharity's unmount mechanism. This may happen due to the user
 * "ejecting" the server from Finder on Mac OS X.
 * Parameters: 'device' is the /etc/mnttab "device" description of the mount.
 */
void	mainModUmountAll(void *caller);
/* This function unmounts everything mounted by Sharity.
 * Parameters: 'caller' is the caller's context.
 * Callback: NULL
 */
btint	mainModUserForMtabDevice(btu8 *device);
/* This function returns the user-ID of the user who mounted the /etc/mnttab
 * device 'device'. If the mount is not found, -1 is returned.
 */
guiObject_t	*mainModRegisterDataSource(void *dsource, guiClass_t *class, guiDsourceMethod_t *methods, btint numMethods);
/* This function (in a certain sense) creates an instance of a GUI object. The
 * object's look is described by 'class' whereas the functionality behind it
 * is defined with the set of 'methods', an array of 'numMethod' functions.
 * Note that this function does not bring up the GUI object in the user
 * interface. You have to call the appropriate client IPC functions for that.
 * Parameters: 'dsource' is an opaque pointer that will be passed to the
 *    data source methods on behalf of this object, 'class' describes the GUI
 *    object's look, 'methods' is a set of functions which will be called from
 *    the GUI to display data and 'numMethods' is the number of functions
 *    passed in 'methods'.
 * Returns: a representation of the GUI object instance, which can be used
 *    in other functions.
 */
void	mainModUnregisterDataSource(guiObject_t *guiObj);
/* This function unregisters and destroys a GUI object representation
 * perviously created with mainModRegisterDataSource().
 * Parameters: 'guiObject' is the object previously created with
 *    mainModRegisterDataSource().
 */
void	mainModRegisterList(btu8 *shortName, btint objectId);
/* This function registers a MAIN_GUI_CLASS_LIST object by name for lookup
 * by the UI tools. The commandline tools need to know more about the user
 * interface objects than only their type and description. They therefore
 * look up particular objects that are known to exist by name.
 * Parameters: 'shortName' is the name under which the object is published and
 *    'objectId' is the object's identifier.
 */
void	mainModAdvertizeConfig(guiObject_t *guiObj, btint precedence);
/* This function is one of the methods to get a GUI object on the user
 * interface's screen. It stores the object and if a GUI application starts up,
 * it queries the list of advertized objects and displays them.
 * Parameters: 'guiObj' is the GUI object representation obtained with
 *    mainModRegisterDataSource(), 'precedence' is an integer specifying where
 *    in the tabbed view of the user interface this object should be displayed.
 *    The smaller this number, the closer it will be to the left, the bigger
 *    the number, the more it will go to the right.
 */
void	mainModGuiUpdate(guiObject_t *go);
/* This function informs the user interface that data in a given GUI object
 * has changed. The GUI application is then expected to reload the data
 * through the data source methods. If 'go' is NULL, all objects are updated.
 * Parameters: 'go' is the object that was modified.
 */
struct notificationCenter	*mainModMountNotification(void);
struct notificationCenter	*mainModUmountNotification(void);
/* The above two functions can be used by modules that need to stay up to date
 * about current mounts. Both functions return a pointer to a notification
 * center which fires under the respective condition. The parameter of the
 * notification is the normalized path of the mountpoint.
 * Returns: A non-retained notificationCenter object.
 */
void	mainModRegisterForUiTerminate(mainModUiTermFunction_t fn);
void	mainModUnregisterForUiTerminate(mainModUiTermFunction_t fn);
/* These two functions allow a module to register/unregister a callback
 * function for the event that a user interface application quits. The module
 * has the possibility to bring up a dialog and/or to prevent termination
 * (if at all possible).
 * Note: It's not guaranteed that this function is called upon termination.
 * It's up to the UI application to call it. It's more a service FOR the UI
 * provided BY the daemon modules, not vice versa.
 * Parameters: The callback function 'fn' which is called upon the event. 'fn'
 *    itself must call back to the calling procedure when ready.
 */

/* ------------------------------------------------------------------------- */

#endif /* __mainmodule_H_INCLUDED__ */
