/*
 * Name: sharity.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-10
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: sharity.c,v 1.14 2002/05/06 15:06:13 cs Exp $
 */


#include "stdheaders.h"
#include "ipcclient.h"
#include "argparser.h"
#include "maininterface.h"
#include "cifsinterface.h"
#include "exception.h"
#include "ipc.h"
#include "errcodes.h"
#include "util.h"
#include "memalloc.h"
#include "system.h"

/*
General Description:
This is currently a test version that allows a user to mount with any
combination of frontend and backend modules.
*/

static char	*progName;

/* ------------------------------------------------------------------------- */

static void	requestReceivedCallback(void *obj, void *arg);

static btint	setupConnection(void *callbackContext)
{
ipccModuleInfo_t	*mi;
ipccRes_t			res;
btint				rval = 0;

	ipccConnect();
	ipccSetRequestHandler(requestReceivedCallback, callbackContext);
	ipccSetProtocolVersion(0, MAIN_IPCFN_NUMFUNCTIONS, MAIN_IPC_PROTOCOL_VERSION);
	mi = ipccGetModuleInfo(&res, "cifsFile");
	if(res.err.class == 0){	/* if we have the CIFS module compiled in */
		ipccSetProtocolVersion(mi->fnBase, mi->fnCount, CIFS_IPC_PROTOCOL_VERSION);
	}
	ipccRegisterForCallback(&res, "main");
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}
	return rval;
}

/* ------------------------------------------------------------------------- */

static void	shortHelp(argParser_t *ap, btu8 *argInfo)
{
	printf("usage: %s %s\n", progName, argInfo);
	printf("       %s -v     for version information\n", progName);
	printf("       %s -man   for the full manual page\n", progName);
	printf("valid options are:\n");
	argpPrintHelp(ap, stdout);
}

/* ------------------------------------------------------------------------- */

static btint	mountImp(btint argc, char **argv)
{
btbool		isReadonly, fakeMode;
argParser_t	*ap;
ipccRes_t	res;
int			rval = 0;

	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");
	argpAddOption(ap, 'r', NULL, 0, "Mount readonly");
	argpAddOption(ap, 'f', NULL, 0, "Mount in fake mode (not connecting backend)");
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		shortHelp(ap, "<frontend module> <mountpoint> <backend module> <backend path> [<options>]");
		return 0;
	}
	if(argpArgumentCount(ap) < 4){
		fprintf(stderr, "Not enough arguments.\n");
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		rval = 1;
		goto doReturn;
	}
	if(argpArgumentCount(ap) > 4){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	isReadonly = argpOptionIsSet(ap, 'r');
	fakeMode = argpOptionIsSet(ap, 'f');
	if((rval = setupConnection(NULL)) != 0)
		goto doReturn;
	ipccMount(&res, argpArgumentAtIndex(ap, 0), argpArgumentAtIndex(ap, 1), argpArgumentAtIndex(ap, 2), argpArgumentAtIndex(ap, 3), isReadonly, fakeMode);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	umountImp(btint argc, char **argv)
{
btbool		fakeMode, all;
argParser_t	*ap;
ipccRes_t	res;
int			rval = 0;

	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");
	argpAddOption(ap, 'f', NULL, 0, "Unmount in fake mode (not syncing to backend)");
	argpAddOption(ap, 'a', NULL, 0, "Unmount everything that\'s mounted");
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		shortHelp(ap, "<mountpoint> [<options>]");
		return 0;
	}
	all = argpOptionIsSet(ap, 'a');
	if(!all && argpArgumentCount(ap) < 1){
		fprintf(stderr, "Not enough arguments.\n");
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		rval = 1;
		goto doReturn;
	}
	if(argpArgumentCount(ap) > (all ? 0 : 1)){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	fakeMode = argpOptionIsSet(ap, 'f');
	if((rval = setupConnection(NULL)) != 0)
		goto doReturn;
	if(all){
		ipccUmount(&res, NULL, fakeMode);
	}else{
		ipccUmount(&res, argpArgumentAtIndex(ap, 0), fakeMode);
	}
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	storemntImp(btint argc, char **argv)
{
btbool		doDelete;
argParser_t	*ap;
ipccRes_t	res;
int			rval = 0;

	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");
	argpAddOption(ap, 'd', NULL, 0, "delete stored record");
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		shortHelp(ap, "<mountpoint> [<options>]");
		return 0;
	}
	if(argpArgumentCount(ap) < 1){
		fprintf(stderr, "Not enough arguments.\n");
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		rval = 1;
		goto doReturn;
	}
	if(argpArgumentCount(ap) > 1){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	doDelete = argpOptionIsSet(ap, 'd');
	if((rval = setupConnection(NULL)) != 0)
		goto doReturn;
	ipccStoremnt(&res, argpArgumentAtIndex(ap, 0), doDelete);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static void	printListRow(ipcMsg_t *msg, btint columns, btint *width, btbool isRaw, btu8 *fieldSeparator)
{
btint	i, w;
btu8	buffer[100], *p;

	ipcGetAscii(msg);	/* rowId */
	for(i=0;i<columns;i++){
		if((p = ipcGetAscii(msg)) == NULL)
			p = "";
		if(isRaw){
			printf("%s", p);
			if(i+1 < columns){
				printf("%s", fieldSeparator);
			}
		}else{
			w = width[i] > 0 ? width[i] - 1 : width[i];
			strncpy(buffer, p, w);
			buffer[w] = 0;	/* always terminate with 0-byte */
			printf("%-*s ", w, buffer);
		}
	}
	putc('\n', stdout);
}

static btint	listPrimitive(btu8 *listName, btbool rawMode, btu8 *fieldSeparator)
{
ipccRes_t	res;
btint		*width = NULL, w, type, columns = 0, i, cnt, listId, rval = 0;
ipcMsg_t	*msg;
btu8		buffer[100], **titles = NULL, *objectName;

	ipccLookupList(&res, listName, &listId);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
		goto doReturn;
	}
	ipccGetObjectDescription(&res, listId);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
		goto doReturn;
	}
	msg = ipccCurrentRxMessage();
	objectName = ipcGetAscii(msg);
	type = ipcGetInt(msg);
	if(type != MAIN_GUI_CLASS_LIST){
		fprintf(stderr, "Object \"%s\" is not a list.\n", listName);
		rval = 1;
		goto doReturn;
	}
	columns = cnt = ipcGetInt(msg);
	titles = memCalloc(columns * sizeof(btu8 *));
	width = memCalloc(columns * sizeof(btint));
	for(i=0;i<cnt;i++){
		titles[i] = utilStrDup(ipcGetAscii(msg));
		ipcGetInt(msg);		/* alignment */
		width[i] = ipcGetInt(msg);
		ipcGetAscii(msg);	/* help text */
	}
	msg = ipccCurrentRxMessage();
	ipccStartProc(msg, MAIN_IPCFN_DSOURCE_EXEC);
	ipcAddInt(msg, listId);
	ipcAddInt(msg, MAIN_GUI_METHOD_GET);
	ipccRpc(msg);
	res.err.class = ipcGetInt(msg);
	res.err.code = ipcGetInt(msg);
	res.errmsg = ipcGetAscii(msg);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
		goto doReturn;
	}
	if(!rawMode){
		for(i=0;i<columns;i++){
			width[i] = (width[i] * 4) / 5;
			if(width[i] > 80)
				width[i] = 80;
			w = width[i] > 0 ? width[i] - 1 : width[i];
			strncpy(buffer, titles[i], w);
			buffer[w] = 0;	/* always terminate with 0-byte */
			printf("%-*s ", w, buffer);
		}
		putc('\n', stdout);
		for(i=0;i<79;i++)
			putc('-', stdout);
		putc('\n', stdout);
	}
	if(fieldSeparator == NULL)
		fieldSeparator = ";";
	cnt = ipcGetInt(msg);
	for(i=0;i<cnt;i++){
		printListRow(msg, columns, width, rawMode, fieldSeparator);
	}
doReturn:
	if(titles != NULL){
		for(i=0;i<columns;i++)
			memFree(titles[i]);
		memFree(titles);
	}
	if(width != NULL)
		memFree(width);
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	listImp(btint argc, char **argv)
{
argParser_t	*ap;
btbool		rawMode;
btu8		*fieldSeparator;
btint		rval;

	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");
	argpAddOption(ap, 'r', NULL, 0, "Raw output format (useful for scripts)");
	argpAddOption(ap, 's', "<separator>", 0, "Set string used to separate table entries");
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		shortHelp(ap, "<list-name> [<options>]");
		return 0;
	}
	if(argpArgumentCount(ap) < 1){
		fprintf(stderr, "Not enough arguments.\n");
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpArgumentCount(ap) > 1){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	fieldSeparator = argpOptionStringVal(ap, 's');
	rawMode = argpOptionIsSet(ap, 'r');
	if((rval = setupConnection(NULL)) != 0)
		return rval;
	return listPrimitive(argpArgumentAtIndex(ap, 0), rawMode, fieldSeparator);
}

/* ------------------------------------------------------------------------- */

static void	displayError(btint sid, ipcMsg_t *msg)
{
ipccRes_t	err;

	err.err.class = ipcGetInt(msg);
	err.err.code = ipcGetInt(msg);
	err.errmsg = ipcGetAscii(msg);
	ipccPrintErr(&err);
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
}

/* ------------------------------------------------------------------------- */

static void	askData(btint sid, ipcMsg_t *msg)
{
btu8	*prompt, buffer[1024], *p;
btbool	withEcho;
btint	l;

	prompt = ipcGetAscii(msg);
	withEcho = ipcGetInt(msg);
	if(withEcho){
		fprintf(stdout, "%s", prompt);
		p = fgets(buffer, ELEMENTS(buffer), stdin);
		if(p != NULL){
			buffer[ELEMENTS(buffer)-1] = 0;	/* always 0-terminate */
			l = strlen(buffer);
			while(l > 0 && (buffer[l-1] == '\n' || buffer[l-1] == '\r')){
				buffer[--l] = 0;
			}
			p = buffer;
		}
	}else{
		p = sysGetPass(prompt, buffer, ELEMENTS(buffer));
	}
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
	ipcAddAscii(msg, p);
}

/* ------------------------------------------------------------------------- */

typedef struct passwordContext{
	btu8	*username;
	btu8	*domain;
	btu8	*password;
	btbool	pwFromStdin;
	btbool	dontAskPw;
	btbool	allowUnencrypted;
	btbool	daemonAlreadyAsked;
}passwordContext_t;

static void	askPassword(passwordContext_t *ctx, btint sid, ipcMsg_t *msg)
{
btu8	*host, buffer[1024], *p;

	if(ctx->password == NULL && !ctx->dontAskPw){
		if(ctx->pwFromStdin){
			if((p = fgets(buffer, ELEMENTS(buffer), stdin)) != NULL){
				if((p = strrchr(buffer, '\n')) != NULL)
					*p = 0;
				p = buffer;
			}
		}else{
			host = ipcGetAscii(msg);
			if(strlen(host) > 128)
				host[128] = 0;	/* emergency brake */
			sprintf(buffer, "%s\'s password on %s: ", ctx->username, host);
			p = sysGetPass(buffer, buffer, ELEMENTS(buffer));
		}
	}else{	/* password given on commandline or stdin */
		if(ctx->daemonAlreadyAsked){
			p = NULL;	/* = user abort to avoid infinite loop */
		}else{
			if(ctx->password != NULL)
				p = ctx->password;
			else
				p = "";
		}
	}
	ctx->daemonAlreadyAsked = 1;
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	if(p == NULL){
		ipcAddInt(msg, ERRCLS_FRAMEWORK);	/* errclass */
		ipcAddInt(msg, ERR_FRWusrabort);	/* errcode */
	}else{
		ipcAddInt(msg, 0);	/* err class */
		ipcAddInt(msg, 0);	/* err code */
		ipcAddAscii(msg, ctx->username);
		ipcAddAscii(msg, p);
		ipcAddAscii(msg, ctx->domain);
		ipcAddInt(msg, ctx->allowUnencrypted);
	}
}

/* ------------------------------------------------------------------------- */

static void	askSharePassword(passwordContext_t *ctx, btint sid, ipcMsg_t *msg)
{
btu8	*share, buffer[1024], *p;

	if(ctx->password == NULL && !ctx->dontAskPw){
		if(ctx->pwFromStdin){
			if((p = fgets(buffer, ELEMENTS(buffer), stdin)) != NULL){
				if((p = strrchr(buffer, '\n')) != NULL)
					*p = 0;
				p = buffer;
			}
		}else{
			share = ipcGetAscii(msg);
			if(strlen(share) > 128)
				share[128] = 0;	/* emergency brake */
			sprintf(buffer, "Share Password for %s: ", share);
			p = sysGetPass(buffer, buffer, ELEMENTS(buffer));
		}
	}else{
		if(ctx->daemonAlreadyAsked){
			p = NULL;	/* = user abort to avoid infinite loop */
		}else{
			if(ctx->password != NULL)
				p = ctx->password;
			else
				p = "";
		}
	}
	ctx->daemonAlreadyAsked = 1;
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	if(p == NULL){
		ipcAddInt(msg, ERRCLS_FRAMEWORK);	/* errclass */
		ipcAddInt(msg, ERR_FRWusrabort);	/* errcode */
	}else{
		ipcAddInt(msg, 0);	/* err class */
		ipcAddInt(msg, 0);	/* err code */
		ipcAddAscii(msg, p);
		ipcAddInt(msg, ctx->allowUnencrypted);
	}
}

/* ------------------------------------------------------------------------- */

static void	requestReceivedCallback(void *obj, void *arg)
{
ipccRequestData_t	*data = arg;
ipcMsg_t			*msg;
btbool				didReply = 1;

	if(arg == NULL){	/* broken pipe to daemon */
		return;			/* we don't care much */
	}
	msg = data->msg;
	ipcGetInt(msg);	/* skip uid */
	ipcGetInt(msg);	/* skip gid */
	switch(data->procId){
	case MAIN_IPCCFN_ASK_PASSWORD:
		if(obj == NULL)
			didReply = 0;
		else
			askPassword(obj, data->sid, msg);
		break;
	case MAIN_IPCCFN_ASK_SHAREPASSWD:
		if(obj == NULL)
			didReply = 0;
		else
			askSharePassword(obj, data->sid, msg);
		break;
	case MAIN_IPCCFN_ASK_DATA:
		askData(data->sid, msg);
		break;
	case MAIN_IPCCFN_DISPLAY_ERROR:
		displayError(data->sid, msg);
		break;
	default:
		didReply = 0;
		break;
	}
	if(!didReply){	/* function not supported, construct error reply */
		ipcMsgClear(msg);
		ipcAddInt(msg, 0);	/* space for total length */
		ipcAddInt(msg, data->protocolVersion);
		ipcAddInt(msg, data->sid);
		ipcAddInt(msg, -1);
		ipcAddInt(msg, ipccUserId());
		ipcAddInt(msg, getgid());
		ipcAddInt(msg, ERRCLS_FRAMEWORK);
		ipcAddInt(msg, ERR_FRWipcnotimpl);
	}
}

/* ------------------------------------------------------------------------- */

static btint	cifsmountImp(btint argc, char **argv)
{
btbool				isReadonly, fakeMode;
argParser_t			*ap;
ipccRes_t			res;
int					rval = 0;
struct passwd		*pwentry;
passwordContext_t	pwContext;
btu8				*frontend, *mp, buffer[MAXPATHLEN + 1];

	memset(&pwContext, 0, sizeof(pwContext));
	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");
	argpAddOption(ap, 'r', NULL, 0, "Mount readonly");
	argpAddOption(ap, 'f', NULL, 0, "Mount in fake mode (not connecting server)");
	argpAddOption(ap, 'm', "<frontend>", 0, "Use the given frontend for mounting");
	argpAddOption(ap, 'U', "<username>", 0, "Login on server as this user");
	argpAddOption(ap, 'D', "<domain>", 0, "Send this domain name to server");
	argpAddOption(ap, 'P', "<password>", 0, "Password given in commandline");
	argpAddOption(ap, 'S', NULL, 0, "Read password from standard input (implies -N)");
	argpAddOption(ap, 'N', NULL, 0, "Don\'t prompt for a password");
	argpAddOption(ap, 'u', NULL, 0, "Allow sending password unencrypted");
	
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		shortHelp(ap, "//<server>/<share> <mountpoint> [<options>]");
		return 0;
	}
	if(argpArgumentCount(ap) < 2){
		fprintf(stderr, "Not enough arguments.\n");
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		rval = 1;
		goto doReturn;
	}
	if(argpArgumentCount(ap) > 2){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	isReadonly = argpOptionIsSet(ap, 'r');
	fakeMode = argpOptionIsSet(ap, 'f');
	frontend = argpOptionStringVal(ap, 'm');	/* may be NULL */
	pwContext.username = argpOptionStringVal(ap, 'U');
	if(pwContext.username == NULL)
		if((pwContext.username = getenv("USER")) == NULL)
			if((pwContext.username = getenv("LOGNAME")) == NULL)
				if((pwentry = getpwuid(ipccUserId())) != NULL)
					pwContext.username = pwentry->pw_name;
	if(pwContext.username == NULL){
		fprintf(stderr, "%s: can\'t figure out which user to connect\n", progName);
		rval = 1;
		goto doReturn;
	}
	pwContext.domain = argpOptionStringVal(ap, 'D');
	pwContext.password = argpOptionStringVal(ap, 'P');
	pwContext.allowUnencrypted = argpOptionIsSet(ap, 'u');
	pwContext.pwFromStdin = argpOptionIsSet(ap, 'S');
	pwContext.dontAskPw = argpOptionIsSet(ap, 'N');
	if((rval = setupConnection(&pwContext)) != 0)
		goto doReturn;
	/* cope with relative pathes etc. */
	if((mp = sysRealpath(argpArgumentAtIndex(ap, 1), buffer)) == NULL)
		mp = argpArgumentAtIndex(ap, 1);	/* try with user specified value anyway */
	ipccMount(&res, frontend, mp, "cifsFile", argpArgumentAtIndex(ap, 0), isReadonly, fakeMode);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	cifsloginImp(btint argc, char **argv)
{
argParser_t			*ap;
ipccRes_t			res;
int					bslen, rval = 0;
struct passwd		*pwentry;
passwordContext_t	pwContext;
btu8				slash[8], backslash[8], normServer[256], normService[512];
btu8				*service, *server, *share;

	memset(&pwContext, 0, sizeof(pwContext));
	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");
	argpAddOption(ap, 'U', "<username>", 0, "Login on server as this user");
	argpAddOption(ap, 'D', "<domain>", 0, "Send this domain name to server");
	argpAddOption(ap, 'P', "<password>", 0, "Password given in commandline");
	argpAddOption(ap, 'S', NULL, 0, "Read password from standard input (implies -N)");
	argpAddOption(ap, 'N', NULL, 0, "Don\'t prompt for a password");
	argpAddOption(ap, 'u', NULL, 0, "Allow sending password unencrypted");
	
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		printf("usage: %s <server> [<options>]\n", progName);
		printf("       %s //<server>/<share> [<options>]\n", progName);
		printf("       %s -v     for version information\n", progName);
		printf("       %s -man   for the full manual page\n", progName);
		printf("valid options are:\n");
		argpPrintHelp(ap, stdout);
		return 0;
	}
	if(argpArgumentCount(ap) < 1){
		fprintf(stderr, "Not enough arguments.\n");
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		rval = 1;
		goto doReturn;
	}
	if(argpArgumentCount(ap) > 1){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	pwContext.username = argpOptionStringVal(ap, 'U');
	if(pwContext.username == NULL)
		if((pwContext.username = getenv("USER")) == NULL)
			if((pwContext.username = getenv("LOGNAME")) == NULL)
				if((pwentry = getpwuid(ipccUserId())) != NULL)
					pwContext.username = pwentry->pw_name;
	if(pwContext.username == NULL){
		fprintf(stderr, "%s: can\'t figure out which user to connect\n", progName);
		rval = 1;
		goto doReturn;
	}
	pwContext.domain = argpOptionStringVal(ap, 'D');
	pwContext.password = argpOptionStringVal(ap, 'P');
	pwContext.allowUnencrypted = argpOptionIsSet(ap, 'u');
	pwContext.pwFromStdin = argpOptionIsSet(ap, 'S');
	pwContext.dontAskPw = argpOptionIsSet(ap, 'N');
	if((rval = setupConnection(&pwContext)) != 0)
		goto doReturn;
	service = argpArgumentAtIndex(ap, 0);
	ipccSlashChars(&res, slash, backslash, sizeof(slash));
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
		goto doReturn;
	}
	bslen = strlen(backslash);
	if(strstr(service, slash) != NULL || strstr(service, backslash) != NULL){
		utilParseService(service, normServer, ELEMENTS(normServer), normService, ELEMENTS(normService), slash, strlen(slash), backslash, bslen);
		server = normServer;
		share = normService + 2 * bslen;
		share = strstr(share, backslash);
		if(share != NULL)
			share += bslen;
	}else{
		server = service;
		share = NULL;
	}
	ipccCifsLogin(&res, server, share);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	cifslistImp(btint argc, char **argv)
{
argParser_t	*ap;
int			rval = 0;
btbool		loginsOnly, mountsOnly, rawFormat;
btu8		*fieldSeparator;

	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");	
	argpAddOption(ap, 'l', NULL, 0, "List logins only");	
	argpAddOption(ap, 'm', NULL, 0, "List mounts only");	
	argpAddOption(ap, 'r', NULL, 0, "Raw output format (useful for scripts)");
	argpAddOption(ap, 's', "<separator>", 0, "Set string used to separate table entries");
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		shortHelp(ap, "[<options>]");
		return 0;
	}
	if(argpArgumentCount(ap) > 0){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	loginsOnly = argpOptionIsSet(ap, 'l');
	mountsOnly = argpOptionIsSet(ap, 'm');
	rawFormat = argpOptionIsSet(ap, 'r');
	fieldSeparator = argpOptionStringVal(ap, 's');
	if(loginsOnly && mountsOnly)
		loginsOnly = mountsOnly = 0;	/* list both in case of conflict */
	if((rval = setupConnection(NULL)) != 0)
		goto doReturn;
	if(!loginsOnly)
		rval = listPrimitive(MAIN_LIST_MOUNTS, rawFormat, fieldSeparator);
	if(!loginsOnly && !mountsOnly)
		printf("==============================================================================\n");
	if(!mountsOnly && rval == 0)
		rval = listPrimitive(CIFS_LIST_LOGINS, rawFormat, fieldSeparator);
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	cifslogoutImp(btint argc, char **argv)
{
argParser_t	*ap;
ipccRes_t	res;
int			rval = 0;

	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");	
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		shortHelp(ap, "<server> [<options>]");
		return 0;
	}
	if(argpArgumentCount(ap) < 1){
		fprintf(stderr, "Not enough arguments.\n");
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		rval = 1;
		goto doReturn;
	}
	if(argpArgumentCount(ap) > 1){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	if((rval = setupConnection(NULL)) != 0)
		goto doReturn;
	ipccCifsLogout(&res, argpArgumentAtIndex(ap, 0));
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static void	printLicense(void)
{
btu8		lstring[8], *licenseType, *expiry;
bti32		maxServers, maxClients, i;
btbool		srvCountLocal;
ipccRes_t	res;

	ipccCifsGetLicense(&res, lstring, &licenseType, &maxClients, &maxServers, &srvCountLocal, &expiry);
	if(res.err.class == 0){
		printf("%24s: ", "License Identifier");
		for(i=0;i<8;i++){
			printf("%02X", lstring[i]);
			if(i == 0 || i == 3)
				printf("-");
		}
		printf("\n");
		printf("%24s: %s\n", "License Type", licenseType);
		if(maxClients != -1)
			printf("%24s: %d\n", "Max. active Clients", maxClients);
		else
			printf("%24s: unlimited\n", "Max. active Clients");
		if(maxServers != -1)
			printf("%24s: %d (%s)\n", "Max. active Servers", maxServers, srvCountLocal ? "locally counted" : "network wide");
		else
			printf("%24s: unlimited\n", "Max. active Servers");
		printf("%24s: %s\n", "License Expiry", expiry);
	}else{
		ipccPrintErr(&res);
	}
}

static void	printLoad(void)
{
btu8		lstring[8], sMaxServers[32], sMaxClients[32], *licenseType, *expiry;
bti32		maxServers, maxClients;
btint		servers, clients;
btbool		scountIsLocal = 0;
ipccRes_t	res;

	ipccCifsGetLicense(&res, lstring, &licenseType, &maxClients, &maxServers, &scountIsLocal, &expiry);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		exit(1);
	}
	ipccCifsCurrentLicenseLoad(&res, &servers, &clients);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		exit(1);
	}
	if(maxClients == -1)
		sprintf(sMaxClients, "unlimited");
	else
		sprintf(sMaxClients, "%d", maxClients);
	if(maxServers == -1)
		sprintf(sMaxServers, "unlimited");
	else
		sprintf(sMaxServers, "%d", maxServers);
	printf("servers mounted: %d (of maximum %s%s)\n", servers, sMaxServers, scountIsLocal ? " locally counted" : "");
	printf("clients active:  %d (of maximum %s)\n", clients, sMaxClients);
}

static void	printList(void)
{
struct ipcMsg	*msg;
btint			numClients, numServers, i, j;
btu32			ip;
btu8			*name;
ipccRes_t		res;

	ipccCifsCurrentServers(&res);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		exit(1);
	}
	msg = ipccCurrentRxMessage();
	numClients = ipcGetInt(msg);
	for(i=0;i<numClients;i++){
		ip = ipcGetInt(msg);
		name = ipcGetAscii(msg);
		numServers = ipcGetInt(msg);
		if(ip == 0){
			printf("\nclient \"%s\" (localhost): %d server%s\n", name, numServers, numServers == 1 ? "":"s");
		}else{
			printf("\nclient \"%s\" (%d.%d.%d.%d): %d server%s\n", name, ip>>24, (ip>>16)&0xff, (ip>>8)&0xff, ip&0xff, numServers, numServers == 1 ? "":"s");
		}
		for(j=0;j<numServers;j++){
			name = ipcGetAscii(msg);
			printf("    server \"%s\"\n", name);
		}
	}
}

static void	setLicense(btu8 *userData)
{
btint		i, l;
btu8		ldata[12], buffer[24], *p = userData;
ipccRes_t	res;

	for(i=0;i<24;p++){
		if(*p == 0){
			fprintf(stderr, "%s: too short for license string\n", userData);
			exit(1);
		}
		if(!isalnum(*p))	/* skip any non-alphanumeric chars */
			continue;
		if(!isxdigit(*p)){
			fprintf(stderr, "%s: contains illegal characters\n", userData);
			exit(1);
		}
		buffer[i++] = *p;
	}
	for(i=0;i<12;i++){
		sscanf(&buffer[2 * i], "%02x", &l);
		ldata[i] = l;
	}
	ipccCifsSetLicense(&res, ldata);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		exit(1);
	}else{
		printf("Licensing succeeded, Key Properties:\n");
		printLicense();
	}
}

static void	licenseHelp(btu8 *name)
{
	fprintf(stderr, "\n");
	fprintf(stderr, "usage:  %s -p     prints license identifier\n", name);
	fprintf(stderr, "        %s -l     prints current client and server load\n", name);
	fprintf(stderr, "        %s -ll    prints license utilisation\n", name);
	fprintf(stderr, "        %s <key>  sets license string to <key>\n", name);
	fprintf(stderr, "        %s -h     prints short help\n", name);
}

static btint	cifslicenseImp(btint argc, char **argv)
{
btint	rval;

	if(argc != 2){
		licenseHelp(progName);
		return 1;
	}
	if(strcmp(argv[1], "-h") == 0){
		licenseHelp(progName);
		return 0;
	}
	if((rval = setupConnection(NULL)) != 0)
		return rval;
	if(strcmp(argv[1], "-p") == 0){
		printLicense();
	}else if(strcmp(argv[1], "-l") == 0){
		printLoad();
	}else if(strcmp(argv[1], "-ll") == 0){
		printLoad();
		printList();
	}else{
		setLicense(argv[1]);
	}
	return 0;
}

/* ------------------------------------------------------------------------- */

static btint	cifsstoreImp(btint argc, char **argv)
{
argParser_t	*ap;
ipccRes_t	res;
btbool		doDelete;
int			rval = 0;

	ap = argpNew();
	argpAddOption(ap, 'h', NULL, 0, "Print short help and exit");	
	argpAddOption(ap, 'd', NULL, 0, "Delete stored login record");	
	if(!argpParse(ap, argc, argv, stderr)){
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		return 1;
	}
	if(argpOptionIsSet(ap, 'h')){
		shortHelp(ap, "<server> [<options>]");
		return 0;
	}
	if(argpArgumentCount(ap) < 1){
		fprintf(stderr, "Not enough arguments.\n");
		fprintf(stderr, "Type \'%s -h\' for a short help.\n", progName);
		rval = 1;
		goto doReturn;
	}
	if(argpArgumentCount(ap) > 1){
		fprintf(stderr, "Warning: excess arguments ignored\n");
	}
	doDelete = argpOptionIsSet(ap, 'd');
	if((rval = setupConnection(NULL)) != 0)
		goto doReturn;
	ipccCifsStore(&res, argpArgumentAtIndex(ap, 0), doDelete);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	genericVersionImp(btint argc, char **argv)
{
ipccRes_t	res;
btint		cnt, i, rval = 0;
ipcMsg_t	*msg;
btu8		*modName;
bti32		implVersion, beta;

	if((rval = setupConnection(NULL)) != 0)
		goto doReturn;
	ipccListModules(&res);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}else{
		msg = ipccCurrentRxMessage();
		cnt = ipcGetInt(msg);
		for(i=0;i<cnt;i++){
			ipcGetInt(msg);	/* type */
			modName = ipcGetAscii(msg);	/* name */
			ipcGetInt(msg);	/* interface version */
			implVersion = ipcGetInt(msg);	/* implementation version */
			beta = ipcGetInt(msg);	/* beta version */
			ipcGetInt(msg);	/* function base */
			ipcGetInt(msg);	/* function count */
			if(beta != 0)
				printf("%30s: version %2d.%d beta %d\n", modName, implVersion / 1000, implVersion % 1000, beta);
			else
				printf("%30s: version %2d.%d\n", modName, implVersion / 1000, implVersion % 1000);
		}
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	genericManualImp(btint argc, char **argv)
{
ipccRes_t	res;
btint		rval = 0;
btu8		*p, helpItem[256], *helpText = "";
FILE		*fp;

	if(argc < 1){
		fprintf(stderr, "internal error: no command name found\n");
		return 1;
	}
	if((rval = setupConnection(NULL)) != 0)
		goto doReturn;
	p = strrchr(argv[0], '/');
	if(p == NULL)
		p = argv[0];
	else
		p++;
	if(strlen(p) > 200)
		p[200] = 0;	/* emergency brake */
	sprintf(helpItem, "man%s", p);
	ipccHelpForKey(&res, helpItem, &helpText);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		rval = 1;
	}else{
		fp = popen("more", "w");
		if(fp == NULL){
			fprintf(stderr, "cannot pipe through more, printing to terminal\n");
			printf("%s", helpText);
		}else{
			fprintf(fp, "%s", helpText);
			pclose(fp);
		}
	}
doReturn:
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	genericHelpImp(btint argc, char **argv);

static struct{
	char	*name;
	btint	(*function)(btint argc, char **argv);
}	commands[] = {
	{"-h", genericHelpImp},
	{"mount", mountImp},
	{"umount", umountImp},
	{"storemnt", storemntImp},
	{"list", listImp},
	{"cifsmount", cifsmountImp},
	{"cifsumount", umountImp},
	{"cifslogin", cifsloginImp},
	{"cifslogout", cifslogoutImp},
	{"cifslist", cifslistImp},
	{"cifslicense", cifslicenseImp},
	{"cifsstore", cifsstoreImp},
	{"cifsstoremnt", storemntImp},
};

static btint	genericHelpImp(btint argc, char **argv)
{
btint	i;

	fprintf(stderr, "Available commands are:\n");
	for(i=1;i<ELEMENTS(commands);i++){
		fprintf(stderr, "\t%s\n", commands[i].name);
	}
	fprintf(stderr, "For more help on a particular command use option \'-h\' with the command.\n");
	return 0;
}

/* ------------------------------------------------------------------------- */

int		main(volatile int argc, char ** volatile argv)
{
volatile btint	i, cmd, rval = 0;
static char		executableName[256];
char			*p;

	if(excInternalError == NULL)	/* reference symbol to fight liker bug on Nextstep 3.3 */
		printf("huh, symbol undefined?\n");
	/* first look up our name in table, maybe started with symlink */
	p = strrchr(argv[0], '/');
	if(p == NULL)
		p = argv[0];
	else
		p++;
	progName = p;
	if(argc == 2 && strcmp(argv[1], "-v") == 0){
		genericVersionImp(argc, argv);
		return 0;
	}else if(argc == 2 && strcmp(argv[1], "-man") == 0){
		genericManualImp(argc, argv);
		return 0;
	}
	for(i=0;i<ELEMENTS(commands);i++){
		if(strcmp(p, commands[i].name) == 0)
			break;
	}
	if(i >= ELEMENTS(commands)){
		if(argc < 2){
			fprintf(stderr, "Usage: %s <command> [<options>]\n", argv[0]);
			fprintf(stderr, "Type %s -h   for a list of commands\n", argv[0]);
			fprintf(stderr, "     %s -man for the manual page\n", argv[0]);
			exit(1);
		}
		if((p = strrchr(argv[0], '/')) == NULL)
			p = argv[0];
		else
			p++;
		if(strlen(p) > 100)
			p[100] = 0;	/* emergency brake */
		if(strlen(argv[1]) > 100)
			argv[1][100] = 0;	/* another emergency brake */
		sprintf(executableName, "%s %s", p, argv[1]);
		progName = executableName;
		for(i=0;i<ELEMENTS(commands);i++){
			if(strcmp(argv[1], commands[i].name) == 0)
				break;
		}
		argc--;
		argv++;
	}
	cmd = i;
	if(cmd < ELEMENTS(commands)){
		if(argc == 2 && strcmp(argv[1], "-v") == 0){
			genericVersionImp(argc, argv);
		}else if(argc == 2 && strcmp(argv[1], "-man") == 0){
			genericManualImp(argc, argv);
		}else{
EXC_TRY;
			rval = commands[cmd].function(argc, argv);
EXC_CATCH;
			fprintf(stderr, "%s caught exception: %s (%s)\n", argv[0], EXC_NAME, EXC_MSG);
			exit(1);
EXC_ENDTRY;
		}
	}else{
		fprintf(stderr, "Command \"%s\" not recognized\n", argv[0]);
		fprintf(stderr, "Type %s -h for a list of commands\n", argv[-1]);
		exit(1);
	}
	return rval;
}

/* ------------------------------------------------------------------------- */

