/*
 * Name: guiMain.c
 * Project: CIFS Client GUI
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-26
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: guiMain.c,v 1.16 2001/05/05 22:29:23 cs Exp $
 */

#include "stdheaders.h"
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#define	initstate	math_initstate	/* work around gcc-header incompatibility */
#include <math.h>
#undef initstate
#include "basictypes.h"
#include "guiHelper.h"
#include "ipc.h"
#include "ipcclient.h"
#include "maininterface.h"
#include "array.h"
#include "hashtable.h"
#include "errcodes.h"
#include "memalloc.h"
#include "util.h"
#include "../mainFramework/mainconst.h"	/* for MAINCONST_VERSION */

/* ------------------------------------------------------------------------- */

static btbool	cfgIsExpert = 0;		/* user-settable expert mode */
static btbool	cfgUsersMayChange = 0;	/* whether users may change settings */
static btbool	cfgIsRoot = 0;			/* whether user is root */

/* ------------------------------------------------------------------------- */

bti8	*helpExpertMode = "Select for Expert Mode";

bti8	*helpAdd = "Add an Entry";

bti8	*helpDelete = "Delete an Entry";

/* ------------------------------------------------------------------------- */

#define	ENTRY_WIDTH		210
#define	ENTRY_HEIGHT	22

typedef struct guiGtkEntry{
	btint		type;
	btu32		flags;
	btint		permissions;
	btint		fractionDivider;
	btint		fractionMultiplier;
	GtkWidget	*entryWidget;
	GtkWidget	*label;
	GtkWidget	*auxWidget;
	GtkWidget	*mainView;
	btbool		isDirty;		/* whether this item has been changed */
	struct guiGtkView	*context;
}guiGtkEntry_t;

typedef struct guiGtkColumn{
	btint		widthPercent;
}guiGtkColumn_t;

typedef struct guiGtkButton{
	btint		function;	/* correlated to method indices */
	btint		parameter;
	btint		flags;
	GtkWidget	*buttonWidget;
	struct guiGtkView	*context;
}guiGtkButton_t;

typedef struct guiGtkView{
	btint		objectId;
	btint		objectType;
	struct guiGtkDialog	*dialog;
	GtkWidget	*mainView;
	GtkCList	*selectionList;
	GtkWidget	*configParent;	/* parent of either configBox or naLabel */
	GtkWidget	*configBox;
	GtkWidget	*naLabel;
	GtkWidget	*helpButton;
	btu8		*hierarchyItem;
	btint		selectedRow;	/* for list views */
	array_t		*entries;
	array_t		*columns;
	array_t		*buttons;
	btbool		isDirty;
	btbool		needsUpdate;
}guiGtkView_t;

typedef struct guiGtkDialog{
	btint			objectId;
	btbool			runModal;
	GtkWidget		*window;
	guiGtkView_t	*view;
	gint			delayedInitId;
}guiGtkDialog_t;

static array_t		*allViews;
static hashtable_t	*dialogsByObjectId;

/* ------------------------------------------------------------------------- */

static btint	dialogsCompare(void *o1, void *o2)
{
guiGtkDialog_t	*this = o1, *other = o2;

	return this->objectId - other->objectId;
}

static btu32	dialogHashVal(void *obj)
{
guiGtkDialog_t	*this = obj;

	return this->objectId;
}

static collectionObjectI_t	dialogInterface = {
	dialogsCompare,
	dialogHashVal,
	NULL,
	NULL,
};

/* ------------------------------------------------------------------------- */

static void	guiViewUpdateDirty(guiGtkView_t *view);
static void	guiViewDataLoad(guiGtkView_t *view, btbool withDefaults);
static void	guiViewStore(guiGtkView_t *view);
static void	guiViewListLoad(guiGtkView_t *view, btbool forceUpdateData);
static void	guiDialogDestroy(guiGtkDialog_t *dialog);

GtkWidget	*statusBar = NULL;
gint		statusBarContext;
gint		statusBarMessageId = -1;

static btbool	ignoreClistSignals = 0;

/* ------------------------------------------------------------------------- */

static void	setWindowTitle(void *window, gchar *title)
{
gchar	titleBuffer[1024];

	gethostname(titleBuffer, sizeof(titleBuffer));
	strncat(titleBuffer, ": ", sizeof(titleBuffer));
	strncat(titleBuffer, title, sizeof(titleBuffer));
	titleBuffer[sizeof(titleBuffer)-1] = 0;
	gtk_window_set_title(GTK_WINDOW(window), titleBuffer);
}

/* ------------------------------------------------------------------------- */

static inline btbool	haveAccess(btint permFlags, btu8 **denyReason)
{
btbool	allow = 0;

	*denyReason = "";
	if((permFlags & MAIN_GUI_PERM_ANYBODY) != 0)
		allow = 1;
	if((permFlags & MAIN_GUI_PERM_AUTHORIZED) != 0)
		allow |= cfgUsersMayChange || cfgIsRoot;
	if((permFlags & MAIN_GUI_PERM_ROOT) != 0)
		allow |= cfgIsRoot;
	if(permFlags == 0)
		*denyReason = "Not editable, display only";
	else if(!allow)
		*denyReason = "Editable only by root";
	if(allow && (permFlags & MAIN_GUI_PERM_EXPERT) != 0 && !cfgIsExpert){
		*denyReason = "Editable only in expert mode";
		allow = 0;
	}
	return allow;
}

/* ------------------------------------------------------------------------- */

static gint	mouseEnters(GtkWidget *widget, GdkEventCrossing *event, guiGtkEntry_t *entry)
{
btu8	*reason = NULL;

	if(statusBar != NULL){
		if(statusBarMessageId != -1)
			gtk_statusbar_remove(GTK_STATUSBAR(statusBar), statusBarContext, statusBarMessageId);
		statusBarMessageId = -1;
		if(!haveAccess(entry->permissions, &reason)){
			statusBarMessageId = gtk_statusbar_push(GTK_STATUSBAR(statusBar), statusBarContext, reason);
		}
	}
	return FALSE;
}

static gint	mouseLeaves(GtkWidget *widget, GdkEventCrossing *event, guiGtkEntry_t *entry)
{
	if(statusBar != NULL){
		if(statusBarMessageId != -1)
			gtk_statusbar_remove(GTK_STATUSBAR(statusBar), statusBarContext, statusBarMessageId);
		statusBarMessageId = -1;
	}
	return FALSE;
}

static void	enumSelected(GtkWidget *sender, guiGtkEntry_t *entry)
{
	gtk_object_set_user_data(GTK_OBJECT(entry->entryWidget), gtk_object_get_user_data(GTK_OBJECT(sender)));
}

static void	runFileselector(GtkWidget *sender, guiGtkEntry_t *entry)
{
gchar	*name;

	name = guiRunFileselector("Select Path", gtk_entry_get_text(GTK_ENTRY(entry->entryWidget)));
	if(name != NULL){
		gtk_entry_set_text(GTK_ENTRY(entry->entryWidget), name);
		g_free(name);
	}
}

static void	entriesTouched(guiGtkEntry_t *entry)
{
	if(!entry->isDirty){
		entry->isDirty = entry->context->isDirty = 1;
		guiViewUpdateDirty(entry->context);
	}
}

static guiGtkEntry_t	*guiEntryNew(guiGtkView_t *contextView, ipcMsg_t *msg)
{
guiGtkEntry_t	*entry;
btu8			*s, *titleText, *helpText, *unitTxt, scratch[256];
GtkWidget		*cbox = NULL, *ebox, *value = NULL, *button, *menu, *item, *unit;
GtkObject		*adj;
btint			i, numItems;
double			minVal, maxVal, increment;

	entry = memCalloc(sizeof(guiGtkEntry_t));
	entry->context = contextView;
	entry->type = ipcGetInt(msg);
	entry->flags = ipcGetInt(msg);
	entry->permissions = ipcGetInt(msg);
	titleText = ipcGetAscii(msg);
	helpText = ipcGetAscii(msg);
	switch(entry->type){
	case MAIN_GUI_ENTRY_STRING:
	case MAIN_GUI_ENTRY_PASSWORD:
	case MAIN_GUI_ENTRY_FILE:
		cbox = value = gtk_entry_new();
		if(entry->type == MAIN_GUI_ENTRY_PASSWORD)
			gtk_entry_set_visibility(GTK_ENTRY(value), FALSE);
		gtk_signal_connect_object(GTK_OBJECT(value), "changed", GTK_SIGNAL_FUNC(entriesTouched), (void *)entry);
		if(entry->type == MAIN_GUI_ENTRY_FILE){
			entry->auxWidget = button = gtk_button_new_with_label("...");
			gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(runFileselector), entry);
			cbox = guiAlignmentBox(FALSE, 0, value, TRUE, button, TRUE, NULL);
			gtk_widget_set_usize(value, ENTRY_WIDTH - 16, -1);
			gtk_widget_show(button);
		}
		break;
	case MAIN_GUI_ENTRY_ENUM:
		cbox = value = gtk_option_menu_new();
		gtk_widget_set_usize(value, ENTRY_WIDTH, ENTRY_HEIGHT);
		menu = gtk_menu_new();
		numItems = ipcGetInt(msg);
		for(i=0;i<numItems;i++){
			s = ipcGetAscii(msg);
			item = gtk_menu_item_new_with_label(s);
			gtk_object_set_user_data(GTK_OBJECT(item), (gpointer)i);
			gtk_signal_connect(GTK_OBJECT(item), "activate", GTK_SIGNAL_FUNC(enumSelected), entry);
			gtk_signal_connect_object(GTK_OBJECT(item), "activate", GTK_SIGNAL_FUNC(entriesTouched), (void *)entry);
			gtk_menu_append(GTK_MENU(menu), item);
			gtk_widget_show(item);
		}
		gtk_option_menu_set_menu(GTK_OPTION_MENU(value), menu);
		break;
	case MAIN_GUI_ENTRY_NUMBER:
		entry->fractionDivider = ipcGetInt(msg);
		entry->fractionMultiplier = ipcGetInt(msg);
		increment = (double)ipcGetInt(msg) * entry->fractionMultiplier / entry->fractionDivider;
		minVal = (double)(btint)ipcGetInt(msg) * entry->fractionMultiplier / entry->fractionDivider;
		maxVal = (double)(btint)ipcGetInt(msg) * entry->fractionMultiplier / entry->fractionDivider;
		unitTxt = ipcGetAscii(msg);
		adj = gtk_adjustment_new(0, minVal, maxVal, increment, increment * 10, 0.0);
		value = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, increment < 0.999 ? 2 : 0);
		gtk_signal_connect_object(GTK_OBJECT(value), "changed", GTK_SIGNAL_FUNC(entriesTouched), (void *)entry);
		gtk_widget_set_usize(value, ENTRY_WIDTH/2, ENTRY_HEIGHT);
		entry->auxWidget = unit = gtk_label_new(unitTxt);
		gtk_misc_set_alignment(GTK_MISC(unit), 0, 0.5);
		cbox = guiAlignmentBox(FALSE, 5, value, FALSE, unit, TRUE, NULL);
		gtk_widget_show(unit);
		break;
	case MAIN_GUI_ENTRY_BOOL:
		value = gtk_check_button_new();
		gtk_signal_connect_object(GTK_OBJECT(value), "clicked", GTK_SIGNAL_FUNC(entriesTouched), (void *)entry);
		cbox = guiAlignmentBox(FALSE, 0, value, FALSE, NULL);
		break;
	case MAIN_GUI_ENTRY_SEPARATOR:
		cbox = guiAlignmentBox(FALSE, 0, NULL);
		break;
	}
	strncpy(scratch, titleText, sizeof(scratch));
	strncat(scratch, ": ", sizeof(scratch));
	scratch[sizeof(scratch)-1] = 0;	/* ensure 0-termination */
	entry->label = gtk_label_new(scratch);
	gtk_misc_set_alignment(GTK_MISC(entry->label), 1.0, 0.5);
	gtk_widget_set_usize(cbox, ENTRY_WIDTH, ENTRY_HEIGHT);
	ebox = gtk_event_box_new();
	gtk_container_add(GTK_CONTAINER(ebox), guiAlignmentBox(FALSE, 0, entry->label, TRUE, cbox, FALSE, NULL));
	gtk_widget_show(ebox);
	gtk_widget_set_events(ebox, GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
	gtk_signal_connect(GTK_OBJECT(ebox), "enter_notify_event", GTK_SIGNAL_FUNC(mouseEnters), entry);
	gtk_signal_connect(GTK_OBJECT(ebox), "leave_notify_event", GTK_SIGNAL_FUNC(mouseLeaves), entry);
	gtk_widget_show(entry->label);
	if(value != NULL){
		gtk_widget_show(value);
		gtk_signal_connect(GTK_OBJECT(value), "enter_notify_event", GTK_SIGNAL_FUNC(mouseEnters), entry);
		gtk_signal_connect(GTK_OBJECT(value), "leave_notify_event", GTK_SIGNAL_FUNC(mouseLeaves), entry);
	}
	if(helpText != NULL){
		helpText = g_strdup(helpText);
		if(value != NULL)
			guiHelpifyWidget(value, helpText);
		guiHelpifyWidget(ebox, helpText);
	}
	entry->mainView = ebox;
	entry->entryWidget = value;
	return entry;
}

static void	guiEntryDestroy(guiGtkEntry_t *entry)
{
	memFree(entry);
}

static void	guiEntryUpdatePermissions(guiGtkEntry_t *entry)
{
btbool	isEditable;
btu8	*reason;

	isEditable = haveAccess(entry->permissions, &reason);
	switch(entry->type){
	case MAIN_GUI_ENTRY_NUMBER:
		/*gtk_widget_set_sensitive(entry->entryWidget, isEditable);*/
	case MAIN_GUI_ENTRY_FILE:
		if(entry->auxWidget != NULL)	/* aux widget is button or label */
			gtk_widget_set_sensitive(entry->auxWidget, isEditable);
	case MAIN_GUI_ENTRY_STRING:
	case MAIN_GUI_ENTRY_PASSWORD:
		/*gtk_entry_set_editable(GTK_ENTRY(entry->entryWidget), isEditable);*/
		gtk_widget_set_sensitive(entry->entryWidget, isEditable);
		break;
	case MAIN_GUI_ENTRY_ENUM:
	case MAIN_GUI_ENTRY_BOOL:
		gtk_widget_set_sensitive(entry->entryWidget, isEditable);
		break;
	}
	gtk_widget_set_sensitive(entry->label, isEditable);
}

/* ------------------------------------------------------------------------- */

static void	buttonClicked(GtkWidget *sender, guiGtkButton_t *button)
{
btbool		methodCalled = 0;
ipcMsg_t	*msg;
ipccRes_t	res;
btu8		*rowId;

	if(button->context->objectType != MAIN_GUI_CLASS_LIST){
		if(button->function == MAIN_GUI_METHOD_GET){
			guiViewDataLoad(button->context, button->parameter);
			methodCalled = 1;
		}else if(button->function == MAIN_GUI_METHOD_SET){
			guiViewStore(button->context);
			methodCalled = 1;
		}
	}
	if(!methodCalled){
		msg = ipccCurrentRxMessage();
		ipccStartProc(msg, MAIN_IPCFN_DSOURCE_EXEC);
		ipcAddInt(msg, button->context->objectId);
		ipcAddInt(msg, button->function);
		ipcAddInt(msg, button->parameter);
		if(button->context->objectType == MAIN_GUI_CLASS_LIST){
			rowId = gtk_clist_get_row_data(button->context->selectionList, button->context->selectedRow);
			ipcAddAscii(msg, rowId);
		}
		ipccRpc(msg);
		res.err.class = ipcGetInt(msg);
		res.err.code = ipcGetInt(msg);
		res.errmsg = ipcGetAscii(msg);
		if(res.err.class != 0){
			ipccPrintErr(&res);
		}
	}
	if((button->flags & MAIN_GUI_BUTTON_CLOSE_DIALOG) != 0){
		if(button->context->dialog != NULL){
			guiDialogDestroy(button->context->dialog);
		}
	}
}

static guiGtkButton_t	*guiButtonNew(guiGtkView_t *view, ipcMsg_t *msg)
{
guiGtkButton_t	*button;
btu8			*name, *helpText, scratch[256];

	button = memAlloc(sizeof(*button));
	name = ipcGetAscii(msg);
	button->function = ipcGetInt(msg);
	button->parameter = ipcGetInt(msg);
	button->flags = ipcGetInt(msg);
	scratch[0] = ' ';
	scratch[1] = 0;
	strncat(scratch, name, sizeof(scratch));
	strncat(scratch, " ", sizeof(scratch));
	scratch[sizeof(scratch)-1] = 0;	/* ensure 0-termination */
	button->buttonWidget = gtk_button_new_with_label(scratch);
	gtk_container_set_border_width(GTK_CONTAINER(button->buttonWidget), 5);
	gtk_widget_show(button->buttonWidget);
	if((helpText = ipcGetAscii(msg)) != NULL){
		helpText = g_strdup(helpText);
		guiHelpifyWidget(button->buttonWidget, helpText);
	}
	gtk_signal_connect(GTK_OBJECT(button->buttonWidget), "clicked", GTK_SIGNAL_FUNC(buttonClicked), button);
	button->context = view;
	return button;
}

static void	guiButtonDestroy(guiGtkButton_t *button)
{
	memFree(button);
}

/* ------------------------------------------------------------------------- */

static void	addEntrySignal(GtkWidget *sender, guiGtkView_t *view)
{
ipcMsg_t	*msg;
btint		i, cnt;
btu8		*item;
GList		*l = NULL;
ipccRes_t	res;

	msg = ipccCurrentRxMessage();
	ipccStartProc(msg, MAIN_IPCFN_DSOURCE_EXEC);
	ipcAddInt(msg, view->objectId);
	ipcAddInt(msg, MAIN_GUI_METHOD_HLIST);
	ipcAddInt(msg, 1);	/* get all possible items */
	ipccRpc(msg);
	res.err.class = ipcGetInt(msg);
	res.err.code = ipcGetInt(msg);
	res.errmsg = ipcGetAscii(msg);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		return;
	}
	cnt = ipcGetInt(msg);
	for(i=0;i<cnt;i++){
		item = ipcGetAscii(msg);
		l = g_list_append(l, item);
	}
	item = guiRunSelectionBox("Add Item", " Add ", " Cancel ", l);
	g_list_free(l);
	if(item != NULL){
		ipccStartProc(msg, MAIN_IPCFN_DSOURCE_EXEC);
		ipcAddInt(msg, view->objectId);
		ipcAddInt(msg, MAIN_GUI_METHOD_ADDHITEM);
		ipcAddAscii(msg, item);
		ipccRpc(msg);
		res.err.class = ipcGetInt(msg);
		res.err.code = ipcGetInt(msg);
		res.errmsg = ipcGetAscii(msg);
		if(res.err.class != 0){
			ipccPrintErr(&res);
		}
		if(view->hierarchyItem != NULL){
			g_free(view->hierarchyItem);
			view->hierarchyItem = item;
		}
	}
}

static void	deleteEntrySignal(GtkWidget *sender, guiGtkView_t *view)
{
ipccRes_t	res;
ipcMsg_t	*msg;

	if(view->hierarchyItem == NULL)
		return;
	msg = ipccCurrentRxMessage();
	ipccStartProc(msg, MAIN_IPCFN_DSOURCE_EXEC);
	ipcAddInt(msg, view->objectId);
	ipcAddInt(msg, MAIN_GUI_METHOD_DELHITEM);
	ipcAddAscii(msg, view->hierarchyItem);
	ipccRpc(msg);
	res.err.class = ipcGetInt(msg);
	res.err.code = ipcGetInt(msg);
	res.errmsg = ipcGetAscii(msg);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		return;
	}
}

static void	selectionChanged(GtkWidget *clist, gint row, gint column, GdkEventButton *event, guiGtkView_t *view)
{
gchar	*item;

	if(ignoreClistSignals)
		return;
	if(view->hierarchyItem != NULL){
		g_free(view->hierarchyItem);
		view->hierarchyItem = NULL;
	}	
	item = NULL;
	gtk_clist_get_text(view->selectionList, row, 0, &item);
	if(item != NULL)
		view->hierarchyItem = g_strdup(item);
	guiViewDataLoad(view, 0);
}

/* ------------------------------------------------------------------------- */

static void	guiViewConfigNew(guiGtkView_t *view, ipcMsg_t *msg, btbool standAloneWindow)
{
btint			numEntries, i;
guiGtkEntry_t	*entry;
guiGtkButton_t	*button;
GtkWidget		*vbox, *buttonBox, *alignBox, *alignment, *leftBox, *frame, *addEntry, *deleteEntry, *scrollWin;

	if(view->objectType == MAIN_GUI_CLASS_HIERARCHYCFG){
		view->selectionList = GTK_CLIST(gtk_clist_new(1));
		gtk_clist_set_selection_mode(view->selectionList, GTK_SELECTION_BROWSE);
		scrollWin = gtk_scrolled_window_new(NULL, NULL);
		gtk_widget_show(scrollWin);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollWin), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
		gtk_container_add(GTK_CONTAINER(scrollWin), GTK_WIDGET(view->selectionList));
		gtk_clist_set_column_width(view->selectionList, 0, 10);
		gtk_widget_set_usize(GTK_WIDGET(scrollWin), 100, -1);	/* height will be overridden */
		gtk_clist_column_titles_hide(view->selectionList);
		gtk_widget_show(GTK_WIDGET(view->selectionList));
		leftBox = guiAlignmentBox(TRUE, 5, scrollWin, TRUE, guiButtonBar(2, " Add ", &addEntry, " Delete ", &deleteEntry), FALSE, NULL);
		gtk_signal_connect(GTK_OBJECT(addEntry), "clicked", GTK_SIGNAL_FUNC(addEntrySignal), view);
		gtk_signal_connect(GTK_OBJECT(deleteEntry), "clicked", GTK_SIGNAL_FUNC(deleteEntrySignal), view);
		guiHelpifyWidget(addEntry, helpAdd);
		guiHelpifyWidget(deleteEntry, helpDelete);
		gtk_widget_show(GTK_WIDGET(view->selectionList));
		view->naLabel = gtk_label_new("Not Applicable");
		gtk_widget_set_sensitive(view->naLabel, FALSE);
		gtk_widget_ref(view->naLabel);
		gtk_widget_show(view->naLabel);
		gtk_signal_connect(GTK_OBJECT(view->selectionList), "select_row", GTK_SIGNAL_FUNC(selectionChanged), view);
		frame = gtk_frame_new("This Item\'s Configuration");
		gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
		gtk_container_add(GTK_CONTAINER(frame), view->naLabel);
		gtk_widget_show(frame);
		view->configParent = frame;
		view->mainView = guiAlignmentBox(FALSE, 5, leftBox, FALSE, frame, TRUE, NULL);
	}
	vbox = gtk_vbox_new(FALSE, 0);
	numEntries = ipcGetInt(msg);	/* unmarshall the entries */
	for(i=0;i<numEntries;i++){
		entry = guiEntryNew(view, msg);
		arrAddObject(view->entries, entry);
		gtk_box_pack_start(GTK_BOX(vbox), entry->mainView, FALSE, FALSE, 0);
	}
	gtk_widget_show(vbox);
	buttonBox = gtk_hbox_new(TRUE, 0);
	numEntries = ipcGetInt(msg);	/* unmarshall the buttons */
	for(i=0;i<numEntries;i++){
		button = guiButtonNew(view, msg);
		arrAddObject(view->buttons, button);
		if(standAloneWindow){
			GTK_WIDGET_SET_FLAGS(button->buttonWidget, GTK_CAN_DEFAULT);
		}
		gtk_box_pack_start(GTK_BOX(buttonBox), button->buttonWidget, TRUE, TRUE, 0);
	}

	alignBox = gtk_hbox_new(FALSE, 0);
	if(standAloneWindow){
		view->helpButton = gtk_toggle_button_new_with_label(" Help ");
		gtk_container_set_border_width(GTK_CONTAINER(view->helpButton), 5);
		GTK_WIDGET_SET_FLAGS(view->helpButton, GTK_CAN_DEFAULT);
		gtk_widget_show(view->helpButton);
		gtk_box_pack_start(GTK_BOX(alignBox), view->helpButton, FALSE, FALSE, 10);
		guiRegisterHelpTogglebutton(view->helpButton);
	}
	gtk_box_pack_end(GTK_BOX(alignBox), buttonBox, FALSE, FALSE, 0);
	gtk_widget_show(buttonBox);
	gtk_widget_show(alignBox);

	alignment = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
    gtk_container_add(GTK_CONTAINER(alignment), vbox);
	gtk_widget_show(alignment);
	view->configBox = guiAlignmentBox(TRUE, 0, alignment, TRUE, alignBox, FALSE, NULL);
	gtk_widget_ref(view->configBox);
	if(view->objectType != MAIN_GUI_CLASS_HIERARCHYCFG){
		view->mainView = view->configBox;
	}
}

/* ------------------------------------------------------------------------- */

void clistRowSelected(GtkWidget *widget, gint row, gint column, GdkEventButton *event, guiGtkView_t *view)
{
	view->selectedRow = row;
}

static void	setColumnWidths(GtkWidget *widget, GtkAllocation *allocation, guiGtkView_t *view)
{
btint			i, cnt, w;
guiGtkColumn_t	*column;

	cnt = arrObjectCount(view->columns);
	for(i=0;i<cnt;i++){
		column = arrObjectAtIndex(view->columns, i);
		w = (view->selectionList->clist_window_width - 30) * column->widthPercent / 100 - 1;
		gtk_clist_set_column_width(view->selectionList, i, w);
	}
}

static void	guiViewListNew(guiGtkView_t *view, ipcMsg_t *msg)
{
btint			numColumns, i, numButtons, alignment;
btu8			*title, *helpText;
GtkCList		*clist;
guiGtkColumn_t	*column;
guiGtkButton_t	*button;
GtkWidget		*buttonBox, *alignBox, *scrollWin;

	numColumns = ipcGetInt(msg);
	clist = GTK_CLIST(gtk_clist_new(numColumns));
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_BROWSE);
	gtk_clist_column_titles_show(clist);
	scrollWin = gtk_scrolled_window_new(NULL, NULL);
	gtk_widget_show(scrollWin);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollWin), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_container_add(GTK_CONTAINER(scrollWin), GTK_WIDGET(clist));
	gtk_signal_connect(GTK_OBJECT(scrollWin), "size_allocate", GTK_SIGNAL_FUNC(setColumnWidths), view);
	gtk_clist_column_titles_active(clist);
	gtk_signal_connect(GTK_OBJECT(clist), "select_row", GTK_SIGNAL_FUNC(clistRowSelected), view);
	for(i=0;i<numColumns;i++){
		column = memAlloc(sizeof(*column));
		title = ipcGetAscii(msg);
		alignment = ipcGetInt(msg);	/* ignore alignment */
		column->widthPercent = ipcGetInt(msg);
		helpText = ipcGetAscii(msg);
		gtk_clist_set_column_title(clist, i, title);
		gtk_clist_set_column_width(clist, i, column->widthPercent * 4);
		arrAddObject(view->columns, column);
		if(helpText != NULL){
			helpText = g_strdup(helpText);
			guiHelpifyWidget(clist->column[i].button, helpText);
		}
	}
	gtk_widget_show(GTK_WIDGET(clist));

	buttonBox = gtk_hbox_new(TRUE, 0);
	numButtons = ipcGetInt(msg);	/* unmarshall the buttons */
	for(i=0;i<numButtons;i++){
		button = guiButtonNew(view, msg);
		arrAddObject(view->buttons, button);
		gtk_box_pack_start(GTK_BOX(buttonBox), button->buttonWidget, TRUE, TRUE, 0);
	}
	alignBox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_end(GTK_BOX(alignBox), buttonBox, FALSE, FALSE, 0);
	gtk_widget_show(buttonBox);
	gtk_widget_show(alignBox);
	view->selectionList = clist;
	view->mainView = guiAlignmentBox(TRUE, 0, scrollWin, TRUE, alignBox, FALSE, NULL);
}

static void	guiColumnDestroy(guiGtkColumn_t *column)
{
	memFree(column);
}

/* ------------------------------------------------------------------------- */

static void	guiViewUpdatePermissions(guiGtkView_t *view)
{
btint	i, cnt;

	if(view->objectType == MAIN_GUI_CLASS_PLAINCFG || view->objectType == MAIN_GUI_CLASS_HIERARCHYCFG){
		cnt = arrObjectCount(view->entries);
		for(i=0;i<cnt;i++){
			guiEntryUpdatePermissions(arrObjectAtIndex(view->entries, i));
		}
	}
}

static guiGtkView_t	*guiViewNew(btint objectId, ipcMsg_t *msg, btbool standAloneWindow)
{
guiGtkView_t	*view;

	view = memCalloc(sizeof(guiGtkView_t));
	view->objectId = objectId;
	view->objectType = ipcGetInt(msg);
	view->entries = arrNew(NULL, 8);
	view->columns = arrNew(NULL, 6);
	view->buttons = arrNew(NULL, 4);
	switch(view->objectType){
	case MAIN_GUI_CLASS_PLAINCFG:
	case MAIN_GUI_CLASS_HIERARCHYCFG:
		guiViewConfigNew(view, msg, standAloneWindow);
		break;
	case MAIN_GUI_CLASS_LIST:
		guiViewListNew(view, msg);
		break;
	}
	arrAddObject(allViews, view);
	guiViewUpdatePermissions(view);
	return view;
}

/* ------------------------------------------------------------------------- */

static void	guiViewDestroy(guiGtkView_t *view)
{
btint	i, cnt;

	arrRemoveObjectEqualTo(allViews, view);
	if(view->helpButton != NULL){
		guiUnregisterHelpTogglebutton(view->helpButton);
		view->helpButton = NULL;
	}
	if(view->configBox != NULL){
		gtk_widget_unref(view->configBox);
		view->configBox = NULL;
	}
	if(view->naLabel != NULL){
		gtk_widget_unref(view->naLabel);
		view->naLabel = NULL;
	}
	if(view->hierarchyItem != NULL){
		g_free(view->hierarchyItem);
		view->hierarchyItem = NULL;
	}
	if(view->entries != NULL){
		cnt = arrObjectCount(view->entries);
		for(i=0;i<cnt;i++){
			guiEntryDestroy(arrObjectAtIndex(view->entries, i));
		}
		arrFree(view->entries);
		view->entries = NULL;
	}
	if(view->columns != NULL){
		cnt = arrObjectCount(view->columns);
		for(i=0;i<cnt;i++){
			guiColumnDestroy(arrObjectAtIndex(view->columns, i));
		}
		arrFree(view->columns);
		view->columns = NULL;
	}
	if(view->buttons != NULL){
		cnt = arrObjectCount(view->buttons);
		for(i=0;i<cnt;i++){
			guiButtonDestroy(arrObjectAtIndex(view->buttons, i));
		}
		arrFree(view->buttons);
		view->buttons = NULL;
	}
	memFree(view);
}

/* ------------------------------------------------------------------------- */

static void	guiViewUpdateDirty(guiGtkView_t *view)
{
btint			i, cnt;
guiGtkButton_t	*button;

	cnt = arrObjectCount(view->buttons);
	for(i=0;i<cnt;i++){
		button = arrObjectAtIndex(view->buttons, i);
		if((button->flags & MAIN_GUI_BUTTON_ACTIVE_IF_DIRTY) != 0){
			gtk_widget_set_sensitive(GTK_WIDGET(button->buttonWidget), view->isDirty);
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	rowDestroyNotify(gpointer data)
{
	if(data != NULL)
		g_free(data);
}

static btu8	*guiRowLoad(guiGtkView_t *view, ipcMsg_t *msg)
{
gchar	*columns[256];
btint	cnt, i;
btu8	*rowId;

	rowId = ipcGetAscii(msg);
	cnt = arrObjectCount(view->columns);
	for(i=0;i<cnt && i<ELEMENTS(columns);i++){
		columns[i] = ipcGetAscii(msg);
	}
	for(;i<cnt;i++){
		ipcGetAscii(msg);
	}
	i = gtk_clist_append(view->selectionList, columns);
	gtk_clist_set_row_data_full(view->selectionList, i, g_strdup(rowId), rowDestroyNotify);
	return rowId;
}

static void	guiEntryLoad(guiGtkEntry_t *entry, ipcMsg_t *msg)
{
btu8	*s;
btint	i;
double	d;

	switch(entry->type){
	case MAIN_GUI_ENTRY_STRING:
	case MAIN_GUI_ENTRY_PASSWORD:
	case MAIN_GUI_ENTRY_FILE:
		s = ipcGetAscii(msg);
		gtk_entry_set_text(GTK_ENTRY(entry->entryWidget), s);
		break;
	case MAIN_GUI_ENTRY_ENUM:
		i = ipcGetInt(msg);
		gtk_option_menu_set_history(GTK_OPTION_MENU(entry->entryWidget), i);
		gtk_object_set_user_data(GTK_OBJECT(entry->entryWidget), (gpointer)i);
		break;
	case MAIN_GUI_ENTRY_NUMBER:
		i = ipcGetInt(msg);
		d = (double)i * entry->fractionMultiplier / entry->fractionDivider;
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(entry->entryWidget), d);
		break;
	case MAIN_GUI_ENTRY_BOOL:
		i = ipcGetInt(msg);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(entry->entryWidget), i);
		break;
	case MAIN_GUI_ENTRY_SEPARATOR:
		/* no data */
		break;
	}
}

static void	guiViewDataLoad(guiGtkView_t *view, btbool withDefaults)
{
btint			i, cnt, selectRowNum;
guiGtkEntry_t	*entry;
ipcMsg_t		*msg;
ipccRes_t		res;
btu8			*rowId, *selectRow;

	if(view->objectType == MAIN_GUI_CLASS_HIERARCHYCFG && view->hierarchyItem == NULL)
		return;
	msg = ipccCurrentRxMessage();
	ipccStartProc(msg, MAIN_IPCFN_DSOURCE_EXEC);
	ipcAddInt(msg, view->objectId);
	ipcAddInt(msg, MAIN_GUI_METHOD_GET);
	if(view->objectType == MAIN_GUI_CLASS_HIERARCHYCFG){
		ipcAddAscii(msg, view->hierarchyItem);
	}
	if(view->objectType != MAIN_GUI_CLASS_LIST){
		ipcAddInt(msg, withDefaults);
	}
	ipccRpc(msg);
	res.err.class = ipcGetInt(msg);
	res.err.code = ipcGetInt(msg);
	res.errmsg = ipcGetAscii(msg);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		return;
	}
	if(view->objectType == MAIN_GUI_CLASS_LIST){
		gtk_clist_freeze(view->selectionList);
		selectRowNum = view->selectedRow;
		selectRow = g_strdup(gtk_clist_get_row_data(view->selectionList, selectRowNum));
		gtk_clist_clear(view->selectionList);
		cnt = ipcGetInt(msg);
		for(i=0;i<cnt;i++){
			rowId = guiRowLoad(view, msg);
			if(selectRow != NULL && strcmp(selectRow, rowId) == 0)
				selectRowNum = i;
		}
		if(selectRowNum >= cnt)
			selectRowNum = cnt - 1;
		gtk_clist_select_row(view->selectionList, selectRowNum, 0);
		gtk_clist_thaw(view->selectionList);
	}else{
		cnt = arrObjectCount(view->entries);
		for(i=0;i<cnt;i++){
			entry = arrObjectAtIndex(view->entries, i);
			guiEntryLoad(entry, msg);
			entry->isDirty = withDefaults;
		}
	}
	view->isDirty = withDefaults;
	guiViewUpdateDirty(view);
}

static void	guiViewListLoad(guiGtkView_t *view, btbool forceUpdateData)
{
btint			i, cnt, row;
ipcMsg_t		*msg;
ipccRes_t		res;
gchar			*item;
btbool			didSelect, wasNull, selectionChanged = 0;

	if(view->objectType == MAIN_GUI_CLASS_HIERARCHYCFG){
		msg = ipccCurrentRxMessage();
		ipccStartProc(msg, MAIN_IPCFN_DSOURCE_EXEC);
		ipcAddInt(msg, view->objectId);
		ipcAddInt(msg, MAIN_GUI_METHOD_HLIST);
		ipcAddInt(msg, 0);	/* only list configured, not all possible items */
		ipccRpc(msg);
		res.err.class = ipcGetInt(msg);
		res.err.code = ipcGetInt(msg);
		res.errmsg = ipcGetAscii(msg);
		if(res.err.class != 0){
			ipccPrintErr(&res);
			return;
		}
		cnt = ipcGetInt(msg);
		wasNull = view->hierarchyItem == NULL;
		gtk_clist_freeze(view->selectionList);
		gtk_clist_clear(view->selectionList);
		didSelect = 0;
		ignoreClistSignals = 1;
		for(i=0;i<cnt;i++){
			item = ipcGetAscii(msg);
			row = gtk_clist_append(view->selectionList, &item);
			if(view->hierarchyItem != NULL && strcmp(item, view->hierarchyItem) == 0){
				didSelect = 1;
				gtk_clist_select_row(view->selectionList, row, 0);
			}
		}
		if(!didSelect){
			if(view->hierarchyItem != NULL){
				g_free(view->hierarchyItem);
				view->hierarchyItem = NULL;
			}
			if(cnt > 0){
				selectionChanged = 1;
				gtk_clist_select_row(view->selectionList, 0, 0);
				item = NULL;
				gtk_clist_get_text(view->selectionList, 0, 0, &item);
				if(item != NULL)
					view->hierarchyItem = g_strdup(item);
			}
		}
		ignoreClistSignals = 0;
		gtk_clist_thaw(view->selectionList);
		if(wasNull && view->hierarchyItem != NULL){
			gtk_container_remove(GTK_CONTAINER(view->configParent), view->naLabel);
			gtk_container_add(GTK_CONTAINER(view->configParent), view->configBox);
		}else if(!wasNull && view->hierarchyItem == NULL){
			gtk_container_remove(GTK_CONTAINER(view->configParent), view->configBox);
			gtk_container_add(GTK_CONTAINER(view->configParent), view->naLabel);
		}
		if(selectionChanged || forceUpdateData){
			guiViewDataLoad(view, 0);
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	guiEntryStore(guiGtkEntry_t *entry, ipcMsg_t *msg)
{
btu8	*s;
btint	i;
double	d;

	ipcAddInt(msg, entry->isDirty);
	switch(entry->type){
	case MAIN_GUI_ENTRY_STRING:
	case MAIN_GUI_ENTRY_PASSWORD:
	case MAIN_GUI_ENTRY_FILE:
		s = gtk_entry_get_text(GTK_ENTRY(entry->entryWidget));
		ipcAddAscii(msg, s);
		break;
	case MAIN_GUI_ENTRY_ENUM:
		i = (btint)gtk_object_get_user_data(GTK_OBJECT(entry->entryWidget));
		ipcAddInt(msg, i);
		break;
	case MAIN_GUI_ENTRY_NUMBER:
		d = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(entry->entryWidget));
		i = (btint)rint(d / entry->fractionMultiplier * entry->fractionDivider);
		ipcAddInt(msg, i);
		break;
	case MAIN_GUI_ENTRY_BOOL:
		i = GTK_TOGGLE_BUTTON(entry->entryWidget)->active;
		ipcAddInt(msg, i);
		break;
	case MAIN_GUI_ENTRY_SEPARATOR:
		/* no data */
		break;
	}
}

static void	guiViewStore(guiGtkView_t *view)
{
ipcMsg_t		*msg;
btint			cnt, i;
ipccRes_t		res;
guiGtkEntry_t	*entry;
btu32			needXXFlags = 0;

	msg = ipccCurrentRxMessage();
	ipccStartProc(msg, MAIN_IPCFN_DSOURCE_EXEC);
	ipcAddInt(msg, view->objectId);
	ipcAddInt(msg, MAIN_GUI_METHOD_SET);
	if(view->objectType == MAIN_GUI_CLASS_HIERARCHYCFG){
		ipcAddAscii(msg, view->hierarchyItem);
	}
	cnt = arrObjectCount(view->entries);
	for(i=0;i<cnt;i++){
		entry = arrObjectAtIndex(view->entries, i);
		if(entry->isDirty)
			needXXFlags |= entry->flags;
		guiEntryStore(entry, msg);
		entry->isDirty = 0;
	}
	ipccRpc(msg);
	res.err.class = ipcGetInt(msg);
	res.err.code = ipcGetInt(msg);
	res.errmsg = ipcGetAscii(msg);
	if(res.err.class != 0){
		ipccPrintErr(&res);
	}
	view->isDirty = 0;
	guiViewUpdateDirty(view);
	if((needXXFlags & MAIN_GUI_ENTRY_FLG_NEEDS_RESTART) != 0){
		guiRunAlert("OK", NULL, NULL, "Change Requires Restart", "You have changed a setting which becomes active only after a restart of the Sharity daemon. If you don\'t know how to restart the daemon, simply reboot the machine.");
	}else if((needXXFlags & MAIN_GUI_ENTRY_FLG_NEEDS_RECONNECT) != 0){
		guiRunAlert("OK", NULL, NULL, "Change Requires Reconnect", "You have changed a setting which becomes active only after the internal server representation is freed. You can achieve this by unmounting all shares and logging out all users for the server. If in doubt, restart the Sharity daemon or reboot the machine.");
		return;
	}else if((needXXFlags & MAIN_GUI_ENTRY_FLG_NEEDS_REMOUNT) != 0){
		guiRunAlert("OK", NULL, NULL, "Change Requires Remount", "You have changed a setting which becomes active only at the next mount. To activate it on a mounted resource, you must first unmount it and then mount it again.");
	}
}

/* ------------------------------------------------------------------------- */

static void	updateAllPermissions(void)
{
btint		i, cnt;
btu8		*val;
ipccRes_t	res;

	ipccGetCfgValue(&res, "main/usersMayConfigure", 0, &val);
	if(res.err.class != 0){
		ipccPrintErr(&res);
	}else{
		cfgUsersMayChange = val != NULL && utilStrCmp(val, "yes", 0) == 0;
	}
	cnt = arrObjectCount(allViews);
	for(i=0;i<cnt;i++){
		guiViewUpdatePermissions(arrObjectAtIndex(allViews, i));
	}
}

/* ------------------------------------------------------------------------- */

static void	expertModeChanged(GtkWidget *sender)
{
	cfgIsExpert = GTK_TOGGLE_BUTTON(sender)->active;
	updateAllPermissions();
}

/* ------------------------------------------------------------------------- */

static gint		doTerminateId = -1;
static btint	connectionFd;
static gint		connectionSignalId;

static void	doTerminate(void)
{

	gtk_main_quit();
	if(doTerminateId == -1){
		doTerminateId = gtk_idle_add((GtkFunction)doTerminate, NULL);
	}
}

static void	termSignal(int signr)
{
ipccRes_t	res;
btbool		doTerm;

	ipccWillTerminate(&res, MAIN_TERM_LOGOUT, &doTerm);
	fprintf(stderr, "*** Sharity: received signal %d, terminating\n", signr);
	exit(1);
}

/* ------------------------------------------------------------------------- */

static void	guiDialogDestroy(guiGtkDialog_t *dialog)
{
	if(dialog->delayedInitId != -1)
		gtk_idle_remove(dialog->delayedInitId);
	dialog->delayedInitId = -1;
	htabRemoveObjectEqualTo(dialogsByObjectId, dialog);
	if(dialog->runModal && dialog->window != NULL)
		gtk_grab_remove(dialog->window);
	if(dialog->view != NULL){
		guiViewDestroy(dialog->view);
		dialog->view = NULL;
	}
	if(dialog->window != NULL){
		gtk_widget_destroy(dialog->window);
		dialog->window = NULL;
	}
	memFree(dialog);
}

/* ------------------------------------------------------------------------- */

static gint	dialogCloseEvent(GtkWidget *sender, GdkEvent *event, guiGtkDialog_t *dialog)
{
btint			i, cnt;
guiGtkButton_t	*button;

	cnt = arrObjectCount(dialog->view->buttons);
	for(i=0;i<cnt;i++){
		button = arrObjectAtIndex(dialog->view->buttons, i);
		if((button->flags & MAIN_GUI_BUTTON_EQUALS_CLOSEBOX) != 0){
			buttonClicked(button->buttonWidget, button);
			break;
		}
	}
	if(i >= cnt){	/* no appropriate button found, just remove the dialog */
		guiDialogDestroy(dialog);
	}
	return TRUE;	/* is handled */
}

static void	guiDialogInit(guiGtkDialog_t *dialog)
{
ipcMsg_t		*msg;
ipccRes_t		res;
btu8			*objectName;
btint			i, cnt;
guiGtkButton_t	*button;

	gtk_idle_remove(dialog->delayedInitId);
	dialog->delayedInitId = -1;
	ipccGetObjectDescription(&res, dialog->objectId);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		return;
	}
	msg = ipccCurrentRxMessage();
	objectName = ipcGetAscii(msg);
	dialog->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	setWindowTitle(dialog->window, objectName);
	dialog->view = guiViewNew(dialog->objectId, msg, 1);
	dialog->view->dialog = dialog;	/* backward link */
	guiViewListLoad(dialog->view, 0);
	guiViewDataLoad(dialog->view, 0);
	gtk_signal_connect(GTK_OBJECT(dialog->window), "delete_event", GTK_SIGNAL_FUNC(dialogCloseEvent), dialog);
	gtk_container_set_border_width(GTK_CONTAINER(dialog->window), 4);
	gtk_container_add(GTK_CONTAINER(dialog->window), dialog->view->mainView);
	cnt = arrObjectCount(dialog->view->buttons);
	for(i=0;i<cnt;i++){
		button = arrObjectAtIndex(dialog->view->buttons, i);
		if((button->flags & MAIN_GUI_BUTTON_DEFAULT) != 0){
			gtk_window_set_default(GTK_WINDOW(dialog->window), button->buttonWidget);
		}
	}
	if(dialog->runModal)
		gtk_grab_add(dialog->window);	/* run modal */
	gtk_widget_show(dialog->window);
}

static gint	updateViewsId = -1;

static void	updateViews(void)
{
btint			cnt, i;
guiGtkView_t	*view;

	gtk_idle_remove(updateViewsId);
	updateViewsId = -1;
	updateAllPermissions();
	cnt = arrObjectCount(allViews);
	for(i=0;i<cnt;i++){
		view = arrObjectAtIndex(allViews, i);
		if(view->needsUpdate){
			guiViewListLoad(view, 0);
			guiViewDataLoad(view, 0);
			view->needsUpdate = 0;
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	runDialog(bti32 sid, ipcMsg_t *msg)
{
btint			objectId;
guiGtkDialog_t	*dialog;
btbool			runModal;

	objectId = ipcGetInt(msg);
	runModal = (ipcGetInt(msg) & MAIN_DIALOG_FLG_MODAL) != 0;
	dialog = memCalloc(sizeof(guiGtkDialog_t));
	dialog->objectId = objectId;
	dialog->runModal = runModal;
	dialog->delayedInitId = gtk_idle_add((GtkFunction)guiDialogInit, dialog);
	htabAddObject(dialogsByObjectId, dialog);
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
}

static void	closeDialog(bti32 sid, ipcMsg_t *msg)
{
btint			objectId;
guiGtkDialog_t	*dialog;

	objectId = ipcGetInt(msg);
	dialog = htabFindObjectEqualTo(dialogsByObjectId, &objectId);
	if(dialog != NULL){
		guiDialogDestroy(dialog);
	}
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
}


static void	updateNotification(bti32 sid, ipcMsg_t *msg)
{
btint			objectId;
guiGtkView_t	*view;
btint			i, cnt;
hashtable_t		*idList;

	idList = htabNew(NULL, 16, 50);
	cnt = ipcGetInt(msg);
	for(i=0;i<cnt;i++){
		objectId = ipcGetInt(msg);
		htabAddObject(idList, (void *)objectId);
	}
	cnt = arrObjectCount(allViews);
	for(i=0;i<cnt;i++){
		view = arrObjectAtIndex(allViews, i);
		if(htabFindObjectEqualTo(idList, (void *)view->objectId) != NULL)
			view->needsUpdate = 1;
	}
	htabFree(idList);
	if(updateViewsId == -1)
		updateViewsId = gtk_idle_add((GtkFunction)updateViews, NULL);
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
}

/* ------------------------------------------------------------------------- */

static void	runFileselectorDialog(btint sid, ipcMsg_t *msg)
{
btu8	*title, *startPath, *path;

	title = ipcGetAscii(msg);
	startPath = ipcGetAscii(msg);
	path = guiRunFileselector(title, startPath);
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
	ipcAddInt(msg, path != NULL);
	if(path != NULL){
		ipcAddAscii(msg, path);
		g_free(path);
	}
}

/* ------------------------------------------------------------------------- */

static void	runAlert(btint sid, ipcMsg_t *msg)
{
btu8	*title, *text, *button1, *button2, *button3;
btint	res;

	title = ipcGetAscii(msg);
	text = ipcGetAscii(msg);
	button1 = ipcGetAscii(msg);
	button2 = ipcGetAscii(msg);
	button3 = ipcGetAscii(msg);
	res = guiRunAlert(button1, button2, button3, title, "%s", text);
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
	ipcAddInt(msg, res + 1);
}

/* ------------------------------------------------------------------------- */

static void	runSelectionBox(btint sid, ipcMsg_t *msg)
{
btu8	*title, *okButton, *cancelButton, *selected;
GList	*entries = NULL;
btint	i, cnt;

	title = ipcGetAscii(msg);
	okButton = ipcGetAscii(msg);
	cancelButton = ipcGetAscii(msg);
	cnt = ipcGetInt(msg);
	for(i=0;i<cnt;i++){
		entries = g_list_append(entries, ipcGetAscii(msg));
	}
	selected = guiRunSelectionBox(title, okButton, cancelButton, entries);
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
	ipcAddInt(msg, selected != NULL);
	if(selected != NULL){
		ipcAddAscii(msg, selected);
		g_free(selected);
	}
}

/* ------------------------------------------------------------------------- */

static void	displayError(btint sid, ipcMsg_t *msg)
{
ipccRes_t	err;

	err.err.class = ipcGetInt(msg);
	err.err.code = ipcGetInt(msg);
	err.errmsg = ipcGetAscii(msg);
	ipccPrintErr(&err);
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
	ipcAddInt(msg, sid);
	ipcAddInt(msg, -1);	/* this is a reply */
	ipcAddInt(msg, ipccUserId());
	ipcAddInt(msg, getgid());
	ipcAddInt(msg, 0);	/* err class */
	ipcAddInt(msg, 0);	/* err code */
}

/* ------------------------------------------------------------------------- */

static void	requestReceivedCallback(void *obj, void *arg)
{
ipccRequestData_t	*data = arg;
ipcMsg_t			*msg;

	if(arg == NULL){	/* broken pipe to daemon */
		gdk_input_remove(connectionSignalId);
		return;
	}
	msg = data->msg;
	ipcGetInt(msg);	/* skip uid */
	ipcGetInt(msg);	/* skip gid */
	switch(data->procId){
	case MAIN_IPCCFN_RUN_DIALOG:
		runDialog(data->sid, msg);
		break;
	case MAIN_IPCCFN_CLOSE_DIALOG:
		closeDialog(data->sid, msg);
		break;
	case MAIN_IPCCFN_NOTIFY_UPDATE:
		updateNotification(data->sid, msg);
		break;
	case MAIN_IPCCFN_RUN_FILESELECTOR:
		runFileselectorDialog(data->sid, msg);
		break;
	case MAIN_IPCCFN_RUN_ALERT:
		runAlert(data->sid, msg);
		break;
	case MAIN_IPCCFN_RUN_SELECTION:
		runSelectionBox(data->sid, msg);
		break;
	case MAIN_IPCCFN_DISPLAY_ERROR:
		displayError(data->sid, msg);
		break;
	default:	/* function not supported, construct error reply */
		ipcMsgClear(msg);
		ipcAddInt(msg, 0);	/* space for total length */
		ipcAddInt(msg, MAIN_IPC_PROTOCOL_VERSION);
		ipcAddInt(msg, data->sid);
		ipcAddInt(msg, -1);
		ipcAddInt(msg, ipccUserId());
		ipcAddInt(msg, getgid());
		ipcAddInt(msg, ERRCLS_FRAMEWORK);
		ipcAddInt(msg, ERR_FRWipcnotimpl);
		break;
	}
}

static void	requestReceivedSignal(void *ctx, gint fd, GdkInputCondition condition)
{
ipcMsg_t			*msg = ipccCurrentRxMessage();	/* use the currently allocated client's msg */
ipccRequestData_t	data;

	ipccReadMsg(msg);
	data.sid = ipcGetInt(msg);
	data.procId = ipcGetInt(msg);
	if(data.procId == -1){	/* this is a reply, not a request */
		guiRunAlert(NULL, NULL, " OK ", "Protocol Violation", "Daemon sent unexpected reply with SID=%d, Proc-ID=%d\nPlease restart the user interface.", data.sid, data.procId);
		exit(1);
	}
	data.msg = msg;
	requestReceivedCallback(ctx, &data);
	ipccWriteMsg(msg);
}


static gint	mainwindowCloseEvent(GtkWidget *sender, GdkEvent *event, void *dummy)
{
	doTerminate();
	return TRUE;	/* is handled */
}

static bti8	*getHelpText(bti8 *key)
{
ipccRes_t	res;
btu8		*helpText;

	ipccHelpForKey(&res, key, &helpText);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		exit(1);
	}
	return helpText;
}

/* ------------------------------------------------------------------------- */

#include "info.xpm"

static GtkWidget	*infoPage(GtkWidget *window)
{
GtkWidget	*frame, *fixed, *image, *box;
GtkCList	*clist;
GtkStyle	*style;
GdkPixmap	*pixmap;
GdkBitmap	*mask;
ipccRes_t	res;
ipcMsg_t	*msg;
btint		cnt, i, version, beta;
btu8		*modName;
gchar		scratch[256], *row[2];

	
	gtk_widget_realize(window);
    style = gtk_widget_get_style(window);

    pixmap = gdk_pixmap_create_from_xpm_d(window->window,  &mask, &style->bg[GTK_STATE_NORMAL], info_xpm);
    image = gtk_pixmap_new(pixmap, mask);
	gtk_widget_show(image);

	clist = (GtkCList *)gtk_clist_new(2);
	gtk_clist_column_titles_show(clist);
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_MULTIPLE);
	gtk_widget_set_sensitive(GTK_WIDGET(clist), 0);
	box = gtk_scrolled_window_new(NULL, NULL);
	gtk_widget_set_usize(GTK_WIDGET(box), 179, 140);
	gtk_widget_show(box);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(box), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_container_add(GTK_CONTAINER(box), GTK_WIDGET(clist));
	gtk_clist_set_column_title(clist, 0, "Module");
	gtk_clist_set_column_title(clist, 1, "Version");
	gtk_clist_set_column_width(clist, 0, 99);
	ipccListModules(&res);
	if(res.err.class != 0){
		ipccPrintErr(&res);
	}else{
		msg = ipccCurrentRxMessage();
		cnt = ipcGetInt(msg);
		row[0] = "GUI Application";
		row[1] = scratch;
		if(MAINCONST_BETA_VERSION != 0)
			sprintf(scratch, "%d.%d b%d", MAINCONST_VERSION / 1000, MAINCONST_VERSION % 1000, MAINCONST_BETA_VERSION);
		else
			sprintf(scratch, "%d.%d", MAINCONST_VERSION / 1000, MAINCONST_VERSION % 1000);
		gtk_clist_append(clist, row);
		for(i=0;i<cnt;i++){
			ipcGetInt(msg);	/* type */
			modName = ipcGetAscii(msg);	/* name */
			ipcGetInt(msg);	/* interface version */
			version = ipcGetInt(msg);	/* implementation version */
			beta = ipcGetInt(msg);	/* beta version */
			ipcGetInt(msg);	/* function base */
			ipcGetInt(msg);	/* function count */
			if(beta != 0)
				sprintf(scratch, "%d.%d b%d", version / 1000, version % 1000, beta);
			else
				sprintf(scratch, "%d.%d", version / 1000, version % 1000);
			row[0] = modName;
			row[1] = scratch;
			gtk_clist_append(clist, row);
		}
	}
	gtk_widget_show(GTK_WIDGET(clist));

	frame = gtk_alignment_new(0.5, 0.5, 0, 0);
	fixed = gtk_fixed_new();
	gtk_container_add(GTK_CONTAINER(frame), fixed);

	gtk_fixed_put(GTK_FIXED(fixed), image, 0, 0);
	gtk_fixed_put(GTK_FIXED(fixed), box, 302, 143);

	gtk_widget_show(fixed);
	gtk_widget_show(frame);
	return frame;
}

/* ------------------------------------------------------------------------- */

gint main (int argc, gchar **argv)
{                                  
GtkWidget		*window, *notebook, *label, *vbox, *checkbox, *helpButton;
ipccRes_t		res;
ipcMsg_t		*msg;
btint			i, cnt, *ids;
btu8			*objectName;
guiGtkView_t	*contentObject;
btbool			doTerm;
GdkPixmap		*pixmap, *mask;
GtkStyle		*style;
Window			xWin;
XWMHints		*hints;

	gtk_set_locale();	/* necessary for japanese localisation of GTK */
    gtk_init(&argc, &argv);	 
	guiSetAppName("Sharity");
	allViews = arrNew(NULL, 32);
	dialogsByObjectId = htabNew(&dialogInterface, 32, 50);
	connectionFd = ipccConnect();
	signal(SIGHUP, termSignal);
	signal(SIGINT, termSignal);
	signal(SIGQUIT, termSignal);
	signal(SIGTERM, termSignal);
	connectionSignalId = gdk_input_add(connectionFd, GDK_INPUT_READ, requestReceivedSignal, NULL);
	ipccSetRequestHandler(requestReceivedCallback, NULL);
	ipccSetProtocolVersion(0, MAIN_IPCFN_NUMFUNCTIONS, MAIN_IPC_PROTOCOL_VERSION);
	cfgIsRoot = ipccUserId() == 0;
	
	guiSetHelpIntroText(getHelpText("guiHelpIntro"));
	helpExpertMode = g_strdup(getHelpText("guiExpertMode"));
	helpAdd = g_strdup(getHelpText("guiAddEntry"));
	helpDelete = g_strdup(getHelpText("guiDeleteEntry"));

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    setWindowTitle(window, "Sharity Configuration");
	gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(mainwindowCloseEvent), NULL);
    gtk_container_set_border_width(GTK_CONTAINER(window), 4);

	statusBar = gtk_statusbar_new();
	gtk_container_set_border_width(GTK_CONTAINER(statusBar), 4);
	gtk_widget_show(statusBar);
	statusBarContext = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusBar), "info about disable reason");
	
	helpButton = gtk_toggle_button_new_with_label(" Help ");
	gtk_container_set_border_width(GTK_CONTAINER(helpButton), 4);
	guiRegisterHelpTogglebutton(helpButton);
	gtk_widget_show(helpButton);
	checkbox = gtk_check_button_new_with_label("Expert Mode");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbox), cfgIsExpert);
	guiHelpifyWidget(checkbox, helpExpertMode);
	gtk_widget_show(checkbox);
    notebook = gtk_notebook_new();
    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
	vbox = guiAlignmentBox(TRUE, 0,
		notebook, TRUE,
		guiAlignmentBox(FALSE, 20, statusBar, TRUE, helpButton, FALSE, checkbox, FALSE, NULL), FALSE,
		NULL);
    gtk_container_add(GTK_CONTAINER(window), vbox);
    gtk_widget_show(notebook);

	ipccConfigObjects(&res);
	if(res.err.class != 0){
		ipccPrintErr(&res);
		exit(1);
	}
	msg = ipccCurrentRxMessage();
	cnt = ipcGetInt(msg);
	ids = memAlloc(sizeof(btint) * cnt);
	for(i=0;i<cnt;i++){
		ids[i] = ipcGetInt(msg);
	}
	for(i=0;i<cnt;i++){
		ipccGetObjectDescription(&res, ids[i]);
		if(res.err.class != 0){
			ipccPrintErr(&res);
			exit(1);
		}
		msg = ipccCurrentRxMessage();
		objectName = ipcGetAscii(msg);
		contentObject = guiViewNew(ids[i], msg, 0);
		label = gtk_label_new(objectName);
		gtk_widget_show(label);
		gtk_notebook_append_page(GTK_NOTEBOOK(notebook), contentObject->mainView, label);
		guiViewListLoad(contentObject, 0);
		guiViewDataLoad(contentObject, 0);
	}
	memFree(ids);
	updateAllPermissions();
		
	label = gtk_label_new("Sharity\nInfo");
	gtk_widget_show(label);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), infoPage(window), label);

/* the following code sets the initial state for the window
 * to be Iconic. It also sets the pixmap for the icon to
 * be sharity.xpm. --- (Courtesy: David B. Varvell)
 */
	xWin = GDK_WINDOW_XWINDOW(window->window);
	hints = XGetWMHints(GDK_DISPLAY(), xWin);
	hints->initial_state = IconicState;
	XSetWMHints(GDK_DISPLAY(), xWin, hints);
	style = gtk_widget_get_style(window);
	pixmap = gdk_pixmap_create_from_xpm_d(window->window,  &mask, &style->bg[GTK_STATE_NORMAL], guiGetIconXpm());
	gdk_window_set_icon (window->window, NULL, pixmap, mask);

	gtk_widget_realize(window);
	gtk_widget_show(window);
	gtk_signal_connect(GTK_OBJECT(checkbox), "clicked", GTK_SIGNAL_FUNC(expertModeChanged), NULL);
	ipccRegisterForCallback(&res, "main");
	if(res.err.class != 0){
		ipccPrintErr(&res);
		exit(1);
	}
	for(;;){
		gtk_main();
		if(doTerminateId != -1){
			gtk_idle_remove(doTerminateId);	/* understood... we terminate */
			doTerminateId = -1;
		}
		ipccWillTerminate(&res, MAIN_TERM_USRQUIT, &doTerm);
		if(res.err.class != 0){
			ipccPrintErr(&res);
		}else{
			if(!doTerm)
				continue;
		}
		break;
	}

	ipccUnregisterForCallback(&res, "main");	/* may bring up panel before it returns */
	/* ignore result, we don't care */
    return 0;
}

/* ------------------------------------------------------------------------- */
